
/*  Copyright (C) 2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <arpa/inet.h>
#include <fcntl.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <math.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "zscanner/scanner.h"
#include "zscanner/functions.h"
#include "libknot/descriptor.h"

/*! \brief Shorthand for setting warning data. */
#define WARN(err_code) { s->error.code = err_code; }
/*! \brief Shorthand for setting error data. */
#define ERR(err_code) { WARN(err_code); s->error.fatal = true; }
/*! \brief Shorthand for error reset. */
#define NOERR { WARN(ZS_OK); s->error.fatal = false; }

/*!
 * \brief Writes record type number to r_data.
 *
 * \param type		Type number.
 * \param rdata_tail	Position where to write type number to.
 */
static inline void type_num(const uint16_t type, uint8_t **rdata_tail)
{
	*((uint16_t *)*rdata_tail) = htons(type);
	*rdata_tail += 2;
}

/*!
 * \brief Sets bit to bitmap window.
 *
 * \param type		Type number.
 * \param s		Scanner context.
 */
static inline void window_add_bit(const uint16_t type, zs_scanner_t *s) {
	uint8_t win      = type / 256;
	uint8_t bit_pos  = type % 256;
	uint8_t byte_pos = bit_pos / 8;

	((s->windows[win]).bitmap)[byte_pos] |= 128 >> (bit_pos % 8);

	if ((s->windows[win]).length < byte_pos + 1) {
		(s->windows[win]).length = byte_pos + 1;
	}

	if (s->last_window < win) {
		s->last_window = win;
	}
}

// Include scanner file (in Ragel).






__attribute__((visibility("default")))
int zs_init(
	zs_scanner_t *s,
	const char *origin,
	const uint16_t rclass,
	const uint32_t ttl)
{
	if (s == NULL) {
		return -1;
	}

	memset(s, 0, sizeof(*s));

	// Nonzero initial scanner state.
	s->cs = 1127;

	// Reset the file descriptor.
	s->file.descriptor = -1;

	// Use the root zone as origin if not specified.
	if (origin == NULL || strlen(origin) == 0) {
		origin = ".";
	}
	size_t origin_len = strlen(origin);

	// Prepare a zone settings header.
	const char *format;
	if (origin[origin_len - 1] != '.') {
		format = "$ORIGIN %s.\n";
	} else {
		format = "$ORIGIN %s\n";
	}

	char settings[1024];
	int ret = snprintf(settings, sizeof(settings), format, origin);
	if (ret <= 0 || ret >= sizeof(settings)) {
		ERR(ZS_ENOMEM);
		return -1;
	}

	// Parse the settings to set up the scanner origin.
	if (zs_set_input_string(s, settings, ret) != 0 ||
	    zs_parse_all(s) != 0) {
		return -1;
	}

	// Set scanner defaults.
	s->path = strdup(".");
	if (s->path == NULL) {
		ERR(ZS_ENOMEM);
		return -1;
	}
	s->default_class = rclass;
	s->default_ttl = ttl;
	s->line_counter = 1;

	s->state = ZS_STATE_NONE;
	s->process.automatic = false;

	return 0;
}

static void input_deinit(
	zs_scanner_t *s)
{
	// Deinit the file input.
	if (s->file.descriptor != -1) {
		// Unmap the file content.
		if (s->input.start != NULL) {
			munmap((void *)s->input.start,
			       s->input.end - s->input.start);
		}

		// Close the opened file.
		close(s->file.descriptor);
		s->file.descriptor = -1;
		free(s->file.name);
		s->file.name = NULL;
	}

	// Unset the input limits.
	s->input.start   = NULL;
	s->input.current = NULL;
	s->input.end     = NULL;
	s->input.eof     = false;
}

__attribute__((visibility("default")))
void zs_deinit(
	zs_scanner_t *s)
{
	if (s == NULL) {
		return;
	}

	input_deinit(s);
	free(s->path);
}

__attribute__((visibility("default")))
int zs_set_input_string(
	zs_scanner_t *s,
	const char *input,
	size_t size)
{
	if (s == NULL) {
		return -1;
	}

	if (input == NULL) {
		ERR(ZS_EINVAL);
		return -1;
	}

	// Deinit possibly opened file.
	input_deinit(s);

	// Set the scanner input limits.
	s->input.start   = input;
	s->input.current = input;
	s->input.end     = input + size;

	return 0;
}

__attribute__((visibility("default")))
int zs_set_input_file(
	zs_scanner_t *s,
	const char *file_name)
{
	if (s == NULL) {
		return -1;
	}

	if (file_name == NULL) {
		ERR(ZS_EINVAL);
		return -1;
	}

	// Deinit possibly opened file.
	input_deinit(s);

	// Try to open the file.
	s->file.descriptor = open(file_name, O_RDONLY);
	if (s->file.descriptor == -1) {
		ERR(ZS_FILE_OPEN);
		return -1;
	}

	// Check for regular file input.
	struct stat file_stat;
	if (fstat(s->file.descriptor, &file_stat) == -1 ||
	    !S_ISREG(file_stat.st_mode)) {
		ERR(ZS_FILE_INVALID);
		input_deinit(s);
		return -1;
	}

	// Check for empty file (cannot mmap).
	if (file_stat.st_size > 0) {
		// Map the file to the memory.
		char *start = mmap(0, file_stat.st_size, PROT_READ, MAP_SHARED,
		                   s->file.descriptor, 0);
		if (start == MAP_FAILED) {
			ERR(ZS_FILE_MMAP);
			input_deinit(s);
			return -1;
		}

		// Try to set the mapped memory advise to sequential.
		(void)madvise(start, file_stat.st_size, MADV_SEQUENTIAL);

		// Set the scanner input limits.
		s->input.start   = start;
		s->input.current = start;
		s->input.end     = start + file_stat.st_size;
	}

	// Get absolute path of the zone file.
	char *full_name = realpath(file_name, NULL);
	if (full_name != NULL) {
		free(s->path);
		s->path = strdup(dirname(full_name));
		free(full_name);
		if (s->path == NULL) {
			ERR(ZS_ENOMEM);
			input_deinit(s);
			return -1;
		}
	} else {
		ERR(ZS_FILE_PATH);
		input_deinit(s);
		return -1;
	}

	s->file.name = strdup(file_name);
	if (s->file.name == NULL) {
		ERR(ZS_ENOMEM);
		input_deinit(s);
		return -1;
	}

	return 0;
}

__attribute__((visibility("default")))
int zs_set_processing(
	zs_scanner_t *s,
	void (*process_record)(zs_scanner_t *),
	void (*process_error)(zs_scanner_t *),
	void *data)
{
	if (s == NULL) {
		return -1;
	}

	s->process.record = process_record;
	s->process.error = process_error;
	s->process.data = data;

	return 0;
}

static void parse(
	zs_scanner_t *s)
{
	// Restore scanner input limits (Ragel internals).
	const char *p = s->input.current;
	const char *pe = s->input.end;
	const char *eof = s->input.eof ? pe : NULL;

	// Restore state variables (Ragel internals).
	int cs  = s->cs;
	int top = s->top;
	int stack[RAGEL_STACK_SIZE];
	memcpy(stack, s->stack, sizeof(stack));

	// Auxiliary variables which are used in scanner body.
	struct in_addr  addr4;
	struct in6_addr addr6;
	uint32_t timestamp;
	int16_t  window;
	int      ret;

	// Next 2 variables are for better performance.
	// Restoring r_data pointer to next free space.
	uint8_t *rdata_tail = s->r_data + s->r_data_tail;
	// Initialization of the last r_data byte.
	uint8_t *rdata_stop = s->r_data + MAX_RDATA_LENGTH - 1;

	// Write scanner body (in C).

	{
	short _widec;
	if ( p == pe )
		goto _test_eof;
	goto _resume;

_again:
	switch ( cs ) {
		case 1127: goto st1127;
		case 0: goto st0;
		case 1: goto st1;
		case 2: goto st2;
		case 3: goto st3;
		case 4: goto st4;
		case 5: goto st5;
		case 6: goto st6;
		case 7: goto st7;
		case 8: goto st8;
		case 9: goto st9;
		case 10: goto st10;
		case 11: goto st11;
		case 12: goto st12;
		case 13: goto st13;
		case 1128: goto st1128;
		case 14: goto st14;
		case 15: goto st15;
		case 16: goto st16;
		case 17: goto st17;
		case 18: goto st18;
		case 19: goto st19;
		case 20: goto st20;
		case 21: goto st21;
		case 22: goto st22;
		case 23: goto st23;
		case 24: goto st24;
		case 25: goto st25;
		case 26: goto st26;
		case 27: goto st27;
		case 28: goto st28;
		case 29: goto st29;
		case 30: goto st30;
		case 31: goto st31;
		case 32: goto st32;
		case 33: goto st33;
		case 34: goto st34;
		case 35: goto st35;
		case 36: goto st36;
		case 37: goto st37;
		case 38: goto st38;
		case 39: goto st39;
		case 40: goto st40;
		case 41: goto st41;
		case 42: goto st42;
		case 43: goto st43;
		case 44: goto st44;
		case 45: goto st45;
		case 46: goto st46;
		case 47: goto st47;
		case 48: goto st48;
		case 49: goto st49;
		case 50: goto st50;
		case 51: goto st51;
		case 52: goto st52;
		case 53: goto st53;
		case 54: goto st54;
		case 55: goto st55;
		case 56: goto st56;
		case 57: goto st57;
		case 58: goto st58;
		case 59: goto st59;
		case 60: goto st60;
		case 61: goto st61;
		case 62: goto st62;
		case 63: goto st63;
		case 64: goto st64;
		case 65: goto st65;
		case 66: goto st66;
		case 67: goto st67;
		case 68: goto st68;
		case 69: goto st69;
		case 70: goto st70;
		case 71: goto st71;
		case 72: goto st72;
		case 73: goto st73;
		case 74: goto st74;
		case 75: goto st75;
		case 76: goto st76;
		case 77: goto st77;
		case 78: goto st78;
		case 79: goto st79;
		case 80: goto st80;
		case 81: goto st81;
		case 82: goto st82;
		case 83: goto st83;
		case 84: goto st84;
		case 85: goto st85;
		case 86: goto st86;
		case 87: goto st87;
		case 88: goto st88;
		case 89: goto st89;
		case 90: goto st90;
		case 91: goto st91;
		case 92: goto st92;
		case 93: goto st93;
		case 94: goto st94;
		case 95: goto st95;
		case 96: goto st96;
		case 97: goto st97;
		case 98: goto st98;
		case 99: goto st99;
		case 100: goto st100;
		case 101: goto st101;
		case 102: goto st102;
		case 103: goto st103;
		case 104: goto st104;
		case 105: goto st105;
		case 106: goto st106;
		case 107: goto st107;
		case 108: goto st108;
		case 109: goto st109;
		case 110: goto st110;
		case 111: goto st111;
		case 112: goto st112;
		case 113: goto st113;
		case 114: goto st114;
		case 115: goto st115;
		case 116: goto st116;
		case 117: goto st117;
		case 118: goto st118;
		case 119: goto st119;
		case 120: goto st120;
		case 121: goto st121;
		case 122: goto st122;
		case 123: goto st123;
		case 124: goto st124;
		case 125: goto st125;
		case 126: goto st126;
		case 127: goto st127;
		case 128: goto st128;
		case 129: goto st129;
		case 130: goto st130;
		case 131: goto st131;
		case 132: goto st132;
		case 133: goto st133;
		case 134: goto st134;
		case 135: goto st135;
		case 136: goto st136;
		case 137: goto st137;
		case 138: goto st138;
		case 139: goto st139;
		case 140: goto st140;
		case 141: goto st141;
		case 1129: goto st1129;
		case 142: goto st142;
		case 143: goto st143;
		case 144: goto st144;
		case 145: goto st145;
		case 146: goto st146;
		case 147: goto st147;
		case 148: goto st148;
		case 149: goto st149;
		case 150: goto st150;
		case 151: goto st151;
		case 1130: goto st1130;
		case 152: goto st152;
		case 153: goto st153;
		case 154: goto st154;
		case 155: goto st155;
		case 156: goto st156;
		case 157: goto st157;
		case 158: goto st158;
		case 159: goto st159;
		case 1131: goto st1131;
		case 160: goto st160;
		case 161: goto st161;
		case 162: goto st162;
		case 1132: goto st1132;
		case 163: goto st163;
		case 164: goto st164;
		case 165: goto st165;
		case 166: goto st166;
		case 167: goto st167;
		case 168: goto st168;
		case 169: goto st169;
		case 170: goto st170;
		case 171: goto st171;
		case 172: goto st172;
		case 173: goto st173;
		case 1133: goto st1133;
		case 174: goto st174;
		case 175: goto st175;
		case 176: goto st176;
		case 177: goto st177;
		case 1134: goto st1134;
		case 178: goto st178;
		case 179: goto st179;
		case 180: goto st180;
		case 181: goto st181;
		case 182: goto st182;
		case 183: goto st183;
		case 184: goto st184;
		case 185: goto st185;
		case 186: goto st186;
		case 187: goto st187;
		case 188: goto st188;
		case 189: goto st189;
		case 190: goto st190;
		case 191: goto st191;
		case 192: goto st192;
		case 193: goto st193;
		case 1135: goto st1135;
		case 194: goto st194;
		case 195: goto st195;
		case 196: goto st196;
		case 197: goto st197;
		case 198: goto st198;
		case 199: goto st199;
		case 200: goto st200;
		case 201: goto st201;
		case 202: goto st202;
		case 203: goto st203;
		case 204: goto st204;
		case 205: goto st205;
		case 206: goto st206;
		case 207: goto st207;
		case 208: goto st208;
		case 209: goto st209;
		case 1136: goto st1136;
		case 210: goto st210;
		case 211: goto st211;
		case 212: goto st212;
		case 213: goto st213;
		case 214: goto st214;
		case 215: goto st215;
		case 216: goto st216;
		case 217: goto st217;
		case 218: goto st218;
		case 219: goto st219;
		case 220: goto st220;
		case 221: goto st221;
		case 222: goto st222;
		case 223: goto st223;
		case 224: goto st224;
		case 225: goto st225;
		case 226: goto st226;
		case 227: goto st227;
		case 228: goto st228;
		case 229: goto st229;
		case 230: goto st230;
		case 231: goto st231;
		case 232: goto st232;
		case 233: goto st233;
		case 234: goto st234;
		case 235: goto st235;
		case 236: goto st236;
		case 237: goto st237;
		case 238: goto st238;
		case 239: goto st239;
		case 240: goto st240;
		case 241: goto st241;
		case 242: goto st242;
		case 243: goto st243;
		case 244: goto st244;
		case 245: goto st245;
		case 246: goto st246;
		case 247: goto st247;
		case 248: goto st248;
		case 249: goto st249;
		case 250: goto st250;
		case 251: goto st251;
		case 252: goto st252;
		case 253: goto st253;
		case 254: goto st254;
		case 255: goto st255;
		case 256: goto st256;
		case 257: goto st257;
		case 258: goto st258;
		case 259: goto st259;
		case 260: goto st260;
		case 261: goto st261;
		case 262: goto st262;
		case 263: goto st263;
		case 264: goto st264;
		case 265: goto st265;
		case 266: goto st266;
		case 267: goto st267;
		case 268: goto st268;
		case 269: goto st269;
		case 1137: goto st1137;
		case 270: goto st270;
		case 271: goto st271;
		case 1138: goto st1138;
		case 272: goto st272;
		case 273: goto st273;
		case 274: goto st274;
		case 275: goto st275;
		case 276: goto st276;
		case 277: goto st277;
		case 278: goto st278;
		case 279: goto st279;
		case 280: goto st280;
		case 1139: goto st1139;
		case 1140: goto st1140;
		case 281: goto st281;
		case 282: goto st282;
		case 283: goto st283;
		case 284: goto st284;
		case 285: goto st285;
		case 286: goto st286;
		case 287: goto st287;
		case 288: goto st288;
		case 289: goto st289;
		case 290: goto st290;
		case 291: goto st291;
		case 292: goto st292;
		case 293: goto st293;
		case 294: goto st294;
		case 1141: goto st1141;
		case 295: goto st295;
		case 296: goto st296;
		case 297: goto st297;
		case 298: goto st298;
		case 299: goto st299;
		case 300: goto st300;
		case 301: goto st301;
		case 302: goto st302;
		case 303: goto st303;
		case 304: goto st304;
		case 1142: goto st1142;
		case 305: goto st305;
		case 306: goto st306;
		case 307: goto st307;
		case 308: goto st308;
		case 309: goto st309;
		case 310: goto st310;
		case 311: goto st311;
		case 312: goto st312;
		case 313: goto st313;
		case 314: goto st314;
		case 315: goto st315;
		case 316: goto st316;
		case 317: goto st317;
		case 318: goto st318;
		case 1143: goto st1143;
		case 319: goto st319;
		case 320: goto st320;
		case 321: goto st321;
		case 322: goto st322;
		case 323: goto st323;
		case 324: goto st324;
		case 325: goto st325;
		case 1144: goto st1144;
		case 326: goto st326;
		case 327: goto st327;
		case 328: goto st328;
		case 329: goto st329;
		case 330: goto st330;
		case 331: goto st331;
		case 332: goto st332;
		case 333: goto st333;
		case 334: goto st334;
		case 1145: goto st1145;
		case 1146: goto st1146;
		case 1147: goto st1147;
		case 335: goto st335;
		case 336: goto st336;
		case 337: goto st337;
		case 338: goto st338;
		case 339: goto st339;
		case 340: goto st340;
		case 341: goto st341;
		case 342: goto st342;
		case 1148: goto st1148;
		case 1149: goto st1149;
		case 343: goto st343;
		case 344: goto st344;
		case 345: goto st345;
		case 1150: goto st1150;
		case 346: goto st346;
		case 347: goto st347;
		case 348: goto st348;
		case 349: goto st349;
		case 350: goto st350;
		case 351: goto st351;
		case 352: goto st352;
		case 353: goto st353;
		case 354: goto st354;
		case 355: goto st355;
		case 356: goto st356;
		case 357: goto st357;
		case 358: goto st358;
		case 359: goto st359;
		case 360: goto st360;
		case 361: goto st361;
		case 362: goto st362;
		case 363: goto st363;
		case 364: goto st364;
		case 365: goto st365;
		case 366: goto st366;
		case 367: goto st367;
		case 368: goto st368;
		case 369: goto st369;
		case 370: goto st370;
		case 371: goto st371;
		case 372: goto st372;
		case 373: goto st373;
		case 374: goto st374;
		case 375: goto st375;
		case 376: goto st376;
		case 377: goto st377;
		case 378: goto st378;
		case 379: goto st379;
		case 380: goto st380;
		case 381: goto st381;
		case 382: goto st382;
		case 383: goto st383;
		case 384: goto st384;
		case 385: goto st385;
		case 386: goto st386;
		case 387: goto st387;
		case 388: goto st388;
		case 389: goto st389;
		case 390: goto st390;
		case 391: goto st391;
		case 392: goto st392;
		case 393: goto st393;
		case 394: goto st394;
		case 395: goto st395;
		case 396: goto st396;
		case 397: goto st397;
		case 398: goto st398;
		case 399: goto st399;
		case 400: goto st400;
		case 401: goto st401;
		case 402: goto st402;
		case 403: goto st403;
		case 404: goto st404;
		case 405: goto st405;
		case 406: goto st406;
		case 407: goto st407;
		case 408: goto st408;
		case 409: goto st409;
		case 410: goto st410;
		case 411: goto st411;
		case 412: goto st412;
		case 413: goto st413;
		case 414: goto st414;
		case 415: goto st415;
		case 416: goto st416;
		case 417: goto st417;
		case 418: goto st418;
		case 419: goto st419;
		case 420: goto st420;
		case 421: goto st421;
		case 422: goto st422;
		case 423: goto st423;
		case 424: goto st424;
		case 425: goto st425;
		case 426: goto st426;
		case 427: goto st427;
		case 428: goto st428;
		case 429: goto st429;
		case 430: goto st430;
		case 431: goto st431;
		case 432: goto st432;
		case 433: goto st433;
		case 434: goto st434;
		case 435: goto st435;
		case 436: goto st436;
		case 437: goto st437;
		case 438: goto st438;
		case 439: goto st439;
		case 440: goto st440;
		case 441: goto st441;
		case 442: goto st442;
		case 443: goto st443;
		case 444: goto st444;
		case 445: goto st445;
		case 446: goto st446;
		case 447: goto st447;
		case 448: goto st448;
		case 449: goto st449;
		case 450: goto st450;
		case 451: goto st451;
		case 452: goto st452;
		case 453: goto st453;
		case 454: goto st454;
		case 455: goto st455;
		case 456: goto st456;
		case 457: goto st457;
		case 458: goto st458;
		case 459: goto st459;
		case 460: goto st460;
		case 461: goto st461;
		case 462: goto st462;
		case 463: goto st463;
		case 464: goto st464;
		case 465: goto st465;
		case 466: goto st466;
		case 467: goto st467;
		case 468: goto st468;
		case 469: goto st469;
		case 470: goto st470;
		case 471: goto st471;
		case 472: goto st472;
		case 473: goto st473;
		case 474: goto st474;
		case 1151: goto st1151;
		case 1152: goto st1152;
		case 1153: goto st1153;
		case 475: goto st475;
		case 476: goto st476;
		case 477: goto st477;
		case 478: goto st478;
		case 479: goto st479;
		case 1154: goto st1154;
		case 480: goto st480;
		case 481: goto st481;
		case 482: goto st482;
		case 483: goto st483;
		case 1155: goto st1155;
		case 1156: goto st1156;
		case 1157: goto st1157;
		case 484: goto st484;
		case 485: goto st485;
		case 1158: goto st1158;
		case 486: goto st486;
		case 487: goto st487;
		case 488: goto st488;
		case 1159: goto st1159;
		case 489: goto st489;
		case 490: goto st490;
		case 491: goto st491;
		case 492: goto st492;
		case 493: goto st493;
		case 494: goto st494;
		case 495: goto st495;
		case 496: goto st496;
		case 497: goto st497;
		case 498: goto st498;
		case 499: goto st499;
		case 500: goto st500;
		case 501: goto st501;
		case 502: goto st502;
		case 503: goto st503;
		case 504: goto st504;
		case 505: goto st505;
		case 506: goto st506;
		case 507: goto st507;
		case 508: goto st508;
		case 509: goto st509;
		case 510: goto st510;
		case 511: goto st511;
		case 512: goto st512;
		case 513: goto st513;
		case 514: goto st514;
		case 515: goto st515;
		case 516: goto st516;
		case 517: goto st517;
		case 518: goto st518;
		case 519: goto st519;
		case 520: goto st520;
		case 521: goto st521;
		case 522: goto st522;
		case 523: goto st523;
		case 524: goto st524;
		case 525: goto st525;
		case 526: goto st526;
		case 527: goto st527;
		case 528: goto st528;
		case 529: goto st529;
		case 530: goto st530;
		case 531: goto st531;
		case 532: goto st532;
		case 533: goto st533;
		case 534: goto st534;
		case 535: goto st535;
		case 536: goto st536;
		case 537: goto st537;
		case 538: goto st538;
		case 539: goto st539;
		case 540: goto st540;
		case 541: goto st541;
		case 542: goto st542;
		case 543: goto st543;
		case 544: goto st544;
		case 545: goto st545;
		case 546: goto st546;
		case 547: goto st547;
		case 548: goto st548;
		case 549: goto st549;
		case 550: goto st550;
		case 551: goto st551;
		case 552: goto st552;
		case 553: goto st553;
		case 554: goto st554;
		case 555: goto st555;
		case 556: goto st556;
		case 557: goto st557;
		case 558: goto st558;
		case 559: goto st559;
		case 560: goto st560;
		case 561: goto st561;
		case 562: goto st562;
		case 563: goto st563;
		case 564: goto st564;
		case 565: goto st565;
		case 566: goto st566;
		case 567: goto st567;
		case 568: goto st568;
		case 569: goto st569;
		case 570: goto st570;
		case 571: goto st571;
		case 572: goto st572;
		case 573: goto st573;
		case 574: goto st574;
		case 575: goto st575;
		case 576: goto st576;
		case 577: goto st577;
		case 578: goto st578;
		case 579: goto st579;
		case 580: goto st580;
		case 581: goto st581;
		case 582: goto st582;
		case 583: goto st583;
		case 584: goto st584;
		case 585: goto st585;
		case 586: goto st586;
		case 587: goto st587;
		case 588: goto st588;
		case 589: goto st589;
		case 590: goto st590;
		case 591: goto st591;
		case 592: goto st592;
		case 1160: goto st1160;
		case 593: goto st593;
		case 594: goto st594;
		case 595: goto st595;
		case 596: goto st596;
		case 597: goto st597;
		case 598: goto st598;
		case 599: goto st599;
		case 600: goto st600;
		case 601: goto st601;
		case 602: goto st602;
		case 603: goto st603;
		case 604: goto st604;
		case 605: goto st605;
		case 606: goto st606;
		case 607: goto st607;
		case 608: goto st608;
		case 609: goto st609;
		case 610: goto st610;
		case 611: goto st611;
		case 612: goto st612;
		case 613: goto st613;
		case 614: goto st614;
		case 615: goto st615;
		case 616: goto st616;
		case 617: goto st617;
		case 618: goto st618;
		case 619: goto st619;
		case 620: goto st620;
		case 621: goto st621;
		case 622: goto st622;
		case 623: goto st623;
		case 624: goto st624;
		case 625: goto st625;
		case 626: goto st626;
		case 627: goto st627;
		case 628: goto st628;
		case 629: goto st629;
		case 630: goto st630;
		case 631: goto st631;
		case 632: goto st632;
		case 633: goto st633;
		case 1161: goto st1161;
		case 634: goto st634;
		case 635: goto st635;
		case 1162: goto st1162;
		case 636: goto st636;
		case 637: goto st637;
		case 638: goto st638;
		case 639: goto st639;
		case 640: goto st640;
		case 641: goto st641;
		case 642: goto st642;
		case 643: goto st643;
		case 644: goto st644;
		case 645: goto st645;
		case 646: goto st646;
		case 647: goto st647;
		case 648: goto st648;
		case 649: goto st649;
		case 1163: goto st1163;
		case 650: goto st650;
		case 651: goto st651;
		case 652: goto st652;
		case 653: goto st653;
		case 654: goto st654;
		case 655: goto st655;
		case 656: goto st656;
		case 657: goto st657;
		case 658: goto st658;
		case 659: goto st659;
		case 660: goto st660;
		case 661: goto st661;
		case 662: goto st662;
		case 663: goto st663;
		case 664: goto st664;
		case 665: goto st665;
		case 666: goto st666;
		case 667: goto st667;
		case 668: goto st668;
		case 669: goto st669;
		case 670: goto st670;
		case 671: goto st671;
		case 1164: goto st1164;
		case 672: goto st672;
		case 673: goto st673;
		case 674: goto st674;
		case 675: goto st675;
		case 676: goto st676;
		case 1165: goto st1165;
		case 677: goto st677;
		case 678: goto st678;
		case 679: goto st679;
		case 680: goto st680;
		case 681: goto st681;
		case 1166: goto st1166;
		case 682: goto st682;
		case 683: goto st683;
		case 684: goto st684;
		case 685: goto st685;
		case 686: goto st686;
		case 1167: goto st1167;
		case 1168: goto st1168;
		case 1169: goto st1169;
		case 687: goto st687;
		case 688: goto st688;
		case 1170: goto st1170;
		case 689: goto st689;
		case 690: goto st690;
		case 691: goto st691;
		case 692: goto st692;
		case 693: goto st693;
		case 694: goto st694;
		case 695: goto st695;
		case 696: goto st696;
		case 697: goto st697;
		case 698: goto st698;
		case 699: goto st699;
		case 700: goto st700;
		case 701: goto st701;
		case 702: goto st702;
		case 703: goto st703;
		case 704: goto st704;
		case 705: goto st705;
		case 706: goto st706;
		case 707: goto st707;
		case 708: goto st708;
		case 709: goto st709;
		case 710: goto st710;
		case 711: goto st711;
		case 712: goto st712;
		case 713: goto st713;
		case 714: goto st714;
		case 715: goto st715;
		case 1171: goto st1171;
		case 1172: goto st1172;
		case 1173: goto st1173;
		case 716: goto st716;
		case 717: goto st717;
		case 718: goto st718;
		case 1174: goto st1174;
		case 1175: goto st1175;
		case 719: goto st719;
		case 720: goto st720;
		case 721: goto st721;
		case 722: goto st722;
		case 1176: goto st1176;
		case 1177: goto st1177;
		case 723: goto st723;
		case 724: goto st724;
		case 725: goto st725;
		case 726: goto st726;
		case 1178: goto st1178;
		case 1179: goto st1179;
		case 727: goto st727;
		case 728: goto st728;
		case 729: goto st729;
		case 730: goto st730;
		case 731: goto st731;
		case 732: goto st732;
		case 733: goto st733;
		case 734: goto st734;
		case 735: goto st735;
		case 736: goto st736;
		case 737: goto st737;
		case 738: goto st738;
		case 739: goto st739;
		case 740: goto st740;
		case 741: goto st741;
		case 742: goto st742;
		case 743: goto st743;
		case 744: goto st744;
		case 745: goto st745;
		case 746: goto st746;
		case 747: goto st747;
		case 748: goto st748;
		case 749: goto st749;
		case 750: goto st750;
		case 751: goto st751;
		case 1180: goto st1180;
		case 752: goto st752;
		case 753: goto st753;
		case 754: goto st754;
		case 755: goto st755;
		case 756: goto st756;
		case 757: goto st757;
		case 758: goto st758;
		case 759: goto st759;
		case 760: goto st760;
		case 761: goto st761;
		case 762: goto st762;
		case 763: goto st763;
		case 764: goto st764;
		case 765: goto st765;
		case 766: goto st766;
		case 1181: goto st1181;
		case 767: goto st767;
		case 768: goto st768;
		case 769: goto st769;
		case 770: goto st770;
		case 771: goto st771;
		case 772: goto st772;
		case 773: goto st773;
		case 774: goto st774;
		case 775: goto st775;
		case 776: goto st776;
		case 777: goto st777;
		case 778: goto st778;
		case 779: goto st779;
		case 1182: goto st1182;
		case 780: goto st780;
		case 781: goto st781;
		case 782: goto st782;
		case 783: goto st783;
		case 784: goto st784;
		case 785: goto st785;
		case 786: goto st786;
		case 787: goto st787;
		case 788: goto st788;
		case 789: goto st789;
		case 790: goto st790;
		case 1183: goto st1183;
		case 1184: goto st1184;
		case 791: goto st791;
		case 792: goto st792;
		case 793: goto st793;
		case 1185: goto st1185;
		case 794: goto st794;
		case 795: goto st795;
		case 796: goto st796;
		case 797: goto st797;
		case 798: goto st798;
		case 799: goto st799;
		case 800: goto st800;
		case 801: goto st801;
		case 802: goto st802;
		case 803: goto st803;
		case 1186: goto st1186;
		case 1187: goto st1187;
		case 1188: goto st1188;
		case 804: goto st804;
		case 805: goto st805;
		case 806: goto st806;
		case 807: goto st807;
		case 808: goto st808;
		case 809: goto st809;
		case 810: goto st810;
		case 811: goto st811;
		case 812: goto st812;
		case 813: goto st813;
		case 814: goto st814;
		case 1189: goto st1189;
		case 1190: goto st1190;
		case 1191: goto st1191;
		case 815: goto st815;
		case 816: goto st816;
		case 817: goto st817;
		case 818: goto st818;
		case 819: goto st819;
		case 820: goto st820;
		case 821: goto st821;
		case 822: goto st822;
		case 823: goto st823;
		case 824: goto st824;
		case 825: goto st825;
		case 826: goto st826;
		case 1192: goto st1192;
		case 827: goto st827;
		case 828: goto st828;
		case 829: goto st829;
		case 1193: goto st1193;
		case 1194: goto st1194;
		case 830: goto st830;
		case 1195: goto st1195;
		case 1196: goto st1196;
		case 831: goto st831;
		case 1197: goto st1197;
		case 1198: goto st1198;
		case 832: goto st832;
		case 833: goto st833;
		case 834: goto st834;
		case 835: goto st835;
		case 836: goto st836;
		case 837: goto st837;
		case 838: goto st838;
		case 839: goto st839;
		case 840: goto st840;
		case 841: goto st841;
		case 842: goto st842;
		case 843: goto st843;
		case 844: goto st844;
		case 845: goto st845;
		case 846: goto st846;
		case 847: goto st847;
		case 848: goto st848;
		case 849: goto st849;
		case 850: goto st850;
		case 851: goto st851;
		case 852: goto st852;
		case 853: goto st853;
		case 854: goto st854;
		case 855: goto st855;
		case 856: goto st856;
		case 857: goto st857;
		case 858: goto st858;
		case 859: goto st859;
		case 860: goto st860;
		case 861: goto st861;
		case 862: goto st862;
		case 863: goto st863;
		case 864: goto st864;
		case 865: goto st865;
		case 866: goto st866;
		case 867: goto st867;
		case 868: goto st868;
		case 869: goto st869;
		case 870: goto st870;
		case 871: goto st871;
		case 872: goto st872;
		case 873: goto st873;
		case 1199: goto st1199;
		case 874: goto st874;
		case 875: goto st875;
		case 876: goto st876;
		case 877: goto st877;
		case 878: goto st878;
		case 879: goto st879;
		case 880: goto st880;
		case 881: goto st881;
		case 882: goto st882;
		case 883: goto st883;
		case 884: goto st884;
		case 885: goto st885;
		case 886: goto st886;
		case 887: goto st887;
		case 888: goto st888;
		case 889: goto st889;
		case 890: goto st890;
		case 891: goto st891;
		case 892: goto st892;
		case 893: goto st893;
		case 894: goto st894;
		case 895: goto st895;
		case 896: goto st896;
		case 897: goto st897;
		case 898: goto st898;
		case 899: goto st899;
		case 900: goto st900;
		case 901: goto st901;
		case 902: goto st902;
		case 903: goto st903;
		case 904: goto st904;
		case 905: goto st905;
		case 906: goto st906;
		case 907: goto st907;
		case 908: goto st908;
		case 909: goto st909;
		case 910: goto st910;
		case 911: goto st911;
		case 912: goto st912;
		case 913: goto st913;
		case 914: goto st914;
		case 915: goto st915;
		case 916: goto st916;
		case 917: goto st917;
		case 918: goto st918;
		case 919: goto st919;
		case 920: goto st920;
		case 921: goto st921;
		case 922: goto st922;
		case 923: goto st923;
		case 924: goto st924;
		case 925: goto st925;
		case 926: goto st926;
		case 927: goto st927;
		case 928: goto st928;
		case 929: goto st929;
		case 930: goto st930;
		case 931: goto st931;
		case 932: goto st932;
		case 933: goto st933;
		case 934: goto st934;
		case 935: goto st935;
		case 936: goto st936;
		case 937: goto st937;
		case 938: goto st938;
		case 939: goto st939;
		case 940: goto st940;
		case 941: goto st941;
		case 942: goto st942;
		case 943: goto st943;
		case 944: goto st944;
		case 945: goto st945;
		case 946: goto st946;
		case 947: goto st947;
		case 948: goto st948;
		case 949: goto st949;
		case 950: goto st950;
		case 951: goto st951;
		case 952: goto st952;
		case 953: goto st953;
		case 954: goto st954;
		case 955: goto st955;
		case 956: goto st956;
		case 957: goto st957;
		case 958: goto st958;
		case 959: goto st959;
		case 960: goto st960;
		case 961: goto st961;
		case 962: goto st962;
		case 963: goto st963;
		case 964: goto st964;
		case 965: goto st965;
		case 966: goto st966;
		case 967: goto st967;
		case 968: goto st968;
		case 969: goto st969;
		case 970: goto st970;
		case 971: goto st971;
		case 972: goto st972;
		case 973: goto st973;
		case 974: goto st974;
		case 975: goto st975;
		case 976: goto st976;
		case 977: goto st977;
		case 978: goto st978;
		case 979: goto st979;
		case 980: goto st980;
		case 981: goto st981;
		case 982: goto st982;
		case 983: goto st983;
		case 984: goto st984;
		case 985: goto st985;
		case 986: goto st986;
		case 987: goto st987;
		case 988: goto st988;
		case 989: goto st989;
		case 990: goto st990;
		case 991: goto st991;
		case 992: goto st992;
		case 993: goto st993;
		case 994: goto st994;
		case 995: goto st995;
		case 996: goto st996;
		case 997: goto st997;
		case 998: goto st998;
		case 999: goto st999;
		case 1000: goto st1000;
		case 1001: goto st1001;
		case 1002: goto st1002;
		case 1003: goto st1003;
		case 1004: goto st1004;
		case 1005: goto st1005;
		case 1006: goto st1006;
		case 1007: goto st1007;
		case 1008: goto st1008;
		case 1009: goto st1009;
		case 1010: goto st1010;
		case 1011: goto st1011;
		case 1012: goto st1012;
		case 1200: goto st1200;
		case 1013: goto st1013;
		case 1014: goto st1014;
		case 1015: goto st1015;
		case 1016: goto st1016;
		case 1017: goto st1017;
		case 1018: goto st1018;
		case 1019: goto st1019;
		case 1020: goto st1020;
		case 1201: goto st1201;
		case 1021: goto st1021;
		case 1022: goto st1022;
		case 1023: goto st1023;
		case 1024: goto st1024;
		case 1025: goto st1025;
		case 1202: goto st1202;
		case 1026: goto st1026;
		case 1027: goto st1027;
		case 1028: goto st1028;
		case 1029: goto st1029;
		case 1030: goto st1030;
		case 1031: goto st1031;
		case 1032: goto st1032;
		case 1033: goto st1033;
		case 1034: goto st1034;
		case 1035: goto st1035;
		case 1036: goto st1036;
		case 1037: goto st1037;
		case 1038: goto st1038;
		case 1039: goto st1039;
		case 1040: goto st1040;
		case 1041: goto st1041;
		case 1042: goto st1042;
		case 1043: goto st1043;
		case 1203: goto st1203;
		case 1044: goto st1044;
		case 1045: goto st1045;
		case 1046: goto st1046;
		case 1047: goto st1047;
		case 1048: goto st1048;
		case 1049: goto st1049;
		case 1050: goto st1050;
		case 1051: goto st1051;
		case 1052: goto st1052;
		case 1053: goto st1053;
		case 1054: goto st1054;
		case 1055: goto st1055;
		case 1056: goto st1056;
		case 1057: goto st1057;
		case 1058: goto st1058;
		case 1059: goto st1059;
		case 1060: goto st1060;
		case 1061: goto st1061;
		case 1062: goto st1062;
		case 1204: goto st1204;
		case 1063: goto st1063;
		case 1064: goto st1064;
		case 1065: goto st1065;
		case 1066: goto st1066;
		case 1067: goto st1067;
		case 1068: goto st1068;
		case 1069: goto st1069;
		case 1070: goto st1070;
		case 1071: goto st1071;
		case 1072: goto st1072;
		case 1073: goto st1073;
		case 1074: goto st1074;
		case 1075: goto st1075;
		case 1076: goto st1076;
		case 1077: goto st1077;
		case 1205: goto st1205;
		case 1206: goto st1206;
		case 1207: goto st1207;
		case 1078: goto st1078;
		case 1079: goto st1079;
		case 1080: goto st1080;
		case 1081: goto st1081;
		case 1082: goto st1082;
		case 1083: goto st1083;
		case 1084: goto st1084;
		case 1208: goto st1208;
		case 1085: goto st1085;
		case 1086: goto st1086;
		case 1087: goto st1087;
		case 1088: goto st1088;
		case 1089: goto st1089;
		case 1090: goto st1090;
		case 1091: goto st1091;
		case 1092: goto st1092;
		case 1093: goto st1093;
		case 1094: goto st1094;
		case 1095: goto st1095;
		case 1096: goto st1096;
		case 1097: goto st1097;
		case 1209: goto st1209;
		case 1098: goto st1098;
		case 1099: goto st1099;
		case 1100: goto st1100;
		case 1101: goto st1101;
		case 1102: goto st1102;
		case 1103: goto st1103;
		case 1104: goto st1104;
		case 1210: goto st1210;
		case 1105: goto st1105;
		case 1106: goto st1106;
		case 1107: goto st1107;
		case 1108: goto st1108;
		case 1109: goto st1109;
		case 1110: goto st1110;
		case 1211: goto st1211;
		case 1111: goto st1111;
		case 1112: goto st1112;
		case 1113: goto st1113;
		case 1114: goto st1114;
		case 1115: goto st1115;
		case 1116: goto st1116;
		case 1212: goto st1212;
		case 1117: goto st1117;
		case 1118: goto st1118;
		case 1119: goto st1119;
		case 1120: goto st1120;
		case 1121: goto st1121;
		case 1122: goto st1122;
		case 1123: goto st1123;
		case 1124: goto st1124;
		case 1213: goto st1213;
		case 1125: goto st1125;
		case 1126: goto st1126;
	default: break;
	}

	if ( ++p == pe )
		goto _test_eof;
_resume:
	switch ( cs )
	{
tr20:
	{
		s->line_counter++;
	}
	goto st1127;
tr83:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1127; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1127; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1127;
tr89:
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1127; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1127; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1127;
tr92:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1127; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1127; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1127;
tr665:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1127;
tr759:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1127; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1127; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1127;
tr874:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1127; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1127; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	{
		s->buffer[s->buffer_length++] = 0;
	}
	goto st1127;
tr878:
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1127; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1127; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	{
		s->buffer[s->buffer_length++] = 0;
	}
	goto st1127;
tr3614:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->line_counter++;
	}
	goto st1127;
tr3674:
	{
		NOERR;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->line_counter++;
	}
	goto st1127;
st1127:
	if ( ++p == pe )
		goto _test_eof1127;
case 1127:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3609;
		case 32: goto tr3609;
		case 36: goto st152;
		case 40: goto tr3611;
		case 41: goto tr3612;
		case 42: goto tr3613;
		case 92: goto tr3613;
		case 95: goto tr3613;
		case 778: goto tr3614;
		case 827: goto tr3615;
		case 1034: goto tr3616;
		case 1083: goto tr3617;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr3613;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3613;
	} else
		goto tr3613;
	goto tr3608;
tr0:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	goto st0;
tr24:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	goto st0;
tr36:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	goto st0;
tr57:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr71:
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr79:
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	goto st0;
tr85:
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	goto st0;
tr114:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	goto st0;
tr139:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	goto st0;
tr145:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	goto st0;
tr582:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	goto st0;
tr584:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr666:
	{
		ERR(ZS_BAD_DIRECTIVE);
		p--; {goto st268;}
	}
	goto st0;
tr677:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	goto st0;
tr691:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	goto st0;
tr719:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr734:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	goto st0;
tr771:
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	{
		ERR(ZS_BAD_DIRECTIVE);
		p--; {goto st268;}
	}
	goto st0;
tr781:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	goto st0;
tr787:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	goto st0;
tr800:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	goto st0;
tr814:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr904:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	goto st0;
tr915:
	{
		WARN(ZS_BAD_DNAME_CHAR);
		p--; {goto st268;}
	}
	goto st0;
tr928:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_DNAME_CHAR);
		p--; {goto st268;}
	}
	goto st0;
tr936:
	{
		WARN(ZS_BAD_TEXT_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_TEXT);
		p--; {goto st268;}
	}
	goto st0;
tr943:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_TEXT_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_TEXT);
		p--; {goto st268;}
	}
	goto st0;
tr953:
	{
		WARN(ZS_BAD_TEXT);
		p--; {goto st268;}
	}
	goto st0;
tr960:
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st268;}
	}
	goto st0;
tr966:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st268;}
	}
	goto st0;
tr968:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st268;}
	}
	goto st0;
tr980:
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st268;}
	}
	goto st0;
tr989:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st268;}
	}
	goto st0;
tr1004:
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {goto st268;}
	}
	goto st0;
tr1016:
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {goto st268;}
	}
	goto st0;
tr1032:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {goto st268;}
	}
	goto st0;
tr1046:
	{
		ERR(ZS_BAD_INCLUDE_FILENAME);
		p--; {goto st268;}
	}
	goto st0;
tr1047:
	{
		ERR(ZS_BAD_INCLUDE_FILENAME);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	goto st0;
tr1054:
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	goto st0;
tr1064:
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {goto st268;}
	}
	goto st0;
tr1081:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {goto st268;}
	}
	goto st0;
tr1098:
	{
		WARN(ZS_BAD_BASE64_CHAR);
		p--; {goto st268;}
	}
	goto st0;
tr1114:
	{
		WARN(ZS_BAD_BITMAP);
		p--; {goto st268;}
	}
	goto st0;
tr1607:
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr1613:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr1621:
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr1683:
	{
		WARN(ZS_BAD_ALGORITHM);
		p--; {goto st268;}
	}
	goto st0;
tr1805:
	{
		WARN(ZS_BAD_CERT_TYPE);
		p--; {goto st268;}
	}
	goto st0;
tr1858:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr1881:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr1900:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr2066:
	{
		s->long_string = false;
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr2089:
	{
		WARN(ZS_BAD_LOC_DATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr2492:
	{
		WARN(ZS_BAD_APL);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr2507:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_APL);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr2552:
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr2623:
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr2638:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr2652:
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr2675:
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr2699:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr2706:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr2825:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr2879:
	{
		WARN(ZS_BAD_TIMESTAMP_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr3258:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr3359:
	{
		WARN(ZS_BAD_BASE32HEX_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr3526:
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_CHAR_COLON);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr3532:
	{
		WARN(ZS_BAD_CHAR_COLON);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr3538:
	{
		WARN(ZS_BAD_CHAR_DASH);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	goto st0;
tr3608:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	goto st0;
tr3630:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	goto st0;
tr3651:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	goto st0;
tr3668:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		ERR(ZS_BAD_DIRECTIVE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	goto st0;
tr3691:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	goto st0;
st0:
cs = 0;
	goto _out;
tr2:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1;
tr3:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1;
tr3609:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1;
tr3611:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1;
tr3612:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1;
tr3678:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1;
tr3679:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1;
tr3669:
	{
		NOERR;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1;
tr3671:
	{
		NOERR;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1;
tr3672:
	{
		NOERR;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1;
st1:
	if ( ++p == pe )
		goto _test_eof1;
case 1:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1;
		case 32: goto st1;
		case 40: goto tr2;
		case 41: goto tr3;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 85: goto tr19;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 117: goto tr19;
		case 778: goto tr20;
		case 827: goto st151;
		case 1034: goto tr22;
		case 1083: goto st264;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr0;
tr4:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st2;
tr28:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st2;
st2:
	if ( ++p == pe )
		goto _test_eof2;
case 2:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr25;
		case 32: goto tr25;
		case 40: goto tr26;
		case 41: goto tr27;
		case 68: goto tr29;
		case 72: goto tr30;
		case 77: goto tr31;
		case 83: goto st166;
		case 87: goto tr33;
		case 100: goto tr29;
		case 104: goto tr30;
		case 109: goto tr31;
		case 115: goto st166;
		case 119: goto tr33;
		case 1034: goto tr34;
		case 1083: goto tr35;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr28;
	goto tr24;
tr38:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st3;
tr39:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st3;
tr55:
	{
		s->line_counter++;
	}
	goto st3;
tr25:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st3;
tr26:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st3;
tr27:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st3;
tr34:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st3;
tr718:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st3;
tr711:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st3;
tr712:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st3;
tr713:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st3;
tr715:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st3;
st3:
	if ( ++p == pe )
		goto _test_eof3;
case 3:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st3;
		case 32: goto st3;
		case 40: goto tr38;
		case 41: goto tr39;
		case 65: goto tr40;
		case 67: goto tr41;
		case 68: goto tr42;
		case 69: goto tr43;
		case 72: goto tr44;
		case 73: goto tr45;
		case 75: goto tr46;
		case 76: goto tr47;
		case 77: goto tr48;
		case 78: goto tr49;
		case 80: goto tr50;
		case 82: goto tr51;
		case 83: goto tr52;
		case 84: goto tr53;
		case 85: goto tr54;
		case 97: goto tr40;
		case 99: goto tr41;
		case 100: goto tr42;
		case 101: goto tr43;
		case 104: goto tr44;
		case 105: goto tr45;
		case 107: goto tr46;
		case 108: goto tr47;
		case 109: goto tr48;
		case 110: goto tr49;
		case 112: goto tr50;
		case 114: goto tr51;
		case 115: goto tr52;
		case 116: goto tr53;
		case 117: goto tr54;
		case 1034: goto tr55;
		case 1083: goto st169;
	}
	goto tr36;
tr5:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st4;
tr40:
	{
		s->r_class = s->default_class;
	}
	goto st4;
tr621:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st4;
st4:
	if ( ++p == pe )
		goto _test_eof4;
case 4:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr58;
		case 32: goto tr58;
		case 40: goto tr59;
		case 41: goto tr60;
		case 65: goto st230;
		case 70: goto st233;
		case 80: goto st237;
		case 97: goto st230;
		case 102: goto st233;
		case 112: goto st237;
		case 2058: goto tr64;
		case 2107: goto tr65;
		case 2314: goto tr66;
		case 2363: goto tr66;
		case 2570: goto tr67;
		case 2619: goto tr68;
	}
	goto tr57;
tr72:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr73:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr75:
	{
		s->line_counter++;
	}
	goto st5;
tr58:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr59:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr60:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr64:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr96:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr156:
	{ s->r_type = KNOT_RRTYPE_CAA; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr157:
	{ s->r_type = KNOT_RRTYPE_CAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr158:
	{ s->r_type = KNOT_RRTYPE_CAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr159:
	{ s->r_type = KNOT_RRTYPE_CAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr177:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr178:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr179:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr180:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr185:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr186:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr187:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr188:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr195:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr196:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr197:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr198:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr206:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr207:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr208:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr209:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr220:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr221:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr222:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr223:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr232:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr233:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr234:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr235:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr243:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr244:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr245:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr246:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr251:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr252:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr253:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr254:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr264:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr265:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr266:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr267:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr273:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr274:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr275:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr276:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr285:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr286:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr287:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr288:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr326:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr327:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr328:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr329:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr337:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr338:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr339:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr340:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr345:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr346:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr347:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr348:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr358:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr359:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr360:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr361:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr367:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr368:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr369:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr370:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr376:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr377:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr378:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr379:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr384:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr385:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr386:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr387:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr397:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr398:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr399:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr400:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr405:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr406:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr407:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr408:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr419:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr420:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr421:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr422:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr428:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr429:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr430:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr431:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr436:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr437:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr438:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr440:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr446:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr447:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr448:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr450:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr455:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr456:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr457:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr459:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr468:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr469:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr470:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr471:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr478:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr479:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr480:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr481:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr489:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr490:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr491:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr492:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr500:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr501:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr502:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr503:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr508:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr509:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr510:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr511:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr521:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr522:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr523:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr524:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr530:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr531:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr532:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr533:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr539:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr540:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr541:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr542:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr550:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr551:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr552:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr553:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr563:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr564:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr565:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr566:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr572:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr573:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr574:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr575:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr585:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr586:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr587:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr589:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr596:
	{ s->r_type = KNOT_RRTYPE_URI; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr597:
	{ s->r_type = KNOT_RRTYPE_URI; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr598:
	{ s->r_type = KNOT_RRTYPE_URI; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr599:
	{ s->r_type = KNOT_RRTYPE_URI; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr841:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr842:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr843:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr844:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr852:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr853:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr854:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr855:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr861:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr862:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st5;
tr863:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st5;
tr864:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
st5:
	if ( ++p == pe )
		goto _test_eof5;
case 5:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st5;
		case 32: goto st5;
		case 40: goto tr72;
		case 41: goto tr73;
		case 92: goto st9;
		case 2058: goto tr75;
		case 2107: goto st10;
		case 2314: goto tr69;
		case 2363: goto tr69;
		case 2570: goto tr77;
		case 2619: goto tr78;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr69;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr69;
	} else
		goto tr69;
	goto tr71;
tr66:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr69:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr93:
	{ p--; }
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr94:
	{
		switch (s->r_type) {
		// Next types must not have empty rdata.
		case KNOT_RRTYPE_A:
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
		case KNOT_RRTYPE_SOA:
		case KNOT_RRTYPE_HINFO:
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
		case KNOT_RRTYPE_RP:
		case KNOT_RRTYPE_AAAA:
		case KNOT_RRTYPE_LOC:
		case KNOT_RRTYPE_SRV:
		case KNOT_RRTYPE_NAPTR:
		case KNOT_RRTYPE_CERT:
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_SSHFP:
		case KNOT_RRTYPE_IPSECKEY:
		case KNOT_RRTYPE_RRSIG:
		case KNOT_RRTYPE_NSEC:
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_NSEC3:
		case KNOT_RRTYPE_NSEC3PARAM:
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_CDS:
		case KNOT_RRTYPE_CDNSKEY:
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L32:
		case KNOT_RRTYPE_L64:
		case KNOT_RRTYPE_LP:
		case KNOT_RRTYPE_EUI48:
		case KNOT_RRTYPE_EUI64:
		case KNOT_RRTYPE_URI:
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st468;}
		// Next types can have empty rdata.
		case KNOT_RRTYPE_APL:
		default:
			{stack[top++] = 6; goto st477;}
		}
	}
	goto st6;
tr161:
	{ s->r_type = KNOT_RRTYPE_CAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr182:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr190:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr200:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr211:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr225:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr237:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr248:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr256:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr269:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr278:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr290:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr331:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr342:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr350:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr363:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr372:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr381:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr389:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr402:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr410:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr424:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr433:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr442:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr452:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr461:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr473:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr483:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr494:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr505:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr513:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr526:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr535:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr544:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr555:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr568:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr577:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr591:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr601:
	{ s->r_type = KNOT_RRTYPE_URI; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr846:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr857:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
tr866:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 6; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 6; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st6;
st6:
	if ( ++p == pe )
		goto _test_eof6;
case 6:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr80;
		case 32: goto tr80;
		case 40: goto tr81;
		case 41: goto tr82;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr83;
		case 1083: goto tr84;
	}
	goto tr79;
tr87:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st7;
tr88:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st7;
tr80:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st7;
tr81:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st7;
tr82:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st7;
st7:
	if ( ++p == pe )
		goto _test_eof7;
case 7:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr87;
		case 41: goto tr88;
		case 778: goto tr89;
		case 827: goto st8;
		case 1034: goto tr89;
		case 1083: goto st8;
	}
	goto tr85;
tr84:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st8;
tr91:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st8;
tr760:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st8;
st8:
	if ( ++p == pe )
		goto _test_eof8;
case 8:
	if ( (*p) == 10 )
		goto tr92;
	goto tr91;
st9:
	if ( ++p == pe )
		goto _test_eof9;
case 9:
	if ( (*p) == 35 )
		goto tr94;
	goto tr93;
tr65:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr95:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st10;
tr160:
	{ s->r_type = KNOT_RRTYPE_CAA; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr181:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr189:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr199:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr210:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr224:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr236:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr247:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr255:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr268:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr277:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr289:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr330:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr341:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr349:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr362:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr371:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr380:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr388:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr401:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr409:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr423:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr432:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr441:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr451:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr460:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr472:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr482:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr493:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr504:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr512:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr525:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr534:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr543:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr554:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr567:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr576:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr590:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr600:
	{ s->r_type = KNOT_RRTYPE_URI; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr845:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr856:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr865:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
st10:
	if ( ++p == pe )
		goto _test_eof10;
case 10:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr96;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr95;
	goto tr71;
tr67:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr77:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr162:
	{ s->r_type = KNOT_RRTYPE_CAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr183:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr191:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr201:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr212:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr226:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr238:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr249:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr257:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr270:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr279:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr291:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr332:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr343:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr351:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr364:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr373:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr382:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr390:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr403:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr411:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr425:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr434:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr443:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr453:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr462:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr474:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr484:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr495:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr506:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr514:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr527:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr536:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr545:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr556:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr569:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr578:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr592:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr602:
	{ s->r_type = KNOT_RRTYPE_URI; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr847:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr858:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
tr867:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 11; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 11; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st11;
st11:
	if ( ++p == pe )
		goto _test_eof11;
case 11:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr97;
		case 32: goto tr97;
		case 40: goto tr98;
		case 41: goto tr99;
		case 92: goto st9;
		case 1802: goto tr83;
		case 1851: goto tr84;
		case 2058: goto tr83;
		case 2107: goto tr100;
		case 2314: goto tr101;
		case 2363: goto tr102;
		case 2570: goto tr83;
		case 2619: goto tr103;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr69;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr69;
	} else
		goto tr69;
	goto tr79;
tr105:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st12;
tr106:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st12;
tr97:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st12;
tr98:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st12;
tr99:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st12;
st12:
	if ( ++p == pe )
		goto _test_eof12;
case 12:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st12;
		case 32: goto st12;
		case 40: goto tr105;
		case 41: goto tr106;
		case 92: goto st9;
		case 1802: goto tr89;
		case 1851: goto st8;
		case 2058: goto tr89;
		case 2107: goto st13;
		case 2314: goto tr108;
		case 2363: goto tr109;
		case 2570: goto tr89;
		case 2619: goto tr110;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr69;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr69;
	} else
		goto tr69;
	goto tr79;
tr100:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st13;
tr111:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st13;
tr167:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st13;
tr739:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st13;
st13:
	if ( ++p == pe )
		goto _test_eof13;
case 13:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr92;
		case 1034: goto tr112;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr111;
	} else if ( _widec >= 640 )
		goto tr91;
	goto tr79;
tr112:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1128; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1128; goto _out;}
		}
	}
	goto st1128;
tr747:
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1128; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1128; goto _out;}
		}
	}
	goto st1128;
tr738:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1128; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1128; goto _out;}
		}
	}
	goto st1128;
st1128:
	if ( ++p == pe )
		goto _test_eof1128;
case 1128:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr3618;
		case 32: goto tr3618;
		case 36: goto tr3619;
		case 40: goto tr3620;
		case 41: goto tr3621;
		case 42: goto tr3622;
		case 58: goto tr69;
		case 92: goto tr3623;
		case 95: goto tr3622;
		case 1802: goto tr3614;
		case 1851: goto tr3615;
		case 2058: goto tr3624;
		case 2107: goto tr3625;
		case 2314: goto tr3626;
		case 2363: goto tr3627;
		case 2570: goto tr3628;
		case 2619: goto tr3629;
	}
	if ( _widec < 60 ) {
		if ( _widec < 11 ) {
			if ( _widec <= 8 )
				goto tr69;
		} else if ( _widec > 44 ) {
			if ( 45 <= _widec && _widec <= 57 )
				goto tr3622;
		} else
			goto tr69;
	} else if ( _widec > 63 ) {
		if ( _widec < 91 ) {
			if ( 64 <= _widec && _widec <= 90 )
				goto tr3622;
		} else if ( _widec > 96 ) {
			if ( _widec > 122 ) {
				if ( 123 <= _widec )
					goto tr69;
			} else if ( _widec >= 97 )
				goto tr3622;
		} else
			goto tr69;
	} else
		goto tr69;
	goto tr781;
tr115:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st14;
tr116:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st14;
tr3618:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st14;
tr3620:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st14;
tr3621:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st14;
st14:
	if ( ++p == pe )
		goto _test_eof14;
case 14:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st14;
		case 32: goto st14;
		case 40: goto tr115;
		case 41: goto tr116;
		case 58: goto tr69;
		case 65: goto tr118;
		case 67: goto tr119;
		case 68: goto tr120;
		case 69: goto tr121;
		case 72: goto tr122;
		case 73: goto tr123;
		case 75: goto tr124;
		case 76: goto tr125;
		case 77: goto tr126;
		case 78: goto tr127;
		case 80: goto tr128;
		case 82: goto tr129;
		case 83: goto tr130;
		case 84: goto tr131;
		case 85: goto tr132;
		case 92: goto st9;
		case 97: goto tr118;
		case 99: goto tr119;
		case 100: goto tr120;
		case 101: goto tr121;
		case 104: goto tr122;
		case 105: goto tr123;
		case 107: goto tr124;
		case 108: goto tr125;
		case 109: goto tr126;
		case 110: goto tr127;
		case 112: goto tr128;
		case 114: goto tr129;
		case 115: goto tr130;
		case 116: goto tr131;
		case 117: goto tr132;
		case 1802: goto tr20;
		case 1851: goto st151;
		case 2058: goto tr133;
		case 2107: goto st239;
		case 2314: goto tr135;
		case 2363: goto tr136;
		case 2570: goto tr137;
		case 2619: goto tr138;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr69;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr69;
		} else if ( _widec >= 48 )
			goto tr117;
	} else
		goto tr69;
	goto tr114;
tr117:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 15; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 15; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 15; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 15; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 15; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 15; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 15; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 15; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 15; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 15; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 15; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 15; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 15; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 15; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 15; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 15; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 15; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 15; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 15; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 15; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 15; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 15; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 15; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 15; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 15; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 15; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 15; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 15; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 15; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st15;
st15:
	if ( ++p == pe )
		goto _test_eof15;
case 15:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr140;
		case 32: goto tr140;
		case 40: goto tr141;
		case 41: goto tr142;
		case 68: goto tr29;
		case 72: goto tr30;
		case 77: goto tr31;
		case 83: goto st166;
		case 87: goto tr33;
		case 100: goto tr29;
		case 104: goto tr30;
		case 109: goto tr31;
		case 115: goto st166;
		case 119: goto tr33;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr143;
		case 1083: goto tr144;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr28;
	goto tr139;
tr147:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st16;
tr148:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st16;
tr140:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st16;
tr141:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st16;
tr142:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st16;
st16:
	if ( ++p == pe )
		goto _test_eof16;
case 16:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st16;
		case 32: goto st16;
		case 40: goto tr147;
		case 41: goto tr148;
		case 65: goto tr40;
		case 67: goto tr41;
		case 68: goto tr42;
		case 69: goto tr43;
		case 72: goto tr44;
		case 73: goto tr45;
		case 75: goto tr46;
		case 76: goto tr47;
		case 77: goto tr48;
		case 78: goto tr49;
		case 80: goto tr50;
		case 82: goto tr51;
		case 83: goto tr52;
		case 84: goto tr53;
		case 85: goto tr54;
		case 97: goto tr40;
		case 99: goto tr41;
		case 100: goto tr42;
		case 101: goto tr43;
		case 104: goto tr44;
		case 105: goto tr45;
		case 107: goto tr46;
		case 108: goto tr47;
		case 109: goto tr48;
		case 110: goto tr49;
		case 112: goto tr50;
		case 114: goto tr51;
		case 115: goto tr52;
		case 116: goto tr53;
		case 117: goto tr54;
		case 778: goto tr89;
		case 827: goto st8;
		case 1034: goto tr149;
		case 1083: goto st267;
	}
	goto tr145;
tr6:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st17;
tr41:
	{
		s->r_class = s->default_class;
	}
	goto st17;
tr622:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st17;
st17:
	if ( ++p == pe )
		goto _test_eof17;
case 17:
	switch( (*p) ) {
		case 65: goto st18;
		case 68: goto st22;
		case 69: goto st29;
		case 78: goto st32;
		case 97: goto st18;
		case 100: goto st22;
		case 101: goto st29;
		case 110: goto st32;
	}
	goto tr36;
st18:
	if ( ++p == pe )
		goto _test_eof18;
case 18:
	switch( (*p) ) {
		case 65: goto st19;
		case 97: goto st19;
	}
	goto tr36;
st19:
	if ( ++p == pe )
		goto _test_eof19;
case 19:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr156;
		case 32: goto tr156;
		case 40: goto tr157;
		case 41: goto tr158;
		case 2058: goto tr159;
		case 2107: goto tr160;
		case 2314: goto tr161;
		case 2363: goto tr161;
		case 2570: goto tr162;
		case 2619: goto tr163;
	}
	goto tr57;
tr68:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr78:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr163:
	{ s->r_type = KNOT_RRTYPE_CAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr184:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr192:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr202:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr213:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr227:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr239:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr250:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr258:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr271:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr280:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr292:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr333:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr344:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr352:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr365:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr374:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr383:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr391:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr404:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr412:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr426:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr435:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr444:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr454:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr463:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr475:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr485:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr496:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr507:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr515:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr528:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr537:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr546:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr557:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr570:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr579:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr593:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr603:
	{ s->r_type = KNOT_RRTYPE_URI; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr848:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr859:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
tr868:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 20; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 20; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 20; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 20; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 20; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 20; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 20; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 20; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 20; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 20; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 20; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 20; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 20; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 20; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 20; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 20; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 20; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 20; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 20; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 20; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 20; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 20; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 20; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 20; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 20; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 20; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 20; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 20; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 20; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st20;
st20:
	if ( ++p == pe )
		goto _test_eof20;
case 20:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr80;
		case 778: goto tr83;
		case 800: goto tr80;
		case 808: goto tr81;
		case 809: goto tr82;
		case 827: goto tr84;
		case 1033: goto tr164;
		case 1034: goto tr83;
		case 1056: goto tr164;
		case 1064: goto tr165;
		case 1065: goto tr166;
		case 1083: goto tr167;
	}
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr95;
	goto tr79;
tr168:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st21;
tr164:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st21;
tr165:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st21;
tr166:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st21;
tr169:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st21;
tr170:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st21;
st21:
	if ( ++p == pe )
		goto _test_eof21;
case 21:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st7;
		case 778: goto tr89;
		case 800: goto st7;
		case 808: goto tr87;
		case 809: goto tr88;
		case 827: goto st8;
		case 1033: goto tr168;
		case 1034: goto tr89;
		case 1056: goto tr168;
		case 1064: goto tr169;
		case 1065: goto tr170;
		case 1083: goto tr111;
	}
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr95;
	goto tr79;
st22:
	if ( ++p == pe )
		goto _test_eof22;
case 22:
	switch( (*p) ) {
		case 78: goto st23;
		case 83: goto st28;
		case 110: goto st23;
		case 115: goto st28;
	}
	goto tr36;
st23:
	if ( ++p == pe )
		goto _test_eof23;
case 23:
	switch( (*p) ) {
		case 83: goto st24;
		case 115: goto st24;
	}
	goto tr36;
st24:
	if ( ++p == pe )
		goto _test_eof24;
case 24:
	switch( (*p) ) {
		case 75: goto st25;
		case 107: goto st25;
	}
	goto tr36;
st25:
	if ( ++p == pe )
		goto _test_eof25;
case 25:
	switch( (*p) ) {
		case 69: goto st26;
		case 101: goto st26;
	}
	goto tr36;
st26:
	if ( ++p == pe )
		goto _test_eof26;
case 26:
	switch( (*p) ) {
		case 89: goto st27;
		case 121: goto st27;
	}
	goto tr36;
st27:
	if ( ++p == pe )
		goto _test_eof27;
case 27:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr177;
		case 32: goto tr177;
		case 40: goto tr178;
		case 41: goto tr179;
		case 2058: goto tr180;
		case 2107: goto tr181;
		case 2314: goto tr182;
		case 2363: goto tr182;
		case 2570: goto tr183;
		case 2619: goto tr184;
	}
	goto tr57;
st28:
	if ( ++p == pe )
		goto _test_eof28;
case 28:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr185;
		case 32: goto tr185;
		case 40: goto tr186;
		case 41: goto tr187;
		case 2058: goto tr188;
		case 2107: goto tr189;
		case 2314: goto tr190;
		case 2363: goto tr190;
		case 2570: goto tr191;
		case 2619: goto tr192;
	}
	goto tr57;
st29:
	if ( ++p == pe )
		goto _test_eof29;
case 29:
	switch( (*p) ) {
		case 82: goto st30;
		case 114: goto st30;
	}
	goto tr36;
st30:
	if ( ++p == pe )
		goto _test_eof30;
case 30:
	switch( (*p) ) {
		case 84: goto st31;
		case 116: goto st31;
	}
	goto tr36;
st31:
	if ( ++p == pe )
		goto _test_eof31;
case 31:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr195;
		case 32: goto tr195;
		case 40: goto tr196;
		case 41: goto tr197;
		case 2058: goto tr198;
		case 2107: goto tr199;
		case 2314: goto tr200;
		case 2363: goto tr200;
		case 2570: goto tr201;
		case 2619: goto tr202;
	}
	goto tr57;
st32:
	if ( ++p == pe )
		goto _test_eof32;
case 32:
	switch( (*p) ) {
		case 65: goto st33;
		case 97: goto st33;
	}
	goto tr36;
st33:
	if ( ++p == pe )
		goto _test_eof33;
case 33:
	switch( (*p) ) {
		case 77: goto st34;
		case 109: goto st34;
	}
	goto tr36;
st34:
	if ( ++p == pe )
		goto _test_eof34;
case 34:
	switch( (*p) ) {
		case 69: goto st35;
		case 101: goto st35;
	}
	goto tr36;
st35:
	if ( ++p == pe )
		goto _test_eof35;
case 35:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr206;
		case 32: goto tr206;
		case 40: goto tr207;
		case 41: goto tr208;
		case 2058: goto tr209;
		case 2107: goto tr210;
		case 2314: goto tr211;
		case 2363: goto tr211;
		case 2570: goto tr212;
		case 2619: goto tr213;
	}
	goto tr57;
tr7:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st36;
tr42:
	{
		s->r_class = s->default_class;
	}
	goto st36;
tr623:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st36;
st36:
	if ( ++p == pe )
		goto _test_eof36;
case 36:
	switch( (*p) ) {
		case 72: goto st37;
		case 78: goto st41;
		case 83: goto st49;
		case 104: goto st37;
		case 110: goto st41;
		case 115: goto st49;
	}
	goto tr36;
st37:
	if ( ++p == pe )
		goto _test_eof37;
case 37:
	switch( (*p) ) {
		case 67: goto st38;
		case 99: goto st38;
	}
	goto tr36;
st38:
	if ( ++p == pe )
		goto _test_eof38;
case 38:
	switch( (*p) ) {
		case 73: goto st39;
		case 105: goto st39;
	}
	goto tr36;
st39:
	if ( ++p == pe )
		goto _test_eof39;
case 39:
	switch( (*p) ) {
		case 68: goto st40;
		case 100: goto st40;
	}
	goto tr36;
st40:
	if ( ++p == pe )
		goto _test_eof40;
case 40:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr220;
		case 32: goto tr220;
		case 40: goto tr221;
		case 41: goto tr222;
		case 2058: goto tr223;
		case 2107: goto tr224;
		case 2314: goto tr225;
		case 2363: goto tr225;
		case 2570: goto tr226;
		case 2619: goto tr227;
	}
	goto tr57;
st41:
	if ( ++p == pe )
		goto _test_eof41;
case 41:
	switch( (*p) ) {
		case 65: goto st42;
		case 83: goto st45;
		case 97: goto st42;
		case 115: goto st45;
	}
	goto tr36;
st42:
	if ( ++p == pe )
		goto _test_eof42;
case 42:
	switch( (*p) ) {
		case 77: goto st43;
		case 109: goto st43;
	}
	goto tr36;
st43:
	if ( ++p == pe )
		goto _test_eof43;
case 43:
	switch( (*p) ) {
		case 69: goto st44;
		case 101: goto st44;
	}
	goto tr36;
st44:
	if ( ++p == pe )
		goto _test_eof44;
case 44:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr232;
		case 32: goto tr232;
		case 40: goto tr233;
		case 41: goto tr234;
		case 2058: goto tr235;
		case 2107: goto tr236;
		case 2314: goto tr237;
		case 2363: goto tr237;
		case 2570: goto tr238;
		case 2619: goto tr239;
	}
	goto tr57;
st45:
	if ( ++p == pe )
		goto _test_eof45;
case 45:
	switch( (*p) ) {
		case 75: goto st46;
		case 107: goto st46;
	}
	goto tr36;
st46:
	if ( ++p == pe )
		goto _test_eof46;
case 46:
	switch( (*p) ) {
		case 69: goto st47;
		case 101: goto st47;
	}
	goto tr36;
st47:
	if ( ++p == pe )
		goto _test_eof47;
case 47:
	switch( (*p) ) {
		case 89: goto st48;
		case 121: goto st48;
	}
	goto tr36;
st48:
	if ( ++p == pe )
		goto _test_eof48;
case 48:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr243;
		case 32: goto tr243;
		case 40: goto tr244;
		case 41: goto tr245;
		case 2058: goto tr246;
		case 2107: goto tr247;
		case 2314: goto tr248;
		case 2363: goto tr248;
		case 2570: goto tr249;
		case 2619: goto tr250;
	}
	goto tr57;
st49:
	if ( ++p == pe )
		goto _test_eof49;
case 49:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr251;
		case 32: goto tr251;
		case 40: goto tr252;
		case 41: goto tr253;
		case 2058: goto tr254;
		case 2107: goto tr255;
		case 2314: goto tr256;
		case 2363: goto tr256;
		case 2570: goto tr257;
		case 2619: goto tr258;
	}
	goto tr57;
tr8:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st50;
tr43:
	{
		s->r_class = s->default_class;
	}
	goto st50;
tr624:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st50;
st50:
	if ( ++p == pe )
		goto _test_eof50;
case 50:
	switch( (*p) ) {
		case 85: goto st51;
		case 117: goto st51;
	}
	goto tr36;
st51:
	if ( ++p == pe )
		goto _test_eof51;
case 51:
	switch( (*p) ) {
		case 73: goto st52;
		case 105: goto st52;
	}
	goto tr36;
st52:
	if ( ++p == pe )
		goto _test_eof52;
case 52:
	switch( (*p) ) {
		case 52: goto st53;
		case 54: goto st55;
	}
	goto tr36;
st53:
	if ( ++p == pe )
		goto _test_eof53;
case 53:
	if ( (*p) == 56 )
		goto st54;
	goto tr36;
st54:
	if ( ++p == pe )
		goto _test_eof54;
case 54:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr264;
		case 32: goto tr264;
		case 40: goto tr265;
		case 41: goto tr266;
		case 2058: goto tr267;
		case 2107: goto tr268;
		case 2314: goto tr269;
		case 2363: goto tr269;
		case 2570: goto tr270;
		case 2619: goto tr271;
	}
	goto tr57;
st55:
	if ( ++p == pe )
		goto _test_eof55;
case 55:
	if ( (*p) == 52 )
		goto st56;
	goto tr36;
st56:
	if ( ++p == pe )
		goto _test_eof56;
case 56:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr273;
		case 32: goto tr273;
		case 40: goto tr274;
		case 41: goto tr275;
		case 2058: goto tr276;
		case 2107: goto tr277;
		case 2314: goto tr278;
		case 2363: goto tr278;
		case 2570: goto tr279;
		case 2619: goto tr280;
	}
	goto tr57;
tr9:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st57;
tr44:
	{
		s->r_class = s->default_class;
	}
	goto st57;
tr625:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st57;
st57:
	if ( ++p == pe )
		goto _test_eof57;
case 57:
	switch( (*p) ) {
		case 73: goto st58;
		case 105: goto st58;
	}
	goto tr36;
st58:
	if ( ++p == pe )
		goto _test_eof58;
case 58:
	switch( (*p) ) {
		case 78: goto st59;
		case 110: goto st59;
	}
	goto tr36;
st59:
	if ( ++p == pe )
		goto _test_eof59;
case 59:
	switch( (*p) ) {
		case 70: goto st60;
		case 102: goto st60;
	}
	goto tr36;
st60:
	if ( ++p == pe )
		goto _test_eof60;
case 60:
	switch( (*p) ) {
		case 79: goto st61;
		case 111: goto st61;
	}
	goto tr36;
st61:
	if ( ++p == pe )
		goto _test_eof61;
case 61:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr285;
		case 32: goto tr285;
		case 40: goto tr286;
		case 41: goto tr287;
		case 2058: goto tr288;
		case 2107: goto tr289;
		case 2314: goto tr290;
		case 2363: goto tr290;
		case 2570: goto tr291;
		case 2619: goto tr292;
	}
	goto tr57;
tr45:
	{
		s->r_class = s->default_class;
	}
	goto st62;
st62:
	if ( ++p == pe )
		goto _test_eof62;
case 62:
	switch( (*p) ) {
		case 78: goto st63;
		case 80: goto st66;
		case 110: goto st63;
		case 112: goto st66;
	}
	goto tr36;
st63:
	if ( ++p == pe )
		goto _test_eof63;
case 63:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr295;
		case 32: goto tr295;
		case 40: goto tr296;
		case 41: goto tr297;
		case 1034: goto tr298;
		case 1083: goto tr299;
	}
	goto tr36;
tr301:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st64;
tr302:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st64;
tr318:
	{
		s->line_counter++;
	}
	goto st64;
tr638:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st64;
tr639:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st64;
tr640:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st64;
tr647:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st64;
tr605:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st64;
tr295:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	goto st64;
tr296:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st64;
tr297:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st64;
tr298:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		s->line_counter++;
	}
	goto st64;
tr656:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st64;
tr657:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st64;
tr658:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st64;
tr660:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st64;
st64:
	if ( ++p == pe )
		goto _test_eof64;
case 64:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st64;
		case 32: goto st64;
		case 40: goto tr301;
		case 41: goto tr302;
		case 65: goto st4;
		case 67: goto st17;
		case 68: goto st36;
		case 69: goto st50;
		case 72: goto st57;
		case 73: goto st65;
		case 75: goto st73;
		case 76: goto st77;
		case 77: goto st85;
		case 78: goto st91;
		case 80: goto st107;
		case 82: goto st110;
		case 83: goto st117;
		case 84: goto st128;
		case 85: goto st138;
		case 97: goto st4;
		case 99: goto st17;
		case 100: goto st36;
		case 101: goto st50;
		case 104: goto st57;
		case 105: goto st65;
		case 107: goto st73;
		case 108: goto st77;
		case 109: goto st85;
		case 110: goto st91;
		case 112: goto st107;
		case 114: goto st110;
		case 115: goto st117;
		case 116: goto st128;
		case 117: goto st138;
		case 1034: goto tr318;
		case 1083: goto st141;
	}
	goto tr36;
tr626:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st65;
st65:
	if ( ++p == pe )
		goto _test_eof65;
case 65:
	switch( (*p) ) {
		case 80: goto st66;
		case 112: goto st66;
	}
	goto tr36;
st66:
	if ( ++p == pe )
		goto _test_eof66;
case 66:
	switch( (*p) ) {
		case 83: goto st67;
		case 115: goto st67;
	}
	goto tr36;
st67:
	if ( ++p == pe )
		goto _test_eof67;
case 67:
	switch( (*p) ) {
		case 69: goto st68;
		case 101: goto st68;
	}
	goto tr36;
st68:
	if ( ++p == pe )
		goto _test_eof68;
case 68:
	switch( (*p) ) {
		case 67: goto st69;
		case 99: goto st69;
	}
	goto tr36;
st69:
	if ( ++p == pe )
		goto _test_eof69;
case 69:
	switch( (*p) ) {
		case 75: goto st70;
		case 107: goto st70;
	}
	goto tr36;
st70:
	if ( ++p == pe )
		goto _test_eof70;
case 70:
	switch( (*p) ) {
		case 69: goto st71;
		case 101: goto st71;
	}
	goto tr36;
st71:
	if ( ++p == pe )
		goto _test_eof71;
case 71:
	switch( (*p) ) {
		case 89: goto st72;
		case 121: goto st72;
	}
	goto tr36;
st72:
	if ( ++p == pe )
		goto _test_eof72;
case 72:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr326;
		case 32: goto tr326;
		case 40: goto tr327;
		case 41: goto tr328;
		case 2058: goto tr329;
		case 2107: goto tr330;
		case 2314: goto tr331;
		case 2363: goto tr331;
		case 2570: goto tr332;
		case 2619: goto tr333;
	}
	goto tr57;
tr11:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st73;
tr46:
	{
		s->r_class = s->default_class;
	}
	goto st73;
tr627:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st73;
st73:
	if ( ++p == pe )
		goto _test_eof73;
case 73:
	switch( (*p) ) {
		case 69: goto st74;
		case 88: goto st76;
		case 101: goto st74;
		case 120: goto st76;
	}
	goto tr36;
st74:
	if ( ++p == pe )
		goto _test_eof74;
case 74:
	switch( (*p) ) {
		case 89: goto st75;
		case 121: goto st75;
	}
	goto tr36;
st75:
	if ( ++p == pe )
		goto _test_eof75;
case 75:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr337;
		case 32: goto tr337;
		case 40: goto tr338;
		case 41: goto tr339;
		case 2058: goto tr340;
		case 2107: goto tr341;
		case 2314: goto tr342;
		case 2363: goto tr342;
		case 2570: goto tr343;
		case 2619: goto tr344;
	}
	goto tr57;
st76:
	if ( ++p == pe )
		goto _test_eof76;
case 76:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr345;
		case 32: goto tr345;
		case 40: goto tr346;
		case 41: goto tr347;
		case 2058: goto tr348;
		case 2107: goto tr349;
		case 2314: goto tr350;
		case 2363: goto tr350;
		case 2570: goto tr351;
		case 2619: goto tr352;
	}
	goto tr57;
tr12:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st77;
tr47:
	{
		s->r_class = s->default_class;
	}
	goto st77;
tr628:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st77;
st77:
	if ( ++p == pe )
		goto _test_eof77;
case 77:
	switch( (*p) ) {
		case 51: goto st78;
		case 54: goto st80;
		case 79: goto st82;
		case 80: goto st84;
		case 111: goto st82;
		case 112: goto st84;
	}
	goto tr36;
st78:
	if ( ++p == pe )
		goto _test_eof78;
case 78:
	if ( (*p) == 50 )
		goto st79;
	goto tr36;
st79:
	if ( ++p == pe )
		goto _test_eof79;
case 79:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr358;
		case 32: goto tr358;
		case 40: goto tr359;
		case 41: goto tr360;
		case 2058: goto tr361;
		case 2107: goto tr362;
		case 2314: goto tr363;
		case 2363: goto tr363;
		case 2570: goto tr364;
		case 2619: goto tr365;
	}
	goto tr57;
st80:
	if ( ++p == pe )
		goto _test_eof80;
case 80:
	if ( (*p) == 52 )
		goto st81;
	goto tr36;
st81:
	if ( ++p == pe )
		goto _test_eof81;
case 81:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr367;
		case 32: goto tr367;
		case 40: goto tr368;
		case 41: goto tr369;
		case 2058: goto tr370;
		case 2107: goto tr371;
		case 2314: goto tr372;
		case 2363: goto tr372;
		case 2570: goto tr373;
		case 2619: goto tr374;
	}
	goto tr57;
st82:
	if ( ++p == pe )
		goto _test_eof82;
case 82:
	switch( (*p) ) {
		case 67: goto st83;
		case 99: goto st83;
	}
	goto tr36;
st83:
	if ( ++p == pe )
		goto _test_eof83;
case 83:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr376;
		case 32: goto tr376;
		case 40: goto tr377;
		case 41: goto tr378;
		case 2058: goto tr379;
		case 2107: goto tr380;
		case 2314: goto tr381;
		case 2363: goto tr381;
		case 2570: goto tr382;
		case 2619: goto tr383;
	}
	goto tr57;
st84:
	if ( ++p == pe )
		goto _test_eof84;
case 84:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr384;
		case 32: goto tr384;
		case 40: goto tr385;
		case 41: goto tr386;
		case 2058: goto tr387;
		case 2107: goto tr388;
		case 2314: goto tr389;
		case 2363: goto tr389;
		case 2570: goto tr390;
		case 2619: goto tr391;
	}
	goto tr57;
tr13:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st85;
tr48:
	{
		s->r_class = s->default_class;
	}
	goto st85;
tr629:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st85;
st85:
	if ( ++p == pe )
		goto _test_eof85;
case 85:
	switch( (*p) ) {
		case 73: goto st86;
		case 88: goto st90;
		case 105: goto st86;
		case 120: goto st90;
	}
	goto tr36;
st86:
	if ( ++p == pe )
		goto _test_eof86;
case 86:
	switch( (*p) ) {
		case 78: goto st87;
		case 110: goto st87;
	}
	goto tr36;
st87:
	if ( ++p == pe )
		goto _test_eof87;
case 87:
	switch( (*p) ) {
		case 70: goto st88;
		case 102: goto st88;
	}
	goto tr36;
st88:
	if ( ++p == pe )
		goto _test_eof88;
case 88:
	switch( (*p) ) {
		case 79: goto st89;
		case 111: goto st89;
	}
	goto tr36;
st89:
	if ( ++p == pe )
		goto _test_eof89;
case 89:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr397;
		case 32: goto tr397;
		case 40: goto tr398;
		case 41: goto tr399;
		case 2058: goto tr400;
		case 2107: goto tr401;
		case 2314: goto tr402;
		case 2363: goto tr402;
		case 2570: goto tr403;
		case 2619: goto tr404;
	}
	goto tr57;
st90:
	if ( ++p == pe )
		goto _test_eof90;
case 90:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr405;
		case 32: goto tr405;
		case 40: goto tr406;
		case 41: goto tr407;
		case 2058: goto tr408;
		case 2107: goto tr409;
		case 2314: goto tr410;
		case 2363: goto tr410;
		case 2570: goto tr411;
		case 2619: goto tr412;
	}
	goto tr57;
tr14:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st91;
tr49:
	{
		s->r_class = s->default_class;
	}
	goto st91;
tr630:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st91;
st91:
	if ( ++p == pe )
		goto _test_eof91;
case 91:
	switch( (*p) ) {
		case 65: goto st92;
		case 73: goto st96;
		case 83: goto st98;
		case 97: goto st92;
		case 105: goto st96;
		case 115: goto st98;
	}
	goto tr36;
st92:
	if ( ++p == pe )
		goto _test_eof92;
case 92:
	switch( (*p) ) {
		case 80: goto st93;
		case 112: goto st93;
	}
	goto tr36;
st93:
	if ( ++p == pe )
		goto _test_eof93;
case 93:
	switch( (*p) ) {
		case 84: goto st94;
		case 116: goto st94;
	}
	goto tr36;
st94:
	if ( ++p == pe )
		goto _test_eof94;
case 94:
	switch( (*p) ) {
		case 82: goto st95;
		case 114: goto st95;
	}
	goto tr36;
st95:
	if ( ++p == pe )
		goto _test_eof95;
case 95:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr419;
		case 32: goto tr419;
		case 40: goto tr420;
		case 41: goto tr421;
		case 2058: goto tr422;
		case 2107: goto tr423;
		case 2314: goto tr424;
		case 2363: goto tr424;
		case 2570: goto tr425;
		case 2619: goto tr426;
	}
	goto tr57;
st96:
	if ( ++p == pe )
		goto _test_eof96;
case 96:
	switch( (*p) ) {
		case 68: goto st97;
		case 100: goto st97;
	}
	goto tr36;
st97:
	if ( ++p == pe )
		goto _test_eof97;
case 97:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr428;
		case 32: goto tr428;
		case 40: goto tr429;
		case 41: goto tr430;
		case 2058: goto tr431;
		case 2107: goto tr432;
		case 2314: goto tr433;
		case 2363: goto tr433;
		case 2570: goto tr434;
		case 2619: goto tr435;
	}
	goto tr57;
st98:
	if ( ++p == pe )
		goto _test_eof98;
case 98:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr436;
		case 32: goto tr436;
		case 40: goto tr437;
		case 41: goto tr438;
		case 69: goto st99;
		case 101: goto st99;
		case 2058: goto tr440;
		case 2107: goto tr441;
		case 2314: goto tr442;
		case 2363: goto tr442;
		case 2570: goto tr443;
		case 2619: goto tr444;
	}
	goto tr57;
st99:
	if ( ++p == pe )
		goto _test_eof99;
case 99:
	switch( (*p) ) {
		case 67: goto st100;
		case 99: goto st100;
	}
	goto tr36;
st100:
	if ( ++p == pe )
		goto _test_eof100;
case 100:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr446;
		case 32: goto tr446;
		case 40: goto tr447;
		case 41: goto tr448;
		case 51: goto st101;
		case 2058: goto tr450;
		case 2107: goto tr451;
		case 2314: goto tr452;
		case 2363: goto tr452;
		case 2570: goto tr453;
		case 2619: goto tr454;
	}
	goto tr57;
st101:
	if ( ++p == pe )
		goto _test_eof101;
case 101:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr455;
		case 32: goto tr455;
		case 40: goto tr456;
		case 41: goto tr457;
		case 80: goto st102;
		case 112: goto st102;
		case 2058: goto tr459;
		case 2107: goto tr460;
		case 2314: goto tr461;
		case 2363: goto tr461;
		case 2570: goto tr462;
		case 2619: goto tr463;
	}
	goto tr57;
st102:
	if ( ++p == pe )
		goto _test_eof102;
case 102:
	switch( (*p) ) {
		case 65: goto st103;
		case 97: goto st103;
	}
	goto tr36;
st103:
	if ( ++p == pe )
		goto _test_eof103;
case 103:
	switch( (*p) ) {
		case 82: goto st104;
		case 114: goto st104;
	}
	goto tr36;
st104:
	if ( ++p == pe )
		goto _test_eof104;
case 104:
	switch( (*p) ) {
		case 65: goto st105;
		case 97: goto st105;
	}
	goto tr36;
st105:
	if ( ++p == pe )
		goto _test_eof105;
case 105:
	switch( (*p) ) {
		case 77: goto st106;
		case 109: goto st106;
	}
	goto tr36;
st106:
	if ( ++p == pe )
		goto _test_eof106;
case 106:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr468;
		case 32: goto tr468;
		case 40: goto tr469;
		case 41: goto tr470;
		case 2058: goto tr471;
		case 2107: goto tr472;
		case 2314: goto tr473;
		case 2363: goto tr473;
		case 2570: goto tr474;
		case 2619: goto tr475;
	}
	goto tr57;
tr15:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st107;
tr50:
	{
		s->r_class = s->default_class;
	}
	goto st107;
tr631:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st107;
st107:
	if ( ++p == pe )
		goto _test_eof107;
case 107:
	switch( (*p) ) {
		case 84: goto st108;
		case 116: goto st108;
	}
	goto tr36;
st108:
	if ( ++p == pe )
		goto _test_eof108;
case 108:
	switch( (*p) ) {
		case 82: goto st109;
		case 114: goto st109;
	}
	goto tr36;
st109:
	if ( ++p == pe )
		goto _test_eof109;
case 109:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr478;
		case 32: goto tr478;
		case 40: goto tr479;
		case 41: goto tr480;
		case 2058: goto tr481;
		case 2107: goto tr482;
		case 2314: goto tr483;
		case 2363: goto tr483;
		case 2570: goto tr484;
		case 2619: goto tr485;
	}
	goto tr57;
tr16:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st110;
tr51:
	{
		s->r_class = s->default_class;
	}
	goto st110;
tr632:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st110;
st110:
	if ( ++p == pe )
		goto _test_eof110;
case 110:
	switch( (*p) ) {
		case 80: goto st111;
		case 82: goto st112;
		case 84: goto st116;
		case 112: goto st111;
		case 114: goto st112;
		case 116: goto st116;
	}
	goto tr36;
st111:
	if ( ++p == pe )
		goto _test_eof111;
case 111:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr489;
		case 32: goto tr489;
		case 40: goto tr490;
		case 41: goto tr491;
		case 2058: goto tr492;
		case 2107: goto tr493;
		case 2314: goto tr494;
		case 2363: goto tr494;
		case 2570: goto tr495;
		case 2619: goto tr496;
	}
	goto tr57;
st112:
	if ( ++p == pe )
		goto _test_eof112;
case 112:
	switch( (*p) ) {
		case 83: goto st113;
		case 115: goto st113;
	}
	goto tr36;
st113:
	if ( ++p == pe )
		goto _test_eof113;
case 113:
	switch( (*p) ) {
		case 73: goto st114;
		case 105: goto st114;
	}
	goto tr36;
st114:
	if ( ++p == pe )
		goto _test_eof114;
case 114:
	switch( (*p) ) {
		case 71: goto st115;
		case 103: goto st115;
	}
	goto tr36;
st115:
	if ( ++p == pe )
		goto _test_eof115;
case 115:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr500;
		case 32: goto tr500;
		case 40: goto tr501;
		case 41: goto tr502;
		case 2058: goto tr503;
		case 2107: goto tr504;
		case 2314: goto tr505;
		case 2363: goto tr505;
		case 2570: goto tr506;
		case 2619: goto tr507;
	}
	goto tr57;
st116:
	if ( ++p == pe )
		goto _test_eof116;
case 116:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr508;
		case 32: goto tr508;
		case 40: goto tr509;
		case 41: goto tr510;
		case 2058: goto tr511;
		case 2107: goto tr512;
		case 2314: goto tr513;
		case 2363: goto tr513;
		case 2570: goto tr514;
		case 2619: goto tr515;
	}
	goto tr57;
tr17:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st117;
tr52:
	{
		s->r_class = s->default_class;
	}
	goto st117;
tr633:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st117;
st117:
	if ( ++p == pe )
		goto _test_eof117;
case 117:
	switch( (*p) ) {
		case 79: goto st118;
		case 80: goto st120;
		case 82: goto st122;
		case 83: goto st124;
		case 111: goto st118;
		case 112: goto st120;
		case 114: goto st122;
		case 115: goto st124;
	}
	goto tr36;
st118:
	if ( ++p == pe )
		goto _test_eof118;
case 118:
	switch( (*p) ) {
		case 65: goto st119;
		case 97: goto st119;
	}
	goto tr36;
st119:
	if ( ++p == pe )
		goto _test_eof119;
case 119:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr521;
		case 32: goto tr521;
		case 40: goto tr522;
		case 41: goto tr523;
		case 2058: goto tr524;
		case 2107: goto tr525;
		case 2314: goto tr526;
		case 2363: goto tr526;
		case 2570: goto tr527;
		case 2619: goto tr528;
	}
	goto tr57;
st120:
	if ( ++p == pe )
		goto _test_eof120;
case 120:
	switch( (*p) ) {
		case 70: goto st121;
		case 102: goto st121;
	}
	goto tr36;
st121:
	if ( ++p == pe )
		goto _test_eof121;
case 121:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr530;
		case 32: goto tr530;
		case 40: goto tr531;
		case 41: goto tr532;
		case 2058: goto tr533;
		case 2107: goto tr534;
		case 2314: goto tr535;
		case 2363: goto tr535;
		case 2570: goto tr536;
		case 2619: goto tr537;
	}
	goto tr57;
st122:
	if ( ++p == pe )
		goto _test_eof122;
case 122:
	switch( (*p) ) {
		case 86: goto st123;
		case 118: goto st123;
	}
	goto tr36;
st123:
	if ( ++p == pe )
		goto _test_eof123;
case 123:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr539;
		case 32: goto tr539;
		case 40: goto tr540;
		case 41: goto tr541;
		case 2058: goto tr542;
		case 2107: goto tr543;
		case 2314: goto tr544;
		case 2363: goto tr544;
		case 2570: goto tr545;
		case 2619: goto tr546;
	}
	goto tr57;
st124:
	if ( ++p == pe )
		goto _test_eof124;
case 124:
	switch( (*p) ) {
		case 72: goto st125;
		case 104: goto st125;
	}
	goto tr36;
st125:
	if ( ++p == pe )
		goto _test_eof125;
case 125:
	switch( (*p) ) {
		case 70: goto st126;
		case 102: goto st126;
	}
	goto tr36;
st126:
	if ( ++p == pe )
		goto _test_eof126;
case 126:
	switch( (*p) ) {
		case 80: goto st127;
		case 112: goto st127;
	}
	goto tr36;
st127:
	if ( ++p == pe )
		goto _test_eof127;
case 127:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr550;
		case 32: goto tr550;
		case 40: goto tr551;
		case 41: goto tr552;
		case 2058: goto tr553;
		case 2107: goto tr554;
		case 2314: goto tr555;
		case 2363: goto tr555;
		case 2570: goto tr556;
		case 2619: goto tr557;
	}
	goto tr57;
tr18:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st128;
tr53:
	{
		s->r_class = s->default_class;
	}
	goto st128;
tr634:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st128;
st128:
	if ( ++p == pe )
		goto _test_eof128;
case 128:
	switch( (*p) ) {
		case 76: goto st129;
		case 88: goto st132;
		case 89: goto st134;
		case 108: goto st129;
		case 120: goto st132;
		case 121: goto st134;
	}
	goto tr36;
st129:
	if ( ++p == pe )
		goto _test_eof129;
case 129:
	switch( (*p) ) {
		case 83: goto st130;
		case 115: goto st130;
	}
	goto tr36;
st130:
	if ( ++p == pe )
		goto _test_eof130;
case 130:
	switch( (*p) ) {
		case 65: goto st131;
		case 97: goto st131;
	}
	goto tr36;
st131:
	if ( ++p == pe )
		goto _test_eof131;
case 131:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr563;
		case 32: goto tr563;
		case 40: goto tr564;
		case 41: goto tr565;
		case 2058: goto tr566;
		case 2107: goto tr567;
		case 2314: goto tr568;
		case 2363: goto tr568;
		case 2570: goto tr569;
		case 2619: goto tr570;
	}
	goto tr57;
st132:
	if ( ++p == pe )
		goto _test_eof132;
case 132:
	switch( (*p) ) {
		case 84: goto st133;
		case 116: goto st133;
	}
	goto tr36;
st133:
	if ( ++p == pe )
		goto _test_eof133;
case 133:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr572;
		case 32: goto tr572;
		case 40: goto tr573;
		case 41: goto tr574;
		case 2058: goto tr575;
		case 2107: goto tr576;
		case 2314: goto tr577;
		case 2363: goto tr577;
		case 2570: goto tr578;
		case 2619: goto tr579;
	}
	goto tr57;
st134:
	if ( ++p == pe )
		goto _test_eof134;
case 134:
	switch( (*p) ) {
		case 80: goto st135;
		case 112: goto st135;
	}
	goto tr36;
st135:
	if ( ++p == pe )
		goto _test_eof135;
case 135:
	switch( (*p) ) {
		case 69: goto st136;
		case 101: goto st136;
	}
	goto tr36;
st136:
	if ( ++p == pe )
		goto _test_eof136;
case 136:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr583;
	goto tr582;
tr583:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st137;
tr588:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st137;
st137:
	if ( ++p == pe )
		goto _test_eof137;
case 137:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr585;
		case 32: goto tr585;
		case 40: goto tr586;
		case 41: goto tr587;
		case 2058: goto tr589;
		case 2107: goto tr590;
		case 2314: goto tr591;
		case 2363: goto tr591;
		case 2570: goto tr592;
		case 2619: goto tr593;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr588;
	goto tr584;
tr19:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st138;
tr54:
	{
		s->r_class = s->default_class;
	}
	goto st138;
tr635:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st138;
st138:
	if ( ++p == pe )
		goto _test_eof138;
case 138:
	switch( (*p) ) {
		case 82: goto st139;
		case 114: goto st139;
	}
	goto tr36;
st139:
	if ( ++p == pe )
		goto _test_eof139;
case 139:
	switch( (*p) ) {
		case 73: goto st140;
		case 105: goto st140;
	}
	goto tr36;
st140:
	if ( ++p == pe )
		goto _test_eof140;
case 140:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr596;
		case 32: goto tr596;
		case 40: goto tr597;
		case 41: goto tr598;
		case 2058: goto tr599;
		case 2107: goto tr600;
		case 2314: goto tr601;
		case 2363: goto tr601;
		case 2570: goto tr602;
		case 2619: goto tr603;
	}
	goto tr57;
tr648:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st141;
tr604:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st141;
tr299:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	goto st141;
tr661:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st141;
st141:
	if ( ++p == pe )
		goto _test_eof141;
case 141:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr605;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr604;
	goto tr36;
tr910:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1129; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1129; goto _out;}
		}
	}
	goto st1129;
tr143:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1129; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1129; goto _out;}
		}
	}
	goto st1129;
tr149:
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1129; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1129; goto _out;}
		}
	}
	goto st1129;
st1129:
	if ( ++p == pe )
		goto _test_eof1129;
case 1129:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3631;
		case 32: goto tr3631;
		case 36: goto st152;
		case 40: goto tr3632;
		case 41: goto tr3633;
		case 42: goto tr3613;
		case 65: goto tr3634;
		case 67: goto tr3635;
		case 68: goto tr3636;
		case 69: goto tr3637;
		case 72: goto tr3638;
		case 73: goto tr3639;
		case 75: goto tr3640;
		case 76: goto tr3641;
		case 77: goto tr3642;
		case 78: goto tr3643;
		case 80: goto tr3644;
		case 82: goto tr3645;
		case 83: goto tr3646;
		case 84: goto tr3647;
		case 85: goto tr3648;
		case 92: goto tr3613;
		case 95: goto tr3613;
		case 97: goto tr3634;
		case 99: goto tr3635;
		case 100: goto tr3636;
		case 101: goto tr3637;
		case 104: goto tr3638;
		case 105: goto tr3639;
		case 107: goto tr3640;
		case 108: goto tr3641;
		case 109: goto tr3642;
		case 110: goto tr3643;
		case 112: goto tr3644;
		case 114: goto tr3645;
		case 115: goto tr3646;
		case 116: goto tr3647;
		case 117: goto tr3648;
		case 778: goto tr3614;
		case 827: goto tr3615;
		case 1034: goto tr3649;
		case 1083: goto tr3650;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr3613;
	} else if ( _widec > 90 ) {
		if ( 98 <= _widec && _widec <= 122 )
			goto tr3613;
	} else
		goto tr3613;
	goto tr3630;
tr607:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st142;
tr608:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st142;
tr3631:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st142;
tr3632:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st142;
tr3633:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st142;
st142:
	if ( ++p == pe )
		goto _test_eof142;
case 142:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st142;
		case 32: goto st142;
		case 40: goto tr607;
		case 41: goto tr608;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 85: goto tr19;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 117: goto tr19;
		case 778: goto tr20;
		case 827: goto st151;
		case 1034: goto tr609;
		case 1083: goto st265;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr0;
tr10:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st143;
st143:
	if ( ++p == pe )
		goto _test_eof143;
case 143:
	switch( (*p) ) {
		case 78: goto st144;
		case 80: goto st66;
		case 110: goto st144;
		case 112: goto st66;
	}
	goto tr36;
st144:
	if ( ++p == pe )
		goto _test_eof144;
case 144:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr612;
		case 32: goto tr612;
		case 40: goto tr613;
		case 41: goto tr614;
		case 1034: goto tr615;
		case 1083: goto tr616;
	}
	goto tr36;
tr618:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st145;
tr619:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st145;
tr636:
	{
		s->line_counter++;
	}
	goto st145;
tr663:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st145;
tr612:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	goto st145;
tr613:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st145;
tr614:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st145;
tr615:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		s->line_counter++;
	}
	goto st145;
st145:
	if ( ++p == pe )
		goto _test_eof145;
case 145:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st145;
		case 32: goto st145;
		case 40: goto tr618;
		case 41: goto tr619;
		case 65: goto tr621;
		case 67: goto tr622;
		case 68: goto tr623;
		case 69: goto tr624;
		case 72: goto tr625;
		case 73: goto tr626;
		case 75: goto tr627;
		case 76: goto tr628;
		case 77: goto tr629;
		case 78: goto tr630;
		case 80: goto tr631;
		case 82: goto tr632;
		case 83: goto tr633;
		case 84: goto tr634;
		case 85: goto tr635;
		case 97: goto tr621;
		case 99: goto tr622;
		case 100: goto tr623;
		case 101: goto tr624;
		case 104: goto tr625;
		case 105: goto tr626;
		case 107: goto tr627;
		case 108: goto tr628;
		case 109: goto tr629;
		case 110: goto tr630;
		case 112: goto tr631;
		case 114: goto tr632;
		case 115: goto tr633;
		case 116: goto tr634;
		case 117: goto tr635;
		case 1034: goto tr636;
		case 1083: goto st150;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr620;
	goto tr582;
tr620:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st146;
tr641:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st146;
st146:
	if ( ++p == pe )
		goto _test_eof146;
case 146:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr638;
		case 32: goto tr638;
		case 40: goto tr639;
		case 41: goto tr640;
		case 68: goto tr642;
		case 72: goto tr643;
		case 77: goto tr644;
		case 83: goto st147;
		case 87: goto tr646;
		case 100: goto tr642;
		case 104: goto tr643;
		case 109: goto tr644;
		case 115: goto st147;
		case 119: goto tr646;
		case 1034: goto tr647;
		case 1083: goto tr648;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr641;
	goto tr24;
tr642:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st147;
tr643:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st147;
tr644:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st147;
tr646:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st147;
st147:
	if ( ++p == pe )
		goto _test_eof147;
case 147:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr638;
		case 32: goto tr638;
		case 40: goto tr639;
		case 41: goto tr640;
		case 1034: goto tr647;
		case 1083: goto tr648;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr649;
	goto tr24;
tr650:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st148;
tr649:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st148;
tr659:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st148;
st148:
	if ( ++p == pe )
		goto _test_eof148;
case 148:
	switch( (*p) ) {
		case 68: goto tr651;
		case 72: goto tr652;
		case 77: goto tr653;
		case 83: goto st149;
		case 87: goto tr655;
		case 100: goto tr651;
		case 104: goto tr652;
		case 109: goto tr653;
		case 115: goto st149;
		case 119: goto tr655;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr650;
	goto tr24;
tr651:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st149;
tr652:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st149;
tr653:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st149;
tr655:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st149;
st149:
	if ( ++p == pe )
		goto _test_eof149;
case 149:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr656;
		case 32: goto tr656;
		case 40: goto tr657;
		case 41: goto tr658;
		case 1034: goto tr660;
		case 1083: goto tr661;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr659;
	goto tr24;
tr662:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st150;
tr616:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	goto st150;
st150:
	if ( ++p == pe )
		goto _test_eof150;
case 150:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr663;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr662;
	goto tr36;
tr3615:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st151;
tr664:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st151;
tr3675:
	{
		NOERR;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st151;
st151:
	if ( ++p == pe )
		goto _test_eof151;
case 151:
	if ( (*p) == 10 )
		goto tr665;
	goto tr664;
tr609:
	{
		s->line_counter++;
	}
	goto st1130;
tr908:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1130;
tr799:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1130; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1130; goto _out;}
		}
	}
	goto st1130;
tr796:
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1130; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1130; goto _out;}
		}
	}
	goto st1130;
tr791:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1130; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1130; goto _out;}
		}
	}
	goto st1130;
tr3649:
	{
		s->line_counter++;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1130;
st1130:
	if ( ++p == pe )
		goto _test_eof1130;
case 1130:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3631;
		case 32: goto tr3631;
		case 36: goto st152;
		case 40: goto tr3632;
		case 41: goto tr3633;
		case 42: goto tr3613;
		case 65: goto tr3653;
		case 67: goto tr3654;
		case 68: goto tr3655;
		case 69: goto tr3656;
		case 72: goto tr3657;
		case 73: goto tr3658;
		case 75: goto tr3659;
		case 76: goto tr3660;
		case 77: goto tr3661;
		case 78: goto tr3662;
		case 80: goto tr3663;
		case 82: goto tr3664;
		case 83: goto tr3665;
		case 84: goto tr3666;
		case 85: goto tr3667;
		case 92: goto tr3613;
		case 95: goto tr3613;
		case 97: goto tr3653;
		case 99: goto tr3654;
		case 100: goto tr3655;
		case 101: goto tr3656;
		case 104: goto tr3657;
		case 105: goto tr3658;
		case 107: goto tr3659;
		case 108: goto tr3660;
		case 109: goto tr3661;
		case 110: goto tr3662;
		case 112: goto tr3663;
		case 114: goto tr3664;
		case 115: goto tr3665;
		case 116: goto tr3666;
		case 117: goto tr3667;
		case 778: goto tr3614;
		case 827: goto tr3615;
		case 1034: goto tr3649;
		case 1083: goto tr3650;
	}
	if ( _widec < 48 ) {
		if ( 45 <= _widec && _widec <= 47 )
			goto tr3613;
	} else if ( _widec > 57 ) {
		if ( _widec > 90 ) {
			if ( 98 <= _widec && _widec <= 122 )
				goto tr3613;
		} else if ( _widec >= 64 )
			goto tr3613;
	} else
		goto tr3652;
	goto tr3651;
tr3670:
	{
		NOERR;
	}
	goto st152;
st152:
	if ( ++p == pe )
		goto _test_eof152;
case 152:
	switch( (*p) ) {
		case 73: goto tr667;
		case 79: goto tr668;
		case 84: goto tr669;
		case 105: goto tr667;
		case 111: goto tr668;
		case 116: goto tr669;
	}
	goto tr666;
tr667:
	{
		ERR(ZS_OK);
	}
	goto st153;
st153:
	if ( ++p == pe )
		goto _test_eof153;
case 153:
	switch( (*p) ) {
		case 78: goto st154;
		case 110: goto st154;
	}
	goto tr666;
st154:
	if ( ++p == pe )
		goto _test_eof154;
case 154:
	switch( (*p) ) {
		case 67: goto st155;
		case 99: goto st155;
	}
	goto tr666;
st155:
	if ( ++p == pe )
		goto _test_eof155;
case 155:
	switch( (*p) ) {
		case 76: goto st156;
		case 108: goto st156;
	}
	goto tr666;
st156:
	if ( ++p == pe )
		goto _test_eof156;
case 156:
	switch( (*p) ) {
		case 85: goto st157;
		case 117: goto st157;
	}
	goto tr666;
st157:
	if ( ++p == pe )
		goto _test_eof157;
case 157:
	switch( (*p) ) {
		case 68: goto st158;
		case 100: goto st158;
	}
	goto tr666;
st158:
	if ( ++p == pe )
		goto _test_eof158;
case 158:
	switch( (*p) ) {
		case 69: goto st159;
		case 101: goto st159;
	}
	goto tr666;
st159:
	if ( ++p == pe )
		goto _test_eof159;
case 159:
	switch( (*p) ) {
		case 32: goto tr676;
		case 59: goto tr676;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr676;
	} else if ( (*p) >= 9 )
		goto tr676;
	goto tr666;
tr676:
	{ p--; {stack[top++] = 1131; goto st312;} }
	goto st1131;
tr777:
	{ p--; {stack[top++] = 1131; goto st300;} }
	goto st1131;
tr780:
	{ p--; {stack[top++] = 1131; goto st291;} }
	goto st1131;
st1131:
	if ( ++p == pe )
		goto _test_eof1131;
case 1131:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3669;
		case 32: goto tr3669;
		case 36: goto tr3670;
		case 40: goto tr3671;
		case 41: goto tr3672;
		case 42: goto tr3673;
		case 92: goto tr3673;
		case 95: goto tr3673;
		case 778: goto tr3674;
		case 827: goto tr3675;
		case 1034: goto tr3676;
		case 1083: goto tr3677;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr3673;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3673;
	} else
		goto tr3673;
	goto tr3668;
tr3613:
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 160; goto st270;} }
	goto st160;
tr3673:
	{
		NOERR;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 160; goto st270;} }
	goto st160;
st160:
	if ( ++p == pe )
		goto _test_eof160;
case 160:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr678;
		case 32: goto tr678;
		case 40: goto tr679;
		case 41: goto tr680;
		case 1034: goto tr681;
		case 1083: goto tr682;
	}
	goto tr677;
tr684:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st161;
tr685:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st161;
tr686:
	{
		s->line_counter++;
	}
	goto st161;
tr690:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st161;
tr678:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	goto st161;
tr679:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st161;
tr680:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st161;
tr681:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st161;
st161:
	if ( ++p == pe )
		goto _test_eof161;
case 161:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st161;
		case 32: goto st161;
		case 40: goto tr684;
		case 41: goto tr685;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 85: goto tr19;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 117: goto tr19;
		case 1034: goto tr686;
		case 1083: goto st162;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr582;
tr688:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st162;
tr682:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	goto st162;
st162:
	if ( ++p == pe )
		goto _test_eof162;
case 162:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr690;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr688;
	goto st0;
tr22:
	{
		s->line_counter++;
	}
	goto st1132;
tr906:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1132;
tr755:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1132; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1132; goto _out;}
		}
	}
	goto st1132;
tr752:
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1132; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1132; goto _out;}
		}
	}
	goto st1132;
tr785:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1132; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1132; goto _out;}
		}
	}
	goto st1132;
tr822:
	{
		s->line_counter++;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1132; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1132; goto _out;}
		}
	}
	goto st1132;
tr825:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1132; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1132; goto _out;}
		}
	}
	goto st1132;
tr3616:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1132;
tr3680:
	{
		s->line_counter++;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1132;
tr3676:
	{
		NOERR;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1132;
tr3684:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1132; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1132; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	goto st1132;
tr3708:
	{
		s->line_counter++;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1132; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1132; goto _out;}
		}
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	goto st1132;
st1132:
	if ( ++p == pe )
		goto _test_eof1132;
case 1132:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3609;
		case 32: goto tr3609;
		case 36: goto st152;
		case 40: goto tr3678;
		case 41: goto tr3679;
		case 42: goto tr3613;
		case 65: goto tr3653;
		case 67: goto tr3654;
		case 68: goto tr3655;
		case 69: goto tr3656;
		case 72: goto tr3657;
		case 73: goto tr3658;
		case 75: goto tr3659;
		case 76: goto tr3660;
		case 77: goto tr3661;
		case 78: goto tr3662;
		case 80: goto tr3663;
		case 82: goto tr3664;
		case 83: goto tr3665;
		case 84: goto tr3666;
		case 85: goto tr3667;
		case 92: goto tr3613;
		case 95: goto tr3613;
		case 97: goto tr3653;
		case 99: goto tr3654;
		case 100: goto tr3655;
		case 101: goto tr3656;
		case 104: goto tr3657;
		case 105: goto tr3658;
		case 107: goto tr3659;
		case 108: goto tr3660;
		case 109: goto tr3661;
		case 110: goto tr3662;
		case 112: goto tr3663;
		case 114: goto tr3664;
		case 115: goto tr3665;
		case 116: goto tr3666;
		case 117: goto tr3667;
		case 778: goto tr3614;
		case 827: goto tr3615;
		case 1034: goto tr3680;
		case 1083: goto tr3617;
	}
	if ( _widec < 48 ) {
		if ( 45 <= _widec && _widec <= 47 )
			goto tr3613;
	} else if ( _widec > 57 ) {
		if ( _widec > 90 ) {
			if ( 98 <= _widec && _widec <= 122 )
				goto tr3613;
		} else if ( _widec >= 64 )
			goto tr3613;
	} else
		goto tr3652;
	goto tr3651;
tr3652:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 163; goto st270;} }
	goto st163;
st163:
	if ( ++p == pe )
		goto _test_eof163;
case 163:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr692;
		case 32: goto tr692;
		case 40: goto tr693;
		case 41: goto tr694;
		case 68: goto tr29;
		case 72: goto tr30;
		case 77: goto tr31;
		case 83: goto st166;
		case 87: goto tr33;
		case 100: goto tr29;
		case 104: goto tr30;
		case 109: goto tr31;
		case 115: goto st166;
		case 119: goto tr33;
		case 1034: goto tr695;
		case 1083: goto tr696;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr28;
	goto tr691;
tr698:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st164;
tr699:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st164;
tr700:
	{
		s->line_counter++;
	}
	goto st164;
tr703:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st164;
tr692:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st164;
tr693:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st164;
tr694:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st164;
tr695:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st164;
st164:
	if ( ++p == pe )
		goto _test_eof164;
case 164:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st164;
		case 32: goto st164;
		case 40: goto tr698;
		case 41: goto tr699;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 85: goto tr19;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 117: goto tr19;
		case 1034: goto tr700;
		case 1083: goto st165;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr582;
tr702:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st165;
tr696:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st165;
st165:
	if ( ++p == pe )
		goto _test_eof165;
case 165:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr703;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr702;
	goto tr36;
tr29:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st166;
tr30:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st166;
tr31:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st166;
tr33:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st166;
st166:
	if ( ++p == pe )
		goto _test_eof166;
case 166:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr25;
		case 32: goto tr25;
		case 40: goto tr26;
		case 41: goto tr27;
		case 1034: goto tr34;
		case 1083: goto tr35;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr704;
	goto tr24;
tr705:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st167;
tr704:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st167;
tr714:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st167;
st167:
	if ( ++p == pe )
		goto _test_eof167;
case 167:
	switch( (*p) ) {
		case 68: goto tr706;
		case 72: goto tr707;
		case 77: goto tr708;
		case 83: goto st168;
		case 87: goto tr710;
		case 100: goto tr706;
		case 104: goto tr707;
		case 109: goto tr708;
		case 115: goto st168;
		case 119: goto tr710;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr705;
	goto tr24;
tr706:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st168;
tr707:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st168;
tr708:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st168;
tr710:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st168;
st168:
	if ( ++p == pe )
		goto _test_eof168;
case 168:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr711;
		case 32: goto tr711;
		case 40: goto tr712;
		case 41: goto tr713;
		case 1034: goto tr715;
		case 1083: goto tr716;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr714;
	goto tr24;
tr35:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st169;
tr717:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st169;
tr716:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st169;
st169:
	if ( ++p == pe )
		goto _test_eof169;
case 169:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr718;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr717;
	goto tr36;
tr3634:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 170; goto st270;} }
	goto st170;
tr3653:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 170; goto st270;} }
	goto st170;
st170:
	if ( ++p == pe )
		goto _test_eof170;
case 170:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr720;
		case 32: goto tr720;
		case 40: goto tr721;
		case 41: goto tr722;
		case 65: goto st230;
		case 70: goto st233;
		case 80: goto st237;
		case 97: goto st230;
		case 102: goto st233;
		case 112: goto st237;
		case 2058: goto tr723;
		case 2107: goto tr724;
		case 2314: goto tr66;
		case 2363: goto tr66;
		case 2570: goto tr725;
		case 2619: goto tr726;
	}
	goto tr719;
tr728:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st171;
tr729:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st171;
tr730:
	{
		s->line_counter++;
	}
	goto st171;
tr892:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st171;
tr720:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st171;
tr721:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st171;
tr722:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st171;
tr723:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st171;
st171:
	if ( ++p == pe )
		goto _test_eof171;
case 171:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st171;
		case 32: goto st171;
		case 40: goto tr728;
		case 41: goto tr729;
		case 58: goto tr69;
		case 65: goto tr118;
		case 67: goto tr119;
		case 68: goto tr120;
		case 69: goto tr121;
		case 72: goto tr122;
		case 73: goto tr123;
		case 75: goto tr124;
		case 76: goto tr125;
		case 77: goto tr126;
		case 78: goto tr127;
		case 80: goto tr128;
		case 82: goto tr129;
		case 83: goto tr130;
		case 84: goto tr131;
		case 85: goto tr132;
		case 92: goto st9;
		case 97: goto tr118;
		case 99: goto tr119;
		case 100: goto tr120;
		case 101: goto tr121;
		case 104: goto tr122;
		case 105: goto tr123;
		case 107: goto tr124;
		case 108: goto tr125;
		case 109: goto tr126;
		case 110: goto tr127;
		case 112: goto tr128;
		case 114: goto tr129;
		case 115: goto tr130;
		case 116: goto tr131;
		case 117: goto tr132;
		case 2058: goto tr730;
		case 2107: goto st246;
		case 2314: goto tr69;
		case 2363: goto tr69;
		case 2570: goto tr732;
		case 2619: goto tr733;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr69;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr69;
		} else if ( _widec >= 48 )
			goto tr117;
	} else
		goto tr69;
	goto tr584;
tr118:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 172; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 172; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 172; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 172; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 172; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 172; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 172; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 172; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 172; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 172; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 172; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 172; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 172; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 172; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 172; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 172; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 172; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 172; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 172; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 172; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 172; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 172; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 172; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 172; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 172; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 172; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 172; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 172; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 172; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st172;
st172:
	if ( ++p == pe )
		goto _test_eof172;
case 172:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr735;
		case 32: goto tr735;
		case 40: goto tr736;
		case 41: goto tr737;
		case 65: goto st230;
		case 70: goto st233;
		case 80: goto st237;
		case 97: goto st230;
		case 102: goto st233;
		case 112: goto st237;
		case 1802: goto tr83;
		case 1851: goto tr84;
		case 2058: goto tr738;
		case 2107: goto tr739;
		case 2314: goto tr740;
		case 2363: goto tr741;
		case 2570: goto tr742;
		case 2619: goto tr743;
	}
	goto tr734;
tr745:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st173;
tr746:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st173;
tr735:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st173;
tr736:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st173;
tr737:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st173;
st173:
	if ( ++p == pe )
		goto _test_eof173;
case 173:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st173;
		case 32: goto st173;
		case 40: goto tr745;
		case 41: goto tr746;
		case 92: goto st9;
		case 1802: goto tr89;
		case 1851: goto st8;
		case 2058: goto tr747;
		case 2107: goto st13;
		case 2314: goto tr108;
		case 2363: goto tr109;
		case 2570: goto tr748;
		case 2619: goto tr110;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr69;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr69;
	} else
		goto tr69;
	goto tr79;
tr101:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1133; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1133; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1133; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1133; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1133; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1133; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1133; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1133; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1133; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1133; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1133; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1133; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1133; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1133; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1133; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1133; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1133; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1133; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1133; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1133; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1133; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1133; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1133; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1133; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1133; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1133; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1133; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1133; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1133; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1133; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1133; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1133;
tr108:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1133; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1133; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1133; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1133; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1133; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1133; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1133; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1133; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1133; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1133; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1133; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1133; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1133; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1133; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1133; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1133; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1133; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1133; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1133; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1133; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1133; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1133; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1133; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1133; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1133; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1133; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1133; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1133; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1133; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1133; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1133; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1133;
tr135:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1133; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1133; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1133; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1133; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1133; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1133; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1133; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1133; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1133; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1133; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1133; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1133; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1133; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1133; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1133; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1133; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1133; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1133; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1133; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1133; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1133; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1133; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1133; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1133; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1133; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1133; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1133; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1133; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1133; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1133;
tr740:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1133; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1133; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1133; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1133; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1133; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1133; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1133; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1133; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1133; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1133; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1133; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1133; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1133; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1133; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1133; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1133; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1133; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1133; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1133; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1133; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1133; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1133; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1133; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1133; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1133; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1133; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1133; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1133; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1133; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1133; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1133; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1133;
tr3626:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1133; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1133; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1133; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1133; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1133; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1133; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1133; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1133; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1133; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1133; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1133; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1133; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1133; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1133; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1133; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1133; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1133; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1133; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1133; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1133; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1133; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1133; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1133; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1133; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1133; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1133; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1133; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1133; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1133; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->line_counter++;
	}
	goto st1133;
st1133:
	if ( ++p == pe )
		goto _test_eof1133;
case 1133:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3681;
		case 32: goto tr3681;
		case 36: goto st152;
		case 40: goto tr3682;
		case 41: goto tr3683;
		case 42: goto tr3613;
		case 92: goto tr3613;
		case 95: goto tr3613;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr3684;
		case 1083: goto tr3685;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr3613;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3613;
	} else
		goto tr3613;
	goto tr781;
tr750:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st174;
tr751:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st174;
tr819:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st174;
tr820:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st174;
tr821:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st174;
tr782:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st174;
tr783:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st174;
tr784:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st174;
tr3681:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	goto st174;
tr3682:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	goto st174;
tr3683:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	goto st174;
st174:
	if ( ++p == pe )
		goto _test_eof174;
case 174:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st174;
		case 32: goto st174;
		case 40: goto tr750;
		case 41: goto tr751;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 85: goto tr19;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 117: goto tr19;
		case 778: goto tr89;
		case 827: goto st8;
		case 1034: goto tr752;
		case 1083: goto st175;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr0;
tr823:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st175;
tr754:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st175;
tr828:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st175;
tr786:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st175;
tr3685:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	goto st175;
st175:
	if ( ++p == pe )
		goto _test_eof175;
case 175:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr92;
		case 1034: goto tr755;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr754;
	} else if ( _widec >= 640 )
		goto tr91;
	goto tr85;
tr109:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 176; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 176; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 176; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 176; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 176; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 176; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 176; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 176; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 176; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 176; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 176; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 176; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 176; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 176; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 176; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 176; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 176; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 176; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 176; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 176; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 176; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 176; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 176; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 176; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 176; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 176; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 176; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 176; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 176; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st176;
tr102:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 176; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 176; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 176; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 176; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 176; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 176; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 176; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 176; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 176; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 176; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 176; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 176; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 176; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 176; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 176; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 176; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 176; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 176; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 176; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 176; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 176; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 176; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 176; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 176; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 176; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 176; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 176; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 176; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 176; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st176;
tr741:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 176; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 176; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 176; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 176; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 176; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 176; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 176; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 176; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 176; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 176; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 176; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 176; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 176; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 176; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 176; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 176; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 176; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 176; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 176; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 176; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 176; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 176; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 176; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 176; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 176; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 176; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 176; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 176; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 176; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st176;
st176:
	if ( ++p == pe )
		goto _test_eof176;
case 176:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr756;
		case 32: goto tr756;
		case 40: goto tr757;
		case 41: goto tr758;
		case 778: goto tr759;
		case 827: goto tr760;
		case 1034: goto tr759;
		case 1083: goto tr760;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr91;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr91;
	} else
		goto tr91;
	goto tr79;
tr761:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st177;
tr756:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st177;
tr757:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st177;
tr758:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st177;
tr762:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st177;
tr763:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st177;
st177:
	if ( ++p == pe )
		goto _test_eof177;
case 177:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr761;
		case 32: goto tr761;
		case 40: goto tr762;
		case 41: goto tr763;
		case 778: goto tr92;
		case 827: goto tr91;
		case 1034: goto tr92;
		case 1083: goto tr91;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr91;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr91;
	} else
		goto tr91;
	goto tr85;
tr742:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1134; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1134; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1134; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1134; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1134; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1134; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1134; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1134; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1134; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1134; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1134; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1134; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1134; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1134; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1134; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1134; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1134; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1134; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1134; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1134; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1134; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1134; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1134; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1134; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1134; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1134; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1134; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1134; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1134; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1134; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1134; goto _out;}
		}
	}
	goto st1134;
tr748:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1134; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1134; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1134; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1134; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1134; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1134; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1134; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1134; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1134; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1134; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1134; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1134; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1134; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1134; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1134; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1134; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1134; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1134; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1134; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1134; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1134; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1134; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1134; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1134; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1134; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1134; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1134; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1134; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1134; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1134; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1134; goto _out;}
		}
	}
	goto st1134;
st1134:
	if ( ++p == pe )
		goto _test_eof1134;
case 1134:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr3686;
		case 32: goto tr3686;
		case 36: goto tr3619;
		case 40: goto tr3687;
		case 41: goto tr3688;
		case 42: goto tr3622;
		case 58: goto tr69;
		case 92: goto tr3623;
		case 95: goto tr3622;
		case 1802: goto tr83;
		case 1851: goto tr84;
		case 2058: goto tr3684;
		case 2107: goto tr3689;
		case 2314: goto tr101;
		case 2363: goto tr102;
		case 2570: goto tr3684;
		case 2619: goto tr3690;
	}
	if ( _widec < 60 ) {
		if ( _widec < 11 ) {
			if ( _widec <= 8 )
				goto tr69;
		} else if ( _widec > 44 ) {
			if ( 45 <= _widec && _widec <= 57 )
				goto tr3622;
		} else
			goto tr69;
	} else if ( _widec > 63 ) {
		if ( _widec < 91 ) {
			if ( 64 <= _widec && _widec <= 90 )
				goto tr3622;
		} else if ( _widec > 96 ) {
			if ( _widec > 122 ) {
				if ( 123 <= _widec )
					goto tr69;
			} else if ( _widec >= 97 )
				goto tr3622;
		} else
			goto tr69;
	} else
		goto tr69;
	goto tr781;
tr765:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st178;
tr766:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st178;
tr893:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st178;
tr894:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st178;
tr895:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st178;
tr3686:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	goto st178;
tr3687:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	goto st178;
tr3688:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	goto st178;
st178:
	if ( ++p == pe )
		goto _test_eof178;
case 178:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st178;
		case 32: goto st178;
		case 40: goto tr765;
		case 41: goto tr766;
		case 58: goto tr69;
		case 65: goto tr118;
		case 67: goto tr119;
		case 68: goto tr120;
		case 69: goto tr121;
		case 72: goto tr122;
		case 73: goto tr123;
		case 75: goto tr124;
		case 76: goto tr125;
		case 77: goto tr126;
		case 78: goto tr127;
		case 80: goto tr128;
		case 82: goto tr129;
		case 83: goto tr130;
		case 84: goto tr131;
		case 85: goto tr132;
		case 92: goto st9;
		case 97: goto tr118;
		case 99: goto tr119;
		case 100: goto tr120;
		case 101: goto tr121;
		case 104: goto tr122;
		case 105: goto tr123;
		case 107: goto tr124;
		case 108: goto tr125;
		case 109: goto tr126;
		case 110: goto tr127;
		case 112: goto tr128;
		case 114: goto tr129;
		case 115: goto tr130;
		case 116: goto tr131;
		case 117: goto tr132;
		case 1802: goto tr89;
		case 1851: goto st8;
		case 2058: goto tr752;
		case 2107: goto st193;
		case 2314: goto tr108;
		case 2363: goto tr109;
		case 2570: goto tr752;
		case 2619: goto tr768;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr69;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr69;
		} else if ( _widec >= 48 )
			goto tr117;
	} else
		goto tr69;
	goto tr114;
tr119:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 179; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 179; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 179; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 179; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 179; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 179; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 179; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 179; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 179; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 179; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 179; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 179; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 179; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 179; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 179; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 179; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 179; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 179; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 179; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 179; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 179; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 179; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 179; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 179; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 179; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 179; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 179; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 179; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 179; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st179;
st179:
	if ( ++p == pe )
		goto _test_eof179;
case 179:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr80;
		case 32: goto tr80;
		case 40: goto tr81;
		case 41: goto tr82;
		case 65: goto st18;
		case 68: goto st22;
		case 69: goto st29;
		case 78: goto st32;
		case 97: goto st18;
		case 100: goto st22;
		case 101: goto st29;
		case 110: goto st32;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr83;
		case 1083: goto tr84;
	}
	goto tr734;
tr120:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 180; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 180; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 180; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 180; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 180; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 180; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 180; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 180; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 180; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 180; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 180; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 180; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 180; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 180; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 180; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 180; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 180; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 180; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 180; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 180; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 180; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 180; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 180; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 180; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 180; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 180; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 180; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 180; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 180; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st180;
st180:
	if ( ++p == pe )
		goto _test_eof180;
case 180:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr80;
		case 32: goto tr80;
		case 40: goto tr81;
		case 41: goto tr82;
		case 72: goto st37;
		case 78: goto st41;
		case 83: goto st49;
		case 104: goto st37;
		case 110: goto st41;
		case 115: goto st49;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr83;
		case 1083: goto tr84;
	}
	goto tr734;
tr121:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 181; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 181; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 181; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 181; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 181; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 181; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 181; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 181; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 181; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 181; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 181; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 181; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 181; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 181; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 181; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 181; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 181; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 181; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 181; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 181; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 181; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 181; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 181; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 181; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 181; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 181; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 181; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 181; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 181; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st181;
st181:
	if ( ++p == pe )
		goto _test_eof181;
case 181:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr80;
		case 32: goto tr80;
		case 40: goto tr81;
		case 41: goto tr82;
		case 85: goto st51;
		case 117: goto st51;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr83;
		case 1083: goto tr84;
	}
	goto tr734;
tr122:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 182; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 182; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 182; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 182; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 182; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 182; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 182; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 182; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 182; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 182; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 182; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 182; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 182; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 182; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 182; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 182; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 182; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 182; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 182; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 182; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 182; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 182; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 182; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 182; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 182; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 182; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 182; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 182; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 182; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st182;
st182:
	if ( ++p == pe )
		goto _test_eof182;
case 182:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr80;
		case 32: goto tr80;
		case 40: goto tr81;
		case 41: goto tr82;
		case 73: goto st58;
		case 105: goto st58;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr83;
		case 1083: goto tr84;
	}
	goto tr734;
tr123:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 183; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 183; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 183; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 183; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 183; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 183; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 183; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 183; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 183; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 183; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 183; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 183; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 183; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 183; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 183; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 183; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 183; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 183; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 183; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 183; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 183; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 183; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 183; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 183; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 183; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 183; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 183; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 183; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 183; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st183;
st183:
	if ( ++p == pe )
		goto _test_eof183;
case 183:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr80;
		case 32: goto tr80;
		case 40: goto tr81;
		case 41: goto tr82;
		case 78: goto st144;
		case 80: goto st66;
		case 110: goto st144;
		case 112: goto st66;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr83;
		case 1083: goto tr84;
	}
	goto tr734;
tr124:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 184; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 184; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 184; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 184; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 184; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 184; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 184; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 184; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 184; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 184; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 184; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 184; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 184; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 184; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 184; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 184; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 184; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 184; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 184; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 184; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 184; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 184; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 184; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 184; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 184; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 184; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 184; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 184; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 184; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st184;
st184:
	if ( ++p == pe )
		goto _test_eof184;
case 184:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr80;
		case 32: goto tr80;
		case 40: goto tr81;
		case 41: goto tr82;
		case 69: goto st74;
		case 88: goto st76;
		case 101: goto st74;
		case 120: goto st76;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr83;
		case 1083: goto tr84;
	}
	goto tr734;
tr125:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 185; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 185; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 185; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 185; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 185; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 185; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 185; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 185; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 185; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 185; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 185; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 185; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 185; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 185; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 185; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 185; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 185; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 185; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 185; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 185; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 185; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 185; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 185; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 185; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 185; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 185; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 185; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 185; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 185; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st185;
st185:
	if ( ++p == pe )
		goto _test_eof185;
case 185:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr80;
		case 32: goto tr80;
		case 40: goto tr81;
		case 41: goto tr82;
		case 51: goto st78;
		case 54: goto st80;
		case 79: goto st82;
		case 80: goto st84;
		case 111: goto st82;
		case 112: goto st84;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr83;
		case 1083: goto tr84;
	}
	goto tr734;
tr126:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 186; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 186; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 186; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 186; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 186; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 186; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 186; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 186; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 186; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 186; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 186; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 186; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 186; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 186; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 186; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 186; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 186; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 186; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 186; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 186; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 186; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 186; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 186; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 186; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 186; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 186; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 186; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 186; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 186; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st186;
st186:
	if ( ++p == pe )
		goto _test_eof186;
case 186:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr80;
		case 32: goto tr80;
		case 40: goto tr81;
		case 41: goto tr82;
		case 73: goto st86;
		case 88: goto st90;
		case 105: goto st86;
		case 120: goto st90;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr83;
		case 1083: goto tr84;
	}
	goto tr734;
tr127:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 187; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 187; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 187; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 187; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 187; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 187; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 187; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 187; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 187; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 187; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 187; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 187; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 187; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 187; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 187; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 187; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 187; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 187; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 187; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 187; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 187; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 187; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 187; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 187; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 187; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 187; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 187; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 187; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 187; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st187;
st187:
	if ( ++p == pe )
		goto _test_eof187;
case 187:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr80;
		case 32: goto tr80;
		case 40: goto tr81;
		case 41: goto tr82;
		case 65: goto st92;
		case 73: goto st96;
		case 83: goto st98;
		case 97: goto st92;
		case 105: goto st96;
		case 115: goto st98;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr83;
		case 1083: goto tr84;
	}
	goto tr734;
tr128:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 188; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 188; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 188; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 188; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 188; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 188; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 188; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 188; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 188; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 188; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 188; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 188; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 188; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 188; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 188; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 188; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 188; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 188; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 188; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 188; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 188; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 188; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 188; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 188; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 188; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 188; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 188; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 188; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 188; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st188;
st188:
	if ( ++p == pe )
		goto _test_eof188;
case 188:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr80;
		case 32: goto tr80;
		case 40: goto tr81;
		case 41: goto tr82;
		case 84: goto st108;
		case 116: goto st108;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr83;
		case 1083: goto tr84;
	}
	goto tr734;
tr129:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 189; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 189; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 189; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 189; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 189; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 189; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 189; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 189; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 189; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 189; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 189; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 189; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 189; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 189; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 189; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 189; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 189; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 189; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 189; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 189; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 189; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 189; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 189; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 189; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 189; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 189; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 189; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 189; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 189; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st189;
st189:
	if ( ++p == pe )
		goto _test_eof189;
case 189:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr80;
		case 32: goto tr80;
		case 40: goto tr81;
		case 41: goto tr82;
		case 80: goto st111;
		case 82: goto st112;
		case 84: goto st116;
		case 112: goto st111;
		case 114: goto st112;
		case 116: goto st116;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr83;
		case 1083: goto tr84;
	}
	goto tr734;
tr130:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 190; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 190; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 190; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 190; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 190; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 190; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 190; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 190; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 190; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 190; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 190; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 190; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 190; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 190; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 190; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 190; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 190; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 190; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 190; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 190; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 190; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 190; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 190; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 190; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 190; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 190; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 190; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 190; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 190; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st190;
st190:
	if ( ++p == pe )
		goto _test_eof190;
case 190:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr80;
		case 32: goto tr80;
		case 40: goto tr81;
		case 41: goto tr82;
		case 79: goto st118;
		case 80: goto st120;
		case 82: goto st122;
		case 83: goto st124;
		case 111: goto st118;
		case 112: goto st120;
		case 114: goto st122;
		case 115: goto st124;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr83;
		case 1083: goto tr84;
	}
	goto tr734;
tr131:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 191; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 191; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 191; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 191; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 191; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 191; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 191; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 191; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 191; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 191; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 191; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 191; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 191; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 191; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 191; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 191; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 191; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 191; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 191; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 191; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 191; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 191; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 191; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 191; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 191; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 191; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 191; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 191; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 191; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st191;
st191:
	if ( ++p == pe )
		goto _test_eof191;
case 191:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr80;
		case 32: goto tr80;
		case 40: goto tr81;
		case 41: goto tr82;
		case 76: goto st129;
		case 88: goto st132;
		case 89: goto st134;
		case 108: goto st129;
		case 120: goto st132;
		case 121: goto st134;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr83;
		case 1083: goto tr84;
	}
	goto tr734;
tr132:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 192; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 192; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 192; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 192; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 192; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 192; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 192; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 192; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 192; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 192; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 192; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 192; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 192; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 192; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 192; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 192; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 192; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 192; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 192; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 192; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 192; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 192; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 192; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 192; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 192; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 192; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 192; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 192; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 192; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st192;
st192:
	if ( ++p == pe )
		goto _test_eof192;
case 192:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr80;
		case 32: goto tr80;
		case 40: goto tr81;
		case 41: goto tr82;
		case 82: goto st139;
		case 114: goto st139;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr83;
		case 1083: goto tr84;
	}
	goto tr734;
tr896:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st193;
tr769:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st193;
tr835:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st193;
tr805:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st193;
tr3689:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	goto st193;
st193:
	if ( ++p == pe )
		goto _test_eof193;
case 193:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr92;
		case 1034: goto tr770;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr769;
	} else if ( _widec >= 640 )
		goto tr91;
	goto tr79;
tr133:
	{
		s->line_counter++;
	}
	goto st1135;
tr870:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1135;
tr770:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1135; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1135; goto _out;}
		}
	}
	goto st1135;
tr811:
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1135; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1135; goto _out;}
		}
	}
	goto st1135;
tr804:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1135; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1135; goto _out;}
		}
	}
	goto st1135;
tr3624:
	{
		s->line_counter++;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1135;
st1135:
	if ( ++p == pe )
		goto _test_eof1135;
case 1135:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr3618;
		case 32: goto tr3618;
		case 36: goto tr3619;
		case 40: goto tr3620;
		case 41: goto tr3621;
		case 42: goto tr3622;
		case 58: goto tr69;
		case 65: goto tr3693;
		case 67: goto tr3694;
		case 68: goto tr3695;
		case 69: goto tr3696;
		case 72: goto tr3697;
		case 73: goto tr3698;
		case 75: goto tr3699;
		case 76: goto tr3700;
		case 77: goto tr3701;
		case 78: goto tr3702;
		case 80: goto tr3703;
		case 82: goto tr3704;
		case 83: goto tr3705;
		case 84: goto tr3706;
		case 85: goto tr3707;
		case 92: goto tr3623;
		case 95: goto tr3622;
		case 97: goto tr3693;
		case 99: goto tr3694;
		case 100: goto tr3695;
		case 101: goto tr3696;
		case 104: goto tr3697;
		case 105: goto tr3698;
		case 107: goto tr3699;
		case 108: goto tr3700;
		case 109: goto tr3701;
		case 110: goto tr3702;
		case 112: goto tr3703;
		case 114: goto tr3704;
		case 115: goto tr3705;
		case 116: goto tr3706;
		case 117: goto tr3707;
		case 1802: goto tr3614;
		case 1851: goto tr3615;
		case 2058: goto tr3624;
		case 2107: goto tr3625;
		case 2314: goto tr3626;
		case 2363: goto tr3627;
		case 2570: goto tr3628;
		case 2619: goto tr3629;
	}
	if ( _widec < 60 ) {
		if ( _widec < 11 ) {
			if ( _widec <= 8 )
				goto tr69;
		} else if ( _widec > 44 ) {
			if ( _widec > 47 ) {
				if ( 48 <= _widec && _widec <= 57 )
					goto tr3692;
			} else if ( _widec >= 45 )
				goto tr3622;
		} else
			goto tr69;
	} else if ( _widec > 63 ) {
		if ( _widec < 91 ) {
			if ( 64 <= _widec && _widec <= 90 )
				goto tr3622;
		} else if ( _widec > 96 ) {
			if ( _widec > 122 ) {
				if ( 123 <= _widec )
					goto tr69;
			} else if ( _widec >= 98 )
				goto tr3622;
		} else
			goto tr69;
	} else
		goto tr69;
	goto tr3691;
tr3619:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 194; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 194; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 194; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 194; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 194; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 194; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 194; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 194; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 194; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 194; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 194; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 194; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 194; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 194; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 194; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 194; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 194; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 194; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 194; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 194; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 194; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 194; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 194; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 194; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 194; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 194; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 194; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 194; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 194; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st194;
st194:
	if ( ++p == pe )
		goto _test_eof194;
case 194:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr80;
		case 32: goto tr80;
		case 40: goto tr81;
		case 41: goto tr82;
		case 73: goto tr667;
		case 79: goto tr668;
		case 84: goto tr669;
		case 105: goto tr667;
		case 111: goto tr668;
		case 116: goto tr669;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr83;
		case 1083: goto tr84;
	}
	goto tr771;
tr668:
	{
		ERR(ZS_OK);
	}
	goto st195;
st195:
	if ( ++p == pe )
		goto _test_eof195;
case 195:
	switch( (*p) ) {
		case 82: goto st196;
		case 114: goto st196;
	}
	goto tr666;
st196:
	if ( ++p == pe )
		goto _test_eof196;
case 196:
	switch( (*p) ) {
		case 73: goto st197;
		case 105: goto st197;
	}
	goto tr666;
st197:
	if ( ++p == pe )
		goto _test_eof197;
case 197:
	switch( (*p) ) {
		case 71: goto st198;
		case 103: goto st198;
	}
	goto tr666;
st198:
	if ( ++p == pe )
		goto _test_eof198;
case 198:
	switch( (*p) ) {
		case 73: goto st199;
		case 105: goto st199;
	}
	goto tr666;
st199:
	if ( ++p == pe )
		goto _test_eof199;
case 199:
	switch( (*p) ) {
		case 78: goto st200;
		case 110: goto st200;
	}
	goto tr666;
st200:
	if ( ++p == pe )
		goto _test_eof200;
case 200:
	switch( (*p) ) {
		case 32: goto tr777;
		case 59: goto tr777;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr777;
	} else if ( (*p) >= 9 )
		goto tr777;
	goto tr666;
tr669:
	{
		ERR(ZS_OK);
	}
	goto st201;
st201:
	if ( ++p == pe )
		goto _test_eof201;
case 201:
	switch( (*p) ) {
		case 84: goto st202;
		case 116: goto st202;
	}
	goto tr666;
st202:
	if ( ++p == pe )
		goto _test_eof202;
case 202:
	switch( (*p) ) {
		case 76: goto st203;
		case 108: goto st203;
	}
	goto tr666;
st203:
	if ( ++p == pe )
		goto _test_eof203;
case 203:
	switch( (*p) ) {
		case 32: goto tr780;
		case 59: goto tr780;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr780;
	} else if ( (*p) >= 9 )
		goto tr780;
	goto tr666;
tr3622:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 204; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 204; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 204; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 204; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 204; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 204; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 204; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 204; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 204; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 204; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 204; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 204; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 204; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 204; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 204; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 204; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 204; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 204; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 204; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 204; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 204; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 204; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 204; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 204; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 204; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 204; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 204; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 204; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 204; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 204; goto st270;} }
	goto st204;
st204:
	if ( ++p == pe )
		goto _test_eof204;
case 204:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr782;
		case 32: goto tr782;
		case 40: goto tr783;
		case 41: goto tr784;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr785;
		case 1083: goto tr786;
	}
	goto tr781;
tr3692:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 205; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 205; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 205; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 205; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 205; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 205; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 205; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 205; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 205; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 205; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 205; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 205; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 205; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 205; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 205; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 205; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 205; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 205; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 205; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 205; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 205; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 205; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 205; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 205; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 205; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 205; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 205; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 205; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 205; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 205; goto st270;} }
	goto st205;
st205:
	if ( ++p == pe )
		goto _test_eof205;
case 205:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr788;
		case 32: goto tr788;
		case 40: goto tr789;
		case 41: goto tr790;
		case 68: goto tr29;
		case 72: goto tr30;
		case 77: goto tr31;
		case 83: goto st166;
		case 87: goto tr33;
		case 100: goto tr29;
		case 104: goto tr30;
		case 109: goto tr31;
		case 115: goto st166;
		case 119: goto tr33;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr791;
		case 1083: goto tr792;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr28;
	goto tr787;
tr794:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st206;
tr795:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st206;
tr788:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st206;
tr789:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st206;
tr790:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st206;
st206:
	if ( ++p == pe )
		goto _test_eof206;
case 206:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st206;
		case 32: goto st206;
		case 40: goto tr794;
		case 41: goto tr795;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 85: goto tr19;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 117: goto tr19;
		case 778: goto tr89;
		case 827: goto st8;
		case 1034: goto tr796;
		case 1083: goto st207;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr0;
tr798:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st207;
tr792:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st207;
st207:
	if ( ++p == pe )
		goto _test_eof207;
case 207:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr92;
		case 1034: goto tr799;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr798;
	} else if ( _widec >= 640 )
		goto tr91;
	goto tr145;
tr3693:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 208; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 208; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 208; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 208; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 208; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 208; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 208; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 208; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 208; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 208; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 208; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 208; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 208; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 208; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 208; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 208; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 208; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 208; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 208; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 208; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 208; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 208; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 208; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 208; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 208; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 208; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 208; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 208; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 208; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 208; goto st270;} }
	goto st208;
st208:
	if ( ++p == pe )
		goto _test_eof208;
case 208:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr801;
		case 32: goto tr801;
		case 40: goto tr802;
		case 41: goto tr803;
		case 65: goto st230;
		case 70: goto st233;
		case 80: goto st237;
		case 97: goto st230;
		case 102: goto st233;
		case 112: goto st237;
		case 1802: goto tr83;
		case 1851: goto tr84;
		case 2058: goto tr804;
		case 2107: goto tr805;
		case 2314: goto tr740;
		case 2363: goto tr741;
		case 2570: goto tr806;
		case 2619: goto tr807;
	}
	goto tr800;
tr809:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st209;
tr810:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st209;
tr801:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st209;
tr802:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st209;
tr803:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st209;
st209:
	if ( ++p == pe )
		goto _test_eof209;
case 209:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st209;
		case 32: goto st209;
		case 40: goto tr809;
		case 41: goto tr810;
		case 58: goto tr69;
		case 65: goto tr118;
		case 67: goto tr119;
		case 68: goto tr120;
		case 69: goto tr121;
		case 72: goto tr122;
		case 73: goto tr123;
		case 75: goto tr124;
		case 76: goto tr125;
		case 77: goto tr126;
		case 78: goto tr127;
		case 80: goto tr128;
		case 82: goto tr129;
		case 83: goto tr130;
		case 84: goto tr131;
		case 85: goto tr132;
		case 92: goto st9;
		case 97: goto tr118;
		case 99: goto tr119;
		case 100: goto tr120;
		case 101: goto tr121;
		case 104: goto tr122;
		case 105: goto tr123;
		case 107: goto tr124;
		case 108: goto tr125;
		case 109: goto tr126;
		case 110: goto tr127;
		case 112: goto tr128;
		case 114: goto tr129;
		case 115: goto tr130;
		case 116: goto tr131;
		case 117: goto tr132;
		case 1802: goto tr89;
		case 1851: goto st8;
		case 2058: goto tr811;
		case 2107: goto st193;
		case 2314: goto tr108;
		case 2363: goto tr109;
		case 2570: goto tr812;
		case 2619: goto tr768;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr69;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr69;
		} else if ( _widec >= 48 )
			goto tr117;
	} else
		goto tr69;
	goto tr114;
tr137:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1136; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1136; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1136; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1136; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1136; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1136; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1136; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1136; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1136; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1136; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1136; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1136; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1136; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1136; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1136; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1136; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1136; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1136; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1136; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1136; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1136; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1136; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1136; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1136; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1136; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1136; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1136; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1136; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1136; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st1136;
tr812:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1136; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1136; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1136; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1136; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1136; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1136; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1136; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1136; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1136; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1136; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1136; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1136; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1136; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1136; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1136; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1136; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1136; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1136; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1136; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1136; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1136; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1136; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1136; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1136; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1136; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1136; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1136; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1136; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1136; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1136; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1136; goto _out;}
		}
	}
	goto st1136;
tr806:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1136; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1136; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1136; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1136; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1136; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1136; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1136; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1136; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1136; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1136; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1136; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1136; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1136; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1136; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1136; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1136; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1136; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1136; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1136; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1136; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1136; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1136; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1136; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1136; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1136; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1136; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1136; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1136; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1136; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->state = ZS_STATE_DATA;

		// Execute the record callback.
		if (s->process.automatic) {
			if (s->process.record != NULL) {
				s->process.record(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; cs = 1136; goto _out;}
				}
			}
		} else {
			// Return if external processing.
			p--; {p++; cs = 1136; goto _out;}
		}
	}
	goto st1136;
tr3628:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1136; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1136; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1136; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1136; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1136; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1136; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1136; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1136; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1136; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1136; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1136; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1136; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1136; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1136; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1136; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1136; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1136; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1136; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1136; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1136; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1136; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1136; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1136; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1136; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1136; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1136; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1136; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 1136; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 1136; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st1136;
st1136:
	if ( ++p == pe )
		goto _test_eof1136;
case 1136:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr3686;
		case 32: goto tr3686;
		case 36: goto tr3619;
		case 40: goto tr3687;
		case 41: goto tr3688;
		case 42: goto tr3622;
		case 58: goto tr69;
		case 65: goto tr3693;
		case 67: goto tr3694;
		case 68: goto tr3695;
		case 69: goto tr3696;
		case 72: goto tr3697;
		case 73: goto tr3698;
		case 75: goto tr3699;
		case 76: goto tr3700;
		case 77: goto tr3701;
		case 78: goto tr3702;
		case 80: goto tr3703;
		case 82: goto tr3704;
		case 83: goto tr3705;
		case 84: goto tr3706;
		case 85: goto tr3707;
		case 92: goto tr3623;
		case 95: goto tr3622;
		case 97: goto tr3693;
		case 99: goto tr3694;
		case 100: goto tr3695;
		case 101: goto tr3696;
		case 104: goto tr3697;
		case 105: goto tr3698;
		case 107: goto tr3699;
		case 108: goto tr3700;
		case 109: goto tr3701;
		case 110: goto tr3702;
		case 112: goto tr3703;
		case 114: goto tr3704;
		case 115: goto tr3705;
		case 116: goto tr3706;
		case 117: goto tr3707;
		case 1802: goto tr83;
		case 1851: goto tr84;
		case 2058: goto tr3708;
		case 2107: goto tr3689;
		case 2314: goto tr101;
		case 2363: goto tr102;
		case 2570: goto tr3708;
		case 2619: goto tr3690;
	}
	if ( _widec < 60 ) {
		if ( _widec < 11 ) {
			if ( _widec <= 8 )
				goto tr69;
		} else if ( _widec > 44 ) {
			if ( _widec > 47 ) {
				if ( 48 <= _widec && _widec <= 57 )
					goto tr3692;
			} else if ( _widec >= 45 )
				goto tr3622;
		} else
			goto tr69;
	} else if ( _widec > 63 ) {
		if ( _widec < 91 ) {
			if ( 64 <= _widec && _widec <= 90 )
				goto tr3622;
		} else if ( _widec > 96 ) {
			if ( _widec > 122 ) {
				if ( 123 <= _widec )
					goto tr69;
			} else if ( _widec >= 98 )
				goto tr3622;
		} else
			goto tr69;
	} else
		goto tr69;
	goto tr3691;
tr3694:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 210; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 210; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 210; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 210; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 210; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 210; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 210; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 210; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 210; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 210; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 210; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 210; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 210; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 210; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 210; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 210; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 210; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 210; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 210; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 210; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 210; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 210; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 210; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 210; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 210; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 210; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 210; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 210; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 210; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 210; goto st270;} }
	goto st210;
st210:
	if ( ++p == pe )
		goto _test_eof210;
case 210:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr782;
		case 32: goto tr782;
		case 40: goto tr783;
		case 41: goto tr784;
		case 65: goto st18;
		case 68: goto st22;
		case 69: goto st29;
		case 78: goto st32;
		case 97: goto st18;
		case 100: goto st22;
		case 101: goto st29;
		case 110: goto st32;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr785;
		case 1083: goto tr786;
	}
	goto tr800;
tr3695:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 211; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 211; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 211; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 211; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 211; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 211; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 211; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 211; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 211; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 211; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 211; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 211; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 211; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 211; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 211; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 211; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 211; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 211; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 211; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 211; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 211; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 211; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 211; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 211; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 211; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 211; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 211; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 211; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 211; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 211; goto st270;} }
	goto st211;
st211:
	if ( ++p == pe )
		goto _test_eof211;
case 211:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr782;
		case 32: goto tr782;
		case 40: goto tr783;
		case 41: goto tr784;
		case 72: goto st37;
		case 78: goto st41;
		case 83: goto st49;
		case 104: goto st37;
		case 110: goto st41;
		case 115: goto st49;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr785;
		case 1083: goto tr786;
	}
	goto tr800;
tr3696:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 212; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 212; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 212; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 212; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 212; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 212; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 212; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 212; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 212; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 212; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 212; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 212; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 212; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 212; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 212; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 212; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 212; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 212; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 212; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 212; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 212; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 212; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 212; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 212; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 212; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 212; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 212; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 212; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 212; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 212; goto st270;} }
	goto st212;
st212:
	if ( ++p == pe )
		goto _test_eof212;
case 212:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr782;
		case 32: goto tr782;
		case 40: goto tr783;
		case 41: goto tr784;
		case 85: goto st51;
		case 117: goto st51;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr785;
		case 1083: goto tr786;
	}
	goto tr800;
tr3697:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 213; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 213; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 213; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 213; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 213; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 213; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 213; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 213; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 213; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 213; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 213; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 213; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 213; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 213; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 213; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 213; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 213; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 213; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 213; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 213; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 213; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 213; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 213; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 213; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 213; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 213; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 213; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 213; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 213; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 213; goto st270;} }
	goto st213;
st213:
	if ( ++p == pe )
		goto _test_eof213;
case 213:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr782;
		case 32: goto tr782;
		case 40: goto tr783;
		case 41: goto tr784;
		case 73: goto st58;
		case 105: goto st58;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr785;
		case 1083: goto tr786;
	}
	goto tr800;
tr3698:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 214; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 214; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 214; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 214; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 214; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 214; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 214; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 214; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 214; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 214; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 214; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 214; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 214; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 214; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 214; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 214; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 214; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 214; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 214; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 214; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 214; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 214; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 214; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 214; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 214; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 214; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 214; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 214; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 214; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 214; goto st270;} }
	goto st214;
st214:
	if ( ++p == pe )
		goto _test_eof214;
case 214:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr782;
		case 32: goto tr782;
		case 40: goto tr783;
		case 41: goto tr784;
		case 78: goto st144;
		case 80: goto st66;
		case 110: goto st144;
		case 112: goto st66;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr785;
		case 1083: goto tr786;
	}
	goto tr800;
tr3699:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 215; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 215; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 215; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 215; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 215; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 215; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 215; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 215; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 215; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 215; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 215; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 215; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 215; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 215; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 215; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 215; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 215; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 215; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 215; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 215; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 215; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 215; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 215; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 215; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 215; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 215; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 215; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 215; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 215; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 215; goto st270;} }
	goto st215;
st215:
	if ( ++p == pe )
		goto _test_eof215;
case 215:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr782;
		case 32: goto tr782;
		case 40: goto tr783;
		case 41: goto tr784;
		case 69: goto st74;
		case 88: goto st76;
		case 101: goto st74;
		case 120: goto st76;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr785;
		case 1083: goto tr786;
	}
	goto tr800;
tr3700:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 216; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 216; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 216; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 216; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 216; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 216; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 216; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 216; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 216; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 216; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 216; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 216; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 216; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 216; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 216; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 216; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 216; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 216; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 216; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 216; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 216; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 216; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 216; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 216; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 216; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 216; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 216; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 216; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 216; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 216; goto st270;} }
	goto st216;
st216:
	if ( ++p == pe )
		goto _test_eof216;
case 216:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr782;
		case 32: goto tr782;
		case 40: goto tr783;
		case 41: goto tr784;
		case 51: goto st78;
		case 54: goto st80;
		case 79: goto st82;
		case 80: goto st84;
		case 111: goto st82;
		case 112: goto st84;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr785;
		case 1083: goto tr786;
	}
	goto tr800;
tr3701:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 217; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 217; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 217; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 217; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 217; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 217; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 217; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 217; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 217; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 217; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 217; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 217; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 217; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 217; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 217; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 217; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 217; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 217; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 217; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 217; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 217; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 217; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 217; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 217; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 217; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 217; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 217; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 217; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 217; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 217; goto st270;} }
	goto st217;
st217:
	if ( ++p == pe )
		goto _test_eof217;
case 217:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr782;
		case 32: goto tr782;
		case 40: goto tr783;
		case 41: goto tr784;
		case 73: goto st86;
		case 88: goto st90;
		case 105: goto st86;
		case 120: goto st90;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr785;
		case 1083: goto tr786;
	}
	goto tr800;
tr3702:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 218; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 218; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 218; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 218; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 218; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 218; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 218; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 218; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 218; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 218; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 218; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 218; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 218; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 218; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 218; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 218; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 218; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 218; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 218; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 218; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 218; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 218; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 218; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 218; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 218; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 218; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 218; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 218; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 218; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 218; goto st270;} }
	goto st218;
st218:
	if ( ++p == pe )
		goto _test_eof218;
case 218:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr782;
		case 32: goto tr782;
		case 40: goto tr783;
		case 41: goto tr784;
		case 65: goto st92;
		case 73: goto st96;
		case 83: goto st98;
		case 97: goto st92;
		case 105: goto st96;
		case 115: goto st98;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr785;
		case 1083: goto tr786;
	}
	goto tr800;
tr3703:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 219; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 219; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 219; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 219; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 219; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 219; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 219; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 219; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 219; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 219; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 219; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 219; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 219; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 219; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 219; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 219; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 219; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 219; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 219; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 219; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 219; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 219; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 219; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 219; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 219; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 219; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 219; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 219; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 219; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 219; goto st270;} }
	goto st219;
st219:
	if ( ++p == pe )
		goto _test_eof219;
case 219:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr782;
		case 32: goto tr782;
		case 40: goto tr783;
		case 41: goto tr784;
		case 84: goto st108;
		case 116: goto st108;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr785;
		case 1083: goto tr786;
	}
	goto tr800;
tr3704:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 220; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 220; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 220; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 220; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 220; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 220; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 220; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 220; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 220; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 220; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 220; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 220; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 220; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 220; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 220; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 220; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 220; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 220; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 220; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 220; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 220; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 220; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 220; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 220; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 220; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 220; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 220; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 220; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 220; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 220; goto st270;} }
	goto st220;
st220:
	if ( ++p == pe )
		goto _test_eof220;
case 220:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr782;
		case 32: goto tr782;
		case 40: goto tr783;
		case 41: goto tr784;
		case 80: goto st111;
		case 82: goto st112;
		case 84: goto st116;
		case 112: goto st111;
		case 114: goto st112;
		case 116: goto st116;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr785;
		case 1083: goto tr786;
	}
	goto tr800;
tr3705:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 221; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 221; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 221; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 221; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 221; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 221; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 221; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 221; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 221; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 221; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 221; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 221; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 221; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 221; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 221; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 221; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 221; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 221; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 221; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 221; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 221; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 221; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 221; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 221; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 221; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 221; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 221; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 221; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 221; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 221; goto st270;} }
	goto st221;
st221:
	if ( ++p == pe )
		goto _test_eof221;
case 221:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr782;
		case 32: goto tr782;
		case 40: goto tr783;
		case 41: goto tr784;
		case 79: goto st118;
		case 80: goto st120;
		case 82: goto st122;
		case 83: goto st124;
		case 111: goto st118;
		case 112: goto st120;
		case 114: goto st122;
		case 115: goto st124;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr785;
		case 1083: goto tr786;
	}
	goto tr800;
tr3706:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 222; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 222; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 222; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 222; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 222; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 222; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 222; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 222; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 222; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 222; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 222; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 222; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 222; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 222; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 222; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 222; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 222; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 222; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 222; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 222; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 222; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 222; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 222; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 222; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 222; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 222; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 222; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 222; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 222; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 222; goto st270;} }
	goto st222;
st222:
	if ( ++p == pe )
		goto _test_eof222;
case 222:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr782;
		case 32: goto tr782;
		case 40: goto tr783;
		case 41: goto tr784;
		case 76: goto st129;
		case 88: goto st132;
		case 89: goto st134;
		case 108: goto st129;
		case 120: goto st132;
		case 121: goto st134;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr785;
		case 1083: goto tr786;
	}
	goto tr800;
tr3707:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 223; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 223; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 223; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 223; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 223; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 223; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 223; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 223; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 223; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 223; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 223; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 223; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 223; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 223; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 223; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 223; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 223; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 223; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 223; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 223; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 223; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 223; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 223; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 223; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 223; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 223; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 223; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 223; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 223; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 223; goto st270;} }
	goto st223;
st223:
	if ( ++p == pe )
		goto _test_eof223;
case 223:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr782;
		case 32: goto tr782;
		case 40: goto tr783;
		case 41: goto tr784;
		case 82: goto st139;
		case 114: goto st139;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr785;
		case 1083: goto tr786;
	}
	goto tr800;
tr3623:
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 224; goto st270;} }
	goto st224;
st224:
	if ( ++p == pe )
		goto _test_eof224;
case 224:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr813;
		case 32: goto tr813;
		case 35: goto tr94;
		case 40: goto tr815;
		case 41: goto tr816;
		case 778: goto tr93;
		case 827: goto tr93;
		case 1034: goto tr817;
		case 1083: goto tr818;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr93;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr93;
	} else
		goto tr93;
	goto tr814;
tr813:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ p--; }
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 225; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 225; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 225; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 225; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 225; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 225; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 225; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 225; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 225; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 225; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 225; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 225; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 225; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 225; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 225; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 225; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 225; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 225; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 225; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 225; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 225; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 225; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 225; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 225; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 225; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 225; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 225; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 225; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 225; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st225;
tr815:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{ p--; }
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 225; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 225; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 225; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 225; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 225; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 225; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 225; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 225; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 225; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 225; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 225; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 225; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 225; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 225; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 225; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 225; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 225; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 225; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 225; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 225; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 225; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 225; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 225; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 225; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 225; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 225; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 225; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 225; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 225; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st225;
tr816:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{ p--; }
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 225; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 225; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 225; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 225; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 225; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 225; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 225; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 225; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 225; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 225; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 225; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 225; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 225; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 225; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 225; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 225; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 225; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 225; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 225; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 225; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 225; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 225; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 225; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 225; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 225; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 225; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 225; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 225; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 225; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st225;
tr817:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{ p--; }
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 225; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 225; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 225; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 225; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 225; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 225; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 225; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 225; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 225; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 225; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 225; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 225; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 225; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 225; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 225; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 225; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 225; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 225; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 225; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 225; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 225; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 225; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 225; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 225; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 225; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 225; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 225; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 225; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 225; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st225;
st225:
	if ( ++p == pe )
		goto _test_eof225;
case 225:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr819;
		case 32: goto tr819;
		case 40: goto tr820;
		case 41: goto tr821;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 85: goto tr19;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 117: goto tr19;
		case 778: goto tr83;
		case 827: goto tr84;
		case 1034: goto tr822;
		case 1083: goto tr823;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr114;
tr818:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ p--; }
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 226; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 226; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 226; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 226; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 226; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 226; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 226; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 226; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 226; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 226; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 226; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 226; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 226; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 226; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 226; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 226; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 226; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 226; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 226; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 226; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 226; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 226; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 226; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 226; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 226; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 226; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 226; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 226; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 226; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st226;
st226:
	if ( ++p == pe )
		goto _test_eof226;
case 226:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr80;
		case 778: goto tr83;
		case 800: goto tr80;
		case 808: goto tr81;
		case 809: goto tr82;
		case 827: goto tr84;
		case 1033: goto tr824;
		case 1034: goto tr825;
		case 1056: goto tr824;
		case 1064: goto tr826;
		case 1065: goto tr827;
		case 1083: goto tr828;
	}
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr688;
	goto tr79;
tr829:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st227;
tr824:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st227;
tr826:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st227;
tr827:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st227;
tr830:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st227;
tr831:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st227;
st227:
	if ( ++p == pe )
		goto _test_eof227;
case 227:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st7;
		case 778: goto tr89;
		case 800: goto st7;
		case 808: goto tr87;
		case 809: goto tr88;
		case 827: goto st8;
		case 1033: goto tr829;
		case 1034: goto tr755;
		case 1056: goto tr829;
		case 1064: goto tr830;
		case 1065: goto tr831;
		case 1083: goto tr754;
	}
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr688;
	goto tr85;
tr768:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 228; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 228; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 228; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 228; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 228; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 228; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 228; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 228; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 228; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 228; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 228; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 228; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 228; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 228; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 228; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 228; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 228; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 228; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 228; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 228; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 228; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 228; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 228; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 228; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 228; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 228; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 228; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 228; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 228; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st228;
tr897:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 228; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 228; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 228; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 228; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 228; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 228; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 228; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 228; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 228; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 228; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 228; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 228; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 228; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 228; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 228; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 228; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 228; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 228; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 228; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 228; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 228; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 228; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 228; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 228; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 228; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 228; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 228; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 228; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 228; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st228;
tr807:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 228; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 228; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 228; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 228; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 228; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 228; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 228; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 228; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 228; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 228; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 228; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 228; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 228; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 228; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 228; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 228; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 228; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 228; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 228; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 228; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 228; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 228; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 228; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 228; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 228; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 228; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 228; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 228; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 228; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st228;
tr3690:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 228; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 228; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 228; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 228; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 228; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 228; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 228; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 228; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 228; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 228; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 228; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 228; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 228; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 228; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 228; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 228; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 228; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 228; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 228; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 228; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 228; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 228; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 228; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 228; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 228; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 228; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 228; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 228; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 228; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	goto st228;
st228:
	if ( ++p == pe )
		goto _test_eof228;
case 228:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr756;
		case 778: goto tr759;
		case 800: goto tr756;
		case 808: goto tr757;
		case 809: goto tr758;
		case 827: goto tr760;
		case 1033: goto tr832;
		case 1034: goto tr825;
		case 1056: goto tr832;
		case 1064: goto tr833;
		case 1065: goto tr834;
		case 1083: goto tr835;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr769;
	} else if ( _widec >= 640 )
		goto tr91;
	goto tr79;
tr836:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st229;
tr832:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st229;
tr833:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st229;
tr834:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st229;
tr837:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st229;
tr838:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st229;
st229:
	if ( ++p == pe )
		goto _test_eof229;
case 229:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr761;
		case 778: goto tr92;
		case 800: goto tr761;
		case 808: goto tr762;
		case 809: goto tr763;
		case 1033: goto tr836;
		case 1034: goto tr755;
		case 1056: goto tr836;
		case 1064: goto tr837;
		case 1065: goto tr838;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr769;
	} else if ( _widec >= 640 )
		goto tr91;
	goto tr79;
st230:
	if ( ++p == pe )
		goto _test_eof230;
case 230:
	switch( (*p) ) {
		case 65: goto st231;
		case 97: goto st231;
	}
	goto tr36;
st231:
	if ( ++p == pe )
		goto _test_eof231;
case 231:
	switch( (*p) ) {
		case 65: goto st232;
		case 97: goto st232;
	}
	goto tr36;
st232:
	if ( ++p == pe )
		goto _test_eof232;
case 232:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr841;
		case 32: goto tr841;
		case 40: goto tr842;
		case 41: goto tr843;
		case 2058: goto tr844;
		case 2107: goto tr845;
		case 2314: goto tr846;
		case 2363: goto tr846;
		case 2570: goto tr847;
		case 2619: goto tr848;
	}
	goto tr57;
st233:
	if ( ++p == pe )
		goto _test_eof233;
case 233:
	switch( (*p) ) {
		case 83: goto st234;
		case 115: goto st234;
	}
	goto tr36;
st234:
	if ( ++p == pe )
		goto _test_eof234;
case 234:
	switch( (*p) ) {
		case 68: goto st235;
		case 100: goto st235;
	}
	goto tr36;
st235:
	if ( ++p == pe )
		goto _test_eof235;
case 235:
	switch( (*p) ) {
		case 66: goto st236;
		case 98: goto st236;
	}
	goto tr36;
st236:
	if ( ++p == pe )
		goto _test_eof236;
case 236:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr852;
		case 32: goto tr852;
		case 40: goto tr853;
		case 41: goto tr854;
		case 2058: goto tr855;
		case 2107: goto tr856;
		case 2314: goto tr857;
		case 2363: goto tr857;
		case 2570: goto tr858;
		case 2619: goto tr859;
	}
	goto tr57;
st237:
	if ( ++p == pe )
		goto _test_eof237;
case 237:
	switch( (*p) ) {
		case 76: goto st238;
		case 108: goto st238;
	}
	goto tr36;
st238:
	if ( ++p == pe )
		goto _test_eof238;
case 238:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr861;
		case 32: goto tr861;
		case 40: goto tr862;
		case 41: goto tr863;
		case 2058: goto tr864;
		case 2107: goto tr865;
		case 2314: goto tr866;
		case 2363: goto tr866;
		case 2570: goto tr867;
		case 2619: goto tr868;
	}
	goto tr57;
tr869:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st239;
tr3625:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st239;
st239:
	if ( ++p == pe )
		goto _test_eof239;
case 239:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr665;
		case 1034: goto tr870;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr869;
	} else if ( _widec >= 640 )
		goto tr664;
	goto tr79;
tr136:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 240; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 240; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 240; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 240; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 240; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 240; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 240; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 240; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 240; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 240; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 240; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 240; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 240; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 240; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 240; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 240; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 240; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 240; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 240; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 240; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 240; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 240; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 240; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 240; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 240; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 240; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 240; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 240; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 240; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st240;
tr3627:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 240; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 240; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 240; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 240; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 240; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 240; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 240; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 240; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 240; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 240; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 240; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 240; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 240; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 240; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 240; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 240; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 240; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 240; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 240; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 240; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 240; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 240; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 240; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 240; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 240; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 240; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 240; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 240; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 240; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st240;
st240:
	if ( ++p == pe )
		goto _test_eof240;
case 240:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr871;
		case 32: goto tr871;
		case 40: goto tr872;
		case 41: goto tr873;
		case 778: goto tr874;
		case 827: goto tr760;
		case 1034: goto tr874;
		case 1083: goto tr760;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr664;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr664;
	} else
		goto tr664;
	goto tr79;
tr875:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st241;
tr871:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st241;
tr872:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st241;
tr873:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st241;
tr876:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st241;
tr877:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st241;
st241:
	if ( ++p == pe )
		goto _test_eof241;
case 241:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr875;
		case 32: goto tr875;
		case 40: goto tr876;
		case 41: goto tr877;
		case 778: goto tr878;
		case 827: goto tr91;
		case 1034: goto tr878;
		case 1083: goto tr91;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr664;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr664;
	} else
		goto tr664;
	goto tr85;
tr138:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 242; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 242; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 242; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 242; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 242; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 242; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 242; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 242; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 242; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 242; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 242; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 242; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 242; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 242; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 242; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 242; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 242; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 242; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 242; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 242; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 242; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 242; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 242; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 242; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 242; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 242; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 242; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 242; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 242; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st242;
tr3629:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 242; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 242; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 242; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 242; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 242; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 242; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 242; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 242; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 242; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 242; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 242; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 242; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 242; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 242; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 242; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 242; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 242; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 242; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 242; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 242; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 242; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 242; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 242; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 242; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 242; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 242; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 242; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 242; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 242; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st242;
st242:
	if ( ++p == pe )
		goto _test_eof242;
case 242:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr871;
		case 778: goto tr874;
		case 800: goto tr871;
		case 808: goto tr872;
		case 809: goto tr873;
		case 827: goto tr760;
		case 1033: goto tr879;
		case 1034: goto tr825;
		case 1056: goto tr879;
		case 1064: goto tr880;
		case 1065: goto tr881;
		case 1083: goto tr835;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr869;
	} else if ( _widec >= 640 )
		goto tr664;
	goto tr79;
tr882:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st243;
tr879:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st243;
tr880:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st243;
tr881:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st243;
tr883:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st243;
tr884:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st243;
st243:
	if ( ++p == pe )
		goto _test_eof243;
case 243:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr875;
		case 778: goto tr878;
		case 800: goto tr875;
		case 808: goto tr876;
		case 809: goto tr877;
		case 827: goto tr91;
		case 1033: goto tr882;
		case 1034: goto tr755;
		case 1056: goto tr882;
		case 1064: goto tr883;
		case 1065: goto tr884;
		case 1083: goto tr769;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr869;
	} else if ( _widec >= 640 )
		goto tr664;
	goto tr79;
tr110:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 244; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 244; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 244; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 244; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 244; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 244; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 244; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 244; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 244; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 244; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 244; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 244; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 244; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 244; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 244; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 244; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 244; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 244; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 244; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 244; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 244; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 244; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 244; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 244; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 244; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 244; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 244; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 244; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 244; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st244;
tr103:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 244; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 244; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 244; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 244; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 244; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 244; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 244; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 244; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 244; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 244; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 244; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 244; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 244; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 244; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 244; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 244; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 244; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 244; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 244; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 244; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 244; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 244; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 244; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 244; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 244; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 244; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 244; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 244; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 244; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st244;
tr743:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 244; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 244; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 244; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 244; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 244; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 244; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 244; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 244; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 244; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 244; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 244; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 244; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 244; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 244; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 244; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 244; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 244; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 244; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 244; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 244; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 244; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 244; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 244; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 244; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 244; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 244; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 244; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 244; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 244; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st244;
st244:
	if ( ++p == pe )
		goto _test_eof244;
case 244:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr756;
		case 778: goto tr759;
		case 800: goto tr756;
		case 808: goto tr757;
		case 809: goto tr758;
		case 827: goto tr760;
		case 1033: goto tr885;
		case 1034: goto tr759;
		case 1056: goto tr885;
		case 1064: goto tr886;
		case 1065: goto tr887;
		case 1083: goto tr167;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr111;
	} else if ( _widec >= 640 )
		goto tr91;
	goto tr79;
tr888:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st245;
tr885:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st245;
tr886:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st245;
tr887:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st245;
tr889:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st245;
tr890:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st245;
st245:
	if ( ++p == pe )
		goto _test_eof245;
case 245:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr761;
		case 778: goto tr92;
		case 800: goto tr761;
		case 808: goto tr762;
		case 809: goto tr763;
		case 1033: goto tr888;
		case 1034: goto tr92;
		case 1056: goto tr888;
		case 1064: goto tr889;
		case 1065: goto tr890;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr111;
	} else if ( _widec >= 640 )
		goto tr91;
	goto tr79;
tr891:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st246;
tr724:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st246;
st246:
	if ( ++p == pe )
		goto _test_eof246;
case 246:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr892;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr891;
	goto tr71;
tr732:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 247; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 247; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 247; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 247; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 247; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 247; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 247; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 247; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 247; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 247; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 247; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 247; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 247; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 247; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 247; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 247; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 247; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 247; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 247; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 247; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 247; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 247; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 247; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 247; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 247; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 247; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 247; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 247; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 247; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st247;
tr725:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 247; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 247; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 247; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 247; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 247; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 247; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 247; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 247; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 247; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 247; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 247; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 247; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 247; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 247; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 247; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 247; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 247; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 247; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 247; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 247; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 247; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 247; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 247; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 247; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 247; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 247; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 247; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 247; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 247; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st247;
st247:
	if ( ++p == pe )
		goto _test_eof247;
case 247:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr893;
		case 32: goto tr893;
		case 40: goto tr894;
		case 41: goto tr895;
		case 58: goto tr69;
		case 65: goto tr118;
		case 67: goto tr119;
		case 68: goto tr120;
		case 69: goto tr121;
		case 72: goto tr122;
		case 73: goto tr123;
		case 75: goto tr124;
		case 76: goto tr125;
		case 77: goto tr126;
		case 78: goto tr127;
		case 80: goto tr128;
		case 82: goto tr129;
		case 83: goto tr130;
		case 84: goto tr131;
		case 85: goto tr132;
		case 92: goto st9;
		case 97: goto tr118;
		case 99: goto tr119;
		case 100: goto tr120;
		case 101: goto tr121;
		case 104: goto tr122;
		case 105: goto tr123;
		case 107: goto tr124;
		case 108: goto tr125;
		case 109: goto tr126;
		case 110: goto tr127;
		case 112: goto tr128;
		case 114: goto tr129;
		case 115: goto tr130;
		case 116: goto tr131;
		case 117: goto tr132;
		case 1802: goto tr83;
		case 1851: goto tr84;
		case 2058: goto tr822;
		case 2107: goto tr896;
		case 2314: goto tr101;
		case 2363: goto tr102;
		case 2570: goto tr822;
		case 2619: goto tr897;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr69;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr69;
		} else if ( _widec >= 48 )
			goto tr117;
	} else
		goto tr69;
	goto tr114;
tr733:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 248; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 248; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 248; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 248; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 248; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 248; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 248; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 248; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 248; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 248; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 248; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 248; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 248; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 248; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 248; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 248; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 248; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 248; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 248; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 248; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 248; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 248; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 248; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 248; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 248; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 248; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 248; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 248; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 248; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st248;
tr726:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 248; goto st632;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 248; goto st634;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 248; goto st636;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 248; goto st668;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 248; goto st673;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 248; goto st678;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 248; goto st683;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 248; goto st687;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 248; goto st689;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 248; goto st744;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 248; goto st755;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 248; goto st772;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 248; goto st783;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 248; goto st794;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 248; goto st807;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 248; goto st817;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 248; goto st856;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 248; goto st1010;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 248; goto st1013;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 248; goto st1024;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 248; goto st1026;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 248; goto st1055;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 248; goto st1068;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 248; goto st1086;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 248; goto st1081;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 248; goto st1099;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 248; goto st1105;}
		case KNOT_RRTYPE_URI:
			{stack[top++] = 248; goto st1111;}
		case KNOT_RRTYPE_CAA:
			{stack[top++] = 248; goto st1119;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st268;}
		}
	}
	goto st248;
st248:
	if ( ++p == pe )
		goto _test_eof248;
case 248:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr80;
		case 778: goto tr83;
		case 800: goto tr80;
		case 808: goto tr81;
		case 809: goto tr82;
		case 827: goto tr84;
		case 1033: goto tr898;
		case 1034: goto tr825;
		case 1056: goto tr898;
		case 1064: goto tr899;
		case 1065: goto tr900;
		case 1083: goto tr835;
	}
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr891;
	goto tr79;
tr901:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st249;
tr898:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st249;
tr899:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st249;
tr900:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st249;
tr902:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st249;
tr903:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st249;
st249:
	if ( ++p == pe )
		goto _test_eof249;
case 249:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st7;
		case 778: goto tr89;
		case 800: goto st7;
		case 808: goto tr87;
		case 809: goto tr88;
		case 827: goto st8;
		case 1033: goto tr901;
		case 1034: goto tr755;
		case 1056: goto tr901;
		case 1064: goto tr902;
		case 1065: goto tr903;
		case 1083: goto tr769;
	}
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr891;
	goto tr79;
tr3635:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 250; goto st270;} }
	goto st250;
tr3654:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 250; goto st270;} }
	goto st250;
st250:
	if ( ++p == pe )
		goto _test_eof250;
case 250:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr678;
		case 32: goto tr678;
		case 40: goto tr679;
		case 41: goto tr680;
		case 65: goto st18;
		case 68: goto st22;
		case 69: goto st29;
		case 78: goto st32;
		case 97: goto st18;
		case 100: goto st22;
		case 101: goto st29;
		case 110: goto st32;
		case 1034: goto tr681;
		case 1083: goto tr682;
	}
	goto tr904;
tr3636:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 251; goto st270;} }
	goto st251;
tr3655:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 251; goto st270;} }
	goto st251;
st251:
	if ( ++p == pe )
		goto _test_eof251;
case 251:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr678;
		case 32: goto tr678;
		case 40: goto tr679;
		case 41: goto tr680;
		case 72: goto st37;
		case 78: goto st41;
		case 83: goto st49;
		case 104: goto st37;
		case 110: goto st41;
		case 115: goto st49;
		case 1034: goto tr681;
		case 1083: goto tr682;
	}
	goto tr904;
tr3637:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 252; goto st270;} }
	goto st252;
tr3656:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 252; goto st270;} }
	goto st252;
st252:
	if ( ++p == pe )
		goto _test_eof252;
case 252:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr678;
		case 32: goto tr678;
		case 40: goto tr679;
		case 41: goto tr680;
		case 85: goto st51;
		case 117: goto st51;
		case 1034: goto tr681;
		case 1083: goto tr682;
	}
	goto tr904;
tr3638:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 253; goto st270;} }
	goto st253;
tr3657:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 253; goto st270;} }
	goto st253;
st253:
	if ( ++p == pe )
		goto _test_eof253;
case 253:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr678;
		case 32: goto tr678;
		case 40: goto tr679;
		case 41: goto tr680;
		case 73: goto st58;
		case 105: goto st58;
		case 1034: goto tr681;
		case 1083: goto tr682;
	}
	goto tr904;
tr3658:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 254; goto st270;} }
	goto st254;
st254:
	if ( ++p == pe )
		goto _test_eof254;
case 254:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr678;
		case 32: goto tr678;
		case 40: goto tr679;
		case 41: goto tr680;
		case 78: goto st144;
		case 80: goto st66;
		case 110: goto st144;
		case 112: goto st66;
		case 1034: goto tr681;
		case 1083: goto tr682;
	}
	goto tr904;
tr3640:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 255; goto st270;} }
	goto st255;
tr3659:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 255; goto st270;} }
	goto st255;
st255:
	if ( ++p == pe )
		goto _test_eof255;
case 255:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr678;
		case 32: goto tr678;
		case 40: goto tr679;
		case 41: goto tr680;
		case 69: goto st74;
		case 88: goto st76;
		case 101: goto st74;
		case 120: goto st76;
		case 1034: goto tr681;
		case 1083: goto tr682;
	}
	goto tr904;
tr3641:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 256; goto st270;} }
	goto st256;
tr3660:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 256; goto st270;} }
	goto st256;
st256:
	if ( ++p == pe )
		goto _test_eof256;
case 256:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr678;
		case 32: goto tr678;
		case 40: goto tr679;
		case 41: goto tr680;
		case 51: goto st78;
		case 54: goto st80;
		case 79: goto st82;
		case 80: goto st84;
		case 111: goto st82;
		case 112: goto st84;
		case 1034: goto tr681;
		case 1083: goto tr682;
	}
	goto tr904;
tr3642:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 257; goto st270;} }
	goto st257;
tr3661:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 257; goto st270;} }
	goto st257;
st257:
	if ( ++p == pe )
		goto _test_eof257;
case 257:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr678;
		case 32: goto tr678;
		case 40: goto tr679;
		case 41: goto tr680;
		case 73: goto st86;
		case 88: goto st90;
		case 105: goto st86;
		case 120: goto st90;
		case 1034: goto tr681;
		case 1083: goto tr682;
	}
	goto tr904;
tr3643:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 258; goto st270;} }
	goto st258;
tr3662:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 258; goto st270;} }
	goto st258;
st258:
	if ( ++p == pe )
		goto _test_eof258;
case 258:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr678;
		case 32: goto tr678;
		case 40: goto tr679;
		case 41: goto tr680;
		case 65: goto st92;
		case 73: goto st96;
		case 83: goto st98;
		case 97: goto st92;
		case 105: goto st96;
		case 115: goto st98;
		case 1034: goto tr681;
		case 1083: goto tr682;
	}
	goto tr904;
tr3644:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 259; goto st270;} }
	goto st259;
tr3663:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 259; goto st270;} }
	goto st259;
st259:
	if ( ++p == pe )
		goto _test_eof259;
case 259:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr678;
		case 32: goto tr678;
		case 40: goto tr679;
		case 41: goto tr680;
		case 84: goto st108;
		case 116: goto st108;
		case 1034: goto tr681;
		case 1083: goto tr682;
	}
	goto tr904;
tr3645:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 260; goto st270;} }
	goto st260;
tr3664:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 260; goto st270;} }
	goto st260;
st260:
	if ( ++p == pe )
		goto _test_eof260;
case 260:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr678;
		case 32: goto tr678;
		case 40: goto tr679;
		case 41: goto tr680;
		case 80: goto st111;
		case 82: goto st112;
		case 84: goto st116;
		case 112: goto st111;
		case 114: goto st112;
		case 116: goto st116;
		case 1034: goto tr681;
		case 1083: goto tr682;
	}
	goto tr904;
tr3646:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 261; goto st270;} }
	goto st261;
tr3665:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 261; goto st270;} }
	goto st261;
st261:
	if ( ++p == pe )
		goto _test_eof261;
case 261:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr678;
		case 32: goto tr678;
		case 40: goto tr679;
		case 41: goto tr680;
		case 79: goto st118;
		case 80: goto st120;
		case 82: goto st122;
		case 83: goto st124;
		case 111: goto st118;
		case 112: goto st120;
		case 114: goto st122;
		case 115: goto st124;
		case 1034: goto tr681;
		case 1083: goto tr682;
	}
	goto tr904;
tr3647:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 262; goto st270;} }
	goto st262;
tr3666:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 262; goto st270;} }
	goto st262;
st262:
	if ( ++p == pe )
		goto _test_eof262;
case 262:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr678;
		case 32: goto tr678;
		case 40: goto tr679;
		case 41: goto tr680;
		case 76: goto st129;
		case 88: goto st132;
		case 89: goto st134;
		case 108: goto st129;
		case 120: goto st132;
		case 121: goto st134;
		case 1034: goto tr681;
		case 1083: goto tr682;
	}
	goto tr904;
tr3648:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 263; goto st270;} }
	goto st263;
tr3667:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 263; goto st270;} }
	goto st263;
st263:
	if ( ++p == pe )
		goto _test_eof263;
case 263:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr678;
		case 32: goto tr678;
		case 40: goto tr679;
		case 41: goto tr680;
		case 82: goto st139;
		case 114: goto st139;
		case 1034: goto tr681;
		case 1083: goto tr682;
	}
	goto tr904;
tr905:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st264;
tr3617:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st264;
tr3677:
	{
		NOERR;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st264;
st264:
	if ( ++p == pe )
		goto _test_eof264;
case 264:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr665;
		case 1034: goto tr906;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr905;
	} else if ( _widec >= 640 )
		goto tr664;
	goto tr85;
tr907:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st265;
tr3650:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st265;
st265:
	if ( ++p == pe )
		goto _test_eof265;
case 265:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr665;
		case 1034: goto tr908;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr907;
	} else if ( _widec >= 640 )
		goto tr664;
	goto tr145;
tr3639:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 266; goto st270;} }
	goto st266;
st266:
	if ( ++p == pe )
		goto _test_eof266;
case 266:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr678;
		case 32: goto tr678;
		case 40: goto tr679;
		case 41: goto tr680;
		case 78: goto st63;
		case 80: goto st66;
		case 110: goto st63;
		case 112: goto st66;
		case 1034: goto tr681;
		case 1083: goto tr682;
	}
	goto tr904;
tr909:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st267;
tr144:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st267;
st267:
	if ( ++p == pe )
		goto _test_eof267;
case 267:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr92;
		case 1034: goto tr910;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr909;
	} else if ( _widec >= 640 )
		goto tr91;
	goto tr145;
st268:
	if ( ++p == pe )
		goto _test_eof268;
case 268:
	if ( (*p) == 10 )
		goto tr912;
	goto tr911;
tr911:
	{
		s->buffer_length = 0;
	}
	{
		if ((*p) == '\r') {
			ERR(ZS_DOS_NEWLINE);
		}

		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st269;
tr913:
	{
		if ((*p) == '\r') {
			ERR(ZS_DOS_NEWLINE);
		}

		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st269;
st269:
	if ( ++p == pe )
		goto _test_eof269;
case 269:
	if ( (*p) == 10 )
		goto tr914;
	goto tr913;
tr912:
	cs = 1137;
	{
		s->buffer_length = 0;
	}
	{
		// Terminate the error context string.
		s->buffer[s->buffer_length++] = 0;

		// Error counter incrementation.
		s->error.counter++;

		// Initialize the fcall stack.
		top = 0;

		// Reset the multiline context.
		s->multiline = false;

		s->state = ZS_STATE_ERROR;

		// Execute the error callback.
		if (s->process.automatic) {
			if (s->process.error != NULL) {
				s->process.error(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; goto _out;}
				}
			}

			// Stop the scanner if fatal error.
			if (s->error.fatal) {
				{p++; goto _out;}
			}
			{goto st1127;}
		} else {
			// Return if external processing.
			p--; cs = 1127; {p++; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto _again;
tr914:
	cs = 1137;
	{
		// Terminate the error context string.
		s->buffer[s->buffer_length++] = 0;

		// Error counter incrementation.
		s->error.counter++;

		// Initialize the fcall stack.
		top = 0;

		// Reset the multiline context.
		s->multiline = false;

		s->state = ZS_STATE_ERROR;

		// Execute the error callback.
		if (s->process.automatic) {
			if (s->process.error != NULL) {
				s->process.error(s);

				// Stop if required from the callback.
				if (s->state == ZS_STATE_STOP) {
					{p++; goto _out;}
				}
			}

			// Stop the scanner if fatal error.
			if (s->error.fatal) {
				{p++; goto _out;}
			}
			{goto st1127;}
		} else {
			// Return if external processing.
			p--; cs = 1127; {p++; goto _out;}
		}
	}
	{
		s->line_counter++;
	}
	goto _again;
st1137:
	if ( ++p == pe )
		goto _test_eof1137;
case 1137:
	goto st0;
st270:
	if ( ++p == pe )
		goto _test_eof270;
case 270:
	switch( (*p) ) {
		case 42: goto tr916;
		case 46: goto tr917;
		case 64: goto st278;
		case 92: goto tr919;
		case 95: goto tr916;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr916;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr916;
	} else
		goto tr916;
	goto tr915;
tr916:
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st271;
tr921:
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st271;
tr925:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st271;
tr932:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st271;
st271:
	if ( ++p == pe )
		goto _test_eof271;
case 271:
	switch( (*p) ) {
		case 32: goto tr920;
		case 42: goto tr921;
		case 46: goto tr922;
		case 59: goto tr920;
		case 92: goto st273;
		case 95: goto tr921;
	}
	if ( (*p) < 45 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr920;
		} else if ( (*p) >= 9 )
			goto tr920;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 90 ) {
			if ( 97 <= (*p) && (*p) <= 122 )
				goto tr921;
		} else if ( (*p) >= 65 )
			goto tr921;
	} else
		goto tr921;
	goto tr915;
tr920:
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		// Check for (relative + origin) dname length overflow.
		if (s->dname_tmp_length + s->zone_origin_length <= MAX_DNAME_LENGTH) {
			memcpy(s->dname + s->dname_tmp_length,
			       s->zone_origin,
			       s->zone_origin_length);

			s->dname_tmp_length += s->zone_origin_length;
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1138;
tr924:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1138;
tr931:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		// Check for (relative + origin) dname length overflow.
		if (s->dname_tmp_length + s->zone_origin_length <= MAX_DNAME_LENGTH) {
			memcpy(s->dname + s->dname_tmp_length,
			       s->zone_origin,
			       s->zone_origin_length);

			s->dname_tmp_length += s->zone_origin_length;
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1138;
tr935:
	{
		// Copy already verified zone origin.
		memcpy(s->dname,
		       s->zone_origin,
		       s->zone_origin_length);

		s->dname_tmp_length = s->zone_origin_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1138;
st1138:
	if ( ++p == pe )
		goto _test_eof1138;
case 1138:
	goto st0;
tr922:
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st272;
tr933:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st272;
st272:
	if ( ++p == pe )
		goto _test_eof272;
case 272:
	switch( (*p) ) {
		case 32: goto tr924;
		case 42: goto tr925;
		case 45: goto tr925;
		case 59: goto tr924;
		case 92: goto tr926;
		case 95: goto tr925;
	}
	if ( (*p) < 47 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr924;
		} else if ( (*p) >= 9 )
			goto tr924;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 90 ) {
			if ( 97 <= (*p) && (*p) <= 122 )
				goto tr925;
		} else if ( (*p) >= 65 )
			goto tr925;
	} else
		goto tr925;
	goto tr915;
tr919:
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st273;
tr926:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st273;
tr934:
	{
		s->dname_tmp_length++;
	}
	goto st273;
st273:
	if ( ++p == pe )
		goto _test_eof273;
case 273:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr927;
	goto tr921;
tr927:
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length] = 0;
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st274;
st274:
	if ( ++p == pe )
		goto _test_eof274;
case 274:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr929;
	goto tr928;
tr929:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st275;
st275:
	if ( ++p == pe )
		goto _test_eof275;
case 275:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr930;
	goto tr928;
tr930:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st276;
st276:
	if ( ++p == pe )
		goto _test_eof276;
case 276:
	switch( (*p) ) {
		case 32: goto tr931;
		case 42: goto tr932;
		case 46: goto tr933;
		case 59: goto tr931;
		case 92: goto tr934;
		case 95: goto tr932;
	}
	if ( (*p) < 45 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr931;
		} else if ( (*p) >= 9 )
			goto tr931;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 90 ) {
			if ( 97 <= (*p) && (*p) <= 122 )
				goto tr932;
		} else if ( (*p) >= 65 )
			goto tr932;
	} else
		goto tr932;
	goto tr928;
tr917:
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	goto st277;
st277:
	if ( ++p == pe )
		goto _test_eof277;
case 277:
	switch( (*p) ) {
		case 32: goto tr924;
		case 59: goto tr924;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr924;
	} else if ( (*p) >= 9 )
		goto tr924;
	goto tr915;
st278:
	if ( ++p == pe )
		goto _test_eof278;
case 278:
	switch( (*p) ) {
		case 32: goto tr935;
		case 59: goto tr935;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr935;
	} else if ( (*p) >= 9 )
		goto tr935;
	goto tr915;
st279:
	if ( ++p == pe )
		goto _test_eof279;
case 279:
	switch( (*p) ) {
		case 34: goto st285;
		case 92: goto st281;
	}
	if ( (*p) > 58 ) {
		if ( 60 <= (*p) && (*p) <= 126 )
			goto tr937;
	} else if ( (*p) >= 33 )
		goto tr937;
	goto tr936;
tr937:
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;

				if (rdata_tail > rdata_stop) {
					WARN(ZS_TEXT_OVERFLOW);
					p--; {goto st268;}
				}
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st280;
tr947:
	{
		rdata_tail++;
	}
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;

				if (rdata_tail > rdata_stop) {
					WARN(ZS_TEXT_OVERFLOW);
					p--; {goto st268;}
				}
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st280;
st280:
	if ( ++p == pe )
		goto _test_eof280;
case 280:
	switch( (*p) ) {
		case 32: goto tr940;
		case 33: goto tr937;
		case 59: goto tr940;
		case 92: goto st281;
	}
	if ( (*p) < 35 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr940;
	} else if ( (*p) > 39 ) {
		if ( (*p) > 41 ) {
			if ( 42 <= (*p) && (*p) <= 126 )
				goto tr937;
		} else if ( (*p) >= 40 )
			goto tr941;
	} else
		goto tr937;
	goto tr936;
tr940:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1139;
tr946:
	{
		rdata_tail++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1139;
st1139:
	if ( ++p == pe )
		goto _test_eof1139;
case 1139:
	goto st0;
tr941:
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;

				if (rdata_tail > rdata_stop) {
					WARN(ZS_TEXT_OVERFLOW);
					p--; {goto st268;}
				}
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1140;
tr948:
	{
		rdata_tail++;
	}
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;

				if (rdata_tail > rdata_stop) {
					WARN(ZS_TEXT_OVERFLOW);
					p--; {goto st268;}
				}
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1140;
st1140:
	if ( ++p == pe )
		goto _test_eof1140;
case 1140:
	switch( (*p) ) {
		case 32: goto tr940;
		case 33: goto tr937;
		case 59: goto tr940;
		case 92: goto st281;
	}
	if ( (*p) < 35 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr940;
	} else if ( (*p) > 39 ) {
		if ( (*p) > 41 ) {
			if ( 42 <= (*p) && (*p) <= 126 )
				goto tr937;
		} else if ( (*p) >= 40 )
			goto tr941;
	} else
		goto tr937;
	goto tr936;
tr949:
	{
		rdata_tail++;
	}
	goto st281;
st281:
	if ( ++p == pe )
		goto _test_eof281;
case 281:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr942;
	goto tr937;
tr942:
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;

				if (rdata_tail > rdata_stop) {
					WARN(ZS_TEXT_OVERFLOW);
					p--; {goto st268;}
				}
			}

			*rdata_tail = 0;
			s->item_length++;
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st282;
st282:
	if ( ++p == pe )
		goto _test_eof282;
case 282:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr944;
	goto tr943;
tr944:
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st283;
st283:
	if ( ++p == pe )
		goto _test_eof283;
case 283:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr945;
	goto tr943;
tr945:
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st284;
st284:
	if ( ++p == pe )
		goto _test_eof284;
case 284:
	switch( (*p) ) {
		case 32: goto tr946;
		case 33: goto tr947;
		case 59: goto tr946;
		case 92: goto tr949;
	}
	if ( (*p) < 35 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr946;
	} else if ( (*p) > 39 ) {
		if ( (*p) > 41 ) {
			if ( 42 <= (*p) && (*p) <= 126 )
				goto tr947;
		} else if ( (*p) >= 40 )
			goto tr948;
	} else
		goto tr947;
	goto tr943;
tr950:
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;

				if (rdata_tail > rdata_stop) {
					WARN(ZS_TEXT_OVERFLOW);
					p--; {goto st268;}
				}
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st285;
tr957:
	{
		rdata_tail++;
	}
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;

				if (rdata_tail > rdata_stop) {
					WARN(ZS_TEXT_OVERFLOW);
					p--; {goto st268;}
				}
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st285;
st285:
	if ( ++p == pe )
		goto _test_eof285;
case 285:
	_widec = (*p);
	if ( 10 <= (*p) && (*p) <= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr950;
		case 34: goto st286;
		case 92: goto st287;
		case 522: goto tr950;
	}
	if ( 32 <= _widec && _widec <= 126 )
		goto tr950;
	goto tr936;
tr958:
	{
		rdata_tail++;
	}
	goto st286;
st286:
	if ( ++p == pe )
		goto _test_eof286;
case 286:
	switch( (*p) ) {
		case 32: goto tr940;
		case 59: goto tr940;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr940;
	} else if ( (*p) >= 9 )
		goto tr940;
	goto tr953;
tr959:
	{
		rdata_tail++;
	}
	goto st287;
st287:
	if ( ++p == pe )
		goto _test_eof287;
case 287:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr954;
	goto tr950;
tr954:
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;

				if (rdata_tail > rdata_stop) {
					WARN(ZS_TEXT_OVERFLOW);
					p--; {goto st268;}
				}
			}

			*rdata_tail = 0;
			s->item_length++;
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st288;
st288:
	if ( ++p == pe )
		goto _test_eof288;
case 288:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr955;
	goto tr943;
tr955:
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st289;
st289:
	if ( ++p == pe )
		goto _test_eof289;
case 289:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr956;
	goto tr943;
tr956:
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st290;
st290:
	if ( ++p == pe )
		goto _test_eof290;
case 290:
	_widec = (*p);
	if ( 10 <= (*p) && (*p) <= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr957;
		case 34: goto tr958;
		case 92: goto tr959;
		case 522: goto tr957;
	}
	if ( 32 <= _widec && _widec <= 126 )
		goto tr957;
	goto tr943;
st291:
	if ( ++p == pe )
		goto _test_eof291;
case 291:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st292;
		case 32: goto st292;
		case 40: goto tr962;
		case 41: goto tr963;
		case 1034: goto tr964;
		case 1083: goto st299;
	}
	goto tr960;
tr962:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st292;
tr963:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st292;
tr964:
	{
		s->line_counter++;
	}
	goto st292;
tr1003:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st292;
st292:
	if ( ++p == pe )
		goto _test_eof292;
case 292:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st292;
		case 32: goto st292;
		case 40: goto tr962;
		case 41: goto tr963;
		case 1034: goto tr964;
		case 1083: goto st299;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr967;
	goto tr966;
tr967:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st293;
tr972:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st293;
st293:
	if ( ++p == pe )
		goto _test_eof293;
case 293:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr969;
		case 32: goto tr969;
		case 40: goto tr970;
		case 41: goto tr971;
		case 68: goto tr973;
		case 72: goto tr974;
		case 77: goto tr975;
		case 83: goto st296;
		case 87: goto tr977;
		case 100: goto tr973;
		case 104: goto tr974;
		case 109: goto tr975;
		case 115: goto st296;
		case 119: goto tr977;
		case 778: goto tr978;
		case 827: goto tr979;
		case 1034: goto tr978;
		case 1083: goto tr979;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr972;
	goto tr968;
tr982:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st294;
tr983:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st294;
tr969:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st294;
tr970:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st294;
tr971:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st294;
tr996:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st294;
tr997:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st294;
tr998:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st294;
st294:
	if ( ++p == pe )
		goto _test_eof294;
case 294:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st294;
		case 32: goto st294;
		case 40: goto tr982;
		case 41: goto tr983;
		case 778: goto tr984;
		case 827: goto st295;
		case 1034: goto tr984;
		case 1083: goto st295;
	}
	goto tr980;
tr978:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->default_ttl = (uint32_t)(s->number64);
		} else {
			ERR(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1141;
tr984:
	{
		if (s->number64 <= UINT32_MAX) {
			s->default_ttl = (uint32_t)(s->number64);
		} else {
			ERR(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1141;
tr987:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->default_ttl = (uint32_t)(s->number64);
		} else {
			ERR(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1141;
tr1000:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->default_ttl = (uint32_t)(s->number64);
		} else {
			ERR(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1141;
st1141:
	if ( ++p == pe )
		goto _test_eof1141;
case 1141:
	goto st0;
tr979:
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st295;
tr986:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st295;
tr1001:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st295;
st295:
	if ( ++p == pe )
		goto _test_eof295;
case 295:
	if ( (*p) == 10 )
		goto tr987;
	goto tr986;
tr973:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st296;
tr974:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st296;
tr975:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st296;
tr977:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st296;
st296:
	if ( ++p == pe )
		goto _test_eof296;
case 296:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr969;
		case 32: goto tr969;
		case 40: goto tr970;
		case 41: goto tr971;
		case 778: goto tr978;
		case 827: goto tr979;
		case 1034: goto tr978;
		case 1083: goto tr979;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr988;
	goto tr968;
tr990:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st297;
tr988:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st297;
tr999:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st297;
st297:
	if ( ++p == pe )
		goto _test_eof297;
case 297:
	switch( (*p) ) {
		case 68: goto tr991;
		case 72: goto tr992;
		case 77: goto tr993;
		case 83: goto st298;
		case 87: goto tr995;
		case 100: goto tr991;
		case 104: goto tr992;
		case 109: goto tr993;
		case 115: goto st298;
		case 119: goto tr995;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr990;
	goto tr989;
tr991:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st298;
tr992:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st298;
tr993:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st298;
tr995:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st298;
st298:
	if ( ++p == pe )
		goto _test_eof298;
case 298:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr996;
		case 32: goto tr996;
		case 40: goto tr997;
		case 41: goto tr998;
		case 778: goto tr1000;
		case 827: goto tr1001;
		case 1034: goto tr1000;
		case 1083: goto tr1001;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr999;
	goto tr968;
tr1002:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st299;
st299:
	if ( ++p == pe )
		goto _test_eof299;
case 299:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1003;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1002;
	goto tr960;
st300:
	if ( ++p == pe )
		goto _test_eof300;
case 300:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st301;
		case 32: goto st301;
		case 40: goto tr1006;
		case 41: goto tr1007;
		case 1034: goto tr1008;
		case 1083: goto st311;
	}
	goto tr1004;
tr1006:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st301;
tr1007:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st301;
tr1008:
	{
		s->line_counter++;
	}
	goto st301;
tr1039:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st301;
st301:
	if ( ++p == pe )
		goto _test_eof301;
case 301:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st301;
		case 32: goto st301;
		case 40: goto tr1006;
		case 41: goto tr1007;
		case 42: goto tr1010;
		case 46: goto tr1011;
		case 92: goto tr1012;
		case 95: goto tr1010;
		case 1034: goto tr1008;
		case 1083: goto st311;
	}
	if ( _widec < 65 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr1010;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1010;
	} else
		goto tr1010;
	goto tr1004;
tr1013:
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st302;
tr1020:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st302;
tr1035:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st302;
tr1010:
	{
		s->dname = s->zone_origin;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st302;
st302:
	if ( ++p == pe )
		goto _test_eof302;
case 302:
	switch( (*p) ) {
		case 42: goto tr1013;
		case 46: goto tr1014;
		case 92: goto st306;
		case 95: goto tr1013;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1013;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1013;
	} else
		goto tr1013;
	goto tr1004;
tr1014:
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st303;
tr1036:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st303;
st303:
	if ( ++p == pe )
		goto _test_eof303;
case 303:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1017;
		case 32: goto tr1017;
		case 40: goto tr1018;
		case 41: goto tr1019;
		case 42: goto tr1020;
		case 45: goto tr1020;
		case 92: goto tr1021;
		case 95: goto tr1020;
		case 778: goto tr1022;
		case 827: goto tr1023;
		case 1034: goto tr1022;
		case 1083: goto tr1023;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr1020;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1020;
	} else
		goto tr1020;
	goto tr1016;
tr1025:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st304;
tr1026:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st304;
tr1017:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st304;
tr1018:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st304;
tr1019:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st304;
st304:
	if ( ++p == pe )
		goto _test_eof304;
case 304:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st304;
		case 32: goto st304;
		case 40: goto tr1025;
		case 41: goto tr1026;
		case 778: goto tr1027;
		case 827: goto st305;
		case 1034: goto tr1027;
		case 1083: goto st305;
	}
	goto tr1016;
tr1022:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		s->zone_origin_length = s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1142;
tr1027:
	{
		s->zone_origin_length = s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1142;
tr1030:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->zone_origin_length = s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1142;
st1142:
	if ( ++p == pe )
		goto _test_eof1142;
case 1142:
	goto st0;
tr1029:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st305;
tr1023:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st305;
st305:
	if ( ++p == pe )
		goto _test_eof305;
case 305:
	if ( (*p) == 10 )
		goto tr1030;
	goto tr1029;
tr1021:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st306;
tr1037:
	{
		s->dname_tmp_length++;
	}
	goto st306;
tr1012:
	{
		s->dname = s->zone_origin;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st306;
st306:
	if ( ++p == pe )
		goto _test_eof306;
case 306:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1031;
	goto tr1013;
tr1031:
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length] = 0;
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st307;
st307:
	if ( ++p == pe )
		goto _test_eof307;
case 307:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1033;
	goto tr1032;
tr1033:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st308;
st308:
	if ( ++p == pe )
		goto _test_eof308;
case 308:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1034;
	goto tr1032;
tr1034:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st309;
st309:
	if ( ++p == pe )
		goto _test_eof309;
case 309:
	switch( (*p) ) {
		case 42: goto tr1035;
		case 46: goto tr1036;
		case 92: goto tr1037;
		case 95: goto tr1035;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1035;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1035;
	} else
		goto tr1035;
	goto tr1032;
tr1011:
	{
		s->dname = s->zone_origin;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	goto st310;
st310:
	if ( ++p == pe )
		goto _test_eof310;
case 310:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1017;
		case 32: goto tr1017;
		case 40: goto tr1018;
		case 41: goto tr1019;
		case 778: goto tr1022;
		case 827: goto tr1023;
		case 1034: goto tr1022;
		case 1083: goto tr1023;
	}
	goto tr1016;
tr1038:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st311;
st311:
	if ( ++p == pe )
		goto _test_eof311;
case 311:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1039;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1038;
	goto tr1004;
st312:
	if ( ++p == pe )
		goto _test_eof312;
case 312:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st313;
		case 32: goto st313;
		case 40: goto tr1041;
		case 41: goto tr1042;
		case 1034: goto tr1043;
		case 1083: goto st328;
	}
	goto st0;
tr1041:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st313;
tr1042:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st313;
tr1043:
	{
		s->line_counter++;
	}
	goto st313;
tr1097:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st313;
st313:
	if ( ++p == pe )
		goto _test_eof313;
case 313:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st313;
		case 32: goto st313;
		case 40: goto tr1041;
		case 41: goto tr1042;
		case 1034: goto tr1043;
		case 1083: goto st328;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1045;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr1045;
	} else
		goto tr1045;
	goto tr1046;
tr1045:
	{
		rdata_tail = s->r_data;
	}
	{ p--; {stack[top++] = 314; goto st279;} }
	goto st314;
st314:
	if ( ++p == pe )
		goto _test_eof314;
case 314:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1048;
		case 32: goto tr1048;
		case 40: goto tr1049;
		case 41: goto tr1050;
		case 778: goto tr1051;
		case 827: goto tr1052;
		case 1034: goto tr1051;
		case 1083: goto tr1053;
	}
	goto tr1047;
tr1056:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st315;
tr1057:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st315;
tr1048:
	{
		size_t len = rdata_tail - s->r_data;
		if (len >= sizeof(s->include_filename)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st268;}
		}

		// Store zero terminated include filename.
		memcpy(s->include_filename, s->r_data, len);
		s->include_filename[len] = '\0';

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st315;
tr1049:
	{
		size_t len = rdata_tail - s->r_data;
		if (len >= sizeof(s->include_filename)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st268;}
		}

		// Store zero terminated include filename.
		memcpy(s->include_filename, s->r_data, len);
		s->include_filename[len] = '\0';

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st315;
tr1050:
	{
		size_t len = rdata_tail - s->r_data;
		if (len >= sizeof(s->include_filename)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st268;}
		}

		// Store zero terminated include filename.
		memcpy(s->include_filename, s->r_data, len);
		s->include_filename[len] = '\0';

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st315;
st315:
	if ( ++p == pe )
		goto _test_eof315;
case 315:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st315;
		case 32: goto st315;
		case 40: goto tr1056;
		case 41: goto tr1057;
		case 42: goto tr1058;
		case 46: goto tr1059;
		case 92: goto tr1060;
		case 95: goto tr1058;
		case 778: goto tr1061;
		case 827: goto st319;
		case 1034: goto tr1061;
		case 1083: goto st325;
	}
	if ( _widec < 65 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr1058;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1058;
	} else
		goto tr1058;
	goto tr1054;
tr1065:
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st316;
tr1071:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st316;
tr1084:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st316;
tr1058:
	{
		s->dname = s->r_data;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st316;
st316:
	if ( ++p == pe )
		goto _test_eof316;
case 316:
	switch( (*p) ) {
		case 42: goto tr1065;
		case 46: goto tr1066;
		case 92: goto st320;
		case 95: goto tr1065;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1065;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1065;
	} else
		goto tr1065;
	goto tr1064;
tr1066:
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st317;
tr1085:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st317;
st317:
	if ( ++p == pe )
		goto _test_eof317;
case 317:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1068;
		case 32: goto tr1068;
		case 40: goto tr1069;
		case 41: goto tr1070;
		case 42: goto tr1071;
		case 45: goto tr1071;
		case 92: goto tr1072;
		case 95: goto tr1071;
		case 778: goto tr1073;
		case 827: goto tr1074;
		case 1034: goto tr1073;
		case 1083: goto tr1074;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr1071;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1071;
	} else
		goto tr1071;
	goto tr1054;
tr1076:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st318;
tr1077:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st318;
tr1068:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->r_data_length = s->dname_tmp_length;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st318;
tr1069:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->r_data_length = s->dname_tmp_length;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st318;
tr1070:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->r_data_length = s->dname_tmp_length;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st318;
st318:
	if ( ++p == pe )
		goto _test_eof318;
case 318:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st318;
		case 32: goto st318;
		case 40: goto tr1076;
		case 41: goto tr1077;
		case 778: goto tr1061;
		case 827: goto st319;
		case 1034: goto tr1061;
		case 1083: goto st319;
	}
	goto tr85;
tr1051:
	cs = 1143;
	{
		size_t len = rdata_tail - s->r_data;
		if (len >= sizeof(s->include_filename)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st268;}
		}

		// Store zero terminated include filename.
		memcpy(s->include_filename, s->r_data, len);
		s->include_filename[len] = '\0';

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		// Extend relative file path.
		if (s->include_filename[0] != '/') {
			ret = snprintf((char *)(s->buffer), sizeof(s->buffer),
			               "%s/%s", s->path, s->include_filename);
			if (ret <= 0 || ret > sizeof(s->buffer)) {
				ERR(ZS_BAD_INCLUDE_FILENAME);
				p--; {goto st268;}
			}
			memcpy(s->include_filename, s->buffer, ret);
		}

		// Origin conversion from wire to text form in \DDD notation.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
			                  (char *)s->buffer);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
			                  (char *)s->buffer);
		}

		// Let the caller to solve the include.
		if (s->process.automatic) {
			// Create new scanner for included zone file.
			zs_scanner_t *ss = malloc(sizeof(zs_scanner_t));
			if (ss == NULL) {
				ERR(ZS_UNPROCESSED_INCLUDE);
				p--; {goto st268;}
			}

			// Parse included zone file.
			if (zs_init(ss, (char *)s->buffer, s->default_class,
			            s->default_ttl) != 0 ||
			    zs_set_input_file(ss, (char *)(s->include_filename)) != 0 ||
			    zs_set_processing(ss, s->process.record, s->process.error,
			                      s->process.data) != 0 ||
			    zs_parse_all(ss) != 0) {
				// File internal errors are handled by error callback.
				if (ss->error.counter > 0) {
					s->error.counter += ss->error.counter;
					ERR(ZS_UNPROCESSED_INCLUDE);
				// General include file error.
				} else {
					ERR(ss->error.code);
				}
				zs_deinit(ss);
				free(ss);
				p--; {goto st268;}
			}
			zs_deinit(ss);
			free(ss);
		} else {
			s->state = ZS_STATE_INCLUDE;
			p--; cs = 1127; {p++; goto _out;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto _again;
tr1061:
	cs = 1143;
	{
		// Extend relative file path.
		if (s->include_filename[0] != '/') {
			ret = snprintf((char *)(s->buffer), sizeof(s->buffer),
			               "%s/%s", s->path, s->include_filename);
			if (ret <= 0 || ret > sizeof(s->buffer)) {
				ERR(ZS_BAD_INCLUDE_FILENAME);
				p--; {goto st268;}
			}
			memcpy(s->include_filename, s->buffer, ret);
		}

		// Origin conversion from wire to text form in \DDD notation.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
			                  (char *)s->buffer);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
			                  (char *)s->buffer);
		}

		// Let the caller to solve the include.
		if (s->process.automatic) {
			// Create new scanner for included zone file.
			zs_scanner_t *ss = malloc(sizeof(zs_scanner_t));
			if (ss == NULL) {
				ERR(ZS_UNPROCESSED_INCLUDE);
				p--; {goto st268;}
			}

			// Parse included zone file.
			if (zs_init(ss, (char *)s->buffer, s->default_class,
			            s->default_ttl) != 0 ||
			    zs_set_input_file(ss, (char *)(s->include_filename)) != 0 ||
			    zs_set_processing(ss, s->process.record, s->process.error,
			                      s->process.data) != 0 ||
			    zs_parse_all(ss) != 0) {
				// File internal errors are handled by error callback.
				if (ss->error.counter > 0) {
					s->error.counter += ss->error.counter;
					ERR(ZS_UNPROCESSED_INCLUDE);
				// General include file error.
				} else {
					ERR(ss->error.code);
				}
				zs_deinit(ss);
				free(ss);
				p--; {goto st268;}
			}
			zs_deinit(ss);
			free(ss);
		} else {
			s->state = ZS_STATE_INCLUDE;
			p--; cs = 1127; {p++; goto _out;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto _again;
tr1073:
	cs = 1143;
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->r_data_length = s->dname_tmp_length;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	{
		// Extend relative file path.
		if (s->include_filename[0] != '/') {
			ret = snprintf((char *)(s->buffer), sizeof(s->buffer),
			               "%s/%s", s->path, s->include_filename);
			if (ret <= 0 || ret > sizeof(s->buffer)) {
				ERR(ZS_BAD_INCLUDE_FILENAME);
				p--; {goto st268;}
			}
			memcpy(s->include_filename, s->buffer, ret);
		}

		// Origin conversion from wire to text form in \DDD notation.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
			                  (char *)s->buffer);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
			                  (char *)s->buffer);
		}

		// Let the caller to solve the include.
		if (s->process.automatic) {
			// Create new scanner for included zone file.
			zs_scanner_t *ss = malloc(sizeof(zs_scanner_t));
			if (ss == NULL) {
				ERR(ZS_UNPROCESSED_INCLUDE);
				p--; {goto st268;}
			}

			// Parse included zone file.
			if (zs_init(ss, (char *)s->buffer, s->default_class,
			            s->default_ttl) != 0 ||
			    zs_set_input_file(ss, (char *)(s->include_filename)) != 0 ||
			    zs_set_processing(ss, s->process.record, s->process.error,
			                      s->process.data) != 0 ||
			    zs_parse_all(ss) != 0) {
				// File internal errors are handled by error callback.
				if (ss->error.counter > 0) {
					s->error.counter += ss->error.counter;
					ERR(ZS_UNPROCESSED_INCLUDE);
				// General include file error.
				} else {
					ERR(ss->error.code);
				}
				zs_deinit(ss);
				free(ss);
				p--; {goto st268;}
			}
			zs_deinit(ss);
			free(ss);
		} else {
			s->state = ZS_STATE_INCLUDE;
			p--; cs = 1127; {p++; goto _out;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto _again;
tr1079:
	cs = 1143;
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		// Extend relative file path.
		if (s->include_filename[0] != '/') {
			ret = snprintf((char *)(s->buffer), sizeof(s->buffer),
			               "%s/%s", s->path, s->include_filename);
			if (ret <= 0 || ret > sizeof(s->buffer)) {
				ERR(ZS_BAD_INCLUDE_FILENAME);
				p--; {goto st268;}
			}
			memcpy(s->include_filename, s->buffer, ret);
		}

		// Origin conversion from wire to text form in \DDD notation.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
			                  (char *)s->buffer);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
			                  (char *)s->buffer);
		}

		// Let the caller to solve the include.
		if (s->process.automatic) {
			// Create new scanner for included zone file.
			zs_scanner_t *ss = malloc(sizeof(zs_scanner_t));
			if (ss == NULL) {
				ERR(ZS_UNPROCESSED_INCLUDE);
				p--; {goto st268;}
			}

			// Parse included zone file.
			if (zs_init(ss, (char *)s->buffer, s->default_class,
			            s->default_ttl) != 0 ||
			    zs_set_input_file(ss, (char *)(s->include_filename)) != 0 ||
			    zs_set_processing(ss, s->process.record, s->process.error,
			                      s->process.data) != 0 ||
			    zs_parse_all(ss) != 0) {
				// File internal errors are handled by error callback.
				if (ss->error.counter > 0) {
					s->error.counter += ss->error.counter;
					ERR(ZS_UNPROCESSED_INCLUDE);
				// General include file error.
				} else {
					ERR(ss->error.code);
				}
				zs_deinit(ss);
				free(ss);
				p--; {goto st268;}
			}
			zs_deinit(ss);
			free(ss);
		} else {
			s->state = ZS_STATE_INCLUDE;
			p--; cs = 1127; {p++; goto _out;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto _again;
st1143:
	if ( ++p == pe )
		goto _test_eof1143;
case 1143:
	goto st0;
tr1078:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st319;
tr1052:
	{
		size_t len = rdata_tail - s->r_data;
		if (len >= sizeof(s->include_filename)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st268;}
		}

		// Store zero terminated include filename.
		memcpy(s->include_filename, s->r_data, len);
		s->include_filename[len] = '\0';

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st319;
tr1074:
	{
		// Enough room for the terminal label is guaranteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->r_data_length = s->dname_tmp_length;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st319;
st319:
	if ( ++p == pe )
		goto _test_eof319;
case 319:
	if ( (*p) == 10 )
		goto tr1079;
	goto tr1078;
tr1072:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st320;
tr1086:
	{
		s->dname_tmp_length++;
	}
	goto st320;
tr1060:
	{
		s->dname = s->r_data;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st320;
st320:
	if ( ++p == pe )
		goto _test_eof320;
case 320:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1080;
	goto tr1065;
tr1080:
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length] = 0;
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st321;
st321:
	if ( ++p == pe )
		goto _test_eof321;
case 321:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1082;
	goto tr1081;
tr1082:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st322;
st322:
	if ( ++p == pe )
		goto _test_eof322;
case 322:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1083;
	goto tr1081;
tr1083:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st323;
st323:
	if ( ++p == pe )
		goto _test_eof323;
case 323:
	switch( (*p) ) {
		case 42: goto tr1084;
		case 46: goto tr1085;
		case 92: goto tr1086;
		case 95: goto tr1084;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1084;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1084;
	} else
		goto tr1084;
	goto tr1081;
tr1059:
	{
		s->dname = s->r_data;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	goto st324;
st324:
	if ( ++p == pe )
		goto _test_eof324;
case 324:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1068;
		case 32: goto tr1068;
		case 40: goto tr1069;
		case 41: goto tr1070;
		case 778: goto tr1073;
		case 827: goto tr1074;
		case 1034: goto tr1073;
		case 1083: goto tr1074;
	}
	goto tr1054;
tr1087:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st325;
tr1053:
	{
		size_t len = rdata_tail - s->r_data;
		if (len >= sizeof(s->include_filename)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st268;}
		}

		// Store zero terminated include filename.
		memcpy(s->include_filename, s->r_data, len);
		s->include_filename[len] = '\0';

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	{
		s->buffer[0] = 0;
		s->buffer_length = 0;
	}
	goto st325;
st325:
	if ( ++p == pe )
		goto _test_eof325;
case 325:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr1079;
		case 1034: goto tr1088;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto tr1087;
	} else if ( _widec >= 640 )
		goto tr1078;
	goto tr85;
tr1088:
	cs = 1144;
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	{
		// Extend relative file path.
		if (s->include_filename[0] != '/') {
			ret = snprintf((char *)(s->buffer), sizeof(s->buffer),
			               "%s/%s", s->path, s->include_filename);
			if (ret <= 0 || ret > sizeof(s->buffer)) {
				ERR(ZS_BAD_INCLUDE_FILENAME);
				p--; {goto st268;}
			}
			memcpy(s->include_filename, s->buffer, ret);
		}

		// Origin conversion from wire to text form in \DDD notation.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
			                  (char *)s->buffer);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
			                  (char *)s->buffer);
		}

		// Let the caller to solve the include.
		if (s->process.automatic) {
			// Create new scanner for included zone file.
			zs_scanner_t *ss = malloc(sizeof(zs_scanner_t));
			if (ss == NULL) {
				ERR(ZS_UNPROCESSED_INCLUDE);
				p--; {goto st268;}
			}

			// Parse included zone file.
			if (zs_init(ss, (char *)s->buffer, s->default_class,
			            s->default_ttl) != 0 ||
			    zs_set_input_file(ss, (char *)(s->include_filename)) != 0 ||
			    zs_set_processing(ss, s->process.record, s->process.error,
			                      s->process.data) != 0 ||
			    zs_parse_all(ss) != 0) {
				// File internal errors are handled by error callback.
				if (ss->error.counter > 0) {
					s->error.counter += ss->error.counter;
					ERR(ZS_UNPROCESSED_INCLUDE);
				// General include file error.
				} else {
					ERR(ss->error.code);
				}
				zs_deinit(ss);
				free(ss);
				p--; {goto st268;}
			}
			zs_deinit(ss);
			free(ss);
		} else {
			s->state = ZS_STATE_INCLUDE;
			p--; cs = 1127; {p++; goto _out;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto _again;
st1144:
	if ( ++p == pe )
		goto _test_eof1144;
case 1144:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st326;
		case 32: goto st326;
		case 40: goto tr1090;
		case 41: goto tr1091;
		case 42: goto tr1058;
		case 46: goto tr1059;
		case 92: goto tr1060;
		case 95: goto tr1058;
		case 1034: goto tr1092;
		case 1083: goto st327;
	}
	if ( _widec < 65 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr1058;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1058;
	} else
		goto tr1058;
	goto tr1064;
tr1090:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st326;
tr1091:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st326;
tr1092:
	{
		s->line_counter++;
	}
	goto st326;
tr1095:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st326;
st326:
	if ( ++p == pe )
		goto _test_eof326;
case 326:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st326;
		case 32: goto st326;
		case 40: goto tr1090;
		case 41: goto tr1091;
		case 42: goto tr1058;
		case 46: goto tr1059;
		case 92: goto tr1060;
		case 95: goto tr1058;
		case 1034: goto tr1092;
		case 1083: goto st327;
	}
	if ( _widec < 65 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr1058;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1058;
	} else
		goto tr1058;
	goto tr1064;
tr1094:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st327;
st327:
	if ( ++p == pe )
		goto _test_eof327;
case 327:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1095;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1094;
	goto st0;
tr1096:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st328;
st328:
	if ( ++p == pe )
		goto _test_eof328;
case 328:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1097;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1096;
	goto st0;
st329:
	if ( ++p == pe )
		goto _test_eof329;
case 329:
	if ( (*p) == 43 )
		goto tr1099;
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr1099;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1099;
	} else
		goto tr1099;
	goto tr1098;
tr1099:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_base64_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st330;
st330:
	if ( ++p == pe )
		goto _test_eof330;
case 330:
	if ( (*p) == 43 )
		goto tr1100;
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr1100;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1100;
	} else
		goto tr1100;
	goto tr1098;
tr1100:
	{
		*(rdata_tail++) += second_left_base64_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = second_right_base64_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st331;
st331:
	if ( ++p == pe )
		goto _test_eof331;
case 331:
	switch( (*p) ) {
		case 43: goto tr1101;
		case 61: goto st335;
	}
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr1101;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1101;
	} else
		goto tr1101;
	goto tr1098;
tr1101:
	{
		*(rdata_tail++) += third_left_base64_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = third_right_base64_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st332;
st332:
	if ( ++p == pe )
		goto _test_eof332;
case 332:
	switch( (*p) ) {
		case 43: goto tr1103;
		case 61: goto st333;
	}
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr1103;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1103;
	} else
		goto tr1103;
	goto tr1098;
tr1105:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st333;
tr1106:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st333;
tr1107:
	{
		s->line_counter++;
	}
	goto st333;
tr1113:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st333;
tr1103:
	{
		*(rdata_tail++) += fourth_base64_to_num[(uint8_t)(*p)];
	}
	goto st333;
st333:
	if ( ++p == pe )
		goto _test_eof333;
case 333:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st333;
		case 32: goto st333;
		case 40: goto tr1105;
		case 41: goto tr1106;
		case 43: goto tr1099;
		case 2058: goto tr1107;
		case 2107: goto st334;
		case 2314: goto tr1109;
		case 2363: goto tr1109;
		case 2570: goto tr1110;
		case 2619: goto tr1111;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr1099;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1099;
	} else
		goto tr1099;
	goto tr1098;
tr1112:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st334;
st334:
	if ( ++p == pe )
		goto _test_eof334;
case 334:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1113;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1112;
	goto tr1098;
tr1109:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1145;
st1145:
	if ( ++p == pe )
		goto _test_eof1145;
case 1145:
	goto st0;
tr1110:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1146;
st1146:
	if ( ++p == pe )
		goto _test_eof1146;
case 1146:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st333;
		case 32: goto st333;
		case 40: goto tr1105;
		case 41: goto tr1106;
		case 43: goto tr1099;
		case 2058: goto tr1107;
		case 2107: goto st334;
		case 2314: goto tr1109;
		case 2363: goto tr1109;
		case 2570: goto tr1110;
		case 2619: goto tr1111;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr1099;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1099;
	} else
		goto tr1099;
	goto tr1098;
tr1111:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1147;
st1147:
	if ( ++p == pe )
		goto _test_eof1147;
case 1147:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1113;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1112;
	goto tr1098;
st335:
	if ( ++p == pe )
		goto _test_eof335;
case 335:
	if ( (*p) == 61 )
		goto st333;
	goto tr1098;
st336:
	if ( ++p == pe )
		goto _test_eof336;
case 336:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1115;
		case 32: goto tr1115;
		case 40: goto tr1116;
		case 41: goto tr1117;
		case 2058: goto tr1118;
		case 2107: goto tr1119;
		case 2314: goto tr1120;
		case 2363: goto tr1120;
		case 2570: goto tr1121;
		case 2619: goto tr1122;
	}
	goto tr1114;
tr1124:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1125:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1141:
	{
		s->line_counter++;
	}
	goto st337;
tr1168:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st337;
tr1115:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	goto st337;
tr1116:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1117:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1118:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		s->line_counter++;
	}
	goto st337;
tr1146:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	goto st337;
tr1147:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1148:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1152:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1159:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	goto st337;
tr1160:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1161:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1162:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1174:
	{ window_add_bit(KNOT_RRTYPE_CAA, s); }
	goto st337;
tr1175:
	{ window_add_bit(KNOT_RRTYPE_CAA, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1176:
	{ window_add_bit(KNOT_RRTYPE_CAA, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1177:
	{ window_add_bit(KNOT_RRTYPE_CAA, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1188:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	goto st337;
tr1189:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1190:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1191:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1196:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	goto st337;
tr1197:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1198:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1199:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1206:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	goto st337;
tr1207:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1208:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1209:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1217:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	goto st337;
tr1218:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1219:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1220:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1231:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	goto st337;
tr1232:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1233:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1234:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1243:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	goto st337;
tr1244:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1245:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1246:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1254:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	goto st337;
tr1255:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1256:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1257:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1262:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	goto st337;
tr1263:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1264:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1265:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1275:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	goto st337;
tr1276:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1277:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1278:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1284:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	goto st337;
tr1285:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1286:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1287:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1296:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	goto st337;
tr1297:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1298:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1299:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1311:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	goto st337;
tr1312:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1313:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1314:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1322:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	goto st337;
tr1323:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1324:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1325:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1330:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	goto st337;
tr1331:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1332:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1333:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1343:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	goto st337;
tr1344:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1345:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1346:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1352:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	goto st337;
tr1353:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1354:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1355:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1361:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	goto st337;
tr1362:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1363:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1364:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1369:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	goto st337;
tr1370:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1371:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1372:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1382:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	goto st337;
tr1383:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1384:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1385:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1390:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	goto st337;
tr1391:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1392:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1393:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1404:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	goto st337;
tr1405:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1406:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1407:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1413:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	goto st337;
tr1414:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1415:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1416:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1421:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	goto st337;
tr1422:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1423:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1425:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1431:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	goto st337;
tr1432:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1433:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1435:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1440:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	goto st337;
tr1441:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1442:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1444:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1453:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	goto st337;
tr1454:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1455:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1456:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1463:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	goto st337;
tr1464:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1465:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1466:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1474:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	goto st337;
tr1475:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1476:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1477:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1485:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	goto st337;
tr1486:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1487:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1488:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1493:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	goto st337;
tr1494:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1495:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1496:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1506:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	goto st337;
tr1507:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1508:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1509:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1515:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	goto st337;
tr1516:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1517:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1518:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1524:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	goto st337;
tr1525:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1526:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1527:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1535:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	goto st337;
tr1536:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1537:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1538:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1548:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	goto st337;
tr1549:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1550:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1551:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1557:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	goto st337;
tr1558:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1559:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1560:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1568:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st337;
tr1569:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1570:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1572:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st337;
tr1579:
	{ window_add_bit(KNOT_RRTYPE_URI, s); }
	goto st337;
tr1580:
	{ window_add_bit(KNOT_RRTYPE_URI, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1581:
	{ window_add_bit(KNOT_RRTYPE_URI, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1582:
	{ window_add_bit(KNOT_RRTYPE_URI, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1590:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	goto st337;
tr1591:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1592:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1593:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		s->line_counter++;
	}
	goto st337;
tr1599:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	goto st337;
tr1600:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st337;
tr1601:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st337;
tr1602:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		s->line_counter++;
	}
	goto st337;
st337:
	if ( ++p == pe )
		goto _test_eof337;
case 337:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st337;
		case 32: goto st337;
		case 40: goto tr1124;
		case 41: goto tr1125;
		case 65: goto st338;
		case 67: goto st343;
		case 68: goto st360;
		case 69: goto st374;
		case 72: goto st381;
		case 73: goto st386;
		case 75: goto st394;
		case 76: goto st398;
		case 77: goto st406;
		case 78: goto st412;
		case 80: goto st428;
		case 82: goto st431;
		case 83: goto st438;
		case 84: goto st449;
		case 85: goto st459;
		case 97: goto st338;
		case 99: goto st343;
		case 100: goto st360;
		case 101: goto st374;
		case 104: goto st381;
		case 105: goto st386;
		case 107: goto st394;
		case 108: goto st398;
		case 109: goto st406;
		case 110: goto st412;
		case 112: goto st428;
		case 114: goto st431;
		case 115: goto st438;
		case 116: goto st449;
		case 117: goto st459;
		case 2058: goto tr1141;
		case 2107: goto st342;
		case 2314: goto tr1143;
		case 2363: goto tr1143;
		case 2570: goto tr1144;
		case 2619: goto tr1145;
	}
	goto tr1114;
st338:
	if ( ++p == pe )
		goto _test_eof338;
case 338:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1146;
		case 32: goto tr1146;
		case 40: goto tr1147;
		case 41: goto tr1148;
		case 65: goto st339;
		case 70: goto st462;
		case 80: goto st466;
		case 97: goto st339;
		case 102: goto st462;
		case 112: goto st466;
		case 2058: goto tr1152;
		case 2107: goto tr1153;
		case 2314: goto tr1154;
		case 2363: goto tr1154;
		case 2570: goto tr1155;
		case 2619: goto tr1156;
	}
	goto tr1114;
st339:
	if ( ++p == pe )
		goto _test_eof339;
case 339:
	switch( (*p) ) {
		case 65: goto st340;
		case 97: goto st340;
	}
	goto tr1114;
st340:
	if ( ++p == pe )
		goto _test_eof340;
case 340:
	switch( (*p) ) {
		case 65: goto st341;
		case 97: goto st341;
	}
	goto tr1114;
st341:
	if ( ++p == pe )
		goto _test_eof341;
case 341:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1159;
		case 32: goto tr1159;
		case 40: goto tr1160;
		case 41: goto tr1161;
		case 2058: goto tr1162;
		case 2107: goto tr1163;
		case 2314: goto tr1164;
		case 2363: goto tr1164;
		case 2570: goto tr1165;
		case 2619: goto tr1166;
	}
	goto tr1114;
tr1167:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st342;
tr1119:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	goto st342;
tr1153:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	goto st342;
tr1163:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	goto st342;
tr1178:
	{ window_add_bit(KNOT_RRTYPE_CAA, s); }
	goto st342;
tr1192:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	goto st342;
tr1200:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	goto st342;
tr1210:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	goto st342;
tr1221:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	goto st342;
tr1235:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	goto st342;
tr1247:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	goto st342;
tr1258:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	goto st342;
tr1266:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	goto st342;
tr1279:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	goto st342;
tr1288:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	goto st342;
tr1300:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	goto st342;
tr1315:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	goto st342;
tr1326:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	goto st342;
tr1334:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	goto st342;
tr1347:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	goto st342;
tr1356:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	goto st342;
tr1365:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	goto st342;
tr1373:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	goto st342;
tr1386:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	goto st342;
tr1394:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	goto st342;
tr1408:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	goto st342;
tr1417:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	goto st342;
tr1426:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	goto st342;
tr1436:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	goto st342;
tr1445:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	goto st342;
tr1457:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	goto st342;
tr1467:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	goto st342;
tr1478:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	goto st342;
tr1489:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	goto st342;
tr1497:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	goto st342;
tr1510:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	goto st342;
tr1519:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	goto st342;
tr1528:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	goto st342;
tr1539:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	goto st342;
tr1552:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	goto st342;
tr1561:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	goto st342;
tr1573:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st342;
tr1583:
	{ window_add_bit(KNOT_RRTYPE_URI, s); }
	goto st342;
tr1594:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	goto st342;
tr1603:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	goto st342;
st342:
	if ( ++p == pe )
		goto _test_eof342;
case 342:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1168;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1167;
	goto tr1114;
tr1120:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1143:
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1154:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1164:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1179:
	{ window_add_bit(KNOT_RRTYPE_CAA, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1193:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1201:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1211:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1222:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1236:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1248:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1259:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1267:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1280:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1289:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1301:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1316:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1327:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1335:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1348:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1357:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1366:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1374:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1387:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1395:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1409:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1418:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1427:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1437:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1446:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1458:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1468:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1479:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1490:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1498:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1511:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1520:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1529:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1540:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1553:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1562:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1574:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1584:
	{ window_add_bit(KNOT_RRTYPE_URI, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1595:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr1604:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
st1148:
	if ( ++p == pe )
		goto _test_eof1148;
case 1148:
	goto st0;
tr1121:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1144:
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1155:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1165:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1180:
	{ window_add_bit(KNOT_RRTYPE_CAA, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1194:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1202:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1212:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1223:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1237:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1249:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1260:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1268:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1281:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1290:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1302:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1317:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1328:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1336:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1349:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1358:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1367:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1375:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1388:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1396:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1410:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1419:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1428:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1438:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1447:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1459:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1469:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1480:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1491:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1499:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1512:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1521:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1530:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1541:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1554:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1563:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1575:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1585:
	{ window_add_bit(KNOT_RRTYPE_URI, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1596:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
tr1605:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
st1149:
	if ( ++p == pe )
		goto _test_eof1149;
case 1149:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st337;
		case 32: goto st337;
		case 40: goto tr1124;
		case 41: goto tr1125;
		case 65: goto st338;
		case 67: goto st343;
		case 68: goto st360;
		case 69: goto st374;
		case 72: goto st381;
		case 73: goto st386;
		case 75: goto st394;
		case 76: goto st398;
		case 77: goto st406;
		case 78: goto st412;
		case 80: goto st428;
		case 82: goto st431;
		case 83: goto st438;
		case 84: goto st449;
		case 85: goto st459;
		case 97: goto st338;
		case 99: goto st343;
		case 100: goto st360;
		case 101: goto st374;
		case 104: goto st381;
		case 105: goto st386;
		case 107: goto st394;
		case 108: goto st398;
		case 109: goto st406;
		case 110: goto st412;
		case 112: goto st428;
		case 114: goto st431;
		case 115: goto st438;
		case 116: goto st449;
		case 117: goto st459;
		case 2058: goto tr1141;
		case 2107: goto st342;
		case 2314: goto tr1143;
		case 2363: goto tr1143;
		case 2570: goto tr1144;
		case 2619: goto tr1145;
	}
	goto tr1114;
st343:
	if ( ++p == pe )
		goto _test_eof343;
case 343:
	switch( (*p) ) {
		case 65: goto st344;
		case 68: goto st346;
		case 69: goto st353;
		case 78: goto st356;
		case 97: goto st344;
		case 100: goto st346;
		case 101: goto st353;
		case 110: goto st356;
	}
	goto tr1114;
st344:
	if ( ++p == pe )
		goto _test_eof344;
case 344:
	switch( (*p) ) {
		case 65: goto st345;
		case 97: goto st345;
	}
	goto tr1114;
st345:
	if ( ++p == pe )
		goto _test_eof345;
case 345:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1174;
		case 32: goto tr1174;
		case 40: goto tr1175;
		case 41: goto tr1176;
		case 2058: goto tr1177;
		case 2107: goto tr1178;
		case 2314: goto tr1179;
		case 2363: goto tr1179;
		case 2570: goto tr1180;
		case 2619: goto tr1181;
	}
	goto tr1114;
tr1122:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1145:
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1156:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1166:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1181:
	{ window_add_bit(KNOT_RRTYPE_CAA, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1195:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1203:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1213:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1224:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1238:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1250:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1261:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1269:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1282:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1291:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1303:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1318:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1329:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1337:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1350:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1359:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1368:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1376:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1389:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1397:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1411:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1420:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1429:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1439:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1448:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1460:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1470:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1481:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1492:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1500:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1513:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1522:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1531:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1542:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1555:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1564:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1576:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1586:
	{ window_add_bit(KNOT_RRTYPE_URI, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1597:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
tr1606:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st268;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
st1150:
	if ( ++p == pe )
		goto _test_eof1150;
case 1150:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1168;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1167;
	goto tr1114;
st346:
	if ( ++p == pe )
		goto _test_eof346;
case 346:
	switch( (*p) ) {
		case 78: goto st347;
		case 83: goto st352;
		case 110: goto st347;
		case 115: goto st352;
	}
	goto tr1114;
st347:
	if ( ++p == pe )
		goto _test_eof347;
case 347:
	switch( (*p) ) {
		case 83: goto st348;
		case 115: goto st348;
	}
	goto tr1114;
st348:
	if ( ++p == pe )
		goto _test_eof348;
case 348:
	switch( (*p) ) {
		case 75: goto st349;
		case 107: goto st349;
	}
	goto tr1114;
st349:
	if ( ++p == pe )
		goto _test_eof349;
case 349:
	switch( (*p) ) {
		case 69: goto st350;
		case 101: goto st350;
	}
	goto tr1114;
st350:
	if ( ++p == pe )
		goto _test_eof350;
case 350:
	switch( (*p) ) {
		case 89: goto st351;
		case 121: goto st351;
	}
	goto tr1114;
st351:
	if ( ++p == pe )
		goto _test_eof351;
case 351:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1188;
		case 32: goto tr1188;
		case 40: goto tr1189;
		case 41: goto tr1190;
		case 2058: goto tr1191;
		case 2107: goto tr1192;
		case 2314: goto tr1193;
		case 2363: goto tr1193;
		case 2570: goto tr1194;
		case 2619: goto tr1195;
	}
	goto tr1114;
st352:
	if ( ++p == pe )
		goto _test_eof352;
case 352:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1196;
		case 32: goto tr1196;
		case 40: goto tr1197;
		case 41: goto tr1198;
		case 2058: goto tr1199;
		case 2107: goto tr1200;
		case 2314: goto tr1201;
		case 2363: goto tr1201;
		case 2570: goto tr1202;
		case 2619: goto tr1203;
	}
	goto tr1114;
st353:
	if ( ++p == pe )
		goto _test_eof353;
case 353:
	switch( (*p) ) {
		case 82: goto st354;
		case 114: goto st354;
	}
	goto tr1114;
st354:
	if ( ++p == pe )
		goto _test_eof354;
case 354:
	switch( (*p) ) {
		case 84: goto st355;
		case 116: goto st355;
	}
	goto tr1114;
st355:
	if ( ++p == pe )
		goto _test_eof355;
case 355:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1206;
		case 32: goto tr1206;
		case 40: goto tr1207;
		case 41: goto tr1208;
		case 2058: goto tr1209;
		case 2107: goto tr1210;
		case 2314: goto tr1211;
		case 2363: goto tr1211;
		case 2570: goto tr1212;
		case 2619: goto tr1213;
	}
	goto tr1114;
st356:
	if ( ++p == pe )
		goto _test_eof356;
case 356:
	switch( (*p) ) {
		case 65: goto st357;
		case 97: goto st357;
	}
	goto tr1114;
st357:
	if ( ++p == pe )
		goto _test_eof357;
case 357:
	switch( (*p) ) {
		case 77: goto st358;
		case 109: goto st358;
	}
	goto tr1114;
st358:
	if ( ++p == pe )
		goto _test_eof358;
case 358:
	switch( (*p) ) {
		case 69: goto st359;
		case 101: goto st359;
	}
	goto tr1114;
st359:
	if ( ++p == pe )
		goto _test_eof359;
case 359:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1217;
		case 32: goto tr1217;
		case 40: goto tr1218;
		case 41: goto tr1219;
		case 2058: goto tr1220;
		case 2107: goto tr1221;
		case 2314: goto tr1222;
		case 2363: goto tr1222;
		case 2570: goto tr1223;
		case 2619: goto tr1224;
	}
	goto tr1114;
st360:
	if ( ++p == pe )
		goto _test_eof360;
case 360:
	switch( (*p) ) {
		case 72: goto st361;
		case 78: goto st365;
		case 83: goto st373;
		case 104: goto st361;
		case 110: goto st365;
		case 115: goto st373;
	}
	goto tr1114;
st361:
	if ( ++p == pe )
		goto _test_eof361;
case 361:
	switch( (*p) ) {
		case 67: goto st362;
		case 99: goto st362;
	}
	goto tr1114;
st362:
	if ( ++p == pe )
		goto _test_eof362;
case 362:
	switch( (*p) ) {
		case 73: goto st363;
		case 105: goto st363;
	}
	goto tr1114;
st363:
	if ( ++p == pe )
		goto _test_eof363;
case 363:
	switch( (*p) ) {
		case 68: goto st364;
		case 100: goto st364;
	}
	goto tr1114;
st364:
	if ( ++p == pe )
		goto _test_eof364;
case 364:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1231;
		case 32: goto tr1231;
		case 40: goto tr1232;
		case 41: goto tr1233;
		case 2058: goto tr1234;
		case 2107: goto tr1235;
		case 2314: goto tr1236;
		case 2363: goto tr1236;
		case 2570: goto tr1237;
		case 2619: goto tr1238;
	}
	goto tr1114;
st365:
	if ( ++p == pe )
		goto _test_eof365;
case 365:
	switch( (*p) ) {
		case 65: goto st366;
		case 83: goto st369;
		case 97: goto st366;
		case 115: goto st369;
	}
	goto tr1114;
st366:
	if ( ++p == pe )
		goto _test_eof366;
case 366:
	switch( (*p) ) {
		case 77: goto st367;
		case 109: goto st367;
	}
	goto tr1114;
st367:
	if ( ++p == pe )
		goto _test_eof367;
case 367:
	switch( (*p) ) {
		case 69: goto st368;
		case 101: goto st368;
	}
	goto tr1114;
st368:
	if ( ++p == pe )
		goto _test_eof368;
case 368:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1243;
		case 32: goto tr1243;
		case 40: goto tr1244;
		case 41: goto tr1245;
		case 2058: goto tr1246;
		case 2107: goto tr1247;
		case 2314: goto tr1248;
		case 2363: goto tr1248;
		case 2570: goto tr1249;
		case 2619: goto tr1250;
	}
	goto tr1114;
st369:
	if ( ++p == pe )
		goto _test_eof369;
case 369:
	switch( (*p) ) {
		case 75: goto st370;
		case 107: goto st370;
	}
	goto tr1114;
st370:
	if ( ++p == pe )
		goto _test_eof370;
case 370:
	switch( (*p) ) {
		case 69: goto st371;
		case 101: goto st371;
	}
	goto tr1114;
st371:
	if ( ++p == pe )
		goto _test_eof371;
case 371:
	switch( (*p) ) {
		case 89: goto st372;
		case 121: goto st372;
	}
	goto tr1114;
st372:
	if ( ++p == pe )
		goto _test_eof372;
case 372:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1254;
		case 32: goto tr1254;
		case 40: goto tr1255;
		case 41: goto tr1256;
		case 2058: goto tr1257;
		case 2107: goto tr1258;
		case 2314: goto tr1259;
		case 2363: goto tr1259;
		case 2570: goto tr1260;
		case 2619: goto tr1261;
	}
	goto tr1114;
st373:
	if ( ++p == pe )
		goto _test_eof373;
case 373:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1262;
		case 32: goto tr1262;
		case 40: goto tr1263;
		case 41: goto tr1264;
		case 2058: goto tr1265;
		case 2107: goto tr1266;
		case 2314: goto tr1267;
		case 2363: goto tr1267;
		case 2570: goto tr1268;
		case 2619: goto tr1269;
	}
	goto tr1114;
st374:
	if ( ++p == pe )
		goto _test_eof374;
case 374:
	switch( (*p) ) {
		case 85: goto st375;
		case 117: goto st375;
	}
	goto tr1114;
st375:
	if ( ++p == pe )
		goto _test_eof375;
case 375:
	switch( (*p) ) {
		case 73: goto st376;
		case 105: goto st376;
	}
	goto tr1114;
st376:
	if ( ++p == pe )
		goto _test_eof376;
case 376:
	switch( (*p) ) {
		case 52: goto st377;
		case 54: goto st379;
	}
	goto tr1114;
st377:
	if ( ++p == pe )
		goto _test_eof377;
case 377:
	if ( (*p) == 56 )
		goto st378;
	goto tr1114;
st378:
	if ( ++p == pe )
		goto _test_eof378;
case 378:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1275;
		case 32: goto tr1275;
		case 40: goto tr1276;
		case 41: goto tr1277;
		case 2058: goto tr1278;
		case 2107: goto tr1279;
		case 2314: goto tr1280;
		case 2363: goto tr1280;
		case 2570: goto tr1281;
		case 2619: goto tr1282;
	}
	goto tr1114;
st379:
	if ( ++p == pe )
		goto _test_eof379;
case 379:
	if ( (*p) == 52 )
		goto st380;
	goto tr1114;
st380:
	if ( ++p == pe )
		goto _test_eof380;
case 380:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1284;
		case 32: goto tr1284;
		case 40: goto tr1285;
		case 41: goto tr1286;
		case 2058: goto tr1287;
		case 2107: goto tr1288;
		case 2314: goto tr1289;
		case 2363: goto tr1289;
		case 2570: goto tr1290;
		case 2619: goto tr1291;
	}
	goto tr1114;
st381:
	if ( ++p == pe )
		goto _test_eof381;
case 381:
	switch( (*p) ) {
		case 73: goto st382;
		case 105: goto st382;
	}
	goto tr1114;
st382:
	if ( ++p == pe )
		goto _test_eof382;
case 382:
	switch( (*p) ) {
		case 78: goto st383;
		case 110: goto st383;
	}
	goto tr1114;
st383:
	if ( ++p == pe )
		goto _test_eof383;
case 383:
	switch( (*p) ) {
		case 70: goto st384;
		case 102: goto st384;
	}
	goto tr1114;
st384:
	if ( ++p == pe )
		goto _test_eof384;
case 384:
	switch( (*p) ) {
		case 79: goto st385;
		case 111: goto st385;
	}
	goto tr1114;
st385:
	if ( ++p == pe )
		goto _test_eof385;
case 385:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1296;
		case 32: goto tr1296;
		case 40: goto tr1297;
		case 41: goto tr1298;
		case 2058: goto tr1299;
		case 2107: goto tr1300;
		case 2314: goto tr1301;
		case 2363: goto tr1301;
		case 2570: goto tr1302;
		case 2619: goto tr1303;
	}
	goto tr1114;
st386:
	if ( ++p == pe )
		goto _test_eof386;
case 386:
	switch( (*p) ) {
		case 80: goto st387;
		case 112: goto st387;
	}
	goto tr1114;
st387:
	if ( ++p == pe )
		goto _test_eof387;
case 387:
	switch( (*p) ) {
		case 83: goto st388;
		case 115: goto st388;
	}
	goto tr1114;
st388:
	if ( ++p == pe )
		goto _test_eof388;
case 388:
	switch( (*p) ) {
		case 69: goto st389;
		case 101: goto st389;
	}
	goto tr1114;
st389:
	if ( ++p == pe )
		goto _test_eof389;
case 389:
	switch( (*p) ) {
		case 67: goto st390;
		case 99: goto st390;
	}
	goto tr1114;
st390:
	if ( ++p == pe )
		goto _test_eof390;
case 390:
	switch( (*p) ) {
		case 75: goto st391;
		case 107: goto st391;
	}
	goto tr1114;
st391:
	if ( ++p == pe )
		goto _test_eof391;
case 391:
	switch( (*p) ) {
		case 69: goto st392;
		case 101: goto st392;
	}
	goto tr1114;
st392:
	if ( ++p == pe )
		goto _test_eof392;
case 392:
	switch( (*p) ) {
		case 89: goto st393;
		case 121: goto st393;
	}
	goto tr1114;
st393:
	if ( ++p == pe )
		goto _test_eof393;
case 393:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1311;
		case 32: goto tr1311;
		case 40: goto tr1312;
		case 41: goto tr1313;
		case 2058: goto tr1314;
		case 2107: goto tr1315;
		case 2314: goto tr1316;
		case 2363: goto tr1316;
		case 2570: goto tr1317;
		case 2619: goto tr1318;
	}
	goto tr1114;
st394:
	if ( ++p == pe )
		goto _test_eof394;
case 394:
	switch( (*p) ) {
		case 69: goto st395;
		case 88: goto st397;
		case 101: goto st395;
		case 120: goto st397;
	}
	goto tr1114;
st395:
	if ( ++p == pe )
		goto _test_eof395;
case 395:
	switch( (*p) ) {
		case 89: goto st396;
		case 121: goto st396;
	}
	goto tr1114;
st396:
	if ( ++p == pe )
		goto _test_eof396;
case 396:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1322;
		case 32: goto tr1322;
		case 40: goto tr1323;
		case 41: goto tr1324;
		case 2058: goto tr1325;
		case 2107: goto tr1326;
		case 2314: goto tr1327;
		case 2363: goto tr1327;
		case 2570: goto tr1328;
		case 2619: goto tr1329;
	}
	goto tr1114;
st397:
	if ( ++p == pe )
		goto _test_eof397;
case 397:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1330;
		case 32: goto tr1330;
		case 40: goto tr1331;
		case 41: goto tr1332;
		case 2058: goto tr1333;
		case 2107: goto tr1334;
		case 2314: goto tr1335;
		case 2363: goto tr1335;
		case 2570: goto tr1336;
		case 2619: goto tr1337;
	}
	goto tr1114;
st398:
	if ( ++p == pe )
		goto _test_eof398;
case 398:
	switch( (*p) ) {
		case 51: goto st399;
		case 54: goto st401;
		case 79: goto st403;
		case 80: goto st405;
		case 111: goto st403;
		case 112: goto st405;
	}
	goto tr1114;
st399:
	if ( ++p == pe )
		goto _test_eof399;
case 399:
	if ( (*p) == 50 )
		goto st400;
	goto tr1114;
st400:
	if ( ++p == pe )
		goto _test_eof400;
case 400:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1343;
		case 32: goto tr1343;
		case 40: goto tr1344;
		case 41: goto tr1345;
		case 2058: goto tr1346;
		case 2107: goto tr1347;
		case 2314: goto tr1348;
		case 2363: goto tr1348;
		case 2570: goto tr1349;
		case 2619: goto tr1350;
	}
	goto tr1114;
st401:
	if ( ++p == pe )
		goto _test_eof401;
case 401:
	if ( (*p) == 52 )
		goto st402;
	goto tr1114;
st402:
	if ( ++p == pe )
		goto _test_eof402;
case 402:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1352;
		case 32: goto tr1352;
		case 40: goto tr1353;
		case 41: goto tr1354;
		case 2058: goto tr1355;
		case 2107: goto tr1356;
		case 2314: goto tr1357;
		case 2363: goto tr1357;
		case 2570: goto tr1358;
		case 2619: goto tr1359;
	}
	goto tr1114;
st403:
	if ( ++p == pe )
		goto _test_eof403;
case 403:
	switch( (*p) ) {
		case 67: goto st404;
		case 99: goto st404;
	}
	goto tr1114;
st404:
	if ( ++p == pe )
		goto _test_eof404;
case 404:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1361;
		case 32: goto tr1361;
		case 40: goto tr1362;
		case 41: goto tr1363;
		case 2058: goto tr1364;
		case 2107: goto tr1365;
		case 2314: goto tr1366;
		case 2363: goto tr1366;
		case 2570: goto tr1367;
		case 2619: goto tr1368;
	}
	goto tr1114;
st405:
	if ( ++p == pe )
		goto _test_eof405;
case 405:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1369;
		case 32: goto tr1369;
		case 40: goto tr1370;
		case 41: goto tr1371;
		case 2058: goto tr1372;
		case 2107: goto tr1373;
		case 2314: goto tr1374;
		case 2363: goto tr1374;
		case 2570: goto tr1375;
		case 2619: goto tr1376;
	}
	goto tr1114;
st406:
	if ( ++p == pe )
		goto _test_eof406;
case 406:
	switch( (*p) ) {
		case 73: goto st407;
		case 88: goto st411;
		case 105: goto st407;
		case 120: goto st411;
	}
	goto tr1114;
st407:
	if ( ++p == pe )
		goto _test_eof407;
case 407:
	switch( (*p) ) {
		case 78: goto st408;
		case 110: goto st408;
	}
	goto tr1114;
st408:
	if ( ++p == pe )
		goto _test_eof408;
case 408:
	switch( (*p) ) {
		case 70: goto st409;
		case 102: goto st409;
	}
	goto tr1114;
st409:
	if ( ++p == pe )
		goto _test_eof409;
case 409:
	switch( (*p) ) {
		case 79: goto st410;
		case 111: goto st410;
	}
	goto tr1114;
st410:
	if ( ++p == pe )
		goto _test_eof410;
case 410:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1382;
		case 32: goto tr1382;
		case 40: goto tr1383;
		case 41: goto tr1384;
		case 2058: goto tr1385;
		case 2107: goto tr1386;
		case 2314: goto tr1387;
		case 2363: goto tr1387;
		case 2570: goto tr1388;
		case 2619: goto tr1389;
	}
	goto tr1114;
st411:
	if ( ++p == pe )
		goto _test_eof411;
case 411:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1390;
		case 32: goto tr1390;
		case 40: goto tr1391;
		case 41: goto tr1392;
		case 2058: goto tr1393;
		case 2107: goto tr1394;
		case 2314: goto tr1395;
		case 2363: goto tr1395;
		case 2570: goto tr1396;
		case 2619: goto tr1397;
	}
	goto tr1114;
st412:
	if ( ++p == pe )
		goto _test_eof412;
case 412:
	switch( (*p) ) {
		case 65: goto st413;
		case 73: goto st417;
		case 83: goto st419;
		case 97: goto st413;
		case 105: goto st417;
		case 115: goto st419;
	}
	goto tr1114;
st413:
	if ( ++p == pe )
		goto _test_eof413;
case 413:
	switch( (*p) ) {
		case 80: goto st414;
		case 112: goto st414;
	}
	goto tr1114;
st414:
	if ( ++p == pe )
		goto _test_eof414;
case 414:
	switch( (*p) ) {
		case 84: goto st415;
		case 116: goto st415;
	}
	goto tr1114;
st415:
	if ( ++p == pe )
		goto _test_eof415;
case 415:
	switch( (*p) ) {
		case 82: goto st416;
		case 114: goto st416;
	}
	goto tr1114;
st416:
	if ( ++p == pe )
		goto _test_eof416;
case 416:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1404;
		case 32: goto tr1404;
		case 40: goto tr1405;
		case 41: goto tr1406;
		case 2058: goto tr1407;
		case 2107: goto tr1408;
		case 2314: goto tr1409;
		case 2363: goto tr1409;
		case 2570: goto tr1410;
		case 2619: goto tr1411;
	}
	goto tr1114;
st417:
	if ( ++p == pe )
		goto _test_eof417;
case 417:
	switch( (*p) ) {
		case 68: goto st418;
		case 100: goto st418;
	}
	goto tr1114;
st418:
	if ( ++p == pe )
		goto _test_eof418;
case 418:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1413;
		case 32: goto tr1413;
		case 40: goto tr1414;
		case 41: goto tr1415;
		case 2058: goto tr1416;
		case 2107: goto tr1417;
		case 2314: goto tr1418;
		case 2363: goto tr1418;
		case 2570: goto tr1419;
		case 2619: goto tr1420;
	}
	goto tr1114;
st419:
	if ( ++p == pe )
		goto _test_eof419;
case 419:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1421;
		case 32: goto tr1421;
		case 40: goto tr1422;
		case 41: goto tr1423;
		case 69: goto st420;
		case 101: goto st420;
		case 2058: goto tr1425;
		case 2107: goto tr1426;
		case 2314: goto tr1427;
		case 2363: goto tr1427;
		case 2570: goto tr1428;
		case 2619: goto tr1429;
	}
	goto tr1114;
st420:
	if ( ++p == pe )
		goto _test_eof420;
case 420:
	switch( (*p) ) {
		case 67: goto st421;
		case 99: goto st421;
	}
	goto tr1114;
st421:
	if ( ++p == pe )
		goto _test_eof421;
case 421:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1431;
		case 32: goto tr1431;
		case 40: goto tr1432;
		case 41: goto tr1433;
		case 51: goto st422;
		case 2058: goto tr1435;
		case 2107: goto tr1436;
		case 2314: goto tr1437;
		case 2363: goto tr1437;
		case 2570: goto tr1438;
		case 2619: goto tr1439;
	}
	goto tr1114;
st422:
	if ( ++p == pe )
		goto _test_eof422;
case 422:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1440;
		case 32: goto tr1440;
		case 40: goto tr1441;
		case 41: goto tr1442;
		case 80: goto st423;
		case 112: goto st423;
		case 2058: goto tr1444;
		case 2107: goto tr1445;
		case 2314: goto tr1446;
		case 2363: goto tr1446;
		case 2570: goto tr1447;
		case 2619: goto tr1448;
	}
	goto tr1114;
st423:
	if ( ++p == pe )
		goto _test_eof423;
case 423:
	switch( (*p) ) {
		case 65: goto st424;
		case 97: goto st424;
	}
	goto tr1114;
st424:
	if ( ++p == pe )
		goto _test_eof424;
case 424:
	switch( (*p) ) {
		case 82: goto st425;
		case 114: goto st425;
	}
	goto tr1114;
st425:
	if ( ++p == pe )
		goto _test_eof425;
case 425:
	switch( (*p) ) {
		case 65: goto st426;
		case 97: goto st426;
	}
	goto tr1114;
st426:
	if ( ++p == pe )
		goto _test_eof426;
case 426:
	switch( (*p) ) {
		case 77: goto st427;
		case 109: goto st427;
	}
	goto tr1114;
st427:
	if ( ++p == pe )
		goto _test_eof427;
case 427:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1453;
		case 32: goto tr1453;
		case 40: goto tr1454;
		case 41: goto tr1455;
		case 2058: goto tr1456;
		case 2107: goto tr1457;
		case 2314: goto tr1458;
		case 2363: goto tr1458;
		case 2570: goto tr1459;
		case 2619: goto tr1460;
	}
	goto tr1114;
st428:
	if ( ++p == pe )
		goto _test_eof428;
case 428:
	switch( (*p) ) {
		case 84: goto st429;
		case 116: goto st429;
	}
	goto tr1114;
st429:
	if ( ++p == pe )
		goto _test_eof429;
case 429:
	switch( (*p) ) {
		case 82: goto st430;
		case 114: goto st430;
	}
	goto tr1114;
st430:
	if ( ++p == pe )
		goto _test_eof430;
case 430:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1463;
		case 32: goto tr1463;
		case 40: goto tr1464;
		case 41: goto tr1465;
		case 2058: goto tr1466;
		case 2107: goto tr1467;
		case 2314: goto tr1468;
		case 2363: goto tr1468;
		case 2570: goto tr1469;
		case 2619: goto tr1470;
	}
	goto tr1114;
st431:
	if ( ++p == pe )
		goto _test_eof431;
case 431:
	switch( (*p) ) {
		case 80: goto st432;
		case 82: goto st433;
		case 84: goto st437;
		case 112: goto st432;
		case 114: goto st433;
		case 116: goto st437;
	}
	goto tr1114;
st432:
	if ( ++p == pe )
		goto _test_eof432;
case 432:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1474;
		case 32: goto tr1474;
		case 40: goto tr1475;
		case 41: goto tr1476;
		case 2058: goto tr1477;
		case 2107: goto tr1478;
		case 2314: goto tr1479;
		case 2363: goto tr1479;
		case 2570: goto tr1480;
		case 2619: goto tr1481;
	}
	goto tr1114;
st433:
	if ( ++p == pe )
		goto _test_eof433;
case 433:
	switch( (*p) ) {
		case 83: goto st434;
		case 115: goto st434;
	}
	goto tr1114;
st434:
	if ( ++p == pe )
		goto _test_eof434;
case 434:
	switch( (*p) ) {
		case 73: goto st435;
		case 105: goto st435;
	}
	goto tr1114;
st435:
	if ( ++p == pe )
		goto _test_eof435;
case 435:
	switch( (*p) ) {
		case 71: goto st436;
		case 103: goto st436;
	}
	goto tr1114;
st436:
	if ( ++p == pe )
		goto _test_eof436;
case 436:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1485;
		case 32: goto tr1485;
		case 40: goto tr1486;
		case 41: goto tr1487;
		case 2058: goto tr1488;
		case 2107: goto tr1489;
		case 2314: goto tr1490;
		case 2363: goto tr1490;
		case 2570: goto tr1491;
		case 2619: goto tr1492;
	}
	goto tr1114;
st437:
	if ( ++p == pe )
		goto _test_eof437;
case 437:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1493;
		case 32: goto tr1493;
		case 40: goto tr1494;
		case 41: goto tr1495;
		case 2058: goto tr1496;
		case 2107: goto tr1497;
		case 2314: goto tr1498;
		case 2363: goto tr1498;
		case 2570: goto tr1499;
		case 2619: goto tr1500;
	}
	goto tr1114;
st438:
	if ( ++p == pe )
		goto _test_eof438;
case 438:
	switch( (*p) ) {
		case 79: goto st439;
		case 80: goto st441;
		case 82: goto st443;
		case 83: goto st445;
		case 111: goto st439;
		case 112: goto st441;
		case 114: goto st443;
		case 115: goto st445;
	}
	goto tr1114;
st439:
	if ( ++p == pe )
		goto _test_eof439;
case 439:
	switch( (*p) ) {
		case 65: goto st440;
		case 97: goto st440;
	}
	goto tr1114;
st440:
	if ( ++p == pe )
		goto _test_eof440;
case 440:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1506;
		case 32: goto tr1506;
		case 40: goto tr1507;
		case 41: goto tr1508;
		case 2058: goto tr1509;
		case 2107: goto tr1510;
		case 2314: goto tr1511;
		case 2363: goto tr1511;
		case 2570: goto tr1512;
		case 2619: goto tr1513;
	}
	goto tr1114;
st441:
	if ( ++p == pe )
		goto _test_eof441;
case 441:
	switch( (*p) ) {
		case 70: goto st442;
		case 102: goto st442;
	}
	goto tr1114;
st442:
	if ( ++p == pe )
		goto _test_eof442;
case 442:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1515;
		case 32: goto tr1515;
		case 40: goto tr1516;
		case 41: goto tr1517;
		case 2058: goto tr1518;
		case 2107: goto tr1519;
		case 2314: goto tr1520;
		case 2363: goto tr1520;
		case 2570: goto tr1521;
		case 2619: goto tr1522;
	}
	goto tr1114;
st443:
	if ( ++p == pe )
		goto _test_eof443;
case 443:
	switch( (*p) ) {
		case 86: goto st444;
		case 118: goto st444;
	}
	goto tr1114;
st444:
	if ( ++p == pe )
		goto _test_eof444;
case 444:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1524;
		case 32: goto tr1524;
		case 40: goto tr1525;
		case 41: goto tr1526;
		case 2058: goto tr1527;
		case 2107: goto tr1528;
		case 2314: goto tr1529;
		case 2363: goto tr1529;
		case 2570: goto tr1530;
		case 2619: goto tr1531;
	}
	goto tr1114;
st445:
	if ( ++p == pe )
		goto _test_eof445;
case 445:
	switch( (*p) ) {
		case 72: goto st446;
		case 104: goto st446;
	}
	goto tr1114;
st446:
	if ( ++p == pe )
		goto _test_eof446;
case 446:
	switch( (*p) ) {
		case 70: goto st447;
		case 102: goto st447;
	}
	goto tr1114;
st447:
	if ( ++p == pe )
		goto _test_eof447;
case 447:
	switch( (*p) ) {
		case 80: goto st448;
		case 112: goto st448;
	}
	goto tr1114;
st448:
	if ( ++p == pe )
		goto _test_eof448;
case 448:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1535;
		case 32: goto tr1535;
		case 40: goto tr1536;
		case 41: goto tr1537;
		case 2058: goto tr1538;
		case 2107: goto tr1539;
		case 2314: goto tr1540;
		case 2363: goto tr1540;
		case 2570: goto tr1541;
		case 2619: goto tr1542;
	}
	goto tr1114;
st449:
	if ( ++p == pe )
		goto _test_eof449;
case 449:
	switch( (*p) ) {
		case 76: goto st450;
		case 88: goto st453;
		case 89: goto st455;
		case 108: goto st450;
		case 120: goto st453;
		case 121: goto st455;
	}
	goto tr1114;
st450:
	if ( ++p == pe )
		goto _test_eof450;
case 450:
	switch( (*p) ) {
		case 83: goto st451;
		case 115: goto st451;
	}
	goto tr1114;
st451:
	if ( ++p == pe )
		goto _test_eof451;
case 451:
	switch( (*p) ) {
		case 65: goto st452;
		case 97: goto st452;
	}
	goto tr1114;
st452:
	if ( ++p == pe )
		goto _test_eof452;
case 452:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1548;
		case 32: goto tr1548;
		case 40: goto tr1549;
		case 41: goto tr1550;
		case 2058: goto tr1551;
		case 2107: goto tr1552;
		case 2314: goto tr1553;
		case 2363: goto tr1553;
		case 2570: goto tr1554;
		case 2619: goto tr1555;
	}
	goto tr1114;
st453:
	if ( ++p == pe )
		goto _test_eof453;
case 453:
	switch( (*p) ) {
		case 84: goto st454;
		case 116: goto st454;
	}
	goto tr1114;
st454:
	if ( ++p == pe )
		goto _test_eof454;
case 454:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1557;
		case 32: goto tr1557;
		case 40: goto tr1558;
		case 41: goto tr1559;
		case 2058: goto tr1560;
		case 2107: goto tr1561;
		case 2314: goto tr1562;
		case 2363: goto tr1562;
		case 2570: goto tr1563;
		case 2619: goto tr1564;
	}
	goto tr1114;
st455:
	if ( ++p == pe )
		goto _test_eof455;
case 455:
	switch( (*p) ) {
		case 80: goto st456;
		case 112: goto st456;
	}
	goto tr1114;
st456:
	if ( ++p == pe )
		goto _test_eof456;
case 456:
	switch( (*p) ) {
		case 69: goto st457;
		case 101: goto st457;
	}
	goto tr1114;
st457:
	if ( ++p == pe )
		goto _test_eof457;
case 457:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1567;
	goto tr1114;
tr1567:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st458;
tr1571:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st458;
st458:
	if ( ++p == pe )
		goto _test_eof458;
case 458:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1568;
		case 32: goto tr1568;
		case 40: goto tr1569;
		case 41: goto tr1570;
		case 2058: goto tr1572;
		case 2107: goto tr1573;
		case 2314: goto tr1574;
		case 2363: goto tr1574;
		case 2570: goto tr1575;
		case 2619: goto tr1576;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1571;
	goto tr1114;
st459:
	if ( ++p == pe )
		goto _test_eof459;
case 459:
	switch( (*p) ) {
		case 82: goto st460;
		case 114: goto st460;
	}
	goto tr1114;
st460:
	if ( ++p == pe )
		goto _test_eof460;
case 460:
	switch( (*p) ) {
		case 73: goto st461;
		case 105: goto st461;
	}
	goto tr1114;
st461:
	if ( ++p == pe )
		goto _test_eof461;
case 461:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1579;
		case 32: goto tr1579;
		case 40: goto tr1580;
		case 41: goto tr1581;
		case 2058: goto tr1582;
		case 2107: goto tr1583;
		case 2314: goto tr1584;
		case 2363: goto tr1584;
		case 2570: goto tr1585;
		case 2619: goto tr1586;
	}
	goto tr1114;
st462:
	if ( ++p == pe )
		goto _test_eof462;
case 462:
	switch( (*p) ) {
		case 83: goto st463;
		case 115: goto st463;
	}
	goto tr1114;
st463:
	if ( ++p == pe )
		goto _test_eof463;
case 463:
	switch( (*p) ) {
		case 68: goto st464;
		case 100: goto st464;
	}
	goto tr1114;
st464:
	if ( ++p == pe )
		goto _test_eof464;
case 464:
	switch( (*p) ) {
		case 66: goto st465;
		case 98: goto st465;
	}
	goto tr1114;
st465:
	if ( ++p == pe )
		goto _test_eof465;
case 465:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1590;
		case 32: goto tr1590;
		case 40: goto tr1591;
		case 41: goto tr1592;
		case 2058: goto tr1593;
		case 2107: goto tr1594;
		case 2314: goto tr1595;
		case 2363: goto tr1595;
		case 2570: goto tr1596;
		case 2619: goto tr1597;
	}
	goto tr1114;
st466:
	if ( ++p == pe )
		goto _test_eof466;
case 466:
	switch( (*p) ) {
		case 76: goto st467;
		case 108: goto st467;
	}
	goto tr1114;
st467:
	if ( ++p == pe )
		goto _test_eof467;
case 467:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1599;
		case 32: goto tr1599;
		case 40: goto tr1600;
		case 41: goto tr1601;
		case 2058: goto tr1602;
		case 2107: goto tr1603;
		case 2314: goto tr1604;
		case 2363: goto tr1604;
		case 2570: goto tr1605;
		case 2619: goto tr1606;
	}
	goto tr1114;
st468:
	if ( ++p == pe )
		goto _test_eof468;
case 468:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st469;
		case 32: goto st469;
		case 40: goto tr1609;
		case 41: goto tr1610;
		case 1034: goto tr1611;
		case 1083: goto st476;
	}
	goto tr1607;
tr1609:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st469;
tr1610:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st469;
tr1611:
	{
		s->line_counter++;
	}
	goto st469;
tr1642:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st469;
st469:
	if ( ++p == pe )
		goto _test_eof469;
case 469:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st469;
		case 32: goto st469;
		case 40: goto tr1609;
		case 41: goto tr1610;
		case 1034: goto tr1611;
		case 1083: goto st476;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1614;
	goto tr1613;
tr1614:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st470;
tr1618:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st470;
st470:
	if ( ++p == pe )
		goto _test_eof470;
case 470:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1615;
		case 32: goto tr1615;
		case 40: goto tr1616;
		case 41: goto tr1617;
		case 1034: goto tr1619;
		case 1083: goto tr1620;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1618;
	goto tr1613;
tr1623:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st471;
tr1624:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st471;
tr1626:
	{
		s->line_counter++;
	}
	goto st471;
tr1640:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st471;
tr1615:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st471;
tr1616:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st471;
tr1617:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st471;
tr1619:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st471;
st471:
	if ( ++p == pe )
		goto _test_eof471;
case 471:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st471;
		case 32: goto st471;
		case 40: goto tr1623;
		case 41: goto tr1624;
		case 1034: goto tr1626;
		case 1083: goto st475;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1625;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1625;
	} else
		goto tr1625;
	goto tr1621;
tr1625:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st472;
st472:
	if ( ++p == pe )
		goto _test_eof472;
case 472:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1628;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr1628;
	} else
		goto tr1628;
	goto tr1621;
tr1630:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st473;
tr1631:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st473;
tr1632:
	{
		s->line_counter++;
	}
	goto st473;
tr1638:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st473;
tr1628:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st473;
st473:
	if ( ++p == pe )
		goto _test_eof473;
case 473:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st473;
		case 32: goto st473;
		case 40: goto tr1630;
		case 41: goto tr1631;
		case 2058: goto tr1632;
		case 2107: goto st474;
		case 2314: goto tr1634;
		case 2363: goto tr1634;
		case 2570: goto tr1635;
		case 2619: goto tr1636;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1625;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1625;
	} else
		goto tr1625;
	goto tr1621;
tr1637:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st474;
st474:
	if ( ++p == pe )
		goto _test_eof474;
case 474:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1638;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1637;
	goto tr1621;
tr1634:
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1151;
st1151:
	if ( ++p == pe )
		goto _test_eof1151;
case 1151:
	goto st0;
tr1635:
	{
		s->line_counter++;
	}
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1152;
st1152:
	if ( ++p == pe )
		goto _test_eof1152;
case 1152:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st473;
		case 32: goto st473;
		case 40: goto tr1630;
		case 41: goto tr1631;
		case 2058: goto tr1632;
		case 2107: goto st474;
		case 2314: goto tr1634;
		case 2363: goto tr1634;
		case 2570: goto tr1635;
		case 2619: goto tr1636;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1625;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1625;
	} else
		goto tr1625;
	goto tr1621;
tr1636:
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1153;
st1153:
	if ( ++p == pe )
		goto _test_eof1153;
case 1153:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1638;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1637;
	goto tr1621;
tr1639:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st475;
tr1620:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st475;
st475:
	if ( ++p == pe )
		goto _test_eof475;
case 475:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1640;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1639;
	goto tr1607;
tr1641:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st476;
st476:
	if ( ++p == pe )
		goto _test_eof476;
case 476:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1642;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1641;
	goto tr1607;
st477:
	if ( ++p == pe )
		goto _test_eof477;
case 477:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st478;
		case 32: goto st478;
		case 40: goto tr1644;
		case 41: goto tr1645;
		case 1034: goto tr1646;
		case 1083: goto st486;
	}
	goto tr1607;
tr1644:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st478;
tr1645:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st478;
tr1646:
	{
		s->line_counter++;
	}
	goto st478;
tr1682:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st478;
st478:
	if ( ++p == pe )
		goto _test_eof478;
case 478:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st478;
		case 32: goto st478;
		case 40: goto tr1644;
		case 41: goto tr1645;
		case 48: goto tr1648;
		case 1034: goto tr1646;
		case 1083: goto st486;
	}
	if ( 49 <= _widec && _widec <= 57 )
		goto tr1649;
	goto tr1613;
tr1648:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st479;
st479:
	if ( ++p == pe )
		goto _test_eof479;
case 479:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1650;
		case 32: goto tr1650;
		case 40: goto tr1651;
		case 41: goto tr1652;
		case 778: goto tr1654;
		case 827: goto tr1654;
		case 1034: goto tr1655;
		case 1083: goto tr1656;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1653;
	goto tr1613;
tr1650:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1154;
tr1651:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1154;
tr1652:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1154;
tr1655:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1154;
st1154:
	if ( ++p == pe )
		goto _test_eof1154;
case 1154:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st480;
		case 32: goto st480;
		case 40: goto tr1658;
		case 41: goto tr1659;
		case 1034: goto tr1661;
		case 1083: goto st484;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1660;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1660;
	} else
		goto tr1660;
	goto tr1621;
tr1658:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st480;
tr1659:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st480;
tr1661:
	{
		s->line_counter++;
	}
	goto st480;
tr1675:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st480;
tr1676:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st480;
tr1677:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st480;
tr1678:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st480;
tr1679:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st480;
st480:
	if ( ++p == pe )
		goto _test_eof480;
case 480:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st480;
		case 32: goto st480;
		case 40: goto tr1658;
		case 41: goto tr1659;
		case 1034: goto tr1661;
		case 1083: goto st484;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1660;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1660;
	} else
		goto tr1660;
	goto tr1621;
tr1660:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st481;
st481:
	if ( ++p == pe )
		goto _test_eof481;
case 481:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1663;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr1663;
	} else
		goto tr1663;
	goto tr1621;
tr1665:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st482;
tr1666:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st482;
tr1667:
	{
		s->line_counter++;
	}
	goto st482;
tr1673:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st482;
tr1663:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st482;
st482:
	if ( ++p == pe )
		goto _test_eof482;
case 482:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st482;
		case 32: goto st482;
		case 40: goto tr1665;
		case 41: goto tr1666;
		case 2058: goto tr1667;
		case 2107: goto st483;
		case 2314: goto tr1669;
		case 2363: goto tr1669;
		case 2570: goto tr1670;
		case 2619: goto tr1671;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1660;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1660;
	} else
		goto tr1660;
	goto tr1621;
tr1672:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st483;
st483:
	if ( ++p == pe )
		goto _test_eof483;
case 483:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1673;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1672;
	goto tr1621;
tr1654:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1155;
tr1669:
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1155;
st1155:
	if ( ++p == pe )
		goto _test_eof1155;
case 1155:
	goto tr1607;
tr1670:
	{
		s->line_counter++;
	}
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1156;
st1156:
	if ( ++p == pe )
		goto _test_eof1156;
case 1156:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st482;
		case 32: goto st482;
		case 40: goto tr1665;
		case 41: goto tr1666;
		case 2058: goto tr1667;
		case 2107: goto st483;
		case 2314: goto tr1669;
		case 2363: goto tr1669;
		case 2570: goto tr1670;
		case 2619: goto tr1671;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1660;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1660;
	} else
		goto tr1660;
	goto tr1621;
tr1671:
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1157;
st1157:
	if ( ++p == pe )
		goto _test_eof1157;
case 1157:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1673;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1672;
	goto tr1621;
tr1674:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st484;
tr1680:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st484;
st484:
	if ( ++p == pe )
		goto _test_eof484;
case 484:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1675;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1674;
	goto tr1607;
tr1649:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st485;
tr1653:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st485;
st485:
	if ( ++p == pe )
		goto _test_eof485;
case 485:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1676;
		case 32: goto tr1676;
		case 40: goto tr1677;
		case 41: goto tr1678;
		case 1034: goto tr1679;
		case 1083: goto tr1680;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1653;
	goto tr1613;
tr1656:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1158;
st1158:
	if ( ++p == pe )
		goto _test_eof1158;
case 1158:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1675;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1674;
	goto tr1607;
tr1681:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st486;
st486:
	if ( ++p == pe )
		goto _test_eof486;
case 486:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1682;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1681;
	goto tr1607;
st487:
	if ( ++p == pe )
		goto _test_eof487;
case 487:
	switch( (*p) ) {
		case 68: goto st489;
		case 69: goto st504;
		case 73: goto st543;
		case 80: goto st551;
		case 82: goto st564;
		case 100: goto st489;
		case 101: goto st504;
		case 105: goto st543;
		case 112: goto st551;
		case 114: goto st564;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1684;
	goto tr1683;
tr1684:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st488;
tr1691:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st488;
st488:
	if ( ++p == pe )
		goto _test_eof488;
case 488:
	switch( (*p) ) {
		case 32: goto tr1690;
		case 59: goto tr1690;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1690;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1691;
	} else
		goto tr1690;
	goto tr1683;
tr1690:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
tr1694:
	{
		*(rdata_tail++) = 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
tr1696:
	{
		*(rdata_tail++) = 3;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
tr1708:
	{
		*(rdata_tail++) = 6;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
tr1718:
	{
		*(rdata_tail++) = 12;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
tr1732:
	{
		*(rdata_tail++) = 13;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
tr1741:
	{
		*(rdata_tail++) = 14;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
tr1748:
	{
		*(rdata_tail++) = 15;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
tr1751:
	{
		*(rdata_tail++) = 16;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
tr1759:
	{
		*(rdata_tail++) = 252;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
tr1770:
	{
		*(rdata_tail++) = 253;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
tr1773:
	{
		*(rdata_tail++) = 254;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
tr1780:
	{
		*(rdata_tail++) = 1;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
tr1786:
	{
		*(rdata_tail++) = 5;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
tr1798:
	{
		*(rdata_tail++) = 7;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
tr1801:
	{
		*(rdata_tail++) = 8;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
tr1804:
	{
		*(rdata_tail++) = 10;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
st1159:
	if ( ++p == pe )
		goto _test_eof1159;
case 1159:
	goto st0;
st489:
	if ( ++p == pe )
		goto _test_eof489;
case 489:
	switch( (*p) ) {
		case 72: goto st490;
		case 83: goto st491;
		case 104: goto st490;
		case 115: goto st491;
	}
	goto tr1683;
st490:
	if ( ++p == pe )
		goto _test_eof490;
case 490:
	switch( (*p) ) {
		case 32: goto tr1694;
		case 59: goto tr1694;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1694;
	} else if ( (*p) >= 9 )
		goto tr1694;
	goto tr1683;
st491:
	if ( ++p == pe )
		goto _test_eof491;
case 491:
	switch( (*p) ) {
		case 65: goto st492;
		case 97: goto st492;
	}
	goto tr1683;
st492:
	if ( ++p == pe )
		goto _test_eof492;
case 492:
	switch( (*p) ) {
		case 32: goto tr1696;
		case 45: goto st493;
		case 59: goto tr1696;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1696;
	} else if ( (*p) >= 9 )
		goto tr1696;
	goto tr1683;
st493:
	if ( ++p == pe )
		goto _test_eof493;
case 493:
	switch( (*p) ) {
		case 78: goto st494;
		case 110: goto st494;
	}
	goto tr1683;
st494:
	if ( ++p == pe )
		goto _test_eof494;
case 494:
	switch( (*p) ) {
		case 83: goto st495;
		case 115: goto st495;
	}
	goto tr1683;
st495:
	if ( ++p == pe )
		goto _test_eof495;
case 495:
	switch( (*p) ) {
		case 69: goto st496;
		case 101: goto st496;
	}
	goto tr1683;
st496:
	if ( ++p == pe )
		goto _test_eof496;
case 496:
	switch( (*p) ) {
		case 67: goto st497;
		case 99: goto st497;
	}
	goto tr1683;
st497:
	if ( ++p == pe )
		goto _test_eof497;
case 497:
	if ( (*p) == 51 )
		goto st498;
	goto tr1683;
st498:
	if ( ++p == pe )
		goto _test_eof498;
case 498:
	if ( (*p) == 45 )
		goto st499;
	goto tr1683;
st499:
	if ( ++p == pe )
		goto _test_eof499;
case 499:
	switch( (*p) ) {
		case 83: goto st500;
		case 115: goto st500;
	}
	goto tr1683;
st500:
	if ( ++p == pe )
		goto _test_eof500;
case 500:
	switch( (*p) ) {
		case 72: goto st501;
		case 104: goto st501;
	}
	goto tr1683;
st501:
	if ( ++p == pe )
		goto _test_eof501;
case 501:
	switch( (*p) ) {
		case 65: goto st502;
		case 97: goto st502;
	}
	goto tr1683;
st502:
	if ( ++p == pe )
		goto _test_eof502;
case 502:
	if ( (*p) == 49 )
		goto st503;
	goto tr1683;
st503:
	if ( ++p == pe )
		goto _test_eof503;
case 503:
	switch( (*p) ) {
		case 32: goto tr1708;
		case 59: goto tr1708;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1708;
	} else if ( (*p) >= 9 )
		goto tr1708;
	goto tr1683;
st504:
	if ( ++p == pe )
		goto _test_eof504;
case 504:
	switch( (*p) ) {
		case 67: goto st505;
		case 68: goto st534;
		case 99: goto st505;
		case 100: goto st534;
	}
	goto tr1683;
st505:
	if ( ++p == pe )
		goto _test_eof505;
case 505:
	switch( (*p) ) {
		case 67: goto st506;
		case 68: goto st512;
		case 99: goto st506;
		case 100: goto st512;
	}
	goto tr1683;
st506:
	if ( ++p == pe )
		goto _test_eof506;
case 506:
	if ( (*p) == 45 )
		goto st507;
	goto tr1683;
st507:
	if ( ++p == pe )
		goto _test_eof507;
case 507:
	switch( (*p) ) {
		case 71: goto st508;
		case 103: goto st508;
	}
	goto tr1683;
st508:
	if ( ++p == pe )
		goto _test_eof508;
case 508:
	switch( (*p) ) {
		case 79: goto st509;
		case 111: goto st509;
	}
	goto tr1683;
st509:
	if ( ++p == pe )
		goto _test_eof509;
case 509:
	switch( (*p) ) {
		case 83: goto st510;
		case 115: goto st510;
	}
	goto tr1683;
st510:
	if ( ++p == pe )
		goto _test_eof510;
case 510:
	switch( (*p) ) {
		case 84: goto st511;
		case 116: goto st511;
	}
	goto tr1683;
st511:
	if ( ++p == pe )
		goto _test_eof511;
case 511:
	switch( (*p) ) {
		case 32: goto tr1718;
		case 59: goto tr1718;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1718;
	} else if ( (*p) >= 9 )
		goto tr1718;
	goto tr1683;
st512:
	if ( ++p == pe )
		goto _test_eof512;
case 512:
	switch( (*p) ) {
		case 83: goto st513;
		case 115: goto st513;
	}
	goto tr1683;
st513:
	if ( ++p == pe )
		goto _test_eof513;
case 513:
	switch( (*p) ) {
		case 65: goto st514;
		case 97: goto st514;
	}
	goto tr1683;
st514:
	if ( ++p == pe )
		goto _test_eof514;
case 514:
	switch( (*p) ) {
		case 80: goto st515;
		case 112: goto st515;
	}
	goto tr1683;
st515:
	if ( ++p == pe )
		goto _test_eof515;
case 515:
	switch( (*p) ) {
		case 50: goto st516;
		case 51: goto st525;
	}
	goto tr1683;
st516:
	if ( ++p == pe )
		goto _test_eof516;
case 516:
	if ( (*p) == 53 )
		goto st517;
	goto tr1683;
st517:
	if ( ++p == pe )
		goto _test_eof517;
case 517:
	if ( (*p) == 54 )
		goto st518;
	goto tr1683;
st518:
	if ( ++p == pe )
		goto _test_eof518;
case 518:
	switch( (*p) ) {
		case 83: goto st519;
		case 115: goto st519;
	}
	goto tr1683;
st519:
	if ( ++p == pe )
		goto _test_eof519;
case 519:
	switch( (*p) ) {
		case 72: goto st520;
		case 104: goto st520;
	}
	goto tr1683;
st520:
	if ( ++p == pe )
		goto _test_eof520;
case 520:
	switch( (*p) ) {
		case 65: goto st521;
		case 97: goto st521;
	}
	goto tr1683;
st521:
	if ( ++p == pe )
		goto _test_eof521;
case 521:
	if ( (*p) == 50 )
		goto st522;
	goto tr1683;
st522:
	if ( ++p == pe )
		goto _test_eof522;
case 522:
	if ( (*p) == 53 )
		goto st523;
	goto tr1683;
st523:
	if ( ++p == pe )
		goto _test_eof523;
case 523:
	if ( (*p) == 54 )
		goto st524;
	goto tr1683;
st524:
	if ( ++p == pe )
		goto _test_eof524;
case 524:
	switch( (*p) ) {
		case 32: goto tr1732;
		case 59: goto tr1732;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1732;
	} else if ( (*p) >= 9 )
		goto tr1732;
	goto tr1683;
st525:
	if ( ++p == pe )
		goto _test_eof525;
case 525:
	if ( (*p) == 56 )
		goto st526;
	goto tr1683;
st526:
	if ( ++p == pe )
		goto _test_eof526;
case 526:
	if ( (*p) == 52 )
		goto st527;
	goto tr1683;
st527:
	if ( ++p == pe )
		goto _test_eof527;
case 527:
	switch( (*p) ) {
		case 83: goto st528;
		case 115: goto st528;
	}
	goto tr1683;
st528:
	if ( ++p == pe )
		goto _test_eof528;
case 528:
	switch( (*p) ) {
		case 72: goto st529;
		case 104: goto st529;
	}
	goto tr1683;
st529:
	if ( ++p == pe )
		goto _test_eof529;
case 529:
	switch( (*p) ) {
		case 65: goto st530;
		case 97: goto st530;
	}
	goto tr1683;
st530:
	if ( ++p == pe )
		goto _test_eof530;
case 530:
	if ( (*p) == 51 )
		goto st531;
	goto tr1683;
st531:
	if ( ++p == pe )
		goto _test_eof531;
case 531:
	if ( (*p) == 56 )
		goto st532;
	goto tr1683;
st532:
	if ( ++p == pe )
		goto _test_eof532;
case 532:
	if ( (*p) == 52 )
		goto st533;
	goto tr1683;
st533:
	if ( ++p == pe )
		goto _test_eof533;
case 533:
	switch( (*p) ) {
		case 32: goto tr1741;
		case 59: goto tr1741;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1741;
	} else if ( (*p) >= 9 )
		goto tr1741;
	goto tr1683;
st534:
	if ( ++p == pe )
		goto _test_eof534;
case 534:
	switch( (*p) ) {
		case 50: goto st535;
		case 52: goto st540;
	}
	goto tr1683;
st535:
	if ( ++p == pe )
		goto _test_eof535;
case 535:
	if ( (*p) == 53 )
		goto st536;
	goto tr1683;
st536:
	if ( ++p == pe )
		goto _test_eof536;
case 536:
	if ( (*p) == 53 )
		goto st537;
	goto tr1683;
st537:
	if ( ++p == pe )
		goto _test_eof537;
case 537:
	if ( (*p) == 49 )
		goto st538;
	goto tr1683;
st538:
	if ( ++p == pe )
		goto _test_eof538;
case 538:
	if ( (*p) == 57 )
		goto st539;
	goto tr1683;
st539:
	if ( ++p == pe )
		goto _test_eof539;
case 539:
	switch( (*p) ) {
		case 32: goto tr1748;
		case 59: goto tr1748;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1748;
	} else if ( (*p) >= 9 )
		goto tr1748;
	goto tr1683;
st540:
	if ( ++p == pe )
		goto _test_eof540;
case 540:
	if ( (*p) == 52 )
		goto st541;
	goto tr1683;
st541:
	if ( ++p == pe )
		goto _test_eof541;
case 541:
	if ( (*p) == 56 )
		goto st542;
	goto tr1683;
st542:
	if ( ++p == pe )
		goto _test_eof542;
case 542:
	switch( (*p) ) {
		case 32: goto tr1751;
		case 59: goto tr1751;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1751;
	} else if ( (*p) >= 9 )
		goto tr1751;
	goto tr1683;
st543:
	if ( ++p == pe )
		goto _test_eof543;
case 543:
	switch( (*p) ) {
		case 78: goto st544;
		case 110: goto st544;
	}
	goto tr1683;
st544:
	if ( ++p == pe )
		goto _test_eof544;
case 544:
	switch( (*p) ) {
		case 68: goto st545;
		case 100: goto st545;
	}
	goto tr1683;
st545:
	if ( ++p == pe )
		goto _test_eof545;
case 545:
	switch( (*p) ) {
		case 73: goto st546;
		case 105: goto st546;
	}
	goto tr1683;
st546:
	if ( ++p == pe )
		goto _test_eof546;
case 546:
	switch( (*p) ) {
		case 82: goto st547;
		case 114: goto st547;
	}
	goto tr1683;
st547:
	if ( ++p == pe )
		goto _test_eof547;
case 547:
	switch( (*p) ) {
		case 69: goto st548;
		case 101: goto st548;
	}
	goto tr1683;
st548:
	if ( ++p == pe )
		goto _test_eof548;
case 548:
	switch( (*p) ) {
		case 67: goto st549;
		case 99: goto st549;
	}
	goto tr1683;
st549:
	if ( ++p == pe )
		goto _test_eof549;
case 549:
	switch( (*p) ) {
		case 84: goto st550;
		case 116: goto st550;
	}
	goto tr1683;
st550:
	if ( ++p == pe )
		goto _test_eof550;
case 550:
	switch( (*p) ) {
		case 32: goto tr1759;
		case 59: goto tr1759;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1759;
	} else if ( (*p) >= 9 )
		goto tr1759;
	goto tr1683;
st551:
	if ( ++p == pe )
		goto _test_eof551;
case 551:
	switch( (*p) ) {
		case 82: goto st552;
		case 114: goto st552;
	}
	goto tr1683;
st552:
	if ( ++p == pe )
		goto _test_eof552;
case 552:
	switch( (*p) ) {
		case 73: goto st553;
		case 105: goto st553;
	}
	goto tr1683;
st553:
	if ( ++p == pe )
		goto _test_eof553;
case 553:
	switch( (*p) ) {
		case 86: goto st554;
		case 118: goto st554;
	}
	goto tr1683;
st554:
	if ( ++p == pe )
		goto _test_eof554;
case 554:
	switch( (*p) ) {
		case 65: goto st555;
		case 97: goto st555;
	}
	goto tr1683;
st555:
	if ( ++p == pe )
		goto _test_eof555;
case 555:
	switch( (*p) ) {
		case 84: goto st556;
		case 116: goto st556;
	}
	goto tr1683;
st556:
	if ( ++p == pe )
		goto _test_eof556;
case 556:
	switch( (*p) ) {
		case 69: goto st557;
		case 101: goto st557;
	}
	goto tr1683;
st557:
	if ( ++p == pe )
		goto _test_eof557;
case 557:
	switch( (*p) ) {
		case 68: goto st558;
		case 79: goto st561;
		case 100: goto st558;
		case 111: goto st561;
	}
	goto tr1683;
st558:
	if ( ++p == pe )
		goto _test_eof558;
case 558:
	switch( (*p) ) {
		case 78: goto st559;
		case 110: goto st559;
	}
	goto tr1683;
st559:
	if ( ++p == pe )
		goto _test_eof559;
case 559:
	switch( (*p) ) {
		case 83: goto st560;
		case 115: goto st560;
	}
	goto tr1683;
st560:
	if ( ++p == pe )
		goto _test_eof560;
case 560:
	switch( (*p) ) {
		case 32: goto tr1770;
		case 59: goto tr1770;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1770;
	} else if ( (*p) >= 9 )
		goto tr1770;
	goto tr1683;
st561:
	if ( ++p == pe )
		goto _test_eof561;
case 561:
	switch( (*p) ) {
		case 73: goto st562;
		case 105: goto st562;
	}
	goto tr1683;
st562:
	if ( ++p == pe )
		goto _test_eof562;
case 562:
	switch( (*p) ) {
		case 68: goto st563;
		case 100: goto st563;
	}
	goto tr1683;
st563:
	if ( ++p == pe )
		goto _test_eof563;
case 563:
	switch( (*p) ) {
		case 32: goto tr1773;
		case 59: goto tr1773;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1773;
	} else if ( (*p) >= 9 )
		goto tr1773;
	goto tr1683;
st564:
	if ( ++p == pe )
		goto _test_eof564;
case 564:
	switch( (*p) ) {
		case 83: goto st565;
		case 115: goto st565;
	}
	goto tr1683;
st565:
	if ( ++p == pe )
		goto _test_eof565;
case 565:
	switch( (*p) ) {
		case 65: goto st566;
		case 97: goto st566;
	}
	goto tr1683;
st566:
	if ( ++p == pe )
		goto _test_eof566;
case 566:
	switch( (*p) ) {
		case 77: goto st567;
		case 83: goto st570;
		case 109: goto st567;
		case 115: goto st570;
	}
	goto tr1683;
st567:
	if ( ++p == pe )
		goto _test_eof567;
case 567:
	switch( (*p) ) {
		case 68: goto st568;
		case 100: goto st568;
	}
	goto tr1683;
st568:
	if ( ++p == pe )
		goto _test_eof568;
case 568:
	if ( (*p) == 53 )
		goto st569;
	goto tr1683;
st569:
	if ( ++p == pe )
		goto _test_eof569;
case 569:
	switch( (*p) ) {
		case 32: goto tr1780;
		case 59: goto tr1780;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1780;
	} else if ( (*p) >= 9 )
		goto tr1780;
	goto tr1683;
st570:
	if ( ++p == pe )
		goto _test_eof570;
case 570:
	switch( (*p) ) {
		case 72: goto st571;
		case 104: goto st571;
	}
	goto tr1683;
st571:
	if ( ++p == pe )
		goto _test_eof571;
case 571:
	switch( (*p) ) {
		case 65: goto st572;
		case 97: goto st572;
	}
	goto tr1683;
st572:
	if ( ++p == pe )
		goto _test_eof572;
case 572:
	switch( (*p) ) {
		case 49: goto st573;
		case 50: goto st585;
		case 53: goto st588;
	}
	goto tr1683;
st573:
	if ( ++p == pe )
		goto _test_eof573;
case 573:
	switch( (*p) ) {
		case 32: goto tr1786;
		case 45: goto st574;
		case 59: goto tr1786;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1786;
	} else if ( (*p) >= 9 )
		goto tr1786;
	goto tr1683;
st574:
	if ( ++p == pe )
		goto _test_eof574;
case 574:
	switch( (*p) ) {
		case 78: goto st575;
		case 110: goto st575;
	}
	goto tr1683;
st575:
	if ( ++p == pe )
		goto _test_eof575;
case 575:
	switch( (*p) ) {
		case 83: goto st576;
		case 115: goto st576;
	}
	goto tr1683;
st576:
	if ( ++p == pe )
		goto _test_eof576;
case 576:
	switch( (*p) ) {
		case 69: goto st577;
		case 101: goto st577;
	}
	goto tr1683;
st577:
	if ( ++p == pe )
		goto _test_eof577;
case 577:
	switch( (*p) ) {
		case 67: goto st578;
		case 99: goto st578;
	}
	goto tr1683;
st578:
	if ( ++p == pe )
		goto _test_eof578;
case 578:
	if ( (*p) == 51 )
		goto st579;
	goto tr1683;
st579:
	if ( ++p == pe )
		goto _test_eof579;
case 579:
	if ( (*p) == 45 )
		goto st580;
	goto tr1683;
st580:
	if ( ++p == pe )
		goto _test_eof580;
case 580:
	switch( (*p) ) {
		case 83: goto st581;
		case 115: goto st581;
	}
	goto tr1683;
st581:
	if ( ++p == pe )
		goto _test_eof581;
case 581:
	switch( (*p) ) {
		case 72: goto st582;
		case 104: goto st582;
	}
	goto tr1683;
st582:
	if ( ++p == pe )
		goto _test_eof582;
case 582:
	switch( (*p) ) {
		case 65: goto st583;
		case 97: goto st583;
	}
	goto tr1683;
st583:
	if ( ++p == pe )
		goto _test_eof583;
case 583:
	if ( (*p) == 49 )
		goto st584;
	goto tr1683;
st584:
	if ( ++p == pe )
		goto _test_eof584;
case 584:
	switch( (*p) ) {
		case 32: goto tr1798;
		case 59: goto tr1798;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1798;
	} else if ( (*p) >= 9 )
		goto tr1798;
	goto tr1683;
st585:
	if ( ++p == pe )
		goto _test_eof585;
case 585:
	if ( (*p) == 53 )
		goto st586;
	goto tr1683;
st586:
	if ( ++p == pe )
		goto _test_eof586;
case 586:
	if ( (*p) == 54 )
		goto st587;
	goto tr1683;
st587:
	if ( ++p == pe )
		goto _test_eof587;
case 587:
	switch( (*p) ) {
		case 32: goto tr1801;
		case 59: goto tr1801;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1801;
	} else if ( (*p) >= 9 )
		goto tr1801;
	goto tr1683;
st588:
	if ( ++p == pe )
		goto _test_eof588;
case 588:
	if ( (*p) == 49 )
		goto st589;
	goto tr1683;
st589:
	if ( ++p == pe )
		goto _test_eof589;
case 589:
	if ( (*p) == 50 )
		goto st590;
	goto tr1683;
st590:
	if ( ++p == pe )
		goto _test_eof590;
case 590:
	switch( (*p) ) {
		case 32: goto tr1804;
		case 59: goto tr1804;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1804;
	} else if ( (*p) >= 9 )
		goto tr1804;
	goto tr1683;
st591:
	if ( ++p == pe )
		goto _test_eof591;
case 591:
	switch( (*p) ) {
		case 65: goto st593;
		case 73: goto st599;
		case 79: goto st616;
		case 80: goto st619;
		case 83: goto st625;
		case 85: goto st629;
		case 97: goto st593;
		case 105: goto st599;
		case 111: goto st616;
		case 112: goto st619;
		case 115: goto st625;
		case 117: goto st629;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1806;
	goto tr1805;
tr1806:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st592;
tr1814:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st592;
st592:
	if ( ++p == pe )
		goto _test_eof592;
case 592:
	switch( (*p) ) {
		case 32: goto tr1813;
		case 59: goto tr1813;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1813;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1814;
	} else
		goto tr1813;
	goto tr1805;
tr1813:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1160;
tr1820:
	{
		*((uint16_t *)rdata_tail) = htons(7);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1160;
tr1829:
	{
		*((uint16_t *)rdata_tail) = htons(8);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1160;
tr1833:
	{
		*((uint16_t *)rdata_tail) = htons(6);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1160;
tr1836:
	{
		*((uint16_t *)rdata_tail) = htons(4);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1160;
tr1840:
	{
		*((uint16_t *)rdata_tail) = htons(5);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1160;
tr1843:
	{
		*((uint16_t *)rdata_tail) = htons(254);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1160;
tr1847:
	{
		*((uint16_t *)rdata_tail) = htons(3);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1160;
tr1850:
	{
		*((uint16_t *)rdata_tail) = htons(1);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1160;
tr1854:
	{
		*((uint16_t *)rdata_tail) = htons(2);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1160;
tr1857:
	{
		*((uint16_t *)rdata_tail) = htons(253);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1160;
st1160:
	if ( ++p == pe )
		goto _test_eof1160;
case 1160:
	goto st0;
st593:
	if ( ++p == pe )
		goto _test_eof593;
case 593:
	switch( (*p) ) {
		case 67: goto st594;
		case 99: goto st594;
	}
	goto tr1805;
st594:
	if ( ++p == pe )
		goto _test_eof594;
case 594:
	switch( (*p) ) {
		case 80: goto st595;
		case 112: goto st595;
	}
	goto tr1805;
st595:
	if ( ++p == pe )
		goto _test_eof595;
case 595:
	switch( (*p) ) {
		case 75: goto st596;
		case 107: goto st596;
	}
	goto tr1805;
st596:
	if ( ++p == pe )
		goto _test_eof596;
case 596:
	switch( (*p) ) {
		case 73: goto st597;
		case 105: goto st597;
	}
	goto tr1805;
st597:
	if ( ++p == pe )
		goto _test_eof597;
case 597:
	switch( (*p) ) {
		case 88: goto st598;
		case 120: goto st598;
	}
	goto tr1805;
st598:
	if ( ++p == pe )
		goto _test_eof598;
case 598:
	switch( (*p) ) {
		case 32: goto tr1820;
		case 59: goto tr1820;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1820;
	} else if ( (*p) >= 9 )
		goto tr1820;
	goto tr1805;
st599:
	if ( ++p == pe )
		goto _test_eof599;
case 599:
	switch( (*p) ) {
		case 65: goto st600;
		case 80: goto st606;
		case 83: goto st612;
		case 97: goto st600;
		case 112: goto st606;
		case 115: goto st612;
	}
	goto tr1805;
st600:
	if ( ++p == pe )
		goto _test_eof600;
case 600:
	switch( (*p) ) {
		case 67: goto st601;
		case 99: goto st601;
	}
	goto tr1805;
st601:
	if ( ++p == pe )
		goto _test_eof601;
case 601:
	switch( (*p) ) {
		case 80: goto st602;
		case 112: goto st602;
	}
	goto tr1805;
st602:
	if ( ++p == pe )
		goto _test_eof602;
case 602:
	switch( (*p) ) {
		case 75: goto st603;
		case 107: goto st603;
	}
	goto tr1805;
st603:
	if ( ++p == pe )
		goto _test_eof603;
case 603:
	switch( (*p) ) {
		case 73: goto st604;
		case 105: goto st604;
	}
	goto tr1805;
st604:
	if ( ++p == pe )
		goto _test_eof604;
case 604:
	switch( (*p) ) {
		case 88: goto st605;
		case 120: goto st605;
	}
	goto tr1805;
st605:
	if ( ++p == pe )
		goto _test_eof605;
case 605:
	switch( (*p) ) {
		case 32: goto tr1829;
		case 59: goto tr1829;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1829;
	} else if ( (*p) >= 9 )
		goto tr1829;
	goto tr1805;
st606:
	if ( ++p == pe )
		goto _test_eof606;
case 606:
	switch( (*p) ) {
		case 71: goto st607;
		case 75: goto st609;
		case 103: goto st607;
		case 107: goto st609;
	}
	goto tr1805;
st607:
	if ( ++p == pe )
		goto _test_eof607;
case 607:
	switch( (*p) ) {
		case 80: goto st608;
		case 112: goto st608;
	}
	goto tr1805;
st608:
	if ( ++p == pe )
		goto _test_eof608;
case 608:
	switch( (*p) ) {
		case 32: goto tr1833;
		case 59: goto tr1833;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1833;
	} else if ( (*p) >= 9 )
		goto tr1833;
	goto tr1805;
st609:
	if ( ++p == pe )
		goto _test_eof609;
case 609:
	switch( (*p) ) {
		case 73: goto st610;
		case 105: goto st610;
	}
	goto tr1805;
st610:
	if ( ++p == pe )
		goto _test_eof610;
case 610:
	switch( (*p) ) {
		case 88: goto st611;
		case 120: goto st611;
	}
	goto tr1805;
st611:
	if ( ++p == pe )
		goto _test_eof611;
case 611:
	switch( (*p) ) {
		case 32: goto tr1836;
		case 59: goto tr1836;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1836;
	} else if ( (*p) >= 9 )
		goto tr1836;
	goto tr1805;
st612:
	if ( ++p == pe )
		goto _test_eof612;
case 612:
	switch( (*p) ) {
		case 80: goto st613;
		case 112: goto st613;
	}
	goto tr1805;
st613:
	if ( ++p == pe )
		goto _test_eof613;
case 613:
	switch( (*p) ) {
		case 75: goto st614;
		case 107: goto st614;
	}
	goto tr1805;
st614:
	if ( ++p == pe )
		goto _test_eof614;
case 614:
	switch( (*p) ) {
		case 73: goto st615;
		case 105: goto st615;
	}
	goto tr1805;
st615:
	if ( ++p == pe )
		goto _test_eof615;
case 615:
	switch( (*p) ) {
		case 32: goto tr1840;
		case 59: goto tr1840;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1840;
	} else if ( (*p) >= 9 )
		goto tr1840;
	goto tr1805;
st616:
	if ( ++p == pe )
		goto _test_eof616;
case 616:
	switch( (*p) ) {
		case 73: goto st617;
		case 105: goto st617;
	}
	goto tr1805;
st617:
	if ( ++p == pe )
		goto _test_eof617;
case 617:
	switch( (*p) ) {
		case 68: goto st618;
		case 100: goto st618;
	}
	goto tr1805;
st618:
	if ( ++p == pe )
		goto _test_eof618;
case 618:
	switch( (*p) ) {
		case 32: goto tr1843;
		case 59: goto tr1843;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1843;
	} else if ( (*p) >= 9 )
		goto tr1843;
	goto tr1805;
st619:
	if ( ++p == pe )
		goto _test_eof619;
case 619:
	switch( (*p) ) {
		case 71: goto st620;
		case 75: goto st622;
		case 103: goto st620;
		case 107: goto st622;
	}
	goto tr1805;
st620:
	if ( ++p == pe )
		goto _test_eof620;
case 620:
	switch( (*p) ) {
		case 80: goto st621;
		case 112: goto st621;
	}
	goto tr1805;
st621:
	if ( ++p == pe )
		goto _test_eof621;
case 621:
	switch( (*p) ) {
		case 32: goto tr1847;
		case 59: goto tr1847;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1847;
	} else if ( (*p) >= 9 )
		goto tr1847;
	goto tr1805;
st622:
	if ( ++p == pe )
		goto _test_eof622;
case 622:
	switch( (*p) ) {
		case 73: goto st623;
		case 105: goto st623;
	}
	goto tr1805;
st623:
	if ( ++p == pe )
		goto _test_eof623;
case 623:
	switch( (*p) ) {
		case 88: goto st624;
		case 120: goto st624;
	}
	goto tr1805;
st624:
	if ( ++p == pe )
		goto _test_eof624;
case 624:
	switch( (*p) ) {
		case 32: goto tr1850;
		case 59: goto tr1850;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1850;
	} else if ( (*p) >= 9 )
		goto tr1850;
	goto tr1805;
st625:
	if ( ++p == pe )
		goto _test_eof625;
case 625:
	switch( (*p) ) {
		case 80: goto st626;
		case 112: goto st626;
	}
	goto tr1805;
st626:
	if ( ++p == pe )
		goto _test_eof626;
case 626:
	switch( (*p) ) {
		case 75: goto st627;
		case 107: goto st627;
	}
	goto tr1805;
st627:
	if ( ++p == pe )
		goto _test_eof627;
case 627:
	switch( (*p) ) {
		case 73: goto st628;
		case 105: goto st628;
	}
	goto tr1805;
st628:
	if ( ++p == pe )
		goto _test_eof628;
case 628:
	switch( (*p) ) {
		case 32: goto tr1854;
		case 59: goto tr1854;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1854;
	} else if ( (*p) >= 9 )
		goto tr1854;
	goto tr1805;
st629:
	if ( ++p == pe )
		goto _test_eof629;
case 629:
	switch( (*p) ) {
		case 82: goto st630;
		case 114: goto st630;
	}
	goto tr1805;
st630:
	if ( ++p == pe )
		goto _test_eof630;
case 630:
	switch( (*p) ) {
		case 73: goto st631;
		case 105: goto st631;
	}
	goto tr1805;
st631:
	if ( ++p == pe )
		goto _test_eof631;
case 631:
	switch( (*p) ) {
		case 32: goto tr1857;
		case 59: goto tr1857;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1857;
	} else if ( (*p) >= 9 )
		goto tr1857;
	goto tr1805;
st632:
	if ( ++p == pe )
		goto _test_eof632;
case 632:
	if ( (*p) == 46 )
		goto tr1859;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1859;
	goto tr1858;
tr1859:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st633;
tr1861:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st633;
st633:
	if ( ++p == pe )
		goto _test_eof633;
case 633:
	switch( (*p) ) {
		case 32: goto tr1860;
		case 46: goto tr1861;
		case 59: goto tr1860;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1860;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1861;
	} else
		goto tr1860;
	goto tr1858;
tr1860:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1161;
st1161:
	if ( ++p == pe )
		goto _test_eof1161;
case 1161:
	goto st0;
st634:
	if ( ++p == pe )
		goto _test_eof634;
case 634:
	switch( (*p) ) {
		case 42: goto tr1862;
		case 92: goto tr1862;
		case 95: goto tr1862;
	}
	if ( (*p) < 64 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1862;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1862;
	} else
		goto tr1862;
	goto tr71;
tr1862:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 635; goto st270;} }
	goto st635;
st635:
	if ( ++p == pe )
		goto _test_eof635;
case 635:
	switch( (*p) ) {
		case 32: goto tr1863;
		case 59: goto tr1863;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1863;
	} else if ( (*p) >= 9 )
		goto tr1863;
	goto tr71;
tr1863:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1162;
st1162:
	if ( ++p == pe )
		goto _test_eof1162;
case 1162:
	goto st0;
st636:
	if ( ++p == pe )
		goto _test_eof636;
case 636:
	switch( (*p) ) {
		case 42: goto tr1864;
		case 92: goto tr1864;
		case 95: goto tr1864;
	}
	if ( (*p) < 64 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1864;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1864;
	} else
		goto tr1864;
	goto tr71;
tr1864:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 637; goto st270;} }
	goto st637;
st637:
	if ( ++p == pe )
		goto _test_eof637;
case 637:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1865;
		case 32: goto tr1865;
		case 40: goto tr1866;
		case 41: goto tr1867;
		case 1034: goto tr1868;
		case 1083: goto tr1869;
	}
	goto tr71;
tr1871:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st638;
tr1872:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st638;
tr1874:
	{
		s->line_counter++;
	}
	goto st638;
tr2018:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st638;
tr1865:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st638;
tr1866:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st638;
tr1867:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st638;
tr1868:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st638;
st638:
	if ( ++p == pe )
		goto _test_eof638;
case 638:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st638;
		case 32: goto st638;
		case 40: goto tr1871;
		case 41: goto tr1872;
		case 42: goto tr1873;
		case 92: goto tr1873;
		case 95: goto tr1873;
		case 1034: goto tr1874;
		case 1083: goto st667;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr1873;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1873;
	} else
		goto tr1873;
	goto tr71;
tr1873:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 639; goto st270;} }
	goto st639;
st639:
	if ( ++p == pe )
		goto _test_eof639;
case 639:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1876;
		case 32: goto tr1876;
		case 40: goto tr1877;
		case 41: goto tr1878;
		case 1034: goto tr1879;
		case 1083: goto tr1880;
	}
	goto tr71;
tr1883:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st640;
tr1884:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st640;
tr1886:
	{
		s->line_counter++;
	}
	goto st640;
tr2016:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st640;
tr1876:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st640;
tr1877:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st640;
tr1878:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st640;
tr1879:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st640;
st640:
	if ( ++p == pe )
		goto _test_eof640;
case 640:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st640;
		case 32: goto st640;
		case 40: goto tr1883;
		case 41: goto tr1884;
		case 1034: goto tr1886;
		case 1083: goto st666;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1885;
	goto tr1881;
tr1885:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st641;
tr1891:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st641;
st641:
	if ( ++p == pe )
		goto _test_eof641;
case 641:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1888;
		case 32: goto tr1888;
		case 40: goto tr1889;
		case 41: goto tr1890;
		case 1034: goto tr1892;
		case 1083: goto tr1893;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1891;
	goto tr1881;
tr1895:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st642;
tr1896:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st642;
tr1898:
	{
		s->line_counter++;
	}
	goto st642;
tr2014:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st642;
tr1888:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st642;
tr1889:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st642;
tr1890:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st642;
tr1892:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st642;
st642:
	if ( ++p == pe )
		goto _test_eof642;
case 642:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st642;
		case 32: goto st642;
		case 40: goto tr1895;
		case 41: goto tr1896;
		case 1034: goto tr1898;
		case 1083: goto st665;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1897;
	goto tr1881;
tr1897:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st643;
tr1904:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st643;
st643:
	if ( ++p == pe )
		goto _test_eof643;
case 643:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1901;
		case 32: goto tr1901;
		case 40: goto tr1902;
		case 41: goto tr1903;
		case 68: goto tr1905;
		case 72: goto tr1906;
		case 77: goto tr1907;
		case 83: goto st662;
		case 87: goto tr1909;
		case 100: goto tr1905;
		case 104: goto tr1906;
		case 109: goto tr1907;
		case 115: goto st662;
		case 119: goto tr1909;
		case 1034: goto tr1910;
		case 1083: goto tr1911;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1904;
	goto tr1900;
tr1913:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st644;
tr1914:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st644;
tr1916:
	{
		s->line_counter++;
	}
	goto st644;
tr1999:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st644;
tr1901:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st644;
tr1902:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st644;
tr1903:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st644;
tr1910:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st644;
tr2007:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st644;
tr2008:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st644;
tr2009:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st644;
tr2011:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st644;
st644:
	if ( ++p == pe )
		goto _test_eof644;
case 644:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st644;
		case 32: goto st644;
		case 40: goto tr1913;
		case 41: goto tr1914;
		case 1034: goto tr1916;
		case 1083: goto st661;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1915;
	goto tr1881;
tr1915:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st645;
tr1921:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st645;
st645:
	if ( ++p == pe )
		goto _test_eof645;
case 645:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1918;
		case 32: goto tr1918;
		case 40: goto tr1919;
		case 41: goto tr1920;
		case 68: goto tr1922;
		case 72: goto tr1923;
		case 77: goto tr1924;
		case 83: goto st658;
		case 87: goto tr1926;
		case 100: goto tr1922;
		case 104: goto tr1923;
		case 109: goto tr1924;
		case 115: goto st658;
		case 119: goto tr1926;
		case 1034: goto tr1927;
		case 1083: goto tr1928;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1921;
	goto tr1900;
tr1930:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st646;
tr1931:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st646;
tr1933:
	{
		s->line_counter++;
	}
	goto st646;
tr1984:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st646;
tr1918:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st646;
tr1919:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st646;
tr1920:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st646;
tr1927:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st646;
tr1992:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st646;
tr1993:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st646;
tr1994:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st646;
tr1996:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st646;
st646:
	if ( ++p == pe )
		goto _test_eof646;
case 646:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st646;
		case 32: goto st646;
		case 40: goto tr1930;
		case 41: goto tr1931;
		case 1034: goto tr1933;
		case 1083: goto st657;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1932;
	goto tr1881;
tr1932:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st647;
tr1938:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st647;
st647:
	if ( ++p == pe )
		goto _test_eof647;
case 647:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1935;
		case 32: goto tr1935;
		case 40: goto tr1936;
		case 41: goto tr1937;
		case 68: goto tr1939;
		case 72: goto tr1940;
		case 77: goto tr1941;
		case 83: goto st654;
		case 87: goto tr1943;
		case 100: goto tr1939;
		case 104: goto tr1940;
		case 109: goto tr1941;
		case 115: goto st654;
		case 119: goto tr1943;
		case 1034: goto tr1944;
		case 1083: goto tr1945;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1938;
	goto tr1900;
tr1947:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st648;
tr1948:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st648;
tr1950:
	{
		s->line_counter++;
	}
	goto st648;
tr1969:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st648;
tr1935:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st648;
tr1936:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st648;
tr1937:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st648;
tr1944:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st648;
tr1977:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st648;
tr1978:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st648;
tr1979:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st648;
tr1981:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st648;
st648:
	if ( ++p == pe )
		goto _test_eof648;
case 648:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st648;
		case 32: goto st648;
		case 40: goto tr1947;
		case 41: goto tr1948;
		case 1034: goto tr1950;
		case 1083: goto st653;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1949;
	goto tr1881;
tr1949:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st649;
tr1953:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st649;
st649:
	if ( ++p == pe )
		goto _test_eof649;
case 649:
	switch( (*p) ) {
		case 32: goto tr1952;
		case 59: goto tr1952;
		case 68: goto tr1954;
		case 72: goto tr1955;
		case 77: goto tr1956;
		case 83: goto st650;
		case 87: goto tr1958;
		case 100: goto tr1954;
		case 104: goto tr1955;
		case 109: goto tr1956;
		case 115: goto st650;
		case 119: goto tr1958;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1952;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1953;
	} else
		goto tr1952;
	goto tr1900;
tr1952:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1163;
tr1966:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1163;
st1163:
	if ( ++p == pe )
		goto _test_eof1163;
case 1163:
	goto st0;
tr1954:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st650;
tr1955:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st650;
tr1956:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st650;
tr1958:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st650;
st650:
	if ( ++p == pe )
		goto _test_eof650;
case 650:
	switch( (*p) ) {
		case 32: goto tr1952;
		case 59: goto tr1952;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1952;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1959;
	} else
		goto tr1952;
	goto tr1900;
tr1960:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st651;
tr1959:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st651;
tr1967:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st651;
st651:
	if ( ++p == pe )
		goto _test_eof651;
case 651:
	switch( (*p) ) {
		case 68: goto tr1961;
		case 72: goto tr1962;
		case 77: goto tr1963;
		case 83: goto st652;
		case 87: goto tr1965;
		case 100: goto tr1961;
		case 104: goto tr1962;
		case 109: goto tr1963;
		case 115: goto st652;
		case 119: goto tr1965;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1960;
	goto tr1900;
tr1961:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st652;
tr1962:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st652;
tr1963:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st652;
tr1965:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st652;
st652:
	if ( ++p == pe )
		goto _test_eof652;
case 652:
	switch( (*p) ) {
		case 32: goto tr1966;
		case 59: goto tr1966;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1966;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1967;
	} else
		goto tr1966;
	goto tr1900;
tr1968:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st653;
tr1945:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st653;
tr1982:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st653;
st653:
	if ( ++p == pe )
		goto _test_eof653;
case 653:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1969;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1968;
	goto tr71;
tr1939:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st654;
tr1940:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st654;
tr1941:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st654;
tr1943:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st654;
st654:
	if ( ++p == pe )
		goto _test_eof654;
case 654:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1935;
		case 32: goto tr1935;
		case 40: goto tr1936;
		case 41: goto tr1937;
		case 1034: goto tr1944;
		case 1083: goto tr1945;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1970;
	goto tr1900;
tr1971:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st655;
tr1970:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st655;
tr1980:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st655;
st655:
	if ( ++p == pe )
		goto _test_eof655;
case 655:
	switch( (*p) ) {
		case 68: goto tr1972;
		case 72: goto tr1973;
		case 77: goto tr1974;
		case 83: goto st656;
		case 87: goto tr1976;
		case 100: goto tr1972;
		case 104: goto tr1973;
		case 109: goto tr1974;
		case 115: goto st656;
		case 119: goto tr1976;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1971;
	goto tr1900;
tr1972:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st656;
tr1973:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st656;
tr1974:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st656;
tr1976:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st656;
st656:
	if ( ++p == pe )
		goto _test_eof656;
case 656:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1977;
		case 32: goto tr1977;
		case 40: goto tr1978;
		case 41: goto tr1979;
		case 1034: goto tr1981;
		case 1083: goto tr1982;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1980;
	goto tr1900;
tr1983:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st657;
tr1928:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st657;
tr1997:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st657;
st657:
	if ( ++p == pe )
		goto _test_eof657;
case 657:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1984;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1983;
	goto tr71;
tr1922:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st658;
tr1923:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st658;
tr1924:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st658;
tr1926:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st658;
st658:
	if ( ++p == pe )
		goto _test_eof658;
case 658:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1918;
		case 32: goto tr1918;
		case 40: goto tr1919;
		case 41: goto tr1920;
		case 1034: goto tr1927;
		case 1083: goto tr1928;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1985;
	goto tr1900;
tr1986:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st659;
tr1985:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st659;
tr1995:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st659;
st659:
	if ( ++p == pe )
		goto _test_eof659;
case 659:
	switch( (*p) ) {
		case 68: goto tr1987;
		case 72: goto tr1988;
		case 77: goto tr1989;
		case 83: goto st660;
		case 87: goto tr1991;
		case 100: goto tr1987;
		case 104: goto tr1988;
		case 109: goto tr1989;
		case 115: goto st660;
		case 119: goto tr1991;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1986;
	goto tr1900;
tr1987:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st660;
tr1988:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st660;
tr1989:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st660;
tr1991:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st660;
st660:
	if ( ++p == pe )
		goto _test_eof660;
case 660:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1992;
		case 32: goto tr1992;
		case 40: goto tr1993;
		case 41: goto tr1994;
		case 1034: goto tr1996;
		case 1083: goto tr1997;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1995;
	goto tr1900;
tr1998:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st661;
tr1911:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st661;
tr2012:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st661;
st661:
	if ( ++p == pe )
		goto _test_eof661;
case 661:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1999;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr1998;
	goto tr71;
tr1905:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st662;
tr1906:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st662;
tr1907:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st662;
tr1909:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st662;
st662:
	if ( ++p == pe )
		goto _test_eof662;
case 662:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1901;
		case 32: goto tr1901;
		case 40: goto tr1902;
		case 41: goto tr1903;
		case 1034: goto tr1910;
		case 1083: goto tr1911;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2000;
	goto tr1900;
tr2001:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st663;
tr2000:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st663;
tr2010:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st663;
st663:
	if ( ++p == pe )
		goto _test_eof663;
case 663:
	switch( (*p) ) {
		case 68: goto tr2002;
		case 72: goto tr2003;
		case 77: goto tr2004;
		case 83: goto st664;
		case 87: goto tr2006;
		case 100: goto tr2002;
		case 104: goto tr2003;
		case 109: goto tr2004;
		case 115: goto st664;
		case 119: goto tr2006;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2001;
	goto tr1900;
tr2002:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st664;
tr2003:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st664;
tr2004:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st664;
tr2006:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st268;}
	              }
	            }
	goto st664;
st664:
	if ( ++p == pe )
		goto _test_eof664;
case 664:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2007;
		case 32: goto tr2007;
		case 40: goto tr2008;
		case 41: goto tr2009;
		case 1034: goto tr2011;
		case 1083: goto tr2012;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2010;
	goto tr1900;
tr2013:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st665;
tr1893:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st665;
st665:
	if ( ++p == pe )
		goto _test_eof665;
case 665:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2014;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2013;
	goto tr71;
tr2015:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st666;
tr1880:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st666;
st666:
	if ( ++p == pe )
		goto _test_eof666;
case 666:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2016;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2015;
	goto tr71;
tr2017:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st667;
tr1869:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st667;
st667:
	if ( ++p == pe )
		goto _test_eof667;
case 667:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2018;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2017;
	goto tr71;
st668:
	if ( ++p == pe )
		goto _test_eof668;
case 668:
	switch( (*p) ) {
		case 32: goto tr71;
		case 59: goto tr71;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr71;
	} else if ( (*p) >= 9 )
		goto tr71;
	goto tr2019;
tr2019:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{ p--; {stack[top++] = 669; goto st279;} }
	goto st669;
st669:
	if ( ++p == pe )
		goto _test_eof669;
case 669:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2020;
		case 32: goto tr2020;
		case 40: goto tr2021;
		case 41: goto tr2022;
		case 1034: goto tr2023;
		case 1083: goto tr2024;
	}
	goto tr71;
tr2027:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st670;
tr2028:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st670;
tr2029:
	{
		s->line_counter++;
	}
	goto st670;
tr2033:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st670;
tr2020:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st670;
tr2021:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st670;
tr2022:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st670;
tr2023:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st670;
st670:
	if ( ++p == pe )
		goto _test_eof670;
case 670:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st670;
		case 32: goto st670;
		case 40: goto tr2027;
		case 41: goto tr2028;
		case 1034: goto tr2029;
		case 1083: goto st672;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2025;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2025;
	} else
		goto tr2025;
	goto tr71;
tr2025:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{ p--; {stack[top++] = 671; goto st279;} }
	goto st671;
st671:
	if ( ++p == pe )
		goto _test_eof671;
case 671:
	switch( (*p) ) {
		case 32: goto tr2031;
		case 59: goto tr2031;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2031;
	} else if ( (*p) >= 9 )
		goto tr2031;
	goto tr71;
tr2031:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1164;
st1164:
	if ( ++p == pe )
		goto _test_eof1164;
case 1164:
	goto st0;
tr2032:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st672;
tr2024:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st672;
st672:
	if ( ++p == pe )
		goto _test_eof672;
case 672:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2033;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2032;
	goto tr71;
st673:
	if ( ++p == pe )
		goto _test_eof673;
case 673:
	switch( (*p) ) {
		case 42: goto tr2034;
		case 92: goto tr2034;
		case 95: goto tr2034;
	}
	if ( (*p) < 64 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr2034;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr2034;
	} else
		goto tr2034;
	goto tr71;
tr2034:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 674; goto st270;} }
	goto st674;
st674:
	if ( ++p == pe )
		goto _test_eof674;
case 674:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2035;
		case 32: goto tr2035;
		case 40: goto tr2036;
		case 41: goto tr2037;
		case 1034: goto tr2038;
		case 1083: goto tr2039;
	}
	goto tr71;
tr2041:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st675;
tr2042:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st675;
tr2044:
	{
		s->line_counter++;
	}
	goto st675;
tr2048:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st675;
tr2035:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st675;
tr2036:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st675;
tr2037:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st675;
tr2038:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st675;
st675:
	if ( ++p == pe )
		goto _test_eof675;
case 675:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st675;
		case 32: goto st675;
		case 40: goto tr2041;
		case 41: goto tr2042;
		case 42: goto tr2043;
		case 92: goto tr2043;
		case 95: goto tr2043;
		case 1034: goto tr2044;
		case 1083: goto st677;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2043;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2043;
	} else
		goto tr2043;
	goto tr71;
tr2043:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 676; goto st270;} }
	goto st676;
st676:
	if ( ++p == pe )
		goto _test_eof676;
case 676:
	switch( (*p) ) {
		case 32: goto tr2046;
		case 59: goto tr2046;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2046;
	} else if ( (*p) >= 9 )
		goto tr2046;
	goto tr71;
tr2046:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1165;
st1165:
	if ( ++p == pe )
		goto _test_eof1165;
case 1165:
	goto st0;
tr2047:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st677;
tr2039:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st677;
st677:
	if ( ++p == pe )
		goto _test_eof677;
case 677:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2048;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2047;
	goto tr71;
st678:
	if ( ++p == pe )
		goto _test_eof678;
case 678:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2049;
	goto tr1881;
tr2049:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st679;
tr2053:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st679;
st679:
	if ( ++p == pe )
		goto _test_eof679;
case 679:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2050;
		case 32: goto tr2050;
		case 40: goto tr2051;
		case 41: goto tr2052;
		case 1034: goto tr2054;
		case 1083: goto tr2055;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2053;
	goto tr1881;
tr2057:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st680;
tr2058:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st680;
tr2060:
	{
		s->line_counter++;
	}
	goto st680;
tr2064:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st680;
tr2050:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st680;
tr2051:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st680;
tr2052:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st680;
tr2054:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st680;
st680:
	if ( ++p == pe )
		goto _test_eof680;
case 680:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st680;
		case 32: goto st680;
		case 40: goto tr2057;
		case 41: goto tr2058;
		case 42: goto tr2059;
		case 92: goto tr2059;
		case 95: goto tr2059;
		case 1034: goto tr2060;
		case 1083: goto st682;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2059;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2059;
	} else
		goto tr2059;
	goto tr71;
tr2059:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 681; goto st270;} }
	goto st681;
st681:
	if ( ++p == pe )
		goto _test_eof681;
case 681:
	switch( (*p) ) {
		case 32: goto tr2062;
		case 59: goto tr2062;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2062;
	} else if ( (*p) >= 9 )
		goto tr2062;
	goto tr71;
tr2062:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1166;
st1166:
	if ( ++p == pe )
		goto _test_eof1166;
case 1166:
	goto st0;
tr2063:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st682;
tr2055:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st682;
st682:
	if ( ++p == pe )
		goto _test_eof682;
case 682:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2064;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2063;
	goto tr71;
st683:
	if ( ++p == pe )
		goto _test_eof683;
case 683:
	switch( (*p) ) {
		case 32: goto tr2066;
		case 59: goto tr2066;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2066;
	} else if ( (*p) >= 9 )
		goto tr2066;
	goto tr2065;
tr2075:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{ p--; {stack[top++] = 684; goto st279;} }
	goto st684;
tr2065:
	{
		s->long_string = true;
	}
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{ p--; {stack[top++] = 684; goto st279;} }
	goto st684;
st684:
	if ( ++p == pe )
		goto _test_eof684;
case 684:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2067;
		case 32: goto tr2067;
		case 40: goto tr2068;
		case 41: goto tr2069;
		case 2058: goto tr2070;
		case 2107: goto tr2071;
		case 2314: goto tr2072;
		case 2363: goto tr2072;
		case 2570: goto tr2073;
		case 2619: goto tr2074;
	}
	goto tr2066;
tr2077:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st685;
tr2078:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st685;
tr2079:
	{
		s->line_counter++;
	}
	goto st685;
tr2085:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st685;
tr2067:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st685;
tr2068:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st685;
tr2069:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st685;
tr2070:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st685;
st685:
	if ( ++p == pe )
		goto _test_eof685;
case 685:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st685;
		case 32: goto st685;
		case 40: goto tr2077;
		case 41: goto tr2078;
		case 2058: goto tr2079;
		case 2107: goto st686;
		case 2314: goto tr2081;
		case 2363: goto tr2081;
		case 2570: goto tr2082;
		case 2619: goto tr2083;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2075;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2075;
	} else
		goto tr2075;
	goto tr2066;
tr2084:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st686;
tr2071:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st686;
st686:
	if ( ++p == pe )
		goto _test_eof686;
case 686:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2085;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2084;
	goto tr2066;
tr2072:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1167;
tr2081:
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1167;
st1167:
	if ( ++p == pe )
		goto _test_eof1167;
case 1167:
	goto st0;
tr2073:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1168;
tr2082:
	{
		s->line_counter++;
	}
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1168;
st1168:
	if ( ++p == pe )
		goto _test_eof1168;
case 1168:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st685;
		case 32: goto st685;
		case 40: goto tr2077;
		case 41: goto tr2078;
		case 2058: goto tr2079;
		case 2107: goto st686;
		case 2314: goto tr2081;
		case 2363: goto tr2081;
		case 2570: goto tr2082;
		case 2619: goto tr2083;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2075;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2075;
	} else
		goto tr2075;
	goto tr2066;
tr2074:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1169;
tr2083:
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1169;
st1169:
	if ( ++p == pe )
		goto _test_eof1169;
case 1169:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2085;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2084;
	goto tr2066;
st687:
	if ( ++p == pe )
		goto _test_eof687;
case 687:
	if ( (*p) == 46 )
		goto tr2086;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 58 )
			goto tr2086;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2086;
	} else
		goto tr2086;
	goto tr1858;
tr2086:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st688;
tr2088:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st688;
st688:
	if ( ++p == pe )
		goto _test_eof688;
case 688:
	switch( (*p) ) {
		case 32: goto tr2087;
		case 46: goto tr2088;
		case 59: goto tr2087;
	}
	if ( (*p) < 48 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr2087;
		} else if ( (*p) >= 9 )
			goto tr2087;
	} else if ( (*p) > 58 ) {
		if ( (*p) > 70 ) {
			if ( 97 <= (*p) && (*p) <= 102 )
				goto tr2088;
		} else if ( (*p) >= 65 )
			goto tr2088;
	} else
		goto tr2088;
	goto tr1858;
tr2087:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1170;
st1170:
	if ( ++p == pe )
		goto _test_eof1170;
case 1170:
	goto st0;
st689:
	if ( ++p == pe )
		goto _test_eof689;
case 689:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2090;
	goto tr2089;
tr2094:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st690;
tr2090:
	{
		memset(&(s->loc), 0, sizeof(s->loc));
		// Defaults.
		s->loc.siz = 100;
		s->loc.vp  = 1000;
		s->loc.hp  = 1000000;
		s->loc.lat_sign  = 1;
		s->loc.long_sign = 1;
		s->loc.alt_sign  = 1;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st690;
st690:
	if ( ++p == pe )
		goto _test_eof690;
case 690:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2091;
		case 32: goto tr2091;
		case 40: goto tr2092;
		case 41: goto tr2093;
		case 1034: goto tr2095;
		case 1083: goto tr2096;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2094;
	goto tr2089;
tr2098:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st691;
tr2099:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st691;
tr2103:
	{
		s->line_counter++;
	}
	goto st691;
tr2346:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st691;
tr2091:
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st691;
tr2092:
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st691;
tr2093:
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st691;
tr2095:
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st691;
st691:
	if ( ++p == pe )
		goto _test_eof691;
case 691:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st691;
		case 32: goto st691;
		case 40: goto tr2098;
		case 41: goto tr2099;
		case 78: goto st696;
		case 83: goto st738;
		case 1034: goto tr2103;
		case 1083: goto st743;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2100;
	goto tr2089;
tr2100:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st692;
tr2108:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st692;
st692:
	if ( ++p == pe )
		goto _test_eof692;
case 692:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2105;
		case 32: goto tr2105;
		case 40: goto tr2106;
		case 41: goto tr2107;
		case 1034: goto tr2109;
		case 1083: goto tr2110;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2108;
	goto tr2089;
tr2112:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st693;
tr2113:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st693;
tr2115:
	{
		s->line_counter++;
	}
	goto st693;
tr2344:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st693;
tr2105:
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st693;
tr2106:
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st693;
tr2107:
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st693;
tr2109:
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st693;
st693:
	if ( ++p == pe )
		goto _test_eof693;
case 693:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st693;
		case 32: goto st693;
		case 40: goto tr2112;
		case 41: goto tr2113;
		case 78: goto st696;
		case 83: goto st738;
		case 1034: goto tr2115;
		case 1083: goto st742;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2114;
	goto tr2089;
tr2121:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st694;
tr2114:
	{
		s->decimals = 3;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st694;
st694:
	if ( ++p == pe )
		goto _test_eof694;
case 694:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2117;
		case 32: goto tr2117;
		case 40: goto tr2118;
		case 41: goto tr2119;
		case 46: goto st740;
		case 1034: goto tr2122;
		case 1083: goto tr2123;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2121;
	goto tr2089;
tr2125:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st695;
tr2126:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st695;
tr2127:
	{
		s->line_counter++;
	}
	goto st695;
tr2335:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st695;
tr2117:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st695;
tr2118:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st695;
tr2119:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st695;
tr2122:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st695;
tr2336:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st695;
tr2337:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st695;
tr2338:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st695;
tr2340:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st695;
st695:
	if ( ++p == pe )
		goto _test_eof695;
case 695:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st695;
		case 32: goto st695;
		case 40: goto tr2125;
		case 41: goto tr2126;
		case 78: goto st696;
		case 83: goto st738;
		case 1034: goto tr2127;
		case 1083: goto st739;
	}
	goto tr2089;
st696:
	if ( ++p == pe )
		goto _test_eof696;
case 696:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st697;
		case 32: goto st697;
		case 40: goto tr2130;
		case 41: goto tr2131;
		case 1034: goto tr2132;
		case 1083: goto st737;
	}
	goto tr2089;
tr2130:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st697;
tr2131:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st697;
tr2132:
	{
		s->line_counter++;
	}
	goto st697;
tr2328:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st697;
tr2329:
	{
		s->loc.lat_sign = -1;
	}
	goto st697;
tr2330:
	{
		s->loc.lat_sign = -1;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st697;
tr2331:
	{
		s->loc.lat_sign = -1;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st697;
tr2332:
	{
		s->loc.lat_sign = -1;
	}
	{
		s->line_counter++;
	}
	goto st697;
st697:
	if ( ++p == pe )
		goto _test_eof697;
case 697:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st697;
		case 32: goto st697;
		case 40: goto tr2130;
		case 41: goto tr2131;
		case 1034: goto tr2132;
		case 1083: goto st737;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2134;
	goto tr2089;
tr2134:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st698;
tr2138:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st698;
st698:
	if ( ++p == pe )
		goto _test_eof698;
case 698:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2135;
		case 32: goto tr2135;
		case 40: goto tr2136;
		case 41: goto tr2137;
		case 1034: goto tr2139;
		case 1083: goto tr2140;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2138;
	goto tr2089;
tr2142:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st699;
tr2143:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st699;
tr2147:
	{
		s->line_counter++;
	}
	goto st699;
tr2326:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st699;
tr2135:
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st699;
tr2136:
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st699;
tr2137:
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st699;
tr2139:
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st699;
st699:
	if ( ++p == pe )
		goto _test_eof699;
case 699:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st699;
		case 32: goto st699;
		case 40: goto tr2142;
		case 41: goto tr2143;
		case 69: goto st704;
		case 87: goto st731;
		case 1034: goto tr2147;
		case 1083: goto st736;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2144;
	goto tr2089;
tr2144:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st700;
tr2152:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st700;
st700:
	if ( ++p == pe )
		goto _test_eof700;
case 700:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2149;
		case 32: goto tr2149;
		case 40: goto tr2150;
		case 41: goto tr2151;
		case 1034: goto tr2153;
		case 1083: goto tr2154;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2152;
	goto tr2089;
tr2156:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st701;
tr2157:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st701;
tr2159:
	{
		s->line_counter++;
	}
	goto st701;
tr2324:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st701;
tr2149:
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st701;
tr2150:
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st701;
tr2151:
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st701;
tr2153:
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st701;
st701:
	if ( ++p == pe )
		goto _test_eof701;
case 701:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st701;
		case 32: goto st701;
		case 40: goto tr2156;
		case 41: goto tr2157;
		case 69: goto st704;
		case 87: goto st731;
		case 1034: goto tr2159;
		case 1083: goto st735;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2158;
	goto tr2089;
tr2165:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st702;
tr2158:
	{
		s->decimals = 3;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st702;
st702:
	if ( ++p == pe )
		goto _test_eof702;
case 702:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2161;
		case 32: goto tr2161;
		case 40: goto tr2162;
		case 41: goto tr2163;
		case 46: goto st733;
		case 1034: goto tr2166;
		case 1083: goto tr2167;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2165;
	goto tr2089;
tr2169:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st703;
tr2170:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st703;
tr2171:
	{
		s->line_counter++;
	}
	goto st703;
tr2315:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st703;
tr2161:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st703;
tr2162:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st703;
tr2163:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st703;
tr2166:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st703;
tr2316:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st703;
tr2317:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st703;
tr2318:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st703;
tr2320:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st703;
st703:
	if ( ++p == pe )
		goto _test_eof703;
case 703:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st703;
		case 32: goto st703;
		case 40: goto tr2169;
		case 41: goto tr2170;
		case 69: goto st704;
		case 87: goto st731;
		case 1034: goto tr2171;
		case 1083: goto st732;
	}
	goto tr2089;
st704:
	if ( ++p == pe )
		goto _test_eof704;
case 704:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st705;
		case 32: goto st705;
		case 40: goto tr2174;
		case 41: goto tr2175;
		case 1034: goto tr2176;
		case 1083: goto st730;
	}
	goto tr2089;
tr2174:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st705;
tr2175:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st705;
tr2176:
	{
		s->line_counter++;
	}
	goto st705;
tr2308:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st705;
tr2309:
	{
		s->loc.long_sign = -1;
	}
	goto st705;
tr2310:
	{
		s->loc.long_sign = -1;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st705;
tr2311:
	{
		s->loc.long_sign = -1;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st705;
tr2312:
	{
		s->loc.long_sign = -1;
	}
	{
		s->line_counter++;
	}
	goto st705;
st705:
	if ( ++p == pe )
		goto _test_eof705;
case 705:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st705;
		case 32: goto st705;
		case 40: goto tr2174;
		case 41: goto tr2175;
		case 45: goto st706;
		case 1034: goto tr2176;
		case 1083: goto st730;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2179;
	goto tr2089;
st706:
	if ( ++p == pe )
		goto _test_eof706;
case 706:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2180;
	goto tr2089;
tr2185:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st707;
tr2179:
	{
		s->decimals = 2;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st707;
tr2180:
	{
		s->loc.alt_sign = -1;
	}
	{
		s->decimals = 2;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st707;
st707:
	if ( ++p == pe )
		goto _test_eof707;
case 707:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2181;
		case 32: goto tr2181;
		case 40: goto tr2182;
		case 41: goto tr2183;
		case 46: goto st727;
		case 109: goto tr2186;
		case 2058: goto tr2187;
		case 2107: goto tr2188;
		case 2314: goto tr2189;
		case 2363: goto tr2189;
		case 2570: goto tr2190;
		case 2619: goto tr2191;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2185;
	goto tr2089;
tr2193:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st708;
tr2194:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st708;
tr2196:
	{
		s->line_counter++;
	}
	goto st708;
tr2295:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st708;
tr2181:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st708;
tr2182:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st708;
tr2183:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st708;
tr2187:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st708;
tr2296:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st708;
tr2297:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st708;
tr2298:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st708;
tr2301:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st708;
st708:
	if ( ++p == pe )
		goto _test_eof708;
case 708:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st708;
		case 32: goto st708;
		case 40: goto tr2193;
		case 41: goto tr2194;
		case 2058: goto tr2196;
		case 2107: goto st726;
		case 2314: goto tr2198;
		case 2363: goto tr2198;
		case 2570: goto tr2199;
		case 2619: goto tr2200;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2195;
	goto tr2089;
tr2205:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st709;
tr2195:
	{
		s->decimals = 2;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st709;
st709:
	if ( ++p == pe )
		goto _test_eof709;
case 709:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2201;
		case 32: goto tr2201;
		case 40: goto tr2202;
		case 41: goto tr2203;
		case 46: goto st723;
		case 109: goto tr2206;
		case 2058: goto tr2207;
		case 2107: goto tr2208;
		case 2314: goto tr2209;
		case 2363: goto tr2209;
		case 2570: goto tr2210;
		case 2619: goto tr2211;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2205;
	goto tr2089;
tr2213:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st710;
tr2214:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st710;
tr2216:
	{
		s->line_counter++;
	}
	goto st710;
tr2282:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st710;
tr2201:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st710;
tr2202:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st710;
tr2203:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st710;
tr2207:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st710;
tr2283:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st710;
tr2284:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st710;
tr2285:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st710;
tr2288:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st710;
st710:
	if ( ++p == pe )
		goto _test_eof710;
case 710:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st710;
		case 32: goto st710;
		case 40: goto tr2213;
		case 41: goto tr2214;
		case 2058: goto tr2216;
		case 2107: goto st722;
		case 2314: goto tr2198;
		case 2363: goto tr2198;
		case 2570: goto tr2218;
		case 2619: goto tr2219;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2215;
	goto tr2089;
tr2224:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st711;
tr2215:
	{
		s->decimals = 2;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st711;
st711:
	if ( ++p == pe )
		goto _test_eof711;
case 711:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2220;
		case 32: goto tr2220;
		case 40: goto tr2221;
		case 41: goto tr2222;
		case 46: goto st719;
		case 109: goto tr2225;
		case 2058: goto tr2226;
		case 2107: goto tr2227;
		case 2314: goto tr2228;
		case 2363: goto tr2228;
		case 2570: goto tr2229;
		case 2619: goto tr2230;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2224;
	goto tr2089;
tr2232:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st712;
tr2233:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st712;
tr2235:
	{
		s->line_counter++;
	}
	goto st712;
tr2269:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st712;
tr2220:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st712;
tr2221:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st712;
tr2222:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st712;
tr2226:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st712;
tr2270:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st712;
tr2271:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st712;
tr2272:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st712;
tr2275:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st712;
st712:
	if ( ++p == pe )
		goto _test_eof712;
case 712:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st712;
		case 32: goto st712;
		case 40: goto tr2232;
		case 41: goto tr2233;
		case 2058: goto tr2235;
		case 2107: goto st718;
		case 2314: goto tr2198;
		case 2363: goto tr2198;
		case 2570: goto tr2237;
		case 2619: goto tr2238;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2234;
	goto tr2089;
tr2243:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st713;
tr2234:
	{
		s->decimals = 2;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st713;
st713:
	if ( ++p == pe )
		goto _test_eof713;
case 713:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2239;
		case 32: goto tr2239;
		case 40: goto tr2240;
		case 41: goto tr2241;
		case 46: goto st716;
		case 109: goto tr2239;
		case 2058: goto tr2244;
		case 2107: goto tr2245;
		case 2314: goto tr2246;
		case 2363: goto tr2246;
		case 2570: goto tr2247;
		case 2619: goto tr2248;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2243;
	goto tr2089;
tr2250:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st714;
tr2251:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st714;
tr2252:
	{
		s->line_counter++;
	}
	goto st714;
tr2257:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st714;
tr2239:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st714;
tr2240:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st714;
tr2241:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st714;
tr2244:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st714;
tr2258:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st714;
tr2259:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st714;
tr2260:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st714;
tr2262:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st714;
st714:
	if ( ++p == pe )
		goto _test_eof714;
case 714:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st714;
		case 32: goto st714;
		case 40: goto tr2250;
		case 41: goto tr2251;
		case 2058: goto tr2252;
		case 2107: goto st715;
		case 2314: goto tr2198;
		case 2363: goto tr2198;
		case 2570: goto tr2254;
		case 2619: goto tr2255;
	}
	goto tr2089;
tr2256:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st715;
tr2245:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st715;
tr2263:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st715;
st715:
	if ( ++p == pe )
		goto _test_eof715;
case 715:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2257;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2256;
	goto tr2089;
tr2189:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1171;
tr2198:
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1171;
tr2209:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1171;
tr2228:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1171;
tr2246:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1171;
tr2264:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1171;
tr2277:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1171;
tr2290:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1171;
tr2303:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1171;
st1171:
	if ( ++p == pe )
		goto _test_eof1171;
case 1171:
	goto st0;
tr2254:
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1172;
tr2247:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1172;
tr2265:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1172;
st1172:
	if ( ++p == pe )
		goto _test_eof1172;
case 1172:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st714;
		case 32: goto st714;
		case 40: goto tr2250;
		case 41: goto tr2251;
		case 2058: goto tr2252;
		case 2107: goto st715;
		case 2314: goto tr2198;
		case 2363: goto tr2198;
		case 2570: goto tr2254;
		case 2619: goto tr2255;
	}
	goto tr2089;
tr2255:
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1173;
tr2248:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1173;
tr2266:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1173;
st1173:
	if ( ++p == pe )
		goto _test_eof1173;
case 1173:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2257;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2256;
	goto tr2089;
st716:
	if ( ++p == pe )
		goto _test_eof716;
case 716:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2258;
		case 32: goto tr2258;
		case 40: goto tr2259;
		case 41: goto tr2260;
		case 109: goto tr2258;
		case 2058: goto tr2262;
		case 2107: goto tr2263;
		case 2314: goto tr2264;
		case 2363: goto tr2264;
		case 2570: goto tr2265;
		case 2619: goto tr2266;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2261;
	goto tr2089;
tr2261:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st717;
tr2267:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st717;
st717:
	if ( ++p == pe )
		goto _test_eof717;
case 717:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2239;
		case 32: goto tr2239;
		case 40: goto tr2240;
		case 41: goto tr2241;
		case 109: goto tr2239;
		case 2058: goto tr2244;
		case 2107: goto tr2245;
		case 2314: goto tr2246;
		case 2363: goto tr2246;
		case 2570: goto tr2247;
		case 2619: goto tr2248;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2267;
	goto tr2089;
tr2268:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st718;
tr2227:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st718;
tr2276:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st718;
st718:
	if ( ++p == pe )
		goto _test_eof718;
case 718:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2269;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2268;
	goto tr2089;
tr2237:
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1174;
tr2229:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1174;
tr2278:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1174;
st1174:
	if ( ++p == pe )
		goto _test_eof1174;
case 1174:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st712;
		case 32: goto st712;
		case 40: goto tr2232;
		case 41: goto tr2233;
		case 2058: goto tr2235;
		case 2107: goto st718;
		case 2314: goto tr2198;
		case 2363: goto tr2198;
		case 2570: goto tr2237;
		case 2619: goto tr2238;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2234;
	goto tr2089;
tr2238:
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1175;
tr2230:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1175;
tr2279:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1175;
st1175:
	if ( ++p == pe )
		goto _test_eof1175;
case 1175:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2269;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2268;
	goto tr2089;
st719:
	if ( ++p == pe )
		goto _test_eof719;
case 719:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2270;
		case 32: goto tr2270;
		case 40: goto tr2271;
		case 41: goto tr2272;
		case 109: goto tr2274;
		case 2058: goto tr2275;
		case 2107: goto tr2276;
		case 2314: goto tr2277;
		case 2363: goto tr2277;
		case 2570: goto tr2278;
		case 2619: goto tr2279;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2273;
	goto tr2089;
tr2273:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st720;
tr2280:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st720;
st720:
	if ( ++p == pe )
		goto _test_eof720;
case 720:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2220;
		case 32: goto tr2220;
		case 40: goto tr2221;
		case 41: goto tr2222;
		case 109: goto tr2225;
		case 2058: goto tr2226;
		case 2107: goto tr2227;
		case 2314: goto tr2228;
		case 2363: goto tr2228;
		case 2570: goto tr2229;
		case 2619: goto tr2230;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2280;
	goto tr2089;
tr2225:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st721;
tr2274:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st721;
st721:
	if ( ++p == pe )
		goto _test_eof721;
case 721:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st712;
		case 32: goto st712;
		case 40: goto tr2232;
		case 41: goto tr2233;
		case 2058: goto tr2235;
		case 2107: goto st718;
		case 2314: goto tr2198;
		case 2363: goto tr2198;
		case 2570: goto tr2237;
		case 2619: goto tr2238;
	}
	goto tr2089;
tr2281:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st722;
tr2208:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st722;
tr2289:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st722;
st722:
	if ( ++p == pe )
		goto _test_eof722;
case 722:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2282;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2281;
	goto tr2089;
tr2218:
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1176;
tr2210:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1176;
tr2291:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1176;
st1176:
	if ( ++p == pe )
		goto _test_eof1176;
case 1176:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st710;
		case 32: goto st710;
		case 40: goto tr2213;
		case 41: goto tr2214;
		case 2058: goto tr2216;
		case 2107: goto st722;
		case 2314: goto tr2198;
		case 2363: goto tr2198;
		case 2570: goto tr2218;
		case 2619: goto tr2219;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2215;
	goto tr2089;
tr2219:
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1177;
tr2211:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1177;
tr2292:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1177;
st1177:
	if ( ++p == pe )
		goto _test_eof1177;
case 1177:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2282;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2281;
	goto tr2089;
st723:
	if ( ++p == pe )
		goto _test_eof723;
case 723:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2283;
		case 32: goto tr2283;
		case 40: goto tr2284;
		case 41: goto tr2285;
		case 109: goto tr2287;
		case 2058: goto tr2288;
		case 2107: goto tr2289;
		case 2314: goto tr2290;
		case 2363: goto tr2290;
		case 2570: goto tr2291;
		case 2619: goto tr2292;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2286;
	goto tr2089;
tr2286:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st724;
tr2293:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st724;
st724:
	if ( ++p == pe )
		goto _test_eof724;
case 724:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2201;
		case 32: goto tr2201;
		case 40: goto tr2202;
		case 41: goto tr2203;
		case 109: goto tr2206;
		case 2058: goto tr2207;
		case 2107: goto tr2208;
		case 2314: goto tr2209;
		case 2363: goto tr2209;
		case 2570: goto tr2210;
		case 2619: goto tr2211;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2293;
	goto tr2089;
tr2206:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st725;
tr2287:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st725;
st725:
	if ( ++p == pe )
		goto _test_eof725;
case 725:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st710;
		case 32: goto st710;
		case 40: goto tr2213;
		case 41: goto tr2214;
		case 2058: goto tr2216;
		case 2107: goto st722;
		case 2314: goto tr2198;
		case 2363: goto tr2198;
		case 2570: goto tr2218;
		case 2619: goto tr2219;
	}
	goto tr2089;
tr2294:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st726;
tr2188:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st726;
tr2302:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st726;
st726:
	if ( ++p == pe )
		goto _test_eof726;
case 726:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2295;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2294;
	goto tr2089;
tr2190:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1178;
tr2199:
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1178;
tr2304:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1178;
st1178:
	if ( ++p == pe )
		goto _test_eof1178;
case 1178:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st708;
		case 32: goto st708;
		case 40: goto tr2193;
		case 41: goto tr2194;
		case 2058: goto tr2196;
		case 2107: goto st726;
		case 2314: goto tr2198;
		case 2363: goto tr2198;
		case 2570: goto tr2199;
		case 2619: goto tr2200;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2195;
	goto tr2089;
tr2191:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1179;
tr2200:
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1179;
tr2305:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1179;
st1179:
	if ( ++p == pe )
		goto _test_eof1179;
case 1179:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2295;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2294;
	goto tr2089;
st727:
	if ( ++p == pe )
		goto _test_eof727;
case 727:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2296;
		case 32: goto tr2296;
		case 40: goto tr2297;
		case 41: goto tr2298;
		case 109: goto tr2300;
		case 2058: goto tr2301;
		case 2107: goto tr2302;
		case 2314: goto tr2303;
		case 2363: goto tr2303;
		case 2570: goto tr2304;
		case 2619: goto tr2305;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2299;
	goto tr2089;
tr2299:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st728;
tr2306:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st728;
st728:
	if ( ++p == pe )
		goto _test_eof728;
case 728:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2181;
		case 32: goto tr2181;
		case 40: goto tr2182;
		case 41: goto tr2183;
		case 109: goto tr2186;
		case 2058: goto tr2187;
		case 2107: goto tr2188;
		case 2314: goto tr2189;
		case 2363: goto tr2189;
		case 2570: goto tr2190;
		case 2619: goto tr2191;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2306;
	goto tr2089;
tr2186:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st729;
tr2300:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st729;
st729:
	if ( ++p == pe )
		goto _test_eof729;
case 729:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st708;
		case 32: goto st708;
		case 40: goto tr2193;
		case 41: goto tr2194;
		case 2058: goto tr2196;
		case 2107: goto st726;
		case 2314: goto tr2198;
		case 2363: goto tr2198;
		case 2570: goto tr2199;
		case 2619: goto tr2200;
	}
	goto tr2089;
tr2307:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st730;
tr2313:
	{
		s->loc.long_sign = -1;
	}
	goto st730;
st730:
	if ( ++p == pe )
		goto _test_eof730;
case 730:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2308;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2307;
	goto tr2089;
st731:
	if ( ++p == pe )
		goto _test_eof731;
case 731:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2309;
		case 32: goto tr2309;
		case 40: goto tr2310;
		case 41: goto tr2311;
		case 1034: goto tr2312;
		case 1083: goto tr2313;
	}
	goto tr2089;
tr2314:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st732;
tr2167:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st732;
tr2321:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st732;
st732:
	if ( ++p == pe )
		goto _test_eof732;
case 732:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2315;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2314;
	goto tr2089;
st733:
	if ( ++p == pe )
		goto _test_eof733;
case 733:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2316;
		case 32: goto tr2316;
		case 40: goto tr2317;
		case 41: goto tr2318;
		case 1034: goto tr2320;
		case 1083: goto tr2321;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2319;
	goto tr2089;
tr2319:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st734;
tr2322:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st734;
st734:
	if ( ++p == pe )
		goto _test_eof734;
case 734:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2161;
		case 32: goto tr2161;
		case 40: goto tr2162;
		case 41: goto tr2163;
		case 1034: goto tr2166;
		case 1083: goto tr2167;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2322;
	goto tr2089;
tr2323:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st735;
tr2154:
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st735;
st735:
	if ( ++p == pe )
		goto _test_eof735;
case 735:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2324;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2323;
	goto tr2089;
tr2325:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st736;
tr2140:
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st736;
st736:
	if ( ++p == pe )
		goto _test_eof736;
case 736:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2326;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2325;
	goto tr2089;
tr2327:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st737;
tr2333:
	{
		s->loc.lat_sign = -1;
	}
	goto st737;
st737:
	if ( ++p == pe )
		goto _test_eof737;
case 737:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2328;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2327;
	goto tr2089;
st738:
	if ( ++p == pe )
		goto _test_eof738;
case 738:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2329;
		case 32: goto tr2329;
		case 40: goto tr2330;
		case 41: goto tr2331;
		case 1034: goto tr2332;
		case 1083: goto tr2333;
	}
	goto tr2089;
tr2334:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st739;
tr2123:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st739;
tr2341:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st739;
st739:
	if ( ++p == pe )
		goto _test_eof739;
case 739:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2335;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2334;
	goto tr2089;
st740:
	if ( ++p == pe )
		goto _test_eof740;
case 740:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2336;
		case 32: goto tr2336;
		case 40: goto tr2337;
		case 41: goto tr2338;
		case 1034: goto tr2340;
		case 1083: goto tr2341;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2339;
	goto tr2089;
tr2339:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st741;
tr2342:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st741;
st741:
	if ( ++p == pe )
		goto _test_eof741;
case 741:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2117;
		case 32: goto tr2117;
		case 40: goto tr2118;
		case 41: goto tr2119;
		case 1034: goto tr2122;
		case 1083: goto tr2123;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2342;
	goto tr2089;
tr2343:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st742;
tr2110:
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st742;
st742:
	if ( ++p == pe )
		goto _test_eof742;
case 742:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2344;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2343;
	goto tr2089;
tr2345:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st743;
tr2096:
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st268;}
		}
	}
	goto st743;
st743:
	if ( ++p == pe )
		goto _test_eof743;
case 743:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2346;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2345;
	goto tr2089;
st744:
	if ( ++p == pe )
		goto _test_eof744;
case 744:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2347;
	goto tr1881;
tr2347:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st745;
tr2351:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st745;
st745:
	if ( ++p == pe )
		goto _test_eof745;
case 745:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2348;
		case 32: goto tr2348;
		case 40: goto tr2349;
		case 41: goto tr2350;
		case 1034: goto tr2352;
		case 1083: goto tr2353;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2351;
	goto tr1881;
tr2355:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st746;
tr2356:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st746;
tr2358:
	{
		s->line_counter++;
	}
	goto st746;
tr2390:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st746;
tr2348:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st746;
tr2349:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st746;
tr2350:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st746;
tr2352:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st746;
st746:
	if ( ++p == pe )
		goto _test_eof746;
case 746:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st746;
		case 32: goto st746;
		case 40: goto tr2355;
		case 41: goto tr2356;
		case 1034: goto tr2358;
		case 1083: goto st754;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2357;
	goto tr1881;
tr2357:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st747;
tr2363:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st747;
st747:
	if ( ++p == pe )
		goto _test_eof747;
case 747:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2360;
		case 32: goto tr2360;
		case 40: goto tr2361;
		case 41: goto tr2362;
		case 1034: goto tr2364;
		case 1083: goto tr2365;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2363;
	goto tr1881;
tr2367:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st748;
tr2368:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st748;
tr2370:
	{
		s->line_counter++;
	}
	goto st748;
tr2388:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st748;
tr2360:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st748;
tr2361:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st748;
tr2362:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st748;
tr2364:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st748;
st748:
	if ( ++p == pe )
		goto _test_eof748;
case 748:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st748;
		case 32: goto st748;
		case 40: goto tr2367;
		case 41: goto tr2368;
		case 1034: goto tr2370;
		case 1083: goto st753;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2369;
	goto tr1881;
tr2369:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st749;
tr2375:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st749;
st749:
	if ( ++p == pe )
		goto _test_eof749;
case 749:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2372;
		case 32: goto tr2372;
		case 40: goto tr2373;
		case 41: goto tr2374;
		case 1034: goto tr2376;
		case 1083: goto tr2377;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2375;
	goto tr1881;
tr2379:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st750;
tr2380:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st750;
tr2382:
	{
		s->line_counter++;
	}
	goto st750;
tr2386:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st750;
tr2372:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st750;
tr2373:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st750;
tr2374:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st750;
tr2376:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st750;
st750:
	if ( ++p == pe )
		goto _test_eof750;
case 750:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st750;
		case 32: goto st750;
		case 40: goto tr2379;
		case 41: goto tr2380;
		case 42: goto tr2381;
		case 92: goto tr2381;
		case 95: goto tr2381;
		case 1034: goto tr2382;
		case 1083: goto st752;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2381;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2381;
	} else
		goto tr2381;
	goto tr71;
tr2381:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 751; goto st270;} }
	goto st751;
st751:
	if ( ++p == pe )
		goto _test_eof751;
case 751:
	switch( (*p) ) {
		case 32: goto tr2384;
		case 59: goto tr2384;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2384;
	} else if ( (*p) >= 9 )
		goto tr2384;
	goto tr71;
tr2384:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1180;
st1180:
	if ( ++p == pe )
		goto _test_eof1180;
case 1180:
	goto st0;
tr2385:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st752;
tr2377:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st752;
st752:
	if ( ++p == pe )
		goto _test_eof752;
case 752:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2386;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2385;
	goto tr71;
tr2387:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st753;
tr2365:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st753;
st753:
	if ( ++p == pe )
		goto _test_eof753;
case 753:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2388;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2387;
	goto tr71;
tr2389:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st754;
tr2353:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st754;
st754:
	if ( ++p == pe )
		goto _test_eof754;
case 754:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2390;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2389;
	goto tr71;
st755:
	if ( ++p == pe )
		goto _test_eof755;
case 755:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2391;
	goto tr1881;
tr2391:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st756;
tr2395:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st756;
st756:
	if ( ++p == pe )
		goto _test_eof756;
case 756:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2392;
		case 32: goto tr2392;
		case 40: goto tr2393;
		case 41: goto tr2394;
		case 1034: goto tr2396;
		case 1083: goto tr2397;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2395;
	goto tr1881;
tr2399:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st757;
tr2400:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st757;
tr2402:
	{
		s->line_counter++;
	}
	goto st757;
tr2459:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st757;
tr2392:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st757;
tr2393:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st757;
tr2394:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st757;
tr2396:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st757;
st757:
	if ( ++p == pe )
		goto _test_eof757;
case 757:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st757;
		case 32: goto st757;
		case 40: goto tr2399;
		case 41: goto tr2400;
		case 1034: goto tr2402;
		case 1083: goto st771;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2401;
	goto tr1881;
tr2401:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st758;
tr2407:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st758;
st758:
	if ( ++p == pe )
		goto _test_eof758;
case 758:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2404;
		case 32: goto tr2404;
		case 40: goto tr2405;
		case 41: goto tr2406;
		case 1034: goto tr2408;
		case 1083: goto tr2409;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2407;
	goto tr1881;
tr2412:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st759;
tr2413:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st759;
tr2414:
	{
		s->line_counter++;
	}
	goto st759;
tr2457:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st759;
tr2404:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st759;
tr2405:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st759;
tr2406:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st759;
tr2408:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st759;
st759:
	if ( ++p == pe )
		goto _test_eof759;
case 759:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st759;
		case 32: goto st759;
		case 40: goto tr2412;
		case 41: goto tr2413;
		case 1034: goto tr2414;
		case 1083: goto st770;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2410;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2410;
	} else
		goto tr2410;
	goto tr71;
tr2410:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{ p--; {stack[top++] = 760; goto st279;} }
	goto st760;
st760:
	if ( ++p == pe )
		goto _test_eof760;
case 760:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2416;
		case 32: goto tr2416;
		case 40: goto tr2417;
		case 41: goto tr2418;
		case 1034: goto tr2419;
		case 1083: goto tr2420;
	}
	goto tr71;
tr2423:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st761;
tr2424:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st761;
tr2425:
	{
		s->line_counter++;
	}
	goto st761;
tr2455:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st761;
tr2416:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st761;
tr2417:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st761;
tr2418:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st761;
tr2419:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st761;
st761:
	if ( ++p == pe )
		goto _test_eof761;
case 761:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st761;
		case 32: goto st761;
		case 40: goto tr2423;
		case 41: goto tr2424;
		case 1034: goto tr2425;
		case 1083: goto st769;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2421;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2421;
	} else
		goto tr2421;
	goto tr71;
tr2421:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{ p--; {stack[top++] = 762; goto st279;} }
	goto st762;
st762:
	if ( ++p == pe )
		goto _test_eof762;
case 762:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2427;
		case 32: goto tr2427;
		case 40: goto tr2428;
		case 41: goto tr2429;
		case 1034: goto tr2430;
		case 1083: goto tr2431;
	}
	goto tr71;
tr2434:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st763;
tr2435:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st763;
tr2436:
	{
		s->line_counter++;
	}
	goto st763;
tr2453:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st763;
tr2427:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st763;
tr2428:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st763;
tr2429:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st763;
tr2430:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st763;
st763:
	if ( ++p == pe )
		goto _test_eof763;
case 763:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st763;
		case 32: goto st763;
		case 40: goto tr2434;
		case 41: goto tr2435;
		case 1034: goto tr2436;
		case 1083: goto st768;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2432;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2432;
	} else
		goto tr2432;
	goto tr71;
tr2432:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{ p--; {stack[top++] = 764; goto st279;} }
	goto st764;
st764:
	if ( ++p == pe )
		goto _test_eof764;
case 764:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2438;
		case 32: goto tr2438;
		case 40: goto tr2439;
		case 41: goto tr2440;
		case 1034: goto tr2441;
		case 1083: goto tr2442;
	}
	goto tr71;
tr2444:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st765;
tr2445:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st765;
tr2447:
	{
		s->line_counter++;
	}
	goto st765;
tr2451:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st765;
tr2438:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st765;
tr2439:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st765;
tr2440:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st765;
tr2441:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st765;
st765:
	if ( ++p == pe )
		goto _test_eof765;
case 765:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st765;
		case 32: goto st765;
		case 40: goto tr2444;
		case 41: goto tr2445;
		case 42: goto tr2446;
		case 92: goto tr2446;
		case 95: goto tr2446;
		case 1034: goto tr2447;
		case 1083: goto st767;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2446;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2446;
	} else
		goto tr2446;
	goto tr71;
tr2446:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 766; goto st270;} }
	goto st766;
st766:
	if ( ++p == pe )
		goto _test_eof766;
case 766:
	switch( (*p) ) {
		case 32: goto tr2449;
		case 59: goto tr2449;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2449;
	} else if ( (*p) >= 9 )
		goto tr2449;
	goto tr71;
tr2449:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1181;
st1181:
	if ( ++p == pe )
		goto _test_eof1181;
case 1181:
	goto st0;
tr2450:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st767;
tr2442:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st767;
st767:
	if ( ++p == pe )
		goto _test_eof767;
case 767:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2451;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2450;
	goto tr71;
tr2452:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st768;
tr2431:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st768;
st768:
	if ( ++p == pe )
		goto _test_eof768;
case 768:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2453;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2452;
	goto tr71;
tr2454:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st769;
tr2420:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st769;
st769:
	if ( ++p == pe )
		goto _test_eof769;
case 769:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2455;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2454;
	goto tr71;
tr2456:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st770;
tr2409:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st770;
st770:
	if ( ++p == pe )
		goto _test_eof770;
case 770:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2457;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2456;
	goto tr71;
tr2458:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st771;
tr2397:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st771;
st771:
	if ( ++p == pe )
		goto _test_eof771;
case 771:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2459;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2458;
	goto tr71;
st772:
	if ( ++p == pe )
		goto _test_eof772;
case 772:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr2460;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr2460;
	} else
		goto tr2460;
	goto tr71;
tr2460:
	{ p--; {stack[top++] = 773; goto st591;} }
	goto st773;
st773:
	if ( ++p == pe )
		goto _test_eof773;
case 773:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st774;
		case 32: goto st774;
		case 40: goto tr2462;
		case 41: goto tr2463;
		case 1034: goto tr2464;
		case 1083: goto st782;
	}
	goto tr71;
tr2462:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st774;
tr2463:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st774;
tr2464:
	{
		s->line_counter++;
	}
	goto st774;
tr2491:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st774;
st774:
	if ( ++p == pe )
		goto _test_eof774;
case 774:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st774;
		case 32: goto st774;
		case 40: goto tr2462;
		case 41: goto tr2463;
		case 1034: goto tr2464;
		case 1083: goto st782;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2466;
	goto tr1881;
tr2466:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st775;
tr2470:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st775;
st775:
	if ( ++p == pe )
		goto _test_eof775;
case 775:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2467;
		case 32: goto tr2467;
		case 40: goto tr2468;
		case 41: goto tr2469;
		case 1034: goto tr2471;
		case 1083: goto tr2472;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2470;
	goto tr1881;
tr2474:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st776;
tr2475:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st776;
tr2477:
	{
		s->line_counter++;
	}
	goto st776;
tr2489:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st776;
tr2467:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st776;
tr2468:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st776;
tr2469:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st776;
tr2471:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st776;
st776:
	if ( ++p == pe )
		goto _test_eof776;
case 776:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st776;
		case 32: goto st776;
		case 40: goto tr2474;
		case 41: goto tr2475;
		case 1034: goto tr2477;
		case 1083: goto st781;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2476;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2476;
	} else
		goto tr2476;
	goto tr71;
tr2476:
	{ p--; {stack[top++] = 777; goto st487;} }
	goto st777;
st777:
	if ( ++p == pe )
		goto _test_eof777;
case 777:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st778;
		case 32: goto st778;
		case 40: goto tr2480;
		case 41: goto tr2481;
		case 1034: goto tr2482;
		case 1083: goto st780;
	}
	goto tr71;
tr2480:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st778;
tr2481:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st778;
tr2482:
	{
		s->line_counter++;
	}
	goto st778;
tr2487:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st778;
st778:
	if ( ++p == pe )
		goto _test_eof778;
case 778:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st778;
		case 32: goto st778;
		case 40: goto tr2480;
		case 41: goto tr2481;
		case 43: goto tr2484;
		case 1034: goto tr2482;
		case 1083: goto st780;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr2484;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2484;
	} else
		goto tr2484;
	goto tr71;
tr2484:
	{ p--; {stack[top++] = 779; goto st329;} }
	goto st779;
st779:
	if ( ++p == pe )
		goto _test_eof779;
case 779:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1152 + ((*p) - -128));
			if (
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1152 + ((*p) - -128));
		if (
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr2485;
		case 1595: goto tr2485;
	}
	goto tr71;
tr2485:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1182;
st1182:
	if ( ++p == pe )
		goto _test_eof1182;
case 1182:
	goto st0;
tr2486:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st780;
st780:
	if ( ++p == pe )
		goto _test_eof780;
case 780:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2487;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2486;
	goto tr71;
tr2488:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st781;
tr2472:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st781;
st781:
	if ( ++p == pe )
		goto _test_eof781;
case 781:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2489;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2488;
	goto tr71;
tr2490:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st782;
st782:
	if ( ++p == pe )
		goto _test_eof782;
case 782:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2491;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2490;
	goto tr71;
tr2495:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st783;
tr2496:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st783;
tr2499:
	{
		s->line_counter++;
	}
	goto st783;
tr2522:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st783;
tr2512:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st268;}
		}
	}
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st268;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	goto st783;
tr2513:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st268;}
		}
	}
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st268;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st783;
tr2514:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st268;}
		}
	}
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st268;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st783;
tr2516:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st268;}
		}
	}
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st268;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		s->line_counter++;
	}
	goto st783;
st783:
	if ( ++p == pe )
		goto _test_eof783;
case 783:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st783;
		case 32: goto st783;
		case 33: goto tr2494;
		case 40: goto tr2495;
		case 41: goto tr2496;
		case 49: goto tr2497;
		case 50: goto tr2498;
		case 2058: goto tr2499;
		case 2107: goto st790;
		case 2314: goto tr2501;
		case 2363: goto tr2501;
		case 2570: goto tr2502;
		case 2619: goto tr2503;
	}
	goto tr2492;
tr2494:
	{
		memset(&(s->apl), 0, sizeof(s->apl));
	}
	{
		s->apl.excl_flag = 128; // dec 128  = bin 10000000.
	}
	goto st784;
st784:
	if ( ++p == pe )
		goto _test_eof784;
case 784:
	switch( (*p) ) {
		case 49: goto tr2504;
		case 50: goto tr2505;
	}
	goto tr2492;
tr2497:
	{
		memset(&(s->apl), 0, sizeof(s->apl));
	}
	{
		s->apl.addr_family = 1;
	}
	goto st785;
tr2504:
	{
		s->apl.addr_family = 1;
	}
	goto st785;
st785:
	if ( ++p == pe )
		goto _test_eof785;
case 785:
	if ( (*p) == 58 )
		goto st786;
	goto tr2492;
st786:
	if ( ++p == pe )
		goto _test_eof786;
case 786:
	if ( (*p) == 46 )
		goto tr2508;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2508;
	goto tr2507;
tr2508:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st787;
tr2509:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st787;
st787:
	if ( ++p == pe )
		goto _test_eof787;
case 787:
	if ( (*p) == 47 )
		goto tr2510;
	if ( 46 <= (*p) && (*p) <= 57 )
		goto tr2509;
	goto tr2507;
tr2510:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	goto st788;
tr2526:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	goto st788;
st788:
	if ( ++p == pe )
		goto _test_eof788;
case 788:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2511;
	goto tr2492;
tr2511:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st789;
tr2515:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st789;
st789:
	if ( ++p == pe )
		goto _test_eof789;
case 789:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2512;
		case 32: goto tr2512;
		case 40: goto tr2513;
		case 41: goto tr2514;
		case 2058: goto tr2516;
		case 2107: goto tr2517;
		case 2314: goto tr2518;
		case 2363: goto tr2518;
		case 2570: goto tr2519;
		case 2619: goto tr2520;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2515;
	goto tr2492;
tr2521:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st790;
tr2517:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st268;}
		}
	}
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st268;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	goto st790;
st790:
	if ( ++p == pe )
		goto _test_eof790;
case 790:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2522;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2521;
	goto tr71;
tr2501:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1183;
tr2518:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st268;}
		}
	}
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st268;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1183;
st1183:
	if ( ++p == pe )
		goto _test_eof1183;
case 1183:
	goto st0;
tr2502:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1184;
tr2519:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st268;}
		}
	}
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st268;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1184;
st1184:
	if ( ++p == pe )
		goto _test_eof1184;
case 1184:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st783;
		case 32: goto st783;
		case 33: goto tr2494;
		case 40: goto tr2495;
		case 41: goto tr2496;
		case 49: goto tr2497;
		case 50: goto tr2498;
		case 2058: goto tr2499;
		case 2107: goto st790;
		case 2314: goto tr2501;
		case 2363: goto tr2501;
		case 2570: goto tr2502;
		case 2619: goto tr2503;
	}
	goto tr2492;
tr2498:
	{
		memset(&(s->apl), 0, sizeof(s->apl));
	}
	{
		s->apl.addr_family = 2;
	}
	goto st791;
tr2505:
	{
		s->apl.addr_family = 2;
	}
	goto st791;
st791:
	if ( ++p == pe )
		goto _test_eof791;
case 791:
	if ( (*p) == 58 )
		goto st792;
	goto tr2492;
st792:
	if ( ++p == pe )
		goto _test_eof792;
case 792:
	if ( (*p) == 46 )
		goto tr2524;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 58 )
			goto tr2524;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2524;
	} else
		goto tr2524;
	goto tr2507;
tr2524:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st793;
tr2525:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st793;
st793:
	if ( ++p == pe )
		goto _test_eof793;
case 793:
	if ( (*p) == 47 )
		goto tr2526;
	if ( (*p) < 65 ) {
		if ( 46 <= (*p) && (*p) <= 58 )
			goto tr2525;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2525;
	} else
		goto tr2525;
	goto tr2507;
tr2503:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1185;
tr2520:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st268;}
		}
	}
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st268;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1185;
st1185:
	if ( ++p == pe )
		goto _test_eof1185;
case 1185:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2522;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2521;
	goto tr71;
st794:
	if ( ++p == pe )
		goto _test_eof794;
case 794:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2527;
	goto tr1881;
tr2527:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st795;
tr2531:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st795;
st795:
	if ( ++p == pe )
		goto _test_eof795;
case 795:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2528;
		case 32: goto tr2528;
		case 40: goto tr2529;
		case 41: goto tr2530;
		case 1034: goto tr2532;
		case 1083: goto tr2533;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2531;
	goto tr1881;
tr2535:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st796;
tr2536:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st796;
tr2538:
	{
		s->line_counter++;
	}
	goto st796;
tr2575:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st796;
tr2528:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st796;
tr2529:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st796;
tr2530:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st796;
tr2532:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st796;
st796:
	if ( ++p == pe )
		goto _test_eof796;
case 796:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st796;
		case 32: goto st796;
		case 40: goto tr2535;
		case 41: goto tr2536;
		case 1034: goto tr2538;
		case 1083: goto st806;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2537;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2537;
	} else
		goto tr2537;
	goto tr71;
tr2537:
	{ p--; {stack[top++] = 797; goto st487;} }
	goto st797;
st797:
	if ( ++p == pe )
		goto _test_eof797;
case 797:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st798;
		case 32: goto st798;
		case 40: goto tr2541;
		case 41: goto tr2542;
		case 1034: goto tr2543;
		case 1083: goto st805;
	}
	goto tr71;
tr2541:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st798;
tr2542:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st798;
tr2543:
	{
		s->line_counter++;
	}
	goto st798;
tr2573:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st798;
st798:
	if ( ++p == pe )
		goto _test_eof798;
case 798:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st798;
		case 32: goto st798;
		case 40: goto tr2541;
		case 41: goto tr2542;
		case 1034: goto tr2543;
		case 1083: goto st805;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2545;
	goto tr1881;
tr2545:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st799;
tr2549:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st799;
st799:
	if ( ++p == pe )
		goto _test_eof799;
case 799:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2546;
		case 32: goto tr2546;
		case 40: goto tr2547;
		case 41: goto tr2548;
		case 1034: goto tr2550;
		case 1083: goto tr2551;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2549;
	goto tr1881;
tr2554:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st800;
tr2555:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st800;
tr2557:
	{
		s->line_counter++;
	}
	goto st800;
tr2571:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st800;
tr2546:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st800;
tr2547:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st800;
tr2548:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st800;
tr2550:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st800;
st800:
	if ( ++p == pe )
		goto _test_eof800;
case 800:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st800;
		case 32: goto st800;
		case 40: goto tr2554;
		case 41: goto tr2555;
		case 1034: goto tr2557;
		case 1083: goto st804;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2556;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2556;
	} else
		goto tr2556;
	goto tr2552;
tr2556:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st801;
st801:
	if ( ++p == pe )
		goto _test_eof801;
case 801:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr2559;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2559;
	} else
		goto tr2559;
	goto tr2552;
tr2561:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st802;
tr2562:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st802;
tr2563:
	{
		s->line_counter++;
	}
	goto st802;
tr2569:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st802;
tr2559:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st802;
st802:
	if ( ++p == pe )
		goto _test_eof802;
case 802:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st802;
		case 32: goto st802;
		case 40: goto tr2561;
		case 41: goto tr2562;
		case 2058: goto tr2563;
		case 2107: goto st803;
		case 2314: goto tr2565;
		case 2363: goto tr2565;
		case 2570: goto tr2566;
		case 2619: goto tr2567;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2556;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2556;
	} else
		goto tr2556;
	goto tr2552;
tr2568:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st803;
st803:
	if ( ++p == pe )
		goto _test_eof803;
case 803:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2569;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2568;
	goto tr2552;
tr2565:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1186;
st1186:
	if ( ++p == pe )
		goto _test_eof1186;
case 1186:
	goto st0;
tr2566:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1187;
st1187:
	if ( ++p == pe )
		goto _test_eof1187;
case 1187:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st802;
		case 32: goto st802;
		case 40: goto tr2561;
		case 41: goto tr2562;
		case 2058: goto tr2563;
		case 2107: goto st803;
		case 2314: goto tr2565;
		case 2363: goto tr2565;
		case 2570: goto tr2566;
		case 2619: goto tr2567;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2556;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2556;
	} else
		goto tr2556;
	goto tr2552;
tr2567:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1188;
st1188:
	if ( ++p == pe )
		goto _test_eof1188;
case 1188:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2569;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2568;
	goto tr2552;
tr2570:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st804;
tr2551:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st804;
st804:
	if ( ++p == pe )
		goto _test_eof804;
case 804:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2571;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2570;
	goto tr71;
tr2572:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st805;
st805:
	if ( ++p == pe )
		goto _test_eof805;
case 805:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2573;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2572;
	goto tr71;
tr2574:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st806;
tr2533:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st806;
st806:
	if ( ++p == pe )
		goto _test_eof806;
case 806:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2575;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2574;
	goto tr71;
st807:
	if ( ++p == pe )
		goto _test_eof807;
case 807:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2576;
	goto tr1881;
tr2576:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st808;
tr2580:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st808;
st808:
	if ( ++p == pe )
		goto _test_eof808;
case 808:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2577;
		case 32: goto tr2577;
		case 40: goto tr2578;
		case 41: goto tr2579;
		case 1034: goto tr2581;
		case 1083: goto tr2582;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2580;
	goto tr1881;
tr2584:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st809;
tr2585:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st809;
tr2587:
	{
		s->line_counter++;
	}
	goto st809;
tr2615:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st809;
tr2577:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st809;
tr2578:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st809;
tr2579:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st809;
tr2581:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st809;
st809:
	if ( ++p == pe )
		goto _test_eof809;
case 809:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st809;
		case 32: goto st809;
		case 40: goto tr2584;
		case 41: goto tr2585;
		case 1034: goto tr2587;
		case 1083: goto st816;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2586;
	goto tr1881;
tr2586:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st810;
tr2592:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st810;
st810:
	if ( ++p == pe )
		goto _test_eof810;
case 810:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2589;
		case 32: goto tr2589;
		case 40: goto tr2590;
		case 41: goto tr2591;
		case 1034: goto tr2593;
		case 1083: goto tr2594;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2592;
	goto tr1881;
tr2596:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st811;
tr2597:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st811;
tr2599:
	{
		s->line_counter++;
	}
	goto st811;
tr2613:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st811;
tr2589:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st811;
tr2590:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st811;
tr2591:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st811;
tr2593:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st811;
st811:
	if ( ++p == pe )
		goto _test_eof811;
case 811:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st811;
		case 32: goto st811;
		case 40: goto tr2596;
		case 41: goto tr2597;
		case 1034: goto tr2599;
		case 1083: goto st815;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2598;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2598;
	} else
		goto tr2598;
	goto tr2552;
tr2598:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st812;
st812:
	if ( ++p == pe )
		goto _test_eof812;
case 812:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr2601;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2601;
	} else
		goto tr2601;
	goto tr2552;
tr2603:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st813;
tr2604:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st813;
tr2605:
	{
		s->line_counter++;
	}
	goto st813;
tr2611:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st813;
tr2601:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st813;
st813:
	if ( ++p == pe )
		goto _test_eof813;
case 813:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st813;
		case 32: goto st813;
		case 40: goto tr2603;
		case 41: goto tr2604;
		case 2058: goto tr2605;
		case 2107: goto st814;
		case 2314: goto tr2607;
		case 2363: goto tr2607;
		case 2570: goto tr2608;
		case 2619: goto tr2609;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2598;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2598;
	} else
		goto tr2598;
	goto tr2552;
tr2610:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st814;
st814:
	if ( ++p == pe )
		goto _test_eof814;
case 814:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2611;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2610;
	goto tr2552;
tr2607:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1189;
st1189:
	if ( ++p == pe )
		goto _test_eof1189;
case 1189:
	goto st0;
tr2608:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1190;
st1190:
	if ( ++p == pe )
		goto _test_eof1190;
case 1190:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st813;
		case 32: goto st813;
		case 40: goto tr2603;
		case 41: goto tr2604;
		case 2058: goto tr2605;
		case 2107: goto st814;
		case 2314: goto tr2607;
		case 2363: goto tr2607;
		case 2570: goto tr2608;
		case 2619: goto tr2609;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2598;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2598;
	} else
		goto tr2598;
	goto tr2552;
tr2609:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1191;
st1191:
	if ( ++p == pe )
		goto _test_eof1191;
case 1191:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2611;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2610;
	goto tr2552;
tr2612:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st815;
tr2594:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st815;
st815:
	if ( ++p == pe )
		goto _test_eof815;
case 815:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2613;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2612;
	goto tr71;
tr2614:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st816;
tr2582:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st816;
st816:
	if ( ++p == pe )
		goto _test_eof816;
case 816:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2615;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2614;
	goto tr71;
st817:
	if ( ++p == pe )
		goto _test_eof817;
case 817:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2616;
	goto tr1881;
tr2616:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st818;
tr2620:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st818;
st818:
	if ( ++p == pe )
		goto _test_eof818;
case 818:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2617;
		case 32: goto tr2617;
		case 40: goto tr2618;
		case 41: goto tr2619;
		case 1034: goto tr2621;
		case 1083: goto tr2622;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2620;
	goto tr1881;
tr2625:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st819;
tr2626:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st819;
tr2631:
	{
		s->line_counter++;
	}
	goto st819;
tr2824:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st819;
tr2617:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st819;
tr2618:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st819;
tr2619:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st819;
tr2621:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st819;
st819:
	if ( ++p == pe )
		goto _test_eof819;
case 819:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st819;
		case 32: goto st819;
		case 40: goto tr2625;
		case 41: goto tr2626;
		case 48: goto tr2627;
		case 49: goto tr2628;
		case 50: goto tr2629;
		case 51: goto tr2630;
		case 1034: goto tr2631;
		case 1083: goto st855;
	}
	goto tr2623;
tr2627:
	{
		*(rdata_tail++) = 0;
	}
	goto st820;
st820:
	if ( ++p == pe )
		goto _test_eof820;
case 820:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st821;
		case 32: goto st821;
		case 40: goto tr2634;
		case 41: goto tr2635;
		case 1034: goto tr2636;
		case 1083: goto st833;
	}
	goto tr2623;
tr2634:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st821;
tr2635:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st821;
tr2636:
	{
		s->line_counter++;
	}
	goto st821;
tr2686:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st821;
st821:
	if ( ++p == pe )
		goto _test_eof821;
case 821:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st821;
		case 32: goto st821;
		case 40: goto tr2634;
		case 41: goto tr2635;
		case 1034: goto tr2636;
		case 1083: goto st833;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2639;
	goto tr2638;
tr2639:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st822;
tr2643:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st822;
st822:
	if ( ++p == pe )
		goto _test_eof822;
case 822:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2640;
		case 32: goto tr2640;
		case 40: goto tr2641;
		case 41: goto tr2642;
		case 1034: goto tr2644;
		case 1083: goto tr2645;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2643;
	goto tr2638;
tr2647:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st823;
tr2648:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st823;
tr2650:
	{
		s->line_counter++;
	}
	goto st823;
tr2684:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st823;
tr2640:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st823;
tr2641:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st823;
tr2642:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st823;
tr2644:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st823;
st823:
	if ( ++p == pe )
		goto _test_eof823;
case 823:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st823;
		case 32: goto st823;
		case 40: goto tr2647;
		case 41: goto tr2648;
		case 46: goto st824;
		case 1034: goto tr2650;
		case 1083: goto st832;
	}
	goto tr2623;
st824:
	if ( ++p == pe )
		goto _test_eof824;
case 824:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 != 0  ) _widec += 1024;
				if (
 s->number64 == 0  ) _widec += 2048;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
				if (
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 != 0  ) _widec += 1024;
				if (
 s->number64 == 0  ) _widec += 2048;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(5760 + ((*p) - -128));
		if (
 s->number64 != 0  ) _widec += 256;
		if (
 s->number64 == 0  ) _widec += 512;
	}
	switch( _widec ) {
		case 6153: goto st825;
		case 6176: goto st825;
		case 6184: goto tr2654;
		case 6185: goto tr2655;
		case 6409: goto st828;
		case 6432: goto st828;
		case 6440: goto tr2657;
		case 6441: goto tr2658;
		case 6665: goto st830;
		case 6688: goto st830;
		case 6696: goto tr2660;
		case 6697: goto tr2661;
		case 9482: goto tr2662;
		case 9531: goto tr2662;
		case 9738: goto tr2662;
		case 9787: goto tr2662;
		case 10250: goto tr2663;
		case 10299: goto st827;
		case 10506: goto tr2662;
		case 10555: goto tr2662;
		case 10762: goto tr2665;
		case 10811: goto tr2666;
		case 11274: goto tr2667;
		case 11323: goto st829;
		case 11530: goto tr2662;
		case 11579: goto tr2662;
		case 11786: goto tr2669;
		case 11835: goto tr2670;
		case 12298: goto tr2671;
		case 12347: goto st831;
		case 12554: goto tr2662;
		case 12603: goto tr2662;
		case 12810: goto tr2673;
		case 12859: goto tr2674;
	}
	goto tr2652;
tr2654:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st825;
tr2655:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st825;
tr2663:
	{
		s->line_counter++;
	}
	goto st825;
tr2678:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st825;
tr2797:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st825;
tr2798:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st825;
tr2799:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st825;
tr2807:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st825;
tr2708:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st825;
tr2709:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st825;
tr2710:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st825;
tr2718:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st825;
tr2753:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st825;
tr2754:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st825;
tr2755:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st825;
tr2763:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st825;
st825:
	if ( ++p == pe )
		goto _test_eof825;
case 825:
	_widec = (*p);
	if ( (*p) < 43 ) {
		if ( (*p) < 32 ) {
			if ( (*p) > 9 ) {
				if ( 10 <= (*p) && (*p) <= 10 ) {
					_widec = (short)(3200 + ((*p) - -128));
					if (
 s->multiline  ) _widec += 256;
					if (
 s->number64 != 0  ) _widec += 512;
				}
			} else if ( (*p) >= 9 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 32 ) {
			if ( (*p) > 40 ) {
				if ( 41 <= (*p) && (*p) <= 41 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 40 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(2688 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
		}
	} else if ( (*p) > 43 ) {
		if ( (*p) < 59 ) {
			if ( (*p) > 47 ) {
				if ( 48 <= (*p) && (*p) <= 57 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 47 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 59 ) {
			if ( (*p) > 90 ) {
				if ( 97 <= (*p) && (*p) <= 122 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 65 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(3200 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(2688 + ((*p) - -128));
		if (
 s->number64 != 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 3081: goto st825;
		case 3104: goto st825;
		case 3112: goto tr2654;
		case 3113: goto tr2655;
		case 3115: goto tr2676;
		case 4106: goto tr2663;
		case 4155: goto st827;
	}
	if ( _widec < 3137 ) {
		if ( 3119 <= _widec && _widec <= 3129 )
			goto tr2676;
	} else if ( _widec > 3162 ) {
		if ( 3169 <= _widec && _widec <= 3194 )
			goto tr2676;
	} else
		goto tr2676;
	goto tr2675;
tr2676:
	{ p--; {stack[top++] = 826; goto st329;} }
	goto st826;
st826:
	if ( ++p == pe )
		goto _test_eof826;
case 826:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1152 + ((*p) - -128));
			if (
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1152 + ((*p) - -128));
		if (
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr2662;
		case 1595: goto tr2662;
	}
	goto tr2675;
tr2662:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1192;
tr2717:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1192;
tr2806:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1192;
tr2762:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1192;
st1192:
	if ( ++p == pe )
		goto _test_eof1192;
case 1192:
	goto st0;
tr2677:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st827;
tr2808:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st827;
tr2719:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st827;
tr2764:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st827;
st827:
	if ( ++p == pe )
		goto _test_eof827;
case 827:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(3200 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(3200 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(3200 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 s->number64 != 0  ) _widec += 512;
	}
	if ( _widec == 4106 )
		goto tr2678;
	if ( 3968 <= _widec && _widec <= 4223 )
		goto tr2677;
	goto tr2675;
tr2657:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st828;
tr2658:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st828;
tr2667:
	{
		s->line_counter++;
	}
	goto st828;
tr2680:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st828;
tr2800:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st828;
tr2801:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st828;
tr2802:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st828;
tr2811:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st828;
tr2711:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st828;
tr2712:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st828;
tr2713:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st828;
tr2722:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st828;
tr2756:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st828;
tr2757:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st828;
tr2758:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st828;
tr2767:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st828;
st828:
	if ( ++p == pe )
		goto _test_eof828;
case 828:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(12928 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 == 0  ) _widec += 1024;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(4224 + ((*p) - -128));
			if (
 s->number64 == 0  ) _widec += 256;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(4224 + ((*p) - -128));
				if (
 s->number64 == 0  ) _widec += 256;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(12928 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 == 0  ) _widec += 1024;
			}
		} else {
			_widec = (short)(4224 + ((*p) - -128));
			if (
 s->number64 == 0  ) _widec += 256;
		}
	} else {
		_widec = (short)(4224 + ((*p) - -128));
		if (
 s->number64 == 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 4617: goto st828;
		case 4640: goto st828;
		case 4648: goto tr2657;
		case 4649: goto tr2658;
		case 13578: goto tr2662;
		case 13627: goto tr2662;
		case 13834: goto tr2662;
		case 13883: goto tr2662;
		case 14346: goto tr2667;
		case 14395: goto st829;
		case 14602: goto tr2662;
		case 14651: goto tr2662;
		case 14858: goto tr2669;
		case 14907: goto tr2670;
	}
	goto tr2675;
tr2679:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st829;
tr2812:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st829;
tr2723:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st829;
tr2768:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st829;
st829:
	if ( ++p == pe )
		goto _test_eof829;
case 829:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(4736 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(4736 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(4736 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 s->number64 == 0  ) _widec += 512;
	}
	if ( _widec == 5642 )
		goto tr2680;
	if ( 5504 <= _widec && _widec <= 5759 )
		goto tr2679;
	goto tr2675;
tr2669:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1193;
tr2724:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1193;
tr2769:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1193;
tr2813:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1193;
st1193:
	if ( ++p == pe )
		goto _test_eof1193;
case 1193:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(12928 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 == 0  ) _widec += 1024;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(4224 + ((*p) - -128));
			if (
 s->number64 == 0  ) _widec += 256;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(4224 + ((*p) - -128));
				if (
 s->number64 == 0  ) _widec += 256;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(12928 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 == 0  ) _widec += 1024;
			}
		} else {
			_widec = (short)(4224 + ((*p) - -128));
			if (
 s->number64 == 0  ) _widec += 256;
		}
	} else {
		_widec = (short)(4224 + ((*p) - -128));
		if (
 s->number64 == 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 4617: goto st828;
		case 4640: goto st828;
		case 4648: goto tr2657;
		case 4649: goto tr2658;
		case 13578: goto tr2662;
		case 13627: goto tr2662;
		case 13834: goto tr2662;
		case 13883: goto tr2662;
		case 14346: goto tr2667;
		case 14395: goto st829;
		case 14602: goto tr2662;
		case 14651: goto tr2662;
		case 14858: goto tr2669;
		case 14907: goto tr2670;
	}
	goto tr2675;
tr2670:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1194;
tr2725:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1194;
tr2814:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1194;
tr2770:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1194;
st1194:
	if ( ++p == pe )
		goto _test_eof1194;
case 1194:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(4736 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(4736 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(4736 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 s->number64 == 0  ) _widec += 512;
	}
	if ( _widec == 5642 )
		goto tr2680;
	if ( 5504 <= _widec && _widec <= 5759 )
		goto tr2679;
	goto tr2675;
tr2660:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st830;
tr2661:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st830;
tr2671:
	{
		s->line_counter++;
	}
	goto st830;
tr2682:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st830;
tr2803:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st830;
tr2804:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st830;
tr2805:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st830;
tr2815:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st830;
tr2714:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st830;
tr2715:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st830;
tr2716:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st830;
tr2726:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st830;
tr2759:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st830;
tr2760:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st830;
tr2761:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st830;
tr2771:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st830;
st830:
	if ( ++p == pe )
		goto _test_eof830;
case 830:
	_widec = (*p);
	if ( (*p) < 43 ) {
		if ( (*p) < 32 ) {
			if ( (*p) > 9 ) {
				if ( 10 <= (*p) && (*p) <= 10 ) {
					_widec = (short)(8832 + ((*p) - -128));
					if (
 s->multiline  ) _widec += 256;
					if (
 !s->multiline  ) _widec += 512;
					if (
 s->number64 != 0  ) _widec += 1024;
					if (
 s->number64 == 0  ) _widec += 2048;
				}
			} else if ( (*p) >= 9 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
				if (
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 32 ) {
			if ( (*p) > 40 ) {
				if ( 41 <= (*p) && (*p) <= 41 ) {
					_widec = (short)(5760 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
					if (
 s->number64 == 0  ) _widec += 512;
				}
			} else if ( (*p) >= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
				if (
 s->number64 == 0  ) _widec += 512;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 43 ) {
		if ( (*p) < 59 ) {
			if ( (*p) > 47 ) {
				if ( 48 <= (*p) && (*p) <= 57 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 47 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 59 ) {
			if ( (*p) > 90 ) {
				if ( 97 <= (*p) && (*p) <= 122 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 65 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(8832 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
			if (
 s->number64 != 0  ) _widec += 1024;
			if (
 s->number64 == 0  ) _widec += 2048;
		}
	} else {
		_widec = (short)(2688 + ((*p) - -128));
		if (
 s->number64 != 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 3115: goto tr2676;
		case 6153: goto st825;
		case 6176: goto st825;
		case 6184: goto tr2654;
		case 6185: goto tr2655;
		case 6409: goto st828;
		case 6432: goto st828;
		case 6440: goto tr2657;
		case 6441: goto tr2658;
		case 6665: goto st830;
		case 6688: goto st830;
		case 6696: goto tr2660;
		case 6697: goto tr2661;
		case 9482: goto tr2662;
		case 9531: goto tr2662;
		case 9738: goto tr2662;
		case 9787: goto tr2662;
		case 10250: goto tr2663;
		case 10299: goto st827;
		case 10506: goto tr2662;
		case 10555: goto tr2662;
		case 10762: goto tr2665;
		case 10811: goto tr2666;
		case 11274: goto tr2667;
		case 11323: goto st829;
		case 11530: goto tr2662;
		case 11579: goto tr2662;
		case 11786: goto tr2669;
		case 11835: goto tr2670;
		case 12298: goto tr2671;
		case 12347: goto st831;
		case 12554: goto tr2662;
		case 12603: goto tr2662;
		case 12810: goto tr2673;
		case 12859: goto tr2674;
	}
	if ( _widec < 3137 ) {
		if ( 3119 <= _widec && _widec <= 3129 )
			goto tr2676;
	} else if ( _widec > 3162 ) {
		if ( 3169 <= _widec && _widec <= 3194 )
			goto tr2676;
	} else
		goto tr2676;
	goto tr2675;
tr2665:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1195;
tr2720:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1195;
tr2765:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1195;
tr2809:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1195;
st1195:
	if ( ++p == pe )
		goto _test_eof1195;
case 1195:
	_widec = (*p);
	if ( (*p) < 43 ) {
		if ( (*p) < 32 ) {
			if ( (*p) > 9 ) {
				if ( 10 <= (*p) && (*p) <= 10 ) {
					_widec = (short)(3200 + ((*p) - -128));
					if (
 s->multiline  ) _widec += 256;
					if (
 s->number64 != 0  ) _widec += 512;
				}
			} else if ( (*p) >= 9 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 32 ) {
			if ( (*p) > 40 ) {
				if ( 41 <= (*p) && (*p) <= 41 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 40 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(2688 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
		}
	} else if ( (*p) > 43 ) {
		if ( (*p) < 59 ) {
			if ( (*p) > 47 ) {
				if ( 48 <= (*p) && (*p) <= 57 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 47 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 59 ) {
			if ( (*p) > 90 ) {
				if ( 97 <= (*p) && (*p) <= 122 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 65 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(3200 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(2688 + ((*p) - -128));
		if (
 s->number64 != 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 3081: goto st825;
		case 3104: goto st825;
		case 3112: goto tr2654;
		case 3113: goto tr2655;
		case 3115: goto tr2676;
		case 4106: goto tr2663;
		case 4155: goto st827;
	}
	if ( _widec < 3137 ) {
		if ( 3119 <= _widec && _widec <= 3129 )
			goto tr2676;
	} else if ( _widec > 3162 ) {
		if ( 3169 <= _widec && _widec <= 3194 )
			goto tr2676;
	} else
		goto tr2676;
	goto tr2675;
tr2666:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1196;
tr2721:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1196;
tr2810:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1196;
tr2766:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1196;
st1196:
	if ( ++p == pe )
		goto _test_eof1196;
case 1196:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(3200 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(3200 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(3200 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 s->number64 != 0  ) _widec += 512;
	}
	if ( _widec == 4106 )
		goto tr2678;
	if ( 3968 <= _widec && _widec <= 4223 )
		goto tr2677;
	goto tr2675;
tr2681:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st831;
tr2816:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st831;
tr2727:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st831;
tr2772:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st831;
st831:
	if ( ++p == pe )
		goto _test_eof831;
case 831:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(6784 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
			if (
 s->number64 == 0  ) _widec += 1024;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(6784 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
			if (
 s->number64 == 0  ) _widec += 1024;
		}
	} else {
		_widec = (short)(6784 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 s->number64 != 0  ) _widec += 512;
		if (
 s->number64 == 0  ) _widec += 1024;
	}
	switch( _widec ) {
		case 7690: goto tr2678;
		case 8202: goto tr2680;
		case 8714: goto tr2682;
	}
	if ( _widec < 8064 ) {
		if ( 7552 <= _widec && _widec <= 7807 )
			goto tr2677;
	} else if ( _widec > 8319 ) {
		if ( 8576 <= _widec && _widec <= 8831 )
			goto tr2681;
	} else
		goto tr2679;
	goto tr2675;
tr2673:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1197;
tr2728:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1197;
tr2773:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1197;
tr2817:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1197;
st1197:
	if ( ++p == pe )
		goto _test_eof1197;
case 1197:
	_widec = (*p);
	if ( (*p) < 43 ) {
		if ( (*p) < 32 ) {
			if ( (*p) > 9 ) {
				if ( 10 <= (*p) && (*p) <= 10 ) {
					_widec = (short)(8832 + ((*p) - -128));
					if (
 s->multiline  ) _widec += 256;
					if (
 !s->multiline  ) _widec += 512;
					if (
 s->number64 != 0  ) _widec += 1024;
					if (
 s->number64 == 0  ) _widec += 2048;
				}
			} else if ( (*p) >= 9 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
				if (
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 32 ) {
			if ( (*p) > 40 ) {
				if ( 41 <= (*p) && (*p) <= 41 ) {
					_widec = (short)(5760 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
					if (
 s->number64 == 0  ) _widec += 512;
				}
			} else if ( (*p) >= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
				if (
 s->number64 == 0  ) _widec += 512;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 43 ) {
		if ( (*p) < 59 ) {
			if ( (*p) > 47 ) {
				if ( 48 <= (*p) && (*p) <= 57 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 47 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 59 ) {
			if ( (*p) > 90 ) {
				if ( 97 <= (*p) && (*p) <= 122 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if (
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 65 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(8832 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
			if (
 s->number64 != 0  ) _widec += 1024;
			if (
 s->number64 == 0  ) _widec += 2048;
		}
	} else {
		_widec = (short)(2688 + ((*p) - -128));
		if (
 s->number64 != 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 3115: goto tr2676;
		case 6153: goto st825;
		case 6176: goto st825;
		case 6184: goto tr2654;
		case 6185: goto tr2655;
		case 6409: goto st828;
		case 6432: goto st828;
		case 6440: goto tr2657;
		case 6441: goto tr2658;
		case 6665: goto st830;
		case 6688: goto st830;
		case 6696: goto tr2660;
		case 6697: goto tr2661;
		case 9482: goto tr2662;
		case 9531: goto tr2662;
		case 9738: goto tr2662;
		case 9787: goto tr2662;
		case 10250: goto tr2663;
		case 10299: goto st827;
		case 10506: goto tr2662;
		case 10555: goto tr2662;
		case 10762: goto tr2665;
		case 10811: goto tr2666;
		case 11274: goto tr2667;
		case 11323: goto st829;
		case 11530: goto tr2662;
		case 11579: goto tr2662;
		case 11786: goto tr2669;
		case 11835: goto tr2670;
		case 12298: goto tr2671;
		case 12347: goto st831;
		case 12554: goto tr2662;
		case 12603: goto tr2662;
		case 12810: goto tr2673;
		case 12859: goto tr2674;
	}
	if ( _widec < 3137 ) {
		if ( 3119 <= _widec && _widec <= 3129 )
			goto tr2676;
	} else if ( _widec > 3162 ) {
		if ( 3169 <= _widec && _widec <= 3194 )
			goto tr2676;
	} else
		goto tr2676;
	goto tr2675;
tr2674:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1198;
tr2729:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1198;
tr2818:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1198;
tr2774:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1198;
st1198:
	if ( ++p == pe )
		goto _test_eof1198;
case 1198:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(6784 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
			if (
 s->number64 == 0  ) _widec += 1024;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(6784 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 s->number64 != 0  ) _widec += 512;
			if (
 s->number64 == 0  ) _widec += 1024;
		}
	} else {
		_widec = (short)(6784 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 s->number64 != 0  ) _widec += 512;
		if (
 s->number64 == 0  ) _widec += 1024;
	}
	switch( _widec ) {
		case 7690: goto tr2678;
		case 8202: goto tr2680;
		case 8714: goto tr2682;
	}
	if ( _widec < 8064 ) {
		if ( 7552 <= _widec && _widec <= 7807 )
			goto tr2677;
	} else if ( _widec > 8319 ) {
		if ( 8576 <= _widec && _widec <= 8831 )
			goto tr2681;
	} else
		goto tr2679;
	goto tr2675;
tr2683:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st832;
tr2645:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st832;
st832:
	if ( ++p == pe )
		goto _test_eof832;
case 832:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2684;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2683;
	goto tr2623;
tr2685:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st833;
st833:
	if ( ++p == pe )
		goto _test_eof833;
case 833:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2686;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2685;
	goto tr2623;
tr2628:
	{
		*(rdata_tail++) = 1;
	}
	goto st834;
st834:
	if ( ++p == pe )
		goto _test_eof834;
case 834:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st835;
		case 32: goto st835;
		case 40: goto tr2688;
		case 41: goto tr2689;
		case 1034: goto tr2690;
		case 1083: goto st840;
	}
	goto tr2623;
tr2688:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st835;
tr2689:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st835;
tr2690:
	{
		s->line_counter++;
	}
	goto st835;
tr2733:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st835;
st835:
	if ( ++p == pe )
		goto _test_eof835;
case 835:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st835;
		case 32: goto st835;
		case 40: goto tr2688;
		case 41: goto tr2689;
		case 1034: goto tr2690;
		case 1083: goto st840;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2692;
	goto tr2638;
tr2692:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st836;
tr2696:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st836;
st836:
	if ( ++p == pe )
		goto _test_eof836;
case 836:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2693;
		case 32: goto tr2693;
		case 40: goto tr2694;
		case 41: goto tr2695;
		case 1034: goto tr2697;
		case 1083: goto tr2698;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2696;
	goto tr2638;
tr2701:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st837;
tr2702:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st837;
tr2704:
	{
		s->line_counter++;
	}
	goto st837;
tr2731:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st837;
tr2693:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st837;
tr2694:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st837;
tr2695:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st837;
tr2697:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st837;
st837:
	if ( ++p == pe )
		goto _test_eof837;
case 837:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st837;
		case 32: goto st837;
		case 40: goto tr2701;
		case 41: goto tr2702;
		case 46: goto tr2703;
		case 1034: goto tr2704;
		case 1083: goto st839;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2703;
	goto tr2699;
tr2703:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st838;
tr2707:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st838;
st838:
	if ( ++p == pe )
		goto _test_eof838;
case 838:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 != 0  ) _widec += 1024;
				if (
 s->number64 == 0  ) _widec += 2048;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
				if (
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 != 0  ) _widec += 1024;
				if (
 s->number64 == 0  ) _widec += 2048;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(5760 + ((*p) - -128));
		if (
 s->number64 != 0  ) _widec += 256;
		if (
 s->number64 == 0  ) _widec += 512;
	}
	switch( _widec ) {
		case 46: goto tr2707;
		case 6153: goto tr2708;
		case 6176: goto tr2708;
		case 6184: goto tr2709;
		case 6185: goto tr2710;
		case 6409: goto tr2711;
		case 6432: goto tr2711;
		case 6440: goto tr2712;
		case 6441: goto tr2713;
		case 6665: goto tr2714;
		case 6688: goto tr2714;
		case 6696: goto tr2715;
		case 6697: goto tr2716;
		case 9482: goto tr2717;
		case 9531: goto tr2717;
		case 9738: goto tr2717;
		case 9787: goto tr2717;
		case 10250: goto tr2718;
		case 10299: goto tr2719;
		case 10506: goto tr2717;
		case 10555: goto tr2717;
		case 10762: goto tr2720;
		case 10811: goto tr2721;
		case 11274: goto tr2722;
		case 11323: goto tr2723;
		case 11530: goto tr2717;
		case 11579: goto tr2717;
		case 11786: goto tr2724;
		case 11835: goto tr2725;
		case 12298: goto tr2726;
		case 12347: goto tr2727;
		case 12554: goto tr2717;
		case 12603: goto tr2717;
		case 12810: goto tr2728;
		case 12859: goto tr2729;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2707;
	goto tr2706;
tr2730:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st839;
tr2698:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st839;
st839:
	if ( ++p == pe )
		goto _test_eof839;
case 839:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2731;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2730;
	goto tr2623;
tr2732:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st840;
st840:
	if ( ++p == pe )
		goto _test_eof840;
case 840:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2733;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2732;
	goto tr2623;
tr2629:
	{
		*(rdata_tail++) = 2;
	}
	goto st841;
st841:
	if ( ++p == pe )
		goto _test_eof841;
case 841:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st842;
		case 32: goto st842;
		case 40: goto tr2735;
		case 41: goto tr2736;
		case 1034: goto tr2737;
		case 1083: goto st847;
	}
	goto tr2623;
tr2735:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st842;
tr2736:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st842;
tr2737:
	{
		s->line_counter++;
	}
	goto st842;
tr2778:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st842;
st842:
	if ( ++p == pe )
		goto _test_eof842;
case 842:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st842;
		case 32: goto st842;
		case 40: goto tr2735;
		case 41: goto tr2736;
		case 1034: goto tr2737;
		case 1083: goto st847;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2739;
	goto tr2638;
tr2739:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st843;
tr2743:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st843;
st843:
	if ( ++p == pe )
		goto _test_eof843;
case 843:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2740;
		case 32: goto tr2740;
		case 40: goto tr2741;
		case 41: goto tr2742;
		case 1034: goto tr2744;
		case 1083: goto tr2745;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2743;
	goto tr2638;
tr2747:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st844;
tr2748:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st844;
tr2750:
	{
		s->line_counter++;
	}
	goto st844;
tr2776:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st844;
tr2740:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st844;
tr2741:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st844;
tr2742:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st844;
tr2744:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st844;
st844:
	if ( ++p == pe )
		goto _test_eof844;
case 844:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st844;
		case 32: goto st844;
		case 40: goto tr2747;
		case 41: goto tr2748;
		case 46: goto tr2749;
		case 1034: goto tr2750;
		case 1083: goto st846;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 58 )
			goto tr2749;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2749;
	} else
		goto tr2749;
	goto tr2699;
tr2749:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st845;
tr2752:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st845;
st845:
	if ( ++p == pe )
		goto _test_eof845;
case 845:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 != 0  ) _widec += 1024;
				if (
 s->number64 == 0  ) _widec += 2048;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
				if (
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 != 0  ) _widec += 1024;
				if (
 s->number64 == 0  ) _widec += 2048;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(5760 + ((*p) - -128));
		if (
 s->number64 != 0  ) _widec += 256;
		if (
 s->number64 == 0  ) _widec += 512;
	}
	switch( _widec ) {
		case 46: goto tr2752;
		case 6153: goto tr2753;
		case 6176: goto tr2753;
		case 6184: goto tr2754;
		case 6185: goto tr2755;
		case 6409: goto tr2756;
		case 6432: goto tr2756;
		case 6440: goto tr2757;
		case 6441: goto tr2758;
		case 6665: goto tr2759;
		case 6688: goto tr2759;
		case 6696: goto tr2760;
		case 6697: goto tr2761;
		case 9482: goto tr2762;
		case 9531: goto tr2762;
		case 9738: goto tr2762;
		case 9787: goto tr2762;
		case 10250: goto tr2763;
		case 10299: goto tr2764;
		case 10506: goto tr2762;
		case 10555: goto tr2762;
		case 10762: goto tr2765;
		case 10811: goto tr2766;
		case 11274: goto tr2767;
		case 11323: goto tr2768;
		case 11530: goto tr2762;
		case 11579: goto tr2762;
		case 11786: goto tr2769;
		case 11835: goto tr2770;
		case 12298: goto tr2771;
		case 12347: goto tr2772;
		case 12554: goto tr2762;
		case 12603: goto tr2762;
		case 12810: goto tr2773;
		case 12859: goto tr2774;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 58 )
			goto tr2752;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2752;
	} else
		goto tr2752;
	goto tr2706;
tr2775:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st846;
tr2745:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st846;
st846:
	if ( ++p == pe )
		goto _test_eof846;
case 846:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2776;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2775;
	goto tr2623;
tr2777:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st847;
st847:
	if ( ++p == pe )
		goto _test_eof847;
case 847:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2778;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2777;
	goto tr2623;
tr2630:
	{
		*(rdata_tail++) = 3;
	}
	goto st848;
st848:
	if ( ++p == pe )
		goto _test_eof848;
case 848:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st849;
		case 32: goto st849;
		case 40: goto tr2780;
		case 41: goto tr2781;
		case 1034: goto tr2782;
		case 1083: goto st854;
	}
	goto tr2623;
tr2780:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st849;
tr2781:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st849;
tr2782:
	{
		s->line_counter++;
	}
	goto st849;
tr2822:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st849;
st849:
	if ( ++p == pe )
		goto _test_eof849;
case 849:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st849;
		case 32: goto st849;
		case 40: goto tr2780;
		case 41: goto tr2781;
		case 1034: goto tr2782;
		case 1083: goto st854;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2784;
	goto tr2638;
tr2784:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st850;
tr2788:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st850;
st850:
	if ( ++p == pe )
		goto _test_eof850;
case 850:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2785;
		case 32: goto tr2785;
		case 40: goto tr2786;
		case 41: goto tr2787;
		case 1034: goto tr2789;
		case 1083: goto tr2790;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2788;
	goto tr2638;
tr2792:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st851;
tr2793:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st851;
tr2795:
	{
		s->line_counter++;
	}
	goto st851;
tr2820:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st851;
tr2785:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st851;
tr2786:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st851;
tr2787:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st851;
tr2789:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st851;
st851:
	if ( ++p == pe )
		goto _test_eof851;
case 851:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st851;
		case 32: goto st851;
		case 40: goto tr2792;
		case 41: goto tr2793;
		case 42: goto tr2794;
		case 92: goto tr2794;
		case 95: goto tr2794;
		case 1034: goto tr2795;
		case 1083: goto st853;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2794;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2794;
	} else
		goto tr2794;
	goto tr2623;
tr2794:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 852; goto st270;} }
	goto st852;
st852:
	if ( ++p == pe )
		goto _test_eof852;
case 852:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 != 0  ) _widec += 1024;
				if (
 s->number64 == 0  ) _widec += 2048;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if (
 s->number64 != 0  ) _widec += 256;
				if (
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if (
 s->multiline  ) _widec += 256;
				if (
 !s->multiline  ) _widec += 512;
				if (
 s->number64 != 0  ) _widec += 1024;
				if (
 s->number64 == 0  ) _widec += 2048;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if (
 s->number64 != 0  ) _widec += 256;
			if (
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(5760 + ((*p) - -128));
		if (
 s->number64 != 0  ) _widec += 256;
		if (
 s->number64 == 0  ) _widec += 512;
	}
	switch( _widec ) {
		case 6153: goto tr2797;
		case 6176: goto tr2797;
		case 6184: goto tr2798;
		case 6185: goto tr2799;
		case 6409: goto tr2800;
		case 6432: goto tr2800;
		case 6440: goto tr2801;
		case 6441: goto tr2802;
		case 6665: goto tr2803;
		case 6688: goto tr2803;
		case 6696: goto tr2804;
		case 6697: goto tr2805;
		case 9482: goto tr2806;
		case 9531: goto tr2806;
		case 9738: goto tr2806;
		case 9787: goto tr2806;
		case 10250: goto tr2807;
		case 10299: goto tr2808;
		case 10506: goto tr2806;
		case 10555: goto tr2806;
		case 10762: goto tr2809;
		case 10811: goto tr2810;
		case 11274: goto tr2811;
		case 11323: goto tr2812;
		case 11530: goto tr2806;
		case 11579: goto tr2806;
		case 11786: goto tr2813;
		case 11835: goto tr2814;
		case 12298: goto tr2815;
		case 12347: goto tr2816;
		case 12554: goto tr2806;
		case 12603: goto tr2806;
		case 12810: goto tr2817;
		case 12859: goto tr2818;
	}
	goto tr2652;
tr2819:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st853;
tr2790:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st853;
st853:
	if ( ++p == pe )
		goto _test_eof853;
case 853:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2820;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2819;
	goto tr2623;
tr2821:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st854;
st854:
	if ( ++p == pe )
		goto _test_eof854;
case 854:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2822;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2821;
	goto tr2623;
tr2823:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st855;
tr2622:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st855;
st855:
	if ( ++p == pe )
		goto _test_eof855;
case 855:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2824;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2823;
	goto tr71;
st856:
	if ( ++p == pe )
		goto _test_eof856;
case 856:
	switch( (*p) ) {
		case 65: goto st857;
		case 67: goto st891;
		case 68: goto st908;
		case 69: goto st922;
		case 72: goto st929;
		case 73: goto st934;
		case 75: goto st942;
		case 76: goto st946;
		case 77: goto st954;
		case 78: goto st960;
		case 80: goto st976;
		case 82: goto st979;
		case 83: goto st986;
		case 84: goto st997;
		case 85: goto st1007;
		case 97: goto st857;
		case 99: goto st891;
		case 100: goto st908;
		case 101: goto st922;
		case 104: goto st929;
		case 105: goto st934;
		case 107: goto st942;
		case 108: goto st946;
		case 109: goto st954;
		case 110: goto st960;
		case 112: goto st976;
		case 114: goto st979;
		case 115: goto st986;
		case 116: goto st997;
		case 117: goto st1007;
	}
	goto tr2825;
st857:
	if ( ++p == pe )
		goto _test_eof857;
case 857:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2841;
		case 32: goto tr2841;
		case 40: goto tr2842;
		case 41: goto tr2843;
		case 65: goto st882;
		case 70: goto st885;
		case 80: goto st889;
		case 97: goto st882;
		case 102: goto st885;
		case 112: goto st889;
		case 1034: goto tr2847;
		case 1083: goto tr2848;
	}
	goto tr2825;
tr2850:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr2851:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr2853:
	{
		s->line_counter++;
	}
	goto st858;
tr2949:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st858;
tr3260:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st858;
tr3261:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3262:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3264:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st858;
tr2841:
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
	goto st858;
tr2842:
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr2843:
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr2847:
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr2952:
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
	goto st858;
tr2953:
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr2954:
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr2955:
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr2960:
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
	goto st858;
tr2961:
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr2962:
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr2963:
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr2966:
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
	goto st858;
tr2967:
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr2968:
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr2969:
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr2976:
	{ type_num(KNOT_RRTYPE_CAA, &rdata_tail); }
	goto st858;
tr2977:
	{ type_num(KNOT_RRTYPE_CAA, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr2978:
	{ type_num(KNOT_RRTYPE_CAA, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr2979:
	{ type_num(KNOT_RRTYPE_CAA, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr2987:
	{ type_num(KNOT_RRTYPE_CDNSKEY, &rdata_tail); }
	goto st858;
tr2988:
	{ type_num(KNOT_RRTYPE_CDNSKEY, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr2989:
	{ type_num(KNOT_RRTYPE_CDNSKEY, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr2990:
	{ type_num(KNOT_RRTYPE_CDNSKEY, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr2992:
	{ type_num(KNOT_RRTYPE_CDS, &rdata_tail); }
	goto st858;
tr2993:
	{ type_num(KNOT_RRTYPE_CDS, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr2994:
	{ type_num(KNOT_RRTYPE_CDS, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr2995:
	{ type_num(KNOT_RRTYPE_CDS, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr2999:
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
	goto st858;
tr3000:
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3001:
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3002:
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3007:
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
	goto st858;
tr3008:
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3009:
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3010:
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3018:
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
	goto st858;
tr3019:
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3020:
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3021:
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3027:
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
	goto st858;
tr3028:
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3029:
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3030:
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3035:
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
	goto st858;
tr3036:
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3037:
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3038:
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3040:
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
	goto st858;
tr3041:
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3042:
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3043:
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3050:
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
	goto st858;
tr3051:
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3052:
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3053:
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3056:
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
	goto st858;
tr3057:
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3058:
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3059:
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3065:
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
	goto st858;
tr3066:
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3067:
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3068:
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3077:
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
	goto st858;
tr3078:
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3079:
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3080:
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3085:
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
	goto st858;
tr3086:
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3087:
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3088:
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3090:
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
	goto st858;
tr3091:
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3092:
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3093:
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3100:
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
	goto st858;
tr3101:
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3102:
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3103:
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3106:
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
	goto st858;
tr3107:
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3108:
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3109:
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3112:
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
	goto st858;
tr3113:
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3114:
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3115:
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3117:
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
	goto st858;
tr3118:
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3119:
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3120:
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3127:
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
	goto st858;
tr3128:
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3129:
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3130:
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3132:
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
	goto st858;
tr3133:
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3134:
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3135:
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3143:
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
	goto st858;
tr3144:
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3145:
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3146:
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3149:
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
	goto st858;
tr3150:
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3151:
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3152:
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3154:
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
	goto st858;
tr3155:
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3156:
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3158:
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3161:
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
	goto st858;
tr3162:
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3163:
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3165:
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3167:
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
	goto st858;
tr3168:
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3169:
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3171:
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3177:
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
	goto st858;
tr3178:
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3179:
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3180:
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3184:
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
	goto st858;
tr3185:
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3186:
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3187:
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3192:
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
	goto st858;
tr3193:
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3194:
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3195:
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3200:
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
	goto st858;
tr3201:
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3202:
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3203:
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3205:
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
	goto st858;
tr3206:
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3207:
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3208:
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3215:
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
	goto st858;
tr3216:
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3217:
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3218:
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3221:
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
	goto st858;
tr3222:
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3223:
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3224:
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3227:
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
	goto st858;
tr3228:
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3229:
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3230:
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3235:
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
	goto st858;
tr3236:
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3237:
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3238:
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3245:
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
	goto st858;
tr3246:
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3247:
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3248:
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3251:
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
	goto st858;
tr3252:
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3253:
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3254:
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
tr3268:
	{ type_num(KNOT_RRTYPE_URI, &rdata_tail); }
	goto st858;
tr3269:
	{ type_num(KNOT_RRTYPE_URI, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st858;
tr3270:
	{ type_num(KNOT_RRTYPE_URI, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st858;
tr3271:
	{ type_num(KNOT_RRTYPE_URI, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st858;
st858:
	if ( ++p == pe )
		goto _test_eof858;
case 858:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st858;
		case 32: goto st858;
		case 40: goto tr2850;
		case 41: goto tr2851;
		case 1034: goto tr2853;
		case 1083: goto st881;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2852;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2852;
	} else
		goto tr2852;
	goto tr71;
tr2852:
	{ p--; {stack[top++] = 859; goto st487;} }
	goto st859;
st859:
	if ( ++p == pe )
		goto _test_eof859;
case 859:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st860;
		case 32: goto st860;
		case 40: goto tr2856;
		case 41: goto tr2857;
		case 1034: goto tr2858;
		case 1083: goto st880;
	}
	goto tr71;
tr2856:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st860;
tr2857:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st860;
tr2858:
	{
		s->line_counter++;
	}
	goto st860;
tr2947:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st860;
st860:
	if ( ++p == pe )
		goto _test_eof860;
case 860:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st860;
		case 32: goto st860;
		case 40: goto tr2856;
		case 41: goto tr2857;
		case 1034: goto tr2858;
		case 1083: goto st880;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2860;
	goto tr1881;
tr2860:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st861;
tr2864:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st861;
st861:
	if ( ++p == pe )
		goto _test_eof861;
case 861:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2861;
		case 32: goto tr2861;
		case 40: goto tr2862;
		case 41: goto tr2863;
		case 1034: goto tr2865;
		case 1083: goto tr2866;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2864;
	goto tr1881;
tr2868:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st862;
tr2869:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st862;
tr2871:
	{
		s->line_counter++;
	}
	goto st862;
tr2945:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st862;
tr2861:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st862;
tr2862:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st862;
tr2863:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st862;
tr2865:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st862;
st862:
	if ( ++p == pe )
		goto _test_eof862;
case 862:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st862;
		case 32: goto st862;
		case 40: goto tr2868;
		case 41: goto tr2869;
		case 1034: goto tr2871;
		case 1083: goto st879;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2870;
	goto tr1881;
tr2870:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st863;
tr2876:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st863;
st863:
	if ( ++p == pe )
		goto _test_eof863;
case 863:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2873;
		case 32: goto tr2873;
		case 40: goto tr2874;
		case 41: goto tr2875;
		case 1034: goto tr2877;
		case 1083: goto tr2878;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2876;
	goto tr1881;
tr2881:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st864;
tr2882:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st864;
tr2884:
	{
		s->line_counter++;
	}
	goto st864;
tr2943:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st864;
tr2873:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st864;
tr2874:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st864;
tr2875:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st864;
tr2877:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st864;
st864:
	if ( ++p == pe )
		goto _test_eof864;
case 864:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st864;
		case 32: goto st864;
		case 40: goto tr2881;
		case 41: goto tr2882;
		case 1034: goto tr2884;
		case 1083: goto st878;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2883;
	goto tr2879;
tr2883:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st865;
tr2889:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st865;
st865:
	if ( ++p == pe )
		goto _test_eof865;
case 865:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2886;
		case 32: goto tr2886;
		case 40: goto tr2887;
		case 41: goto tr2888;
		case 1034: goto tr2890;
		case 1083: goto tr2891;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2889;
	goto tr2879;
tr2893:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st866;
tr2894:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st866;
tr2896:
	{
		s->line_counter++;
	}
	goto st866;
tr2941:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st866;
tr2886:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st268;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st268;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st268;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st268;}
		}
	}
	goto st866;
tr2887:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st268;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st268;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st268;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st866;
tr2888:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st268;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st268;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st268;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st866;
tr2890:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st268;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st268;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st268;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st866;
st866:
	if ( ++p == pe )
		goto _test_eof866;
case 866:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st866;
		case 32: goto st866;
		case 40: goto tr2893;
		case 41: goto tr2894;
		case 1034: goto tr2896;
		case 1083: goto st877;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2895;
	goto tr2879;
tr2895:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st867;
tr2901:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st867;
st867:
	if ( ++p == pe )
		goto _test_eof867;
case 867:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2898;
		case 32: goto tr2898;
		case 40: goto tr2899;
		case 41: goto tr2900;
		case 1034: goto tr2902;
		case 1083: goto tr2903;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2901;
	goto tr2879;
tr2905:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st868;
tr2906:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st868;
tr2908:
	{
		s->line_counter++;
	}
	goto st868;
tr2939:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st868;
tr2898:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st268;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st268;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st268;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st268;}
		}
	}
	goto st868;
tr2899:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st268;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st268;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st268;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st868;
tr2900:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st268;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st268;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st268;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st868;
tr2902:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st268;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st268;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st268;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st868;
st868:
	if ( ++p == pe )
		goto _test_eof868;
case 868:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st868;
		case 32: goto st868;
		case 40: goto tr2905;
		case 41: goto tr2906;
		case 1034: goto tr2908;
		case 1083: goto st876;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2907;
	goto tr1881;
tr2907:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st869;
tr2913:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st869;
st869:
	if ( ++p == pe )
		goto _test_eof869;
case 869:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2910;
		case 32: goto tr2910;
		case 40: goto tr2911;
		case 41: goto tr2912;
		case 1034: goto tr2914;
		case 1083: goto tr2915;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2913;
	goto tr1881;
tr2917:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st870;
tr2918:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st870;
tr2920:
	{
		s->line_counter++;
	}
	goto st870;
tr2937:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st870;
tr2910:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st870;
tr2911:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st870;
tr2912:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st870;
tr2914:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st870;
st870:
	if ( ++p == pe )
		goto _test_eof870;
case 870:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st870;
		case 32: goto st870;
		case 40: goto tr2917;
		case 41: goto tr2918;
		case 42: goto tr2919;
		case 92: goto tr2919;
		case 95: goto tr2919;
		case 1034: goto tr2920;
		case 1083: goto st875;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2919;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2919;
	} else
		goto tr2919;
	goto tr71;
tr2919:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 871; goto st270;} }
	goto st871;
st871:
	if ( ++p == pe )
		goto _test_eof871;
case 871:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2922;
		case 32: goto tr2922;
		case 40: goto tr2923;
		case 41: goto tr2924;
		case 1034: goto tr2925;
		case 1083: goto tr2926;
	}
	goto tr71;
tr2928:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st872;
tr2929:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st872;
tr2931:
	{
		s->line_counter++;
	}
	goto st872;
tr2935:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st872;
tr2922:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st872;
tr2923:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st872;
tr2924:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st872;
tr2925:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st872;
st872:
	if ( ++p == pe )
		goto _test_eof872;
case 872:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st872;
		case 32: goto st872;
		case 40: goto tr2928;
		case 41: goto tr2929;
		case 43: goto tr2930;
		case 1034: goto tr2931;
		case 1083: goto st874;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr2930;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2930;
	} else
		goto tr2930;
	goto tr71;
tr2930:
	{ p--; {stack[top++] = 873; goto st329;} }
	goto st873;
st873:
	if ( ++p == pe )
		goto _test_eof873;
case 873:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1152 + ((*p) - -128));
			if (
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1152 + ((*p) - -128));
		if (
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr2933;
		case 1595: goto tr2933;
	}
	goto tr71;
tr2933:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1199;
st1199:
	if ( ++p == pe )
		goto _test_eof1199;
case 1199:
	goto st0;
tr2934:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st874;
tr2926:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st874;
st874:
	if ( ++p == pe )
		goto _test_eof874;
case 874:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2935;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2934;
	goto tr71;
tr2936:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st875;
tr2915:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st875;
st875:
	if ( ++p == pe )
		goto _test_eof875;
case 875:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2937;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2936;
	goto tr71;
tr2938:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st876;
tr2903:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st268;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st268;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st268;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st268;}
		}
	}
	goto st876;
st876:
	if ( ++p == pe )
		goto _test_eof876;
case 876:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2939;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2938;
	goto tr71;
tr2940:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st877;
tr2891:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st268;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st268;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st268;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st268;}
		}
	}
	goto st877;
st877:
	if ( ++p == pe )
		goto _test_eof877;
case 877:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2941;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2940;
	goto tr71;
tr2942:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st878;
tr2878:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st878;
st878:
	if ( ++p == pe )
		goto _test_eof878;
case 878:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2943;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2942;
	goto tr71;
tr2944:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st879;
tr2866:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st879;
st879:
	if ( ++p == pe )
		goto _test_eof879;
case 879:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2945;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2944;
	goto tr71;
tr2946:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st880;
st880:
	if ( ++p == pe )
		goto _test_eof880;
case 880:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2947;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2946;
	goto tr71;
tr2948:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st881;
tr3265:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st881;
tr2848:
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
	goto st881;
tr2956:
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
	goto st881;
tr2964:
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
	goto st881;
tr2970:
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
	goto st881;
tr2980:
	{ type_num(KNOT_RRTYPE_CAA, &rdata_tail); }
	goto st881;
tr2991:
	{ type_num(KNOT_RRTYPE_CDNSKEY, &rdata_tail); }
	goto st881;
tr2996:
	{ type_num(KNOT_RRTYPE_CDS, &rdata_tail); }
	goto st881;
tr3003:
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
	goto st881;
tr3011:
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
	goto st881;
tr3022:
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
	goto st881;
tr3031:
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
	goto st881;
tr3039:
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
	goto st881;
tr3044:
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
	goto st881;
tr3054:
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
	goto st881;
tr3060:
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
	goto st881;
tr3069:
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
	goto st881;
tr3081:
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
	goto st881;
tr3089:
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
	goto st881;
tr3094:
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
	goto st881;
tr3104:
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
	goto st881;
tr3110:
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
	goto st881;
tr3116:
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
	goto st881;
tr3121:
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
	goto st881;
tr3131:
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
	goto st881;
tr3136:
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
	goto st881;
tr3147:
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
	goto st881;
tr3153:
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
	goto st881;
tr3159:
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
	goto st881;
tr3166:
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
	goto st881;
tr3172:
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
	goto st881;
tr3181:
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
	goto st881;
tr3188:
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
	goto st881;
tr3196:
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
	goto st881;
tr3204:
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
	goto st881;
tr3209:
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
	goto st881;
tr3219:
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
	goto st881;
tr3225:
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
	goto st881;
tr3231:
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
	goto st881;
tr3239:
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
	goto st881;
tr3249:
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
	goto st881;
tr3255:
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
	goto st881;
tr3272:
	{ type_num(KNOT_RRTYPE_URI, &rdata_tail); }
	goto st881;
st881:
	if ( ++p == pe )
		goto _test_eof881;
case 881:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2949;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr2948;
	goto tr71;
st882:
	if ( ++p == pe )
		goto _test_eof882;
case 882:
	switch( (*p) ) {
		case 65: goto st883;
		case 97: goto st883;
	}
	goto tr2825;
st883:
	if ( ++p == pe )
		goto _test_eof883;
case 883:
	switch( (*p) ) {
		case 65: goto st884;
		case 97: goto st884;
	}
	goto tr2825;
st884:
	if ( ++p == pe )
		goto _test_eof884;
case 884:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2952;
		case 32: goto tr2952;
		case 40: goto tr2953;
		case 41: goto tr2954;
		case 1034: goto tr2955;
		case 1083: goto tr2956;
	}
	goto tr2825;
st885:
	if ( ++p == pe )
		goto _test_eof885;
case 885:
	switch( (*p) ) {
		case 83: goto st886;
		case 115: goto st886;
	}
	goto tr2825;
st886:
	if ( ++p == pe )
		goto _test_eof886;
case 886:
	switch( (*p) ) {
		case 68: goto st887;
		case 100: goto st887;
	}
	goto tr2825;
st887:
	if ( ++p == pe )
		goto _test_eof887;
case 887:
	switch( (*p) ) {
		case 66: goto st888;
		case 98: goto st888;
	}
	goto tr2825;
st888:
	if ( ++p == pe )
		goto _test_eof888;
case 888:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2960;
		case 32: goto tr2960;
		case 40: goto tr2961;
		case 41: goto tr2962;
		case 1034: goto tr2963;
		case 1083: goto tr2964;
	}
	goto tr2825;
st889:
	if ( ++p == pe )
		goto _test_eof889;
case 889:
	switch( (*p) ) {
		case 76: goto st890;
		case 108: goto st890;
	}
	goto tr2825;
st890:
	if ( ++p == pe )
		goto _test_eof890;
case 890:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2966;
		case 32: goto tr2966;
		case 40: goto tr2967;
		case 41: goto tr2968;
		case 1034: goto tr2969;
		case 1083: goto tr2970;
	}
	goto tr2825;
st891:
	if ( ++p == pe )
		goto _test_eof891;
case 891:
	switch( (*p) ) {
		case 65: goto st892;
		case 68: goto st894;
		case 69: goto st901;
		case 78: goto st904;
		case 97: goto st892;
		case 100: goto st894;
		case 101: goto st901;
		case 110: goto st904;
	}
	goto tr2825;
st892:
	if ( ++p == pe )
		goto _test_eof892;
case 892:
	switch( (*p) ) {
		case 65: goto st893;
		case 97: goto st893;
	}
	goto tr2825;
st893:
	if ( ++p == pe )
		goto _test_eof893;
case 893:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2976;
		case 32: goto tr2976;
		case 40: goto tr2977;
		case 41: goto tr2978;
		case 1034: goto tr2979;
		case 1083: goto tr2980;
	}
	goto tr2825;
st894:
	if ( ++p == pe )
		goto _test_eof894;
case 894:
	switch( (*p) ) {
		case 78: goto st895;
		case 83: goto st900;
		case 110: goto st895;
		case 115: goto st900;
	}
	goto tr2825;
st895:
	if ( ++p == pe )
		goto _test_eof895;
case 895:
	switch( (*p) ) {
		case 83: goto st896;
		case 115: goto st896;
	}
	goto tr2825;
st896:
	if ( ++p == pe )
		goto _test_eof896;
case 896:
	switch( (*p) ) {
		case 75: goto st897;
		case 107: goto st897;
	}
	goto tr2825;
st897:
	if ( ++p == pe )
		goto _test_eof897;
case 897:
	switch( (*p) ) {
		case 69: goto st898;
		case 101: goto st898;
	}
	goto tr2825;
st898:
	if ( ++p == pe )
		goto _test_eof898;
case 898:
	switch( (*p) ) {
		case 89: goto st899;
		case 121: goto st899;
	}
	goto tr2825;
st899:
	if ( ++p == pe )
		goto _test_eof899;
case 899:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2987;
		case 32: goto tr2987;
		case 40: goto tr2988;
		case 41: goto tr2989;
		case 1034: goto tr2990;
		case 1083: goto tr2991;
	}
	goto tr2825;
st900:
	if ( ++p == pe )
		goto _test_eof900;
case 900:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2992;
		case 32: goto tr2992;
		case 40: goto tr2993;
		case 41: goto tr2994;
		case 1034: goto tr2995;
		case 1083: goto tr2996;
	}
	goto tr2825;
st901:
	if ( ++p == pe )
		goto _test_eof901;
case 901:
	switch( (*p) ) {
		case 82: goto st902;
		case 114: goto st902;
	}
	goto tr2825;
st902:
	if ( ++p == pe )
		goto _test_eof902;
case 902:
	switch( (*p) ) {
		case 84: goto st903;
		case 116: goto st903;
	}
	goto tr2825;
st903:
	if ( ++p == pe )
		goto _test_eof903;
case 903:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2999;
		case 32: goto tr2999;
		case 40: goto tr3000;
		case 41: goto tr3001;
		case 1034: goto tr3002;
		case 1083: goto tr3003;
	}
	goto tr2825;
st904:
	if ( ++p == pe )
		goto _test_eof904;
case 904:
	switch( (*p) ) {
		case 65: goto st905;
		case 97: goto st905;
	}
	goto tr2825;
st905:
	if ( ++p == pe )
		goto _test_eof905;
case 905:
	switch( (*p) ) {
		case 77: goto st906;
		case 109: goto st906;
	}
	goto tr2825;
st906:
	if ( ++p == pe )
		goto _test_eof906;
case 906:
	switch( (*p) ) {
		case 69: goto st907;
		case 101: goto st907;
	}
	goto tr2825;
st907:
	if ( ++p == pe )
		goto _test_eof907;
case 907:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3007;
		case 32: goto tr3007;
		case 40: goto tr3008;
		case 41: goto tr3009;
		case 1034: goto tr3010;
		case 1083: goto tr3011;
	}
	goto tr2825;
st908:
	if ( ++p == pe )
		goto _test_eof908;
case 908:
	switch( (*p) ) {
		case 72: goto st909;
		case 78: goto st913;
		case 83: goto st921;
		case 104: goto st909;
		case 110: goto st913;
		case 115: goto st921;
	}
	goto tr2825;
st909:
	if ( ++p == pe )
		goto _test_eof909;
case 909:
	switch( (*p) ) {
		case 67: goto st910;
		case 99: goto st910;
	}
	goto tr2825;
st910:
	if ( ++p == pe )
		goto _test_eof910;
case 910:
	switch( (*p) ) {
		case 73: goto st911;
		case 105: goto st911;
	}
	goto tr2825;
st911:
	if ( ++p == pe )
		goto _test_eof911;
case 911:
	switch( (*p) ) {
		case 68: goto st912;
		case 100: goto st912;
	}
	goto tr2825;
st912:
	if ( ++p == pe )
		goto _test_eof912;
case 912:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3018;
		case 32: goto tr3018;
		case 40: goto tr3019;
		case 41: goto tr3020;
		case 1034: goto tr3021;
		case 1083: goto tr3022;
	}
	goto tr2825;
st913:
	if ( ++p == pe )
		goto _test_eof913;
case 913:
	switch( (*p) ) {
		case 65: goto st914;
		case 83: goto st917;
		case 97: goto st914;
		case 115: goto st917;
	}
	goto tr2825;
st914:
	if ( ++p == pe )
		goto _test_eof914;
case 914:
	switch( (*p) ) {
		case 77: goto st915;
		case 109: goto st915;
	}
	goto tr2825;
st915:
	if ( ++p == pe )
		goto _test_eof915;
case 915:
	switch( (*p) ) {
		case 69: goto st916;
		case 101: goto st916;
	}
	goto tr2825;
st916:
	if ( ++p == pe )
		goto _test_eof916;
case 916:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3027;
		case 32: goto tr3027;
		case 40: goto tr3028;
		case 41: goto tr3029;
		case 1034: goto tr3030;
		case 1083: goto tr3031;
	}
	goto tr2825;
st917:
	if ( ++p == pe )
		goto _test_eof917;
case 917:
	switch( (*p) ) {
		case 75: goto st918;
		case 107: goto st918;
	}
	goto tr2825;
st918:
	if ( ++p == pe )
		goto _test_eof918;
case 918:
	switch( (*p) ) {
		case 69: goto st919;
		case 101: goto st919;
	}
	goto tr2825;
st919:
	if ( ++p == pe )
		goto _test_eof919;
case 919:
	switch( (*p) ) {
		case 89: goto st920;
		case 121: goto st920;
	}
	goto tr2825;
st920:
	if ( ++p == pe )
		goto _test_eof920;
case 920:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3035;
		case 32: goto tr3035;
		case 40: goto tr3036;
		case 41: goto tr3037;
		case 1034: goto tr3038;
		case 1083: goto tr3039;
	}
	goto tr2825;
st921:
	if ( ++p == pe )
		goto _test_eof921;
case 921:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3040;
		case 32: goto tr3040;
		case 40: goto tr3041;
		case 41: goto tr3042;
		case 1034: goto tr3043;
		case 1083: goto tr3044;
	}
	goto tr2825;
st922:
	if ( ++p == pe )
		goto _test_eof922;
case 922:
	switch( (*p) ) {
		case 85: goto st923;
		case 117: goto st923;
	}
	goto tr2825;
st923:
	if ( ++p == pe )
		goto _test_eof923;
case 923:
	switch( (*p) ) {
		case 73: goto st924;
		case 105: goto st924;
	}
	goto tr2825;
st924:
	if ( ++p == pe )
		goto _test_eof924;
case 924:
	switch( (*p) ) {
		case 52: goto st925;
		case 54: goto st927;
	}
	goto tr2825;
st925:
	if ( ++p == pe )
		goto _test_eof925;
case 925:
	if ( (*p) == 56 )
		goto st926;
	goto tr2825;
st926:
	if ( ++p == pe )
		goto _test_eof926;
case 926:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3050;
		case 32: goto tr3050;
		case 40: goto tr3051;
		case 41: goto tr3052;
		case 1034: goto tr3053;
		case 1083: goto tr3054;
	}
	goto tr2825;
st927:
	if ( ++p == pe )
		goto _test_eof927;
case 927:
	if ( (*p) == 52 )
		goto st928;
	goto tr2825;
st928:
	if ( ++p == pe )
		goto _test_eof928;
case 928:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3056;
		case 32: goto tr3056;
		case 40: goto tr3057;
		case 41: goto tr3058;
		case 1034: goto tr3059;
		case 1083: goto tr3060;
	}
	goto tr2825;
st929:
	if ( ++p == pe )
		goto _test_eof929;
case 929:
	switch( (*p) ) {
		case 73: goto st930;
		case 105: goto st930;
	}
	goto tr2825;
st930:
	if ( ++p == pe )
		goto _test_eof930;
case 930:
	switch( (*p) ) {
		case 78: goto st931;
		case 110: goto st931;
	}
	goto tr2825;
st931:
	if ( ++p == pe )
		goto _test_eof931;
case 931:
	switch( (*p) ) {
		case 70: goto st932;
		case 102: goto st932;
	}
	goto tr2825;
st932:
	if ( ++p == pe )
		goto _test_eof932;
case 932:
	switch( (*p) ) {
		case 79: goto st933;
		case 111: goto st933;
	}
	goto tr2825;
st933:
	if ( ++p == pe )
		goto _test_eof933;
case 933:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3065;
		case 32: goto tr3065;
		case 40: goto tr3066;
		case 41: goto tr3067;
		case 1034: goto tr3068;
		case 1083: goto tr3069;
	}
	goto tr2825;
st934:
	if ( ++p == pe )
		goto _test_eof934;
case 934:
	switch( (*p) ) {
		case 80: goto st935;
		case 112: goto st935;
	}
	goto tr2825;
st935:
	if ( ++p == pe )
		goto _test_eof935;
case 935:
	switch( (*p) ) {
		case 83: goto st936;
		case 115: goto st936;
	}
	goto tr2825;
st936:
	if ( ++p == pe )
		goto _test_eof936;
case 936:
	switch( (*p) ) {
		case 69: goto st937;
		case 101: goto st937;
	}
	goto tr2825;
st937:
	if ( ++p == pe )
		goto _test_eof937;
case 937:
	switch( (*p) ) {
		case 67: goto st938;
		case 99: goto st938;
	}
	goto tr2825;
st938:
	if ( ++p == pe )
		goto _test_eof938;
case 938:
	switch( (*p) ) {
		case 75: goto st939;
		case 107: goto st939;
	}
	goto tr2825;
st939:
	if ( ++p == pe )
		goto _test_eof939;
case 939:
	switch( (*p) ) {
		case 69: goto st940;
		case 101: goto st940;
	}
	goto tr2825;
st940:
	if ( ++p == pe )
		goto _test_eof940;
case 940:
	switch( (*p) ) {
		case 89: goto st941;
		case 121: goto st941;
	}
	goto tr2825;
st941:
	if ( ++p == pe )
		goto _test_eof941;
case 941:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3077;
		case 32: goto tr3077;
		case 40: goto tr3078;
		case 41: goto tr3079;
		case 1034: goto tr3080;
		case 1083: goto tr3081;
	}
	goto tr2825;
st942:
	if ( ++p == pe )
		goto _test_eof942;
case 942:
	switch( (*p) ) {
		case 69: goto st943;
		case 88: goto st945;
		case 101: goto st943;
		case 120: goto st945;
	}
	goto tr2825;
st943:
	if ( ++p == pe )
		goto _test_eof943;
case 943:
	switch( (*p) ) {
		case 89: goto st944;
		case 121: goto st944;
	}
	goto tr2825;
st944:
	if ( ++p == pe )
		goto _test_eof944;
case 944:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3085;
		case 32: goto tr3085;
		case 40: goto tr3086;
		case 41: goto tr3087;
		case 1034: goto tr3088;
		case 1083: goto tr3089;
	}
	goto tr2825;
st945:
	if ( ++p == pe )
		goto _test_eof945;
case 945:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3090;
		case 32: goto tr3090;
		case 40: goto tr3091;
		case 41: goto tr3092;
		case 1034: goto tr3093;
		case 1083: goto tr3094;
	}
	goto tr2825;
st946:
	if ( ++p == pe )
		goto _test_eof946;
case 946:
	switch( (*p) ) {
		case 51: goto st947;
		case 54: goto st949;
		case 79: goto st951;
		case 80: goto st953;
		case 111: goto st951;
		case 112: goto st953;
	}
	goto tr2825;
st947:
	if ( ++p == pe )
		goto _test_eof947;
case 947:
	if ( (*p) == 50 )
		goto st948;
	goto tr2825;
st948:
	if ( ++p == pe )
		goto _test_eof948;
case 948:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3100;
		case 32: goto tr3100;
		case 40: goto tr3101;
		case 41: goto tr3102;
		case 1034: goto tr3103;
		case 1083: goto tr3104;
	}
	goto tr2825;
st949:
	if ( ++p == pe )
		goto _test_eof949;
case 949:
	if ( (*p) == 52 )
		goto st950;
	goto tr2825;
st950:
	if ( ++p == pe )
		goto _test_eof950;
case 950:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3106;
		case 32: goto tr3106;
		case 40: goto tr3107;
		case 41: goto tr3108;
		case 1034: goto tr3109;
		case 1083: goto tr3110;
	}
	goto tr2825;
st951:
	if ( ++p == pe )
		goto _test_eof951;
case 951:
	switch( (*p) ) {
		case 67: goto st952;
		case 99: goto st952;
	}
	goto tr2825;
st952:
	if ( ++p == pe )
		goto _test_eof952;
case 952:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3112;
		case 32: goto tr3112;
		case 40: goto tr3113;
		case 41: goto tr3114;
		case 1034: goto tr3115;
		case 1083: goto tr3116;
	}
	goto tr2825;
st953:
	if ( ++p == pe )
		goto _test_eof953;
case 953:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3117;
		case 32: goto tr3117;
		case 40: goto tr3118;
		case 41: goto tr3119;
		case 1034: goto tr3120;
		case 1083: goto tr3121;
	}
	goto tr2825;
st954:
	if ( ++p == pe )
		goto _test_eof954;
case 954:
	switch( (*p) ) {
		case 73: goto st955;
		case 88: goto st959;
		case 105: goto st955;
		case 120: goto st959;
	}
	goto tr2825;
st955:
	if ( ++p == pe )
		goto _test_eof955;
case 955:
	switch( (*p) ) {
		case 78: goto st956;
		case 110: goto st956;
	}
	goto tr2825;
st956:
	if ( ++p == pe )
		goto _test_eof956;
case 956:
	switch( (*p) ) {
		case 70: goto st957;
		case 102: goto st957;
	}
	goto tr2825;
st957:
	if ( ++p == pe )
		goto _test_eof957;
case 957:
	switch( (*p) ) {
		case 79: goto st958;
		case 111: goto st958;
	}
	goto tr2825;
st958:
	if ( ++p == pe )
		goto _test_eof958;
case 958:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3127;
		case 32: goto tr3127;
		case 40: goto tr3128;
		case 41: goto tr3129;
		case 1034: goto tr3130;
		case 1083: goto tr3131;
	}
	goto tr2825;
st959:
	if ( ++p == pe )
		goto _test_eof959;
case 959:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3132;
		case 32: goto tr3132;
		case 40: goto tr3133;
		case 41: goto tr3134;
		case 1034: goto tr3135;
		case 1083: goto tr3136;
	}
	goto tr2825;
st960:
	if ( ++p == pe )
		goto _test_eof960;
case 960:
	switch( (*p) ) {
		case 65: goto st961;
		case 73: goto st965;
		case 83: goto st967;
		case 97: goto st961;
		case 105: goto st965;
		case 115: goto st967;
	}
	goto tr2825;
st961:
	if ( ++p == pe )
		goto _test_eof961;
case 961:
	switch( (*p) ) {
		case 80: goto st962;
		case 112: goto st962;
	}
	goto tr2825;
st962:
	if ( ++p == pe )
		goto _test_eof962;
case 962:
	switch( (*p) ) {
		case 84: goto st963;
		case 116: goto st963;
	}
	goto tr2825;
st963:
	if ( ++p == pe )
		goto _test_eof963;
case 963:
	switch( (*p) ) {
		case 82: goto st964;
		case 114: goto st964;
	}
	goto tr2825;
st964:
	if ( ++p == pe )
		goto _test_eof964;
case 964:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3143;
		case 32: goto tr3143;
		case 40: goto tr3144;
		case 41: goto tr3145;
		case 1034: goto tr3146;
		case 1083: goto tr3147;
	}
	goto tr2825;
st965:
	if ( ++p == pe )
		goto _test_eof965;
case 965:
	switch( (*p) ) {
		case 68: goto st966;
		case 100: goto st966;
	}
	goto tr2825;
st966:
	if ( ++p == pe )
		goto _test_eof966;
case 966:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3149;
		case 32: goto tr3149;
		case 40: goto tr3150;
		case 41: goto tr3151;
		case 1034: goto tr3152;
		case 1083: goto tr3153;
	}
	goto tr2825;
st967:
	if ( ++p == pe )
		goto _test_eof967;
case 967:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3154;
		case 32: goto tr3154;
		case 40: goto tr3155;
		case 41: goto tr3156;
		case 69: goto st968;
		case 101: goto st968;
		case 1034: goto tr3158;
		case 1083: goto tr3159;
	}
	goto tr2825;
st968:
	if ( ++p == pe )
		goto _test_eof968;
case 968:
	switch( (*p) ) {
		case 67: goto st969;
		case 99: goto st969;
	}
	goto tr2825;
st969:
	if ( ++p == pe )
		goto _test_eof969;
case 969:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3161;
		case 32: goto tr3161;
		case 40: goto tr3162;
		case 41: goto tr3163;
		case 51: goto st970;
		case 1034: goto tr3165;
		case 1083: goto tr3166;
	}
	goto tr2825;
st970:
	if ( ++p == pe )
		goto _test_eof970;
case 970:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3167;
		case 32: goto tr3167;
		case 40: goto tr3168;
		case 41: goto tr3169;
		case 80: goto st971;
		case 112: goto st971;
		case 1034: goto tr3171;
		case 1083: goto tr3172;
	}
	goto tr2825;
st971:
	if ( ++p == pe )
		goto _test_eof971;
case 971:
	switch( (*p) ) {
		case 65: goto st972;
		case 97: goto st972;
	}
	goto tr2825;
st972:
	if ( ++p == pe )
		goto _test_eof972;
case 972:
	switch( (*p) ) {
		case 82: goto st973;
		case 114: goto st973;
	}
	goto tr2825;
st973:
	if ( ++p == pe )
		goto _test_eof973;
case 973:
	switch( (*p) ) {
		case 65: goto st974;
		case 97: goto st974;
	}
	goto tr2825;
st974:
	if ( ++p == pe )
		goto _test_eof974;
case 974:
	switch( (*p) ) {
		case 77: goto st975;
		case 109: goto st975;
	}
	goto tr2825;
st975:
	if ( ++p == pe )
		goto _test_eof975;
case 975:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3177;
		case 32: goto tr3177;
		case 40: goto tr3178;
		case 41: goto tr3179;
		case 1034: goto tr3180;
		case 1083: goto tr3181;
	}
	goto tr2825;
st976:
	if ( ++p == pe )
		goto _test_eof976;
case 976:
	switch( (*p) ) {
		case 84: goto st977;
		case 116: goto st977;
	}
	goto tr2825;
st977:
	if ( ++p == pe )
		goto _test_eof977;
case 977:
	switch( (*p) ) {
		case 82: goto st978;
		case 114: goto st978;
	}
	goto tr2825;
st978:
	if ( ++p == pe )
		goto _test_eof978;
case 978:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3184;
		case 32: goto tr3184;
		case 40: goto tr3185;
		case 41: goto tr3186;
		case 1034: goto tr3187;
		case 1083: goto tr3188;
	}
	goto tr2825;
st979:
	if ( ++p == pe )
		goto _test_eof979;
case 979:
	switch( (*p) ) {
		case 80: goto st980;
		case 82: goto st981;
		case 84: goto st985;
		case 112: goto st980;
		case 114: goto st981;
		case 116: goto st985;
	}
	goto tr2825;
st980:
	if ( ++p == pe )
		goto _test_eof980;
case 980:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3192;
		case 32: goto tr3192;
		case 40: goto tr3193;
		case 41: goto tr3194;
		case 1034: goto tr3195;
		case 1083: goto tr3196;
	}
	goto tr2825;
st981:
	if ( ++p == pe )
		goto _test_eof981;
case 981:
	switch( (*p) ) {
		case 83: goto st982;
		case 115: goto st982;
	}
	goto tr2825;
st982:
	if ( ++p == pe )
		goto _test_eof982;
case 982:
	switch( (*p) ) {
		case 73: goto st983;
		case 105: goto st983;
	}
	goto tr2825;
st983:
	if ( ++p == pe )
		goto _test_eof983;
case 983:
	switch( (*p) ) {
		case 71: goto st984;
		case 103: goto st984;
	}
	goto tr2825;
st984:
	if ( ++p == pe )
		goto _test_eof984;
case 984:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3200;
		case 32: goto tr3200;
		case 40: goto tr3201;
		case 41: goto tr3202;
		case 1034: goto tr3203;
		case 1083: goto tr3204;
	}
	goto tr2825;
st985:
	if ( ++p == pe )
		goto _test_eof985;
case 985:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3205;
		case 32: goto tr3205;
		case 40: goto tr3206;
		case 41: goto tr3207;
		case 1034: goto tr3208;
		case 1083: goto tr3209;
	}
	goto tr2825;
st986:
	if ( ++p == pe )
		goto _test_eof986;
case 986:
	switch( (*p) ) {
		case 79: goto st987;
		case 80: goto st989;
		case 82: goto st991;
		case 83: goto st993;
		case 111: goto st987;
		case 112: goto st989;
		case 114: goto st991;
		case 115: goto st993;
	}
	goto tr2825;
st987:
	if ( ++p == pe )
		goto _test_eof987;
case 987:
	switch( (*p) ) {
		case 65: goto st988;
		case 97: goto st988;
	}
	goto tr2825;
st988:
	if ( ++p == pe )
		goto _test_eof988;
case 988:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3215;
		case 32: goto tr3215;
		case 40: goto tr3216;
		case 41: goto tr3217;
		case 1034: goto tr3218;
		case 1083: goto tr3219;
	}
	goto tr2825;
st989:
	if ( ++p == pe )
		goto _test_eof989;
case 989:
	switch( (*p) ) {
		case 70: goto st990;
		case 102: goto st990;
	}
	goto tr2825;
st990:
	if ( ++p == pe )
		goto _test_eof990;
case 990:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3221;
		case 32: goto tr3221;
		case 40: goto tr3222;
		case 41: goto tr3223;
		case 1034: goto tr3224;
		case 1083: goto tr3225;
	}
	goto tr2825;
st991:
	if ( ++p == pe )
		goto _test_eof991;
case 991:
	switch( (*p) ) {
		case 86: goto st992;
		case 118: goto st992;
	}
	goto tr2825;
st992:
	if ( ++p == pe )
		goto _test_eof992;
case 992:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3227;
		case 32: goto tr3227;
		case 40: goto tr3228;
		case 41: goto tr3229;
		case 1034: goto tr3230;
		case 1083: goto tr3231;
	}
	goto tr2825;
st993:
	if ( ++p == pe )
		goto _test_eof993;
case 993:
	switch( (*p) ) {
		case 72: goto st994;
		case 104: goto st994;
	}
	goto tr2825;
st994:
	if ( ++p == pe )
		goto _test_eof994;
case 994:
	switch( (*p) ) {
		case 70: goto st995;
		case 102: goto st995;
	}
	goto tr2825;
st995:
	if ( ++p == pe )
		goto _test_eof995;
case 995:
	switch( (*p) ) {
		case 80: goto st996;
		case 112: goto st996;
	}
	goto tr2825;
st996:
	if ( ++p == pe )
		goto _test_eof996;
case 996:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3235;
		case 32: goto tr3235;
		case 40: goto tr3236;
		case 41: goto tr3237;
		case 1034: goto tr3238;
		case 1083: goto tr3239;
	}
	goto tr2825;
st997:
	if ( ++p == pe )
		goto _test_eof997;
case 997:
	switch( (*p) ) {
		case 76: goto st998;
		case 88: goto st1001;
		case 89: goto st1003;
		case 108: goto st998;
		case 120: goto st1001;
		case 121: goto st1003;
	}
	goto tr2825;
st998:
	if ( ++p == pe )
		goto _test_eof998;
case 998:
	switch( (*p) ) {
		case 83: goto st999;
		case 115: goto st999;
	}
	goto tr2825;
st999:
	if ( ++p == pe )
		goto _test_eof999;
case 999:
	switch( (*p) ) {
		case 65: goto st1000;
		case 97: goto st1000;
	}
	goto tr2825;
st1000:
	if ( ++p == pe )
		goto _test_eof1000;
case 1000:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3245;
		case 32: goto tr3245;
		case 40: goto tr3246;
		case 41: goto tr3247;
		case 1034: goto tr3248;
		case 1083: goto tr3249;
	}
	goto tr2825;
st1001:
	if ( ++p == pe )
		goto _test_eof1001;
case 1001:
	switch( (*p) ) {
		case 84: goto st1002;
		case 116: goto st1002;
	}
	goto tr2825;
st1002:
	if ( ++p == pe )
		goto _test_eof1002;
case 1002:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3251;
		case 32: goto tr3251;
		case 40: goto tr3252;
		case 41: goto tr3253;
		case 1034: goto tr3254;
		case 1083: goto tr3255;
	}
	goto tr2825;
st1003:
	if ( ++p == pe )
		goto _test_eof1003;
case 1003:
	switch( (*p) ) {
		case 80: goto st1004;
		case 112: goto st1004;
	}
	goto tr2825;
st1004:
	if ( ++p == pe )
		goto _test_eof1004;
case 1004:
	switch( (*p) ) {
		case 69: goto st1005;
		case 101: goto st1005;
	}
	goto tr2825;
st1005:
	if ( ++p == pe )
		goto _test_eof1005;
case 1005:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3259;
	goto tr3258;
tr3259:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1006;
tr3263:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1006;
st1006:
	if ( ++p == pe )
		goto _test_eof1006;
case 1006:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3260;
		case 32: goto tr3260;
		case 40: goto tr3261;
		case 41: goto tr3262;
		case 1034: goto tr3264;
		case 1083: goto tr3265;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3263;
	goto tr3258;
st1007:
	if ( ++p == pe )
		goto _test_eof1007;
case 1007:
	switch( (*p) ) {
		case 82: goto st1008;
		case 114: goto st1008;
	}
	goto tr2825;
st1008:
	if ( ++p == pe )
		goto _test_eof1008;
case 1008:
	switch( (*p) ) {
		case 73: goto st1009;
		case 105: goto st1009;
	}
	goto tr2825;
st1009:
	if ( ++p == pe )
		goto _test_eof1009;
case 1009:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3268;
		case 32: goto tr3268;
		case 40: goto tr3269;
		case 41: goto tr3270;
		case 1034: goto tr3271;
		case 1083: goto tr3272;
	}
	goto tr2825;
st1010:
	if ( ++p == pe )
		goto _test_eof1010;
case 1010:
	switch( (*p) ) {
		case 42: goto tr3273;
		case 92: goto tr3273;
		case 95: goto tr3273;
	}
	if ( (*p) < 64 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr3273;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr3273;
	} else
		goto tr3273;
	goto tr71;
tr3273:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 1011; goto st270;} }
	goto st1011;
st1011:
	if ( ++p == pe )
		goto _test_eof1011;
case 1011:
	switch( (*p) ) {
		case 32: goto tr3274;
		case 59: goto tr3274;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3274;
	} else if ( (*p) >= 9 )
		goto tr3274;
	goto tr71;
tr3274:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{ p--; {stack[top++] = 1012; goto st336;} }
	goto st1012;
st1012:
	if ( ++p == pe )
		goto _test_eof1012;
case 1012:
	switch( (*p) ) {
		case 32: goto tr3275;
		case 59: goto tr3275;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3275;
	} else if ( (*p) >= 9 )
		goto tr3275;
	goto tr71;
tr3275:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1200;
st1200:
	if ( ++p == pe )
		goto _test_eof1200;
case 1200:
	goto st0;
st1013:
	if ( ++p == pe )
		goto _test_eof1013;
case 1013:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3276;
	goto tr1881;
tr3276:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1014;
tr3280:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1014;
st1014:
	if ( ++p == pe )
		goto _test_eof1014;
case 1014:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3277;
		case 32: goto tr3277;
		case 40: goto tr3278;
		case 41: goto tr3279;
		case 1034: goto tr3281;
		case 1083: goto tr3282;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3280;
	goto tr1881;
tr3284:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1015;
tr3285:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1015;
tr3287:
	{
		s->line_counter++;
	}
	goto st1015;
tr3313:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1015;
tr3277:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1015;
tr3278:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1015;
tr3279:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1015;
tr3281:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1015;
st1015:
	if ( ++p == pe )
		goto _test_eof1015;
case 1015:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1015;
		case 32: goto st1015;
		case 40: goto tr3284;
		case 41: goto tr3285;
		case 1034: goto tr3287;
		case 1083: goto st1023;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3286;
	goto tr1881;
tr3286:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1016;
tr3292:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1016;
st1016:
	if ( ++p == pe )
		goto _test_eof1016;
case 1016:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3289;
		case 32: goto tr3289;
		case 40: goto tr3290;
		case 41: goto tr3291;
		case 1034: goto tr3293;
		case 1083: goto tr3294;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3292;
	goto tr1881;
tr3296:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1017;
tr3297:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1017;
tr3299:
	{
		s->line_counter++;
	}
	goto st1017;
tr3311:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1017;
tr3289:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1017;
tr3290:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1017;
tr3291:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1017;
tr3293:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1017;
st1017:
	if ( ++p == pe )
		goto _test_eof1017;
case 1017:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1017;
		case 32: goto st1017;
		case 40: goto tr3296;
		case 41: goto tr3297;
		case 1034: goto tr3299;
		case 1083: goto st1022;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3298;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3298;
	} else
		goto tr3298;
	goto tr71;
tr3298:
	{ p--; {stack[top++] = 1018; goto st487;} }
	goto st1018;
st1018:
	if ( ++p == pe )
		goto _test_eof1018;
case 1018:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1019;
		case 32: goto st1019;
		case 40: goto tr3302;
		case 41: goto tr3303;
		case 1034: goto tr3304;
		case 1083: goto st1021;
	}
	goto tr71;
tr3302:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1019;
tr3303:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1019;
tr3304:
	{
		s->line_counter++;
	}
	goto st1019;
tr3309:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1019;
st1019:
	if ( ++p == pe )
		goto _test_eof1019;
case 1019:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1019;
		case 32: goto st1019;
		case 40: goto tr3302;
		case 41: goto tr3303;
		case 43: goto tr3306;
		case 1034: goto tr3304;
		case 1083: goto st1021;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr3306;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3306;
	} else
		goto tr3306;
	goto tr71;
tr3306:
	{ p--; {stack[top++] = 1020; goto st329;} }
	goto st1020;
st1020:
	if ( ++p == pe )
		goto _test_eof1020;
case 1020:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1152 + ((*p) - -128));
			if (
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1152 + ((*p) - -128));
		if (
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr3307;
		case 1595: goto tr3307;
	}
	goto tr71;
tr3307:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1201;
st1201:
	if ( ++p == pe )
		goto _test_eof1201;
case 1201:
	goto st0;
tr3308:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1021;
st1021:
	if ( ++p == pe )
		goto _test_eof1021;
case 1021:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3309;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3308;
	goto tr71;
tr3310:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1022;
tr3294:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1022;
st1022:
	if ( ++p == pe )
		goto _test_eof1022;
case 1022:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3311;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3310;
	goto tr71;
tr3312:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1023;
tr3282:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1023;
st1023:
	if ( ++p == pe )
		goto _test_eof1023;
case 1023:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3313;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3312;
	goto tr71;
st1024:
	if ( ++p == pe )
		goto _test_eof1024;
case 1024:
	if ( (*p) == 43 )
		goto tr3314;
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr3314;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr3314;
	} else
		goto tr3314;
	goto tr71;
tr3314:
	{ p--; {stack[top++] = 1025; goto st329;} }
	goto st1025;
st1025:
	if ( ++p == pe )
		goto _test_eof1025;
case 1025:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1152 + ((*p) - -128));
			if (
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1152 + ((*p) - -128));
		if (
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr3315;
		case 1595: goto tr3315;
	}
	goto tr71;
tr3315:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1202;
st1202:
	if ( ++p == pe )
		goto _test_eof1202;
case 1202:
	goto st0;
st1026:
	if ( ++p == pe )
		goto _test_eof1026;
case 1026:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3316;
	goto tr1881;
tr3316:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1027;
tr3320:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1027;
st1027:
	if ( ++p == pe )
		goto _test_eof1027;
case 1027:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3317;
		case 32: goto tr3317;
		case 40: goto tr3318;
		case 41: goto tr3319;
		case 1034: goto tr3321;
		case 1083: goto tr3322;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3320;
	goto tr1881;
tr3324:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1028;
tr3325:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1028;
tr3327:
	{
		s->line_counter++;
	}
	goto st1028;
tr3391:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1028;
tr3317:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1028;
tr3318:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1028;
tr3319:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1028;
tr3321:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1028;
st1028:
	if ( ++p == pe )
		goto _test_eof1028;
case 1028:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1028;
		case 32: goto st1028;
		case 40: goto tr3324;
		case 41: goto tr3325;
		case 1034: goto tr3327;
		case 1083: goto st1054;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3326;
	goto tr1881;
tr3326:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1029;
tr3332:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1029;
st1029:
	if ( ++p == pe )
		goto _test_eof1029;
case 1029:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3329;
		case 32: goto tr3329;
		case 40: goto tr3330;
		case 41: goto tr3331;
		case 1034: goto tr3333;
		case 1083: goto tr3334;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3332;
	goto tr1881;
tr3336:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1030;
tr3337:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1030;
tr3339:
	{
		s->line_counter++;
	}
	goto st1030;
tr3389:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1030;
tr3329:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1030;
tr3330:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1030;
tr3331:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1030;
tr3333:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1030;
st1030:
	if ( ++p == pe )
		goto _test_eof1030;
case 1030:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1030;
		case 32: goto st1030;
		case 40: goto tr3336;
		case 41: goto tr3337;
		case 1034: goto tr3339;
		case 1083: goto st1053;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3338;
	goto tr1881;
tr3338:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1031;
tr3344:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1031;
st1031:
	if ( ++p == pe )
		goto _test_eof1031;
case 1031:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3341;
		case 32: goto tr3341;
		case 40: goto tr3342;
		case 41: goto tr3343;
		case 1034: goto tr3345;
		case 1083: goto tr3346;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3344;
	goto tr1881;
tr3348:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1032;
tr3349:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1032;
tr3352:
	{
		s->line_counter++;
	}
	goto st1032;
tr3387:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1032;
tr3341:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1032;
tr3342:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1032;
tr3343:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1032;
tr3345:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1032;
st1032:
	if ( ++p == pe )
		goto _test_eof1032;
case 1032:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1032;
		case 32: goto st1032;
		case 40: goto tr3348;
		case 41: goto tr3349;
		case 45: goto tr3350;
		case 1034: goto tr3352;
		case 1083: goto st1052;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3351;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3351;
	} else
		goto tr3351;
	goto tr2552;
tr3350:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1033;
st1033:
	if ( ++p == pe )
		goto _test_eof1033;
case 1033:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3354;
		case 32: goto tr3354;
		case 40: goto tr3355;
		case 41: goto tr3356;
		case 1034: goto tr3357;
		case 1083: goto tr3358;
	}
	goto tr2552;
tr3361:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1034;
tr3362:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1034;
tr3364:
	{
		s->line_counter++;
	}
	goto st1034;
tr3383:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1034;
tr3354:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1034;
tr3355:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1034;
tr3356:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1034;
tr3357:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1034;
st1034:
	if ( ++p == pe )
		goto _test_eof1034;
case 1034:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1034;
		case 32: goto st1034;
		case 40: goto tr3361;
		case 41: goto tr3362;
		case 1034: goto tr3364;
		case 1083: goto st1049;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3363;
	} else if ( _widec > 86 ) {
		if ( 97 <= _widec && _widec <= 118 )
			goto tr3363;
	} else
		goto tr3363;
	goto tr3359;
tr3363:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1035;
tr3378:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1035;
st1035:
	if ( ++p == pe )
		goto _test_eof1035;
case 1035:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3366;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3366;
	} else
		goto tr3366;
	goto tr3359;
tr3366:
	{
		*(rdata_tail++) += second_left_base32hex_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = second_right_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1036;
st1036:
	if ( ++p == pe )
		goto _test_eof1036;
case 1036:
	if ( (*p) == 61 )
		goto st1047;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3367;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3367;
	} else
		goto tr3367;
	goto tr3359;
tr3367:
	{
		*rdata_tail += third_base32hex_to_num[(uint8_t)(*p)];
	}
	goto st1037;
st1037:
	if ( ++p == pe )
		goto _test_eof1037;
case 1037:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3369;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3369;
	} else
		goto tr3369;
	goto tr3359;
tr3369:
	{
		*(rdata_tail++) += fourth_left_base32hex_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = fourth_right_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1038;
st1038:
	if ( ++p == pe )
		goto _test_eof1038;
case 1038:
	if ( (*p) == 61 )
		goto st1046;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3370;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3370;
	} else
		goto tr3370;
	goto tr3359;
tr3370:
	{
		*(rdata_tail++) += fifth_left_base32hex_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = fifth_right_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1039;
st1039:
	if ( ++p == pe )
		goto _test_eof1039;
case 1039:
	if ( (*p) == 61 )
		goto st1044;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3372;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3372;
	} else
		goto tr3372;
	goto tr3359;
tr3372:
	{
		*rdata_tail += sixth_base32hex_to_num[(uint8_t)(*p)];
	}
	goto st1040;
st1040:
	if ( ++p == pe )
		goto _test_eof1040;
case 1040:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3374;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3374;
	} else
		goto tr3374;
	goto tr3359;
tr3374:
	{
		*(rdata_tail++) += seventh_left_base32hex_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = seventh_right_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1041;
st1041:
	if ( ++p == pe )
		goto _test_eof1041;
case 1041:
	if ( (*p) == 61 )
		goto st1042;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3375;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3375;
	} else
		goto tr3375;
	goto tr3359;
tr3375:
	{
		*(rdata_tail++) += eighth_base32hex_to_num[(uint8_t)(*p)];
	}
	goto st1042;
st1042:
	if ( ++p == pe )
		goto _test_eof1042;
case 1042:
	switch( (*p) ) {
		case 32: goto tr3377;
		case 59: goto tr3377;
	}
	if ( (*p) < 48 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr3377;
		} else if ( (*p) >= 9 )
			goto tr3377;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 86 ) {
			if ( 97 <= (*p) && (*p) <= 118 )
				goto tr3378;
		} else if ( (*p) >= 65 )
			goto tr3378;
	} else
		goto tr3378;
	goto tr3359;
tr3377:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{ p--; {stack[top++] = 1043; goto st336;} }
	goto st1043;
st1043:
	if ( ++p == pe )
		goto _test_eof1043;
case 1043:
	switch( (*p) ) {
		case 32: goto tr3379;
		case 59: goto tr3379;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3379;
	} else if ( (*p) >= 9 )
		goto tr3379;
	goto tr71;
tr3379:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1203;
st1203:
	if ( ++p == pe )
		goto _test_eof1203;
case 1203:
	goto st0;
st1044:
	if ( ++p == pe )
		goto _test_eof1044;
case 1044:
	if ( (*p) == 61 )
		goto st1045;
	goto tr3359;
st1045:
	if ( ++p == pe )
		goto _test_eof1045;
case 1045:
	if ( (*p) == 61 )
		goto st1042;
	goto tr3359;
st1046:
	if ( ++p == pe )
		goto _test_eof1046;
case 1046:
	if ( (*p) == 61 )
		goto st1044;
	goto tr3359;
st1047:
	if ( ++p == pe )
		goto _test_eof1047;
case 1047:
	if ( (*p) == 61 )
		goto st1048;
	goto tr3359;
st1048:
	if ( ++p == pe )
		goto _test_eof1048;
case 1048:
	if ( (*p) == 61 )
		goto st1046;
	goto tr3359;
tr3382:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1049;
tr3358:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1049;
st1049:
	if ( ++p == pe )
		goto _test_eof1049;
case 1049:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3383;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3382;
	goto tr71;
tr3385:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1050;
tr3351:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1050;
st1050:
	if ( ++p == pe )
		goto _test_eof1050;
case 1050:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3384;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3384;
	} else
		goto tr3384;
	goto tr2552;
tr3384:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1051;
st1051:
	if ( ++p == pe )
		goto _test_eof1051;
case 1051:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3354;
		case 32: goto tr3354;
		case 40: goto tr3355;
		case 41: goto tr3356;
		case 1034: goto tr3357;
		case 1083: goto tr3358;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3385;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3385;
	} else
		goto tr3385;
	goto tr2552;
tr3386:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1052;
tr3346:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1052;
st1052:
	if ( ++p == pe )
		goto _test_eof1052;
case 1052:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3387;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3386;
	goto tr71;
tr3388:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1053;
tr3334:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1053;
st1053:
	if ( ++p == pe )
		goto _test_eof1053;
case 1053:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3389;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3388;
	goto tr71;
tr3390:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1054;
tr3322:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1054;
st1054:
	if ( ++p == pe )
		goto _test_eof1054;
case 1054:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3391;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3390;
	goto tr71;
st1055:
	if ( ++p == pe )
		goto _test_eof1055;
case 1055:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3392;
	goto tr1881;
tr3392:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1056;
tr3396:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1056;
st1056:
	if ( ++p == pe )
		goto _test_eof1056;
case 1056:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3393;
		case 32: goto tr3393;
		case 40: goto tr3394;
		case 41: goto tr3395;
		case 1034: goto tr3397;
		case 1083: goto tr3398;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3396;
	goto tr1881;
tr3400:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1057;
tr3401:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1057;
tr3403:
	{
		s->line_counter++;
	}
	goto st1057;
tr3438:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1057;
tr3393:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1057;
tr3394:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1057;
tr3395:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1057;
tr3397:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1057;
st1057:
	if ( ++p == pe )
		goto _test_eof1057;
case 1057:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1057;
		case 32: goto st1057;
		case 40: goto tr3400;
		case 41: goto tr3401;
		case 1034: goto tr3403;
		case 1083: goto st1067;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3402;
	goto tr1881;
tr3402:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1058;
tr3408:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1058;
st1058:
	if ( ++p == pe )
		goto _test_eof1058;
case 1058:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3405;
		case 32: goto tr3405;
		case 40: goto tr3406;
		case 41: goto tr3407;
		case 1034: goto tr3409;
		case 1083: goto tr3410;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3408;
	goto tr1881;
tr3412:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1059;
tr3413:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1059;
tr3415:
	{
		s->line_counter++;
	}
	goto st1059;
tr3436:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1059;
tr3405:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1059;
tr3406:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1059;
tr3407:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1059;
tr3409:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1059;
st1059:
	if ( ++p == pe )
		goto _test_eof1059;
case 1059:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1059;
		case 32: goto st1059;
		case 40: goto tr3412;
		case 41: goto tr3413;
		case 1034: goto tr3415;
		case 1083: goto st1066;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3414;
	goto tr1881;
tr3414:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1060;
tr3420:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1060;
st1060:
	if ( ++p == pe )
		goto _test_eof1060;
case 1060:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3417;
		case 32: goto tr3417;
		case 40: goto tr3418;
		case 41: goto tr3419;
		case 1034: goto tr3421;
		case 1083: goto tr3422;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3420;
	goto tr1881;
tr3424:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1061;
tr3425:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1061;
tr3428:
	{
		s->line_counter++;
	}
	goto st1061;
tr3434:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1061;
tr3417:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1061;
tr3418:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1061;
tr3419:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1061;
tr3421:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1061;
st1061:
	if ( ++p == pe )
		goto _test_eof1061;
case 1061:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1061;
		case 32: goto st1061;
		case 40: goto tr3424;
		case 41: goto tr3425;
		case 45: goto tr3426;
		case 1034: goto tr3428;
		case 1083: goto st1065;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3427;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3427;
	} else
		goto tr3427;
	goto tr2552;
tr3426:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1062;
st1062:
	if ( ++p == pe )
		goto _test_eof1062;
case 1062:
	switch( (*p) ) {
		case 32: goto tr3430;
		case 59: goto tr3430;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3430;
	} else if ( (*p) >= 9 )
		goto tr3430;
	goto tr2552;
tr3430:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1204;
st1204:
	if ( ++p == pe )
		goto _test_eof1204;
case 1204:
	goto st0;
tr3432:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1063;
tr3427:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1063;
st1063:
	if ( ++p == pe )
		goto _test_eof1063;
case 1063:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3431;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3431;
	} else
		goto tr3431;
	goto tr2552;
tr3431:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1064;
st1064:
	if ( ++p == pe )
		goto _test_eof1064;
case 1064:
	switch( (*p) ) {
		case 32: goto tr3430;
		case 59: goto tr3430;
	}
	if ( (*p) < 48 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr3430;
		} else if ( (*p) >= 9 )
			goto tr3430;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 70 ) {
			if ( 97 <= (*p) && (*p) <= 102 )
				goto tr3432;
		} else if ( (*p) >= 65 )
			goto tr3432;
	} else
		goto tr3432;
	goto tr2552;
tr3433:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1065;
tr3422:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1065;
st1065:
	if ( ++p == pe )
		goto _test_eof1065;
case 1065:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3434;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3433;
	goto tr71;
tr3435:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1066;
tr3410:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1066;
st1066:
	if ( ++p == pe )
		goto _test_eof1066;
case 1066:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3436;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3435;
	goto tr71;
tr3437:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1067;
tr3398:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1067;
st1067:
	if ( ++p == pe )
		goto _test_eof1067;
case 1067:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3438;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3437;
	goto tr71;
st1068:
	if ( ++p == pe )
		goto _test_eof1068;
case 1068:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3439;
	goto tr1881;
tr3439:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1069;
tr3443:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1069;
st1069:
	if ( ++p == pe )
		goto _test_eof1069;
case 1069:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3440;
		case 32: goto tr3440;
		case 40: goto tr3441;
		case 41: goto tr3442;
		case 1034: goto tr3444;
		case 1083: goto tr3445;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3443;
	goto tr1881;
tr3447:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1070;
tr3448:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1070;
tr3450:
	{
		s->line_counter++;
	}
	goto st1070;
tr3492:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1070;
tr3440:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1070;
tr3441:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1070;
tr3442:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1070;
tr3444:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1070;
st1070:
	if ( ++p == pe )
		goto _test_eof1070;
case 1070:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1070;
		case 32: goto st1070;
		case 40: goto tr3447;
		case 41: goto tr3448;
		case 1034: goto tr3450;
		case 1083: goto st1080;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3449;
	goto tr1881;
tr3449:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1071;
tr3455:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1071;
st1071:
	if ( ++p == pe )
		goto _test_eof1071;
case 1071:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3452;
		case 32: goto tr3452;
		case 40: goto tr3453;
		case 41: goto tr3454;
		case 1034: goto tr3456;
		case 1083: goto tr3457;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3455;
	goto tr1881;
tr3459:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1072;
tr3460:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1072;
tr3462:
	{
		s->line_counter++;
	}
	goto st1072;
tr3490:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1072;
tr3452:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1072;
tr3453:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1072;
tr3454:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1072;
tr3456:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1072;
st1072:
	if ( ++p == pe )
		goto _test_eof1072;
case 1072:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1072;
		case 32: goto st1072;
		case 40: goto tr3459;
		case 41: goto tr3460;
		case 1034: goto tr3462;
		case 1083: goto st1079;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3461;
	goto tr1881;
tr3461:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1073;
tr3467:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1073;
st1073:
	if ( ++p == pe )
		goto _test_eof1073;
case 1073:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3464;
		case 32: goto tr3464;
		case 40: goto tr3465;
		case 41: goto tr3466;
		case 1034: goto tr3468;
		case 1083: goto tr3469;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3467;
	goto tr1881;
tr3471:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1074;
tr3472:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1074;
tr3474:
	{
		s->line_counter++;
	}
	goto st1074;
tr3488:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1074;
tr3464:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1074;
tr3465:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1074;
tr3466:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1074;
tr3468:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1074;
st1074:
	if ( ++p == pe )
		goto _test_eof1074;
case 1074:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1074;
		case 32: goto st1074;
		case 40: goto tr3471;
		case 41: goto tr3472;
		case 1034: goto tr3474;
		case 1083: goto st1078;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3473;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3473;
	} else
		goto tr3473;
	goto tr2552;
tr3473:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1075;
st1075:
	if ( ++p == pe )
		goto _test_eof1075;
case 1075:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3476;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3476;
	} else
		goto tr3476;
	goto tr2552;
tr3478:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1076;
tr3479:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1076;
tr3480:
	{
		s->line_counter++;
	}
	goto st1076;
tr3486:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1076;
tr3476:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1076;
st1076:
	if ( ++p == pe )
		goto _test_eof1076;
case 1076:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st1076;
		case 32: goto st1076;
		case 40: goto tr3478;
		case 41: goto tr3479;
		case 2058: goto tr3480;
		case 2107: goto st1077;
		case 2314: goto tr3482;
		case 2363: goto tr3482;
		case 2570: goto tr3483;
		case 2619: goto tr3484;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3473;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3473;
	} else
		goto tr3473;
	goto tr2552;
tr3485:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1077;
st1077:
	if ( ++p == pe )
		goto _test_eof1077;
case 1077:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3486;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3485;
	goto tr2552;
tr3482:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1205;
st1205:
	if ( ++p == pe )
		goto _test_eof1205;
case 1205:
	goto st0;
tr3483:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1206;
st1206:
	if ( ++p == pe )
		goto _test_eof1206;
case 1206:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
			if (
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
		if (
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st1076;
		case 32: goto st1076;
		case 40: goto tr3478;
		case 41: goto tr3479;
		case 2058: goto tr3480;
		case 2107: goto st1077;
		case 2314: goto tr3482;
		case 2363: goto tr3482;
		case 2570: goto tr3483;
		case 2619: goto tr3484;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3473;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3473;
	} else
		goto tr3473;
	goto tr2552;
tr3484:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1207;
st1207:
	if ( ++p == pe )
		goto _test_eof1207;
case 1207:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3486;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3485;
	goto tr2552;
tr3487:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1078;
tr3469:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1078;
st1078:
	if ( ++p == pe )
		goto _test_eof1078;
case 1078:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3488;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3487;
	goto tr71;
tr3489:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1079;
tr3457:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1079;
st1079:
	if ( ++p == pe )
		goto _test_eof1079;
case 1079:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3490;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3489;
	goto tr71;
tr3491:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1080;
tr3445:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1080;
st1080:
	if ( ++p == pe )
		goto _test_eof1080;
case 1080:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3492;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3491;
	goto tr71;
st1081:
	if ( ++p == pe )
		goto _test_eof1081;
case 1081:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3493;
	goto tr1881;
tr3493:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1082;
tr3497:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1082;
st1082:
	if ( ++p == pe )
		goto _test_eof1082;
case 1082:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3494;
		case 32: goto tr3494;
		case 40: goto tr3495;
		case 41: goto tr3496;
		case 1034: goto tr3498;
		case 1083: goto tr3499;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3497;
	goto tr1881;
tr3501:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1083;
tr3502:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1083;
tr3504:
	{
		s->line_counter++;
	}
	goto st1083;
tr3509:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1083;
tr3494:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1083;
tr3495:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1083;
tr3496:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1083;
tr3498:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1083;
st1083:
	if ( ++p == pe )
		goto _test_eof1083;
case 1083:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1083;
		case 32: goto st1083;
		case 40: goto tr3501;
		case 41: goto tr3502;
		case 46: goto tr3503;
		case 1034: goto tr3504;
		case 1083: goto st1085;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3503;
	goto tr1858;
tr3503:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1084;
tr3507:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1084;
st1084:
	if ( ++p == pe )
		goto _test_eof1084;
case 1084:
	switch( (*p) ) {
		case 32: goto tr3506;
		case 46: goto tr3507;
		case 59: goto tr3506;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr3506;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3507;
	} else
		goto tr3506;
	goto tr1858;
tr3506:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st268;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1208;
st1208:
	if ( ++p == pe )
		goto _test_eof1208;
case 1208:
	goto st0;
tr3508:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1085;
tr3499:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1085;
st1085:
	if ( ++p == pe )
		goto _test_eof1085;
case 1085:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3509;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3508;
	goto tr71;
st1086:
	if ( ++p == pe )
		goto _test_eof1086;
case 1086:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3510;
	goto tr1881;
tr3510:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1087;
tr3514:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1087;
st1087:
	if ( ++p == pe )
		goto _test_eof1087;
case 1087:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3511;
		case 32: goto tr3511;
		case 40: goto tr3512;
		case 41: goto tr3513;
		case 1034: goto tr3515;
		case 1083: goto tr3516;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3514;
	goto tr1881;
tr3518:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1088;
tr3519:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1088;
tr3521:
	{
		s->line_counter++;
	}
	goto st1088;
tr3535:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1088;
tr3511:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1088;
tr3512:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1088;
tr3513:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1088;
tr3515:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1088;
st1088:
	if ( ++p == pe )
		goto _test_eof1088;
case 1088:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1088;
		case 32: goto st1088;
		case 40: goto tr3518;
		case 41: goto tr3519;
		case 1034: goto tr3521;
		case 1083: goto st1098;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3520;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3520;
	} else
		goto tr3520;
	goto tr2552;
tr3520:
	{
		s->item_length = 0;
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1089;
st1089:
	if ( ++p == pe )
		goto _test_eof1089;
case 1089:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3523;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3523;
	} else
		goto tr3523;
	goto tr2552;
tr3523:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1090;
st1090:
	if ( ++p == pe )
		goto _test_eof1090;
case 1090:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3524;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3524;
	} else
		goto tr3524;
	goto tr2552;
tr3524:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1091;
st1091:
	if ( ++p == pe )
		goto _test_eof1091;
case 1091:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3525;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3525;
	} else
		goto tr3525;
	goto tr2552;
tr3525:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1092;
st1092:
	if ( ++p == pe )
		goto _test_eof1092;
case 1092:
	if ( (*p) == 58 )
		goto tr3527;
	goto tr3526;
tr3527:
	{
		s->item_length++;
	}
	goto st1093;
st1093:
	if ( ++p == pe )
		goto _test_eof1093;
case 1093:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3528;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3528;
	} else
		goto tr3528;
	goto tr2552;
tr3528:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1094;
st1094:
	if ( ++p == pe )
		goto _test_eof1094;
case 1094:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3529;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3529;
	} else
		goto tr3529;
	goto tr2552;
tr3529:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1095;
st1095:
	if ( ++p == pe )
		goto _test_eof1095;
case 1095:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3530;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3530;
	} else
		goto tr3530;
	goto tr2552;
tr3530:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1096;
st1096:
	if ( ++p == pe )
		goto _test_eof1096;
case 1096:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3531;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3531;
	} else
		goto tr3531;
	goto tr2552;
tr3531:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1097;
st1097:
	if ( ++p == pe )
		goto _test_eof1097;
case 1097:
	switch( (*p) ) {
		case 32: goto tr3533;
		case 58: goto tr3527;
		case 59: goto tr3533;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3533;
	} else if ( (*p) >= 9 )
		goto tr3533;
	goto tr3532;
tr3533:
	{
		s->item_length++;
	}
	{
		if (s->item_length != 4) {
			WARN(ZS_BAD_L64_LENGTH);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1209;
st1209:
	if ( ++p == pe )
		goto _test_eof1209;
case 1209:
	goto st0;
tr3534:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1098;
tr3516:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1098;
st1098:
	if ( ++p == pe )
		goto _test_eof1098;
case 1098:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3535;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3534;
	goto tr71;
st1099:
	if ( ++p == pe )
		goto _test_eof1099;
case 1099:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3536;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3536;
	} else
		goto tr3536;
	goto tr2552;
tr3536:
	{
		s->item_length = 0;
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1100;
st1100:
	if ( ++p == pe )
		goto _test_eof1100;
case 1100:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3537;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3537;
	} else
		goto tr3537;
	goto tr2552;
tr3537:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1101;
st1101:
	if ( ++p == pe )
		goto _test_eof1101;
case 1101:
	if ( (*p) == 45 )
		goto tr3539;
	goto tr3538;
tr3539:
	{
		s->item_length++;
	}
	goto st1102;
st1102:
	if ( ++p == pe )
		goto _test_eof1102;
case 1102:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3540;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3540;
	} else
		goto tr3540;
	goto tr2552;
tr3540:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1103;
st1103:
	if ( ++p == pe )
		goto _test_eof1103;
case 1103:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3541;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3541;
	} else
		goto tr3541;
	goto tr2552;
tr3541:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1104;
st1104:
	if ( ++p == pe )
		goto _test_eof1104;
case 1104:
	switch( (*p) ) {
		case 32: goto tr3542;
		case 45: goto tr3539;
		case 59: goto tr3542;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3542;
	} else if ( (*p) >= 9 )
		goto tr3542;
	goto tr3538;
tr3542:
	{
		s->item_length++;
	}
	{
		if (s->item_length != 6) {
			WARN(ZS_BAD_EUI_LENGTH);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1210;
st1210:
	if ( ++p == pe )
		goto _test_eof1210;
case 1210:
	goto st0;
st1105:
	if ( ++p == pe )
		goto _test_eof1105;
case 1105:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3543;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3543;
	} else
		goto tr3543;
	goto tr2552;
tr3543:
	{
		s->item_length = 0;
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1106;
st1106:
	if ( ++p == pe )
		goto _test_eof1106;
case 1106:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3544;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3544;
	} else
		goto tr3544;
	goto tr2552;
tr3544:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1107;
st1107:
	if ( ++p == pe )
		goto _test_eof1107;
case 1107:
	if ( (*p) == 45 )
		goto tr3545;
	goto tr3538;
tr3545:
	{
		s->item_length++;
	}
	goto st1108;
st1108:
	if ( ++p == pe )
		goto _test_eof1108;
case 1108:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3546;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3546;
	} else
		goto tr3546;
	goto tr2552;
tr3546:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1109;
st1109:
	if ( ++p == pe )
		goto _test_eof1109;
case 1109:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3547;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3547;
	} else
		goto tr3547;
	goto tr2552;
tr3547:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1110;
st1110:
	if ( ++p == pe )
		goto _test_eof1110;
case 1110:
	switch( (*p) ) {
		case 32: goto tr3548;
		case 45: goto tr3545;
		case 59: goto tr3548;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3548;
	} else if ( (*p) >= 9 )
		goto tr3548;
	goto tr3538;
tr3548:
	{
		s->item_length++;
	}
	{
		if (s->item_length != 8) {
			WARN(ZS_BAD_EUI_LENGTH);
			p--; {goto st268;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1211;
st1211:
	if ( ++p == pe )
		goto _test_eof1211;
case 1211:
	goto st0;
st1111:
	if ( ++p == pe )
		goto _test_eof1111;
case 1111:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3549;
	goto tr1881;
tr3549:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1112;
tr3553:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1112;
st1112:
	if ( ++p == pe )
		goto _test_eof1112;
case 1112:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3550;
		case 32: goto tr3550;
		case 40: goto tr3551;
		case 41: goto tr3552;
		case 1034: goto tr3554;
		case 1083: goto tr3555;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3553;
	goto tr1881;
tr3557:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1113;
tr3558:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1113;
tr3560:
	{
		s->line_counter++;
	}
	goto st1113;
tr3578:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1113;
tr3550:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1113;
tr3551:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1113;
tr3552:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1113;
tr3554:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1113;
st1113:
	if ( ++p == pe )
		goto _test_eof1113;
case 1113:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1113;
		case 32: goto st1113;
		case 40: goto tr3557;
		case 41: goto tr3558;
		case 1034: goto tr3560;
		case 1083: goto st1118;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3559;
	goto tr1881;
tr3559:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1114;
tr3565:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1114;
st1114:
	if ( ++p == pe )
		goto _test_eof1114;
case 1114:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3562;
		case 32: goto tr3562;
		case 40: goto tr3563;
		case 41: goto tr3564;
		case 1034: goto tr3566;
		case 1083: goto tr3567;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3565;
	goto tr1881;
tr3570:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1115;
tr3571:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1115;
tr3572:
	{
		s->line_counter++;
	}
	goto st1115;
tr3576:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1115;
tr3562:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1115;
tr3563:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1115;
tr3564:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1115;
tr3566:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1115;
st1115:
	if ( ++p == pe )
		goto _test_eof1115;
case 1115:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1115;
		case 32: goto st1115;
		case 40: goto tr3570;
		case 41: goto tr3571;
		case 1034: goto tr3572;
		case 1083: goto st1117;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr3568;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr3568;
	} else
		goto tr3568;
	goto tr71;
tr3568:
	{ p--; {stack[top++] = 1116; goto st279;} }
	goto st1116;
st1116:
	if ( ++p == pe )
		goto _test_eof1116;
case 1116:
	switch( (*p) ) {
		case 32: goto tr3574;
		case 59: goto tr3574;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3574;
	} else if ( (*p) >= 9 )
		goto tr3574;
	goto tr71;
tr3574:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1212;
st1212:
	if ( ++p == pe )
		goto _test_eof1212;
case 1212:
	goto st0;
tr3575:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1117;
tr3567:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1117;
st1117:
	if ( ++p == pe )
		goto _test_eof1117;
case 1117:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3576;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3575;
	goto tr71;
tr3577:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1118;
tr3555:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1118;
st1118:
	if ( ++p == pe )
		goto _test_eof1118;
case 1118:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3578;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3577;
	goto tr71;
st1119:
	if ( ++p == pe )
		goto _test_eof1119;
case 1119:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3579;
	goto tr1881;
tr3579:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1120;
tr3583:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1120;
st1120:
	if ( ++p == pe )
		goto _test_eof1120;
case 1120:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3580;
		case 32: goto tr3580;
		case 40: goto tr3581;
		case 41: goto tr3582;
		case 1034: goto tr3584;
		case 1083: goto tr3585;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3583;
	goto tr1881;
tr3588:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1121;
tr3589:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1121;
tr3590:
	{
		s->line_counter++;
	}
	goto st1121;
tr3607:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1121;
tr3580:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1121;
tr3581:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1121;
tr3582:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1121;
tr3584:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1121;
st1121:
	if ( ++p == pe )
		goto _test_eof1121;
case 1121:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1121;
		case 32: goto st1121;
		case 40: goto tr3588;
		case 41: goto tr3589;
		case 1034: goto tr3590;
		case 1083: goto st1126;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr3586;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr3586;
	} else
		goto tr3586;
	goto tr71;
tr3586:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{ p--; {stack[top++] = 1122; goto st279;} }
	goto st1122;
st1122:
	if ( ++p == pe )
		goto _test_eof1122;
case 1122:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3592;
		case 32: goto tr3592;
		case 40: goto tr3593;
		case 41: goto tr3594;
		case 1034: goto tr3595;
		case 1083: goto tr3596;
	}
	goto tr71;
tr3599:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1123;
tr3600:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1123;
tr3601:
	{
		s->line_counter++;
	}
	goto st1123;
tr3605:
	{
		s->buffer[s->buffer_length++] = 0;
	}
	{
		s->line_counter++;
	}
	goto st1123;
tr3592:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1123;
tr3593:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = true;
	}
	goto st1123;
tr3594:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st268;}
		}
		s->multiline = false;
	}
	goto st1123;
tr3595:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1123;
st1123:
	if ( ++p == pe )
		goto _test_eof1123;
case 1123:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1123;
		case 32: goto st1123;
		case 40: goto tr3599;
		case 41: goto tr3600;
		case 1034: goto tr3601;
		case 1083: goto st1125;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr3597;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr3597;
	} else
		goto tr3597;
	goto tr71;
tr3597:
	{ p--; {stack[top++] = 1124; goto st279;} }
	goto st1124;
st1124:
	if ( ++p == pe )
		goto _test_eof1124;
case 1124:
	switch( (*p) ) {
		case 32: goto tr3603;
		case 59: goto tr3603;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3603;
	} else if ( (*p) >= 9 )
		goto tr3603;
	goto tr71;
tr3603:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1213;
st1213:
	if ( ++p == pe )
		goto _test_eof1213;
case 1213:
	goto st0;
tr3604:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1125;
tr3596:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1125;
st1125:
	if ( ++p == pe )
		goto _test_eof1125;
case 1125:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3605;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3604;
	goto tr71;
tr3606:
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st1126;
tr3585:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st268;}
		}
	}
	goto st1126;
st1126:
	if ( ++p == pe )
		goto _test_eof1126;
case 1126:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if (
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if (
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3607;
	if ( 896 <= _widec && _widec <= 1151 )
		goto tr3606;
	goto tr71;
	}
	_test_eof1127: cs = 1127; goto _test_eof;
	_test_eof1: cs = 1; goto _test_eof;
	_test_eof2: cs = 2; goto _test_eof;
	_test_eof3: cs = 3; goto _test_eof;
	_test_eof4: cs = 4; goto _test_eof;
	_test_eof5: cs = 5; goto _test_eof;
	_test_eof6: cs = 6; goto _test_eof;
	_test_eof7: cs = 7; goto _test_eof;
	_test_eof8: cs = 8; goto _test_eof;
	_test_eof9: cs = 9; goto _test_eof;
	_test_eof10: cs = 10; goto _test_eof;
	_test_eof11: cs = 11; goto _test_eof;
	_test_eof12: cs = 12; goto _test_eof;
	_test_eof13: cs = 13; goto _test_eof;
	_test_eof1128: cs = 1128; goto _test_eof;
	_test_eof14: cs = 14; goto _test_eof;
	_test_eof15: cs = 15; goto _test_eof;
	_test_eof16: cs = 16; goto _test_eof;
	_test_eof17: cs = 17; goto _test_eof;
	_test_eof18: cs = 18; goto _test_eof;
	_test_eof19: cs = 19; goto _test_eof;
	_test_eof20: cs = 20; goto _test_eof;
	_test_eof21: cs = 21; goto _test_eof;
	_test_eof22: cs = 22; goto _test_eof;
	_test_eof23: cs = 23; goto _test_eof;
	_test_eof24: cs = 24; goto _test_eof;
	_test_eof25: cs = 25; goto _test_eof;
	_test_eof26: cs = 26; goto _test_eof;
	_test_eof27: cs = 27; goto _test_eof;
	_test_eof28: cs = 28; goto _test_eof;
	_test_eof29: cs = 29; goto _test_eof;
	_test_eof30: cs = 30; goto _test_eof;
	_test_eof31: cs = 31; goto _test_eof;
	_test_eof32: cs = 32; goto _test_eof;
	_test_eof33: cs = 33; goto _test_eof;
	_test_eof34: cs = 34; goto _test_eof;
	_test_eof35: cs = 35; goto _test_eof;
	_test_eof36: cs = 36; goto _test_eof;
	_test_eof37: cs = 37; goto _test_eof;
	_test_eof38: cs = 38; goto _test_eof;
	_test_eof39: cs = 39; goto _test_eof;
	_test_eof40: cs = 40; goto _test_eof;
	_test_eof41: cs = 41; goto _test_eof;
	_test_eof42: cs = 42; goto _test_eof;
	_test_eof43: cs = 43; goto _test_eof;
	_test_eof44: cs = 44; goto _test_eof;
	_test_eof45: cs = 45; goto _test_eof;
	_test_eof46: cs = 46; goto _test_eof;
	_test_eof47: cs = 47; goto _test_eof;
	_test_eof48: cs = 48; goto _test_eof;
	_test_eof49: cs = 49; goto _test_eof;
	_test_eof50: cs = 50; goto _test_eof;
	_test_eof51: cs = 51; goto _test_eof;
	_test_eof52: cs = 52; goto _test_eof;
	_test_eof53: cs = 53; goto _test_eof;
	_test_eof54: cs = 54; goto _test_eof;
	_test_eof55: cs = 55; goto _test_eof;
	_test_eof56: cs = 56; goto _test_eof;
	_test_eof57: cs = 57; goto _test_eof;
	_test_eof58: cs = 58; goto _test_eof;
	_test_eof59: cs = 59; goto _test_eof;
	_test_eof60: cs = 60; goto _test_eof;
	_test_eof61: cs = 61; goto _test_eof;
	_test_eof62: cs = 62; goto _test_eof;
	_test_eof63: cs = 63; goto _test_eof;
	_test_eof64: cs = 64; goto _test_eof;
	_test_eof65: cs = 65; goto _test_eof;
	_test_eof66: cs = 66; goto _test_eof;
	_test_eof67: cs = 67; goto _test_eof;
	_test_eof68: cs = 68; goto _test_eof;
	_test_eof69: cs = 69; goto _test_eof;
	_test_eof70: cs = 70; goto _test_eof;
	_test_eof71: cs = 71; goto _test_eof;
	_test_eof72: cs = 72; goto _test_eof;
	_test_eof73: cs = 73; goto _test_eof;
	_test_eof74: cs = 74; goto _test_eof;
	_test_eof75: cs = 75; goto _test_eof;
	_test_eof76: cs = 76; goto _test_eof;
	_test_eof77: cs = 77; goto _test_eof;
	_test_eof78: cs = 78; goto _test_eof;
	_test_eof79: cs = 79; goto _test_eof;
	_test_eof80: cs = 80; goto _test_eof;
	_test_eof81: cs = 81; goto _test_eof;
	_test_eof82: cs = 82; goto _test_eof;
	_test_eof83: cs = 83; goto _test_eof;
	_test_eof84: cs = 84; goto _test_eof;
	_test_eof85: cs = 85; goto _test_eof;
	_test_eof86: cs = 86; goto _test_eof;
	_test_eof87: cs = 87; goto _test_eof;
	_test_eof88: cs = 88; goto _test_eof;
	_test_eof89: cs = 89; goto _test_eof;
	_test_eof90: cs = 90; goto _test_eof;
	_test_eof91: cs = 91; goto _test_eof;
	_test_eof92: cs = 92; goto _test_eof;
	_test_eof93: cs = 93; goto _test_eof;
	_test_eof94: cs = 94; goto _test_eof;
	_test_eof95: cs = 95; goto _test_eof;
	_test_eof96: cs = 96; goto _test_eof;
	_test_eof97: cs = 97; goto _test_eof;
	_test_eof98: cs = 98; goto _test_eof;
	_test_eof99: cs = 99; goto _test_eof;
	_test_eof100: cs = 100; goto _test_eof;
	_test_eof101: cs = 101; goto _test_eof;
	_test_eof102: cs = 102; goto _test_eof;
	_test_eof103: cs = 103; goto _test_eof;
	_test_eof104: cs = 104; goto _test_eof;
	_test_eof105: cs = 105; goto _test_eof;
	_test_eof106: cs = 106; goto _test_eof;
	_test_eof107: cs = 107; goto _test_eof;
	_test_eof108: cs = 108; goto _test_eof;
	_test_eof109: cs = 109; goto _test_eof;
	_test_eof110: cs = 110; goto _test_eof;
	_test_eof111: cs = 111; goto _test_eof;
	_test_eof112: cs = 112; goto _test_eof;
	_test_eof113: cs = 113; goto _test_eof;
	_test_eof114: cs = 114; goto _test_eof;
	_test_eof115: cs = 115; goto _test_eof;
	_test_eof116: cs = 116; goto _test_eof;
	_test_eof117: cs = 117; goto _test_eof;
	_test_eof118: cs = 118; goto _test_eof;
	_test_eof119: cs = 119; goto _test_eof;
	_test_eof120: cs = 120; goto _test_eof;
	_test_eof121: cs = 121; goto _test_eof;
	_test_eof122: cs = 122; goto _test_eof;
	_test_eof123: cs = 123; goto _test_eof;
	_test_eof124: cs = 124; goto _test_eof;
	_test_eof125: cs = 125; goto _test_eof;
	_test_eof126: cs = 126; goto _test_eof;
	_test_eof127: cs = 127; goto _test_eof;
	_test_eof128: cs = 128; goto _test_eof;
	_test_eof129: cs = 129; goto _test_eof;
	_test_eof130: cs = 130; goto _test_eof;
	_test_eof131: cs = 131; goto _test_eof;
	_test_eof132: cs = 132; goto _test_eof;
	_test_eof133: cs = 133; goto _test_eof;
	_test_eof134: cs = 134; goto _test_eof;
	_test_eof135: cs = 135; goto _test_eof;
	_test_eof136: cs = 136; goto _test_eof;
	_test_eof137: cs = 137; goto _test_eof;
	_test_eof138: cs = 138; goto _test_eof;
	_test_eof139: cs = 139; goto _test_eof;
	_test_eof140: cs = 140; goto _test_eof;
	_test_eof141: cs = 141; goto _test_eof;
	_test_eof1129: cs = 1129; goto _test_eof;
	_test_eof142: cs = 142; goto _test_eof;
	_test_eof143: cs = 143; goto _test_eof;
	_test_eof144: cs = 144; goto _test_eof;
	_test_eof145: cs = 145; goto _test_eof;
	_test_eof146: cs = 146; goto _test_eof;
	_test_eof147: cs = 147; goto _test_eof;
	_test_eof148: cs = 148; goto _test_eof;
	_test_eof149: cs = 149; goto _test_eof;
	_test_eof150: cs = 150; goto _test_eof;
	_test_eof151: cs = 151; goto _test_eof;
	_test_eof1130: cs = 1130; goto _test_eof;
	_test_eof152: cs = 152; goto _test_eof;
	_test_eof153: cs = 153; goto _test_eof;
	_test_eof154: cs = 154; goto _test_eof;
	_test_eof155: cs = 155; goto _test_eof;
	_test_eof156: cs = 156; goto _test_eof;
	_test_eof157: cs = 157; goto _test_eof;
	_test_eof158: cs = 158; goto _test_eof;
	_test_eof159: cs = 159; goto _test_eof;
	_test_eof1131: cs = 1131; goto _test_eof;
	_test_eof160: cs = 160; goto _test_eof;
	_test_eof161: cs = 161; goto _test_eof;
	_test_eof162: cs = 162; goto _test_eof;
	_test_eof1132: cs = 1132; goto _test_eof;
	_test_eof163: cs = 163; goto _test_eof;
	_test_eof164: cs = 164; goto _test_eof;
	_test_eof165: cs = 165; goto _test_eof;
	_test_eof166: cs = 166; goto _test_eof;
	_test_eof167: cs = 167; goto _test_eof;
	_test_eof168: cs = 168; goto _test_eof;
	_test_eof169: cs = 169; goto _test_eof;
	_test_eof170: cs = 170; goto _test_eof;
	_test_eof171: cs = 171; goto _test_eof;
	_test_eof172: cs = 172; goto _test_eof;
	_test_eof173: cs = 173; goto _test_eof;
	_test_eof1133: cs = 1133; goto _test_eof;
	_test_eof174: cs = 174; goto _test_eof;
	_test_eof175: cs = 175; goto _test_eof;
	_test_eof176: cs = 176; goto _test_eof;
	_test_eof177: cs = 177; goto _test_eof;
	_test_eof1134: cs = 1134; goto _test_eof;
	_test_eof178: cs = 178; goto _test_eof;
	_test_eof179: cs = 179; goto _test_eof;
	_test_eof180: cs = 180; goto _test_eof;
	_test_eof181: cs = 181; goto _test_eof;
	_test_eof182: cs = 182; goto _test_eof;
	_test_eof183: cs = 183; goto _test_eof;
	_test_eof184: cs = 184; goto _test_eof;
	_test_eof185: cs = 185; goto _test_eof;
	_test_eof186: cs = 186; goto _test_eof;
	_test_eof187: cs = 187; goto _test_eof;
	_test_eof188: cs = 188; goto _test_eof;
	_test_eof189: cs = 189; goto _test_eof;
	_test_eof190: cs = 190; goto _test_eof;
	_test_eof191: cs = 191; goto _test_eof;
	_test_eof192: cs = 192; goto _test_eof;
	_test_eof193: cs = 193; goto _test_eof;
	_test_eof1135: cs = 1135; goto _test_eof;
	_test_eof194: cs = 194; goto _test_eof;
	_test_eof195: cs = 195; goto _test_eof;
	_test_eof196: cs = 196; goto _test_eof;
	_test_eof197: cs = 197; goto _test_eof;
	_test_eof198: cs = 198; goto _test_eof;
	_test_eof199: cs = 199; goto _test_eof;
	_test_eof200: cs = 200; goto _test_eof;
	_test_eof201: cs = 201; goto _test_eof;
	_test_eof202: cs = 202; goto _test_eof;
	_test_eof203: cs = 203; goto _test_eof;
	_test_eof204: cs = 204; goto _test_eof;
	_test_eof205: cs = 205; goto _test_eof;
	_test_eof206: cs = 206; goto _test_eof;
	_test_eof207: cs = 207; goto _test_eof;
	_test_eof208: cs = 208; goto _test_eof;
	_test_eof209: cs = 209; goto _test_eof;
	_test_eof1136: cs = 1136; goto _test_eof;
	_test_eof210: cs = 210; goto _test_eof;
	_test_eof211: cs = 211; goto _test_eof;
	_test_eof212: cs = 212; goto _test_eof;
	_test_eof213: cs = 213; goto _test_eof;
	_test_eof214: cs = 214; goto _test_eof;
	_test_eof215: cs = 215; goto _test_eof;
	_test_eof216: cs = 216; goto _test_eof;
	_test_eof217: cs = 217; goto _test_eof;
	_test_eof218: cs = 218; goto _test_eof;
	_test_eof219: cs = 219; goto _test_eof;
	_test_eof220: cs = 220; goto _test_eof;
	_test_eof221: cs = 221; goto _test_eof;
	_test_eof222: cs = 222; goto _test_eof;
	_test_eof223: cs = 223; goto _test_eof;
	_test_eof224: cs = 224; goto _test_eof;
	_test_eof225: cs = 225; goto _test_eof;
	_test_eof226: cs = 226; goto _test_eof;
	_test_eof227: cs = 227; goto _test_eof;
	_test_eof228: cs = 228; goto _test_eof;
	_test_eof229: cs = 229; goto _test_eof;
	_test_eof230: cs = 230; goto _test_eof;
	_test_eof231: cs = 231; goto _test_eof;
	_test_eof232: cs = 232; goto _test_eof;
	_test_eof233: cs = 233; goto _test_eof;
	_test_eof234: cs = 234; goto _test_eof;
	_test_eof235: cs = 235; goto _test_eof;
	_test_eof236: cs = 236; goto _test_eof;
	_test_eof237: cs = 237; goto _test_eof;
	_test_eof238: cs = 238; goto _test_eof;
	_test_eof239: cs = 239; goto _test_eof;
	_test_eof240: cs = 240; goto _test_eof;
	_test_eof241: cs = 241; goto _test_eof;
	_test_eof242: cs = 242; goto _test_eof;
	_test_eof243: cs = 243; goto _test_eof;
	_test_eof244: cs = 244; goto _test_eof;
	_test_eof245: cs = 245; goto _test_eof;
	_test_eof246: cs = 246; goto _test_eof;
	_test_eof247: cs = 247; goto _test_eof;
	_test_eof248: cs = 248; goto _test_eof;
	_test_eof249: cs = 249; goto _test_eof;
	_test_eof250: cs = 250; goto _test_eof;
	_test_eof251: cs = 251; goto _test_eof;
	_test_eof252: cs = 252; goto _test_eof;
	_test_eof253: cs = 253; goto _test_eof;
	_test_eof254: cs = 254; goto _test_eof;
	_test_eof255: cs = 255; goto _test_eof;
	_test_eof256: cs = 256; goto _test_eof;
	_test_eof257: cs = 257; goto _test_eof;
	_test_eof258: cs = 258; goto _test_eof;
	_test_eof259: cs = 259; goto _test_eof;
	_test_eof260: cs = 260; goto _test_eof;
	_test_eof261: cs = 261; goto _test_eof;
	_test_eof262: cs = 262; goto _test_eof;
	_test_eof263: cs = 263; goto _test_eof;
	_test_eof264: cs = 264; goto _test_eof;
	_test_eof265: cs = 265; goto _test_eof;
	_test_eof266: cs = 266; goto _test_eof;
	_test_eof267: cs = 267; goto _test_eof;
	_test_eof268: cs = 268; goto _test_eof;
	_test_eof269: cs = 269; goto _test_eof;
	_test_eof1137: cs = 1137; goto _test_eof;
	_test_eof270: cs = 270; goto _test_eof;
	_test_eof271: cs = 271; goto _test_eof;
	_test_eof1138: cs = 1138; goto _test_eof;
	_test_eof272: cs = 272; goto _test_eof;
	_test_eof273: cs = 273; goto _test_eof;
	_test_eof274: cs = 274; goto _test_eof;
	_test_eof275: cs = 275; goto _test_eof;
	_test_eof276: cs = 276; goto _test_eof;
	_test_eof277: cs = 277; goto _test_eof;
	_test_eof278: cs = 278; goto _test_eof;
	_test_eof279: cs = 279; goto _test_eof;
	_test_eof280: cs = 280; goto _test_eof;
	_test_eof1139: cs = 1139; goto _test_eof;
	_test_eof1140: cs = 1140; goto _test_eof;
	_test_eof281: cs = 281; goto _test_eof;
	_test_eof282: cs = 282; goto _test_eof;
	_test_eof283: cs = 283; goto _test_eof;
	_test_eof284: cs = 284; goto _test_eof;
	_test_eof285: cs = 285; goto _test_eof;
	_test_eof286: cs = 286; goto _test_eof;
	_test_eof287: cs = 287; goto _test_eof;
	_test_eof288: cs = 288; goto _test_eof;
	_test_eof289: cs = 289; goto _test_eof;
	_test_eof290: cs = 290; goto _test_eof;
	_test_eof291: cs = 291; goto _test_eof;
	_test_eof292: cs = 292; goto _test_eof;
	_test_eof293: cs = 293; goto _test_eof;
	_test_eof294: cs = 294; goto _test_eof;
	_test_eof1141: cs = 1141; goto _test_eof;
	_test_eof295: cs = 295; goto _test_eof;
	_test_eof296: cs = 296; goto _test_eof;
	_test_eof297: cs = 297; goto _test_eof;
	_test_eof298: cs = 298; goto _test_eof;
	_test_eof299: cs = 299; goto _test_eof;
	_test_eof300: cs = 300; goto _test_eof;
	_test_eof301: cs = 301; goto _test_eof;
	_test_eof302: cs = 302; goto _test_eof;
	_test_eof303: cs = 303; goto _test_eof;
	_test_eof304: cs = 304; goto _test_eof;
	_test_eof1142: cs = 1142; goto _test_eof;
	_test_eof305: cs = 305; goto _test_eof;
	_test_eof306: cs = 306; goto _test_eof;
	_test_eof307: cs = 307; goto _test_eof;
	_test_eof308: cs = 308; goto _test_eof;
	_test_eof309: cs = 309; goto _test_eof;
	_test_eof310: cs = 310; goto _test_eof;
	_test_eof311: cs = 311; goto _test_eof;
	_test_eof312: cs = 312; goto _test_eof;
	_test_eof313: cs = 313; goto _test_eof;
	_test_eof314: cs = 314; goto _test_eof;
	_test_eof315: cs = 315; goto _test_eof;
	_test_eof316: cs = 316; goto _test_eof;
	_test_eof317: cs = 317; goto _test_eof;
	_test_eof318: cs = 318; goto _test_eof;
	_test_eof1143: cs = 1143; goto _test_eof;
	_test_eof319: cs = 319; goto _test_eof;
	_test_eof320: cs = 320; goto _test_eof;
	_test_eof321: cs = 321; goto _test_eof;
	_test_eof322: cs = 322; goto _test_eof;
	_test_eof323: cs = 323; goto _test_eof;
	_test_eof324: cs = 324; goto _test_eof;
	_test_eof325: cs = 325; goto _test_eof;
	_test_eof1144: cs = 1144; goto _test_eof;
	_test_eof326: cs = 326; goto _test_eof;
	_test_eof327: cs = 327; goto _test_eof;
	_test_eof328: cs = 328; goto _test_eof;
	_test_eof329: cs = 329; goto _test_eof;
	_test_eof330: cs = 330; goto _test_eof;
	_test_eof331: cs = 331; goto _test_eof;
	_test_eof332: cs = 332; goto _test_eof;
	_test_eof333: cs = 333; goto _test_eof;
	_test_eof334: cs = 334; goto _test_eof;
	_test_eof1145: cs = 1145; goto _test_eof;
	_test_eof1146: cs = 1146; goto _test_eof;
	_test_eof1147: cs = 1147; goto _test_eof;
	_test_eof335: cs = 335; goto _test_eof;
	_test_eof336: cs = 336; goto _test_eof;
	_test_eof337: cs = 337; goto _test_eof;
	_test_eof338: cs = 338; goto _test_eof;
	_test_eof339: cs = 339; goto _test_eof;
	_test_eof340: cs = 340; goto _test_eof;
	_test_eof341: cs = 341; goto _test_eof;
	_test_eof342: cs = 342; goto _test_eof;
	_test_eof1148: cs = 1148; goto _test_eof;
	_test_eof1149: cs = 1149; goto _test_eof;
	_test_eof343: cs = 343; goto _test_eof;
	_test_eof344: cs = 344; goto _test_eof;
	_test_eof345: cs = 345; goto _test_eof;
	_test_eof1150: cs = 1150; goto _test_eof;
	_test_eof346: cs = 346; goto _test_eof;
	_test_eof347: cs = 347; goto _test_eof;
	_test_eof348: cs = 348; goto _test_eof;
	_test_eof349: cs = 349; goto _test_eof;
	_test_eof350: cs = 350; goto _test_eof;
	_test_eof351: cs = 351; goto _test_eof;
	_test_eof352: cs = 352; goto _test_eof;
	_test_eof353: cs = 353; goto _test_eof;
	_test_eof354: cs = 354; goto _test_eof;
	_test_eof355: cs = 355; goto _test_eof;
	_test_eof356: cs = 356; goto _test_eof;
	_test_eof357: cs = 357; goto _test_eof;
	_test_eof358: cs = 358; goto _test_eof;
	_test_eof359: cs = 359; goto _test_eof;
	_test_eof360: cs = 360; goto _test_eof;
	_test_eof361: cs = 361; goto _test_eof;
	_test_eof362: cs = 362; goto _test_eof;
	_test_eof363: cs = 363; goto _test_eof;
	_test_eof364: cs = 364; goto _test_eof;
	_test_eof365: cs = 365; goto _test_eof;
	_test_eof366: cs = 366; goto _test_eof;
	_test_eof367: cs = 367; goto _test_eof;
	_test_eof368: cs = 368; goto _test_eof;
	_test_eof369: cs = 369; goto _test_eof;
	_test_eof370: cs = 370; goto _test_eof;
	_test_eof371: cs = 371; goto _test_eof;
	_test_eof372: cs = 372; goto _test_eof;
	_test_eof373: cs = 373; goto _test_eof;
	_test_eof374: cs = 374; goto _test_eof;
	_test_eof375: cs = 375; goto _test_eof;
	_test_eof376: cs = 376; goto _test_eof;
	_test_eof377: cs = 377; goto _test_eof;
	_test_eof378: cs = 378; goto _test_eof;
	_test_eof379: cs = 379; goto _test_eof;
	_test_eof380: cs = 380; goto _test_eof;
	_test_eof381: cs = 381; goto _test_eof;
	_test_eof382: cs = 382; goto _test_eof;
	_test_eof383: cs = 383; goto _test_eof;
	_test_eof384: cs = 384; goto _test_eof;
	_test_eof385: cs = 385; goto _test_eof;
	_test_eof386: cs = 386; goto _test_eof;
	_test_eof387: cs = 387; goto _test_eof;
	_test_eof388: cs = 388; goto _test_eof;
	_test_eof389: cs = 389; goto _test_eof;
	_test_eof390: cs = 390; goto _test_eof;
	_test_eof391: cs = 391; goto _test_eof;
	_test_eof392: cs = 392; goto _test_eof;
	_test_eof393: cs = 393; goto _test_eof;
	_test_eof394: cs = 394; goto _test_eof;
	_test_eof395: cs = 395; goto _test_eof;
	_test_eof396: cs = 396; goto _test_eof;
	_test_eof397: cs = 397; goto _test_eof;
	_test_eof398: cs = 398; goto _test_eof;
	_test_eof399: cs = 399; goto _test_eof;
	_test_eof400: cs = 400; goto _test_eof;
	_test_eof401: cs = 401; goto _test_eof;
	_test_eof402: cs = 402; goto _test_eof;
	_test_eof403: cs = 403; goto _test_eof;
	_test_eof404: cs = 404; goto _test_eof;
	_test_eof405: cs = 405; goto _test_eof;
	_test_eof406: cs = 406; goto _test_eof;
	_test_eof407: cs = 407; goto _test_eof;
	_test_eof408: cs = 408; goto _test_eof;
	_test_eof409: cs = 409; goto _test_eof;
	_test_eof410: cs = 410; goto _test_eof;
	_test_eof411: cs = 411; goto _test_eof;
	_test_eof412: cs = 412; goto _test_eof;
	_test_eof413: cs = 413; goto _test_eof;
	_test_eof414: cs = 414; goto _test_eof;
	_test_eof415: cs = 415; goto _test_eof;
	_test_eof416: cs = 416; goto _test_eof;
	_test_eof417: cs = 417; goto _test_eof;
	_test_eof418: cs = 418; goto _test_eof;
	_test_eof419: cs = 419; goto _test_eof;
	_test_eof420: cs = 420; goto _test_eof;
	_test_eof421: cs = 421; goto _test_eof;
	_test_eof422: cs = 422; goto _test_eof;
	_test_eof423: cs = 423; goto _test_eof;
	_test_eof424: cs = 424; goto _test_eof;
	_test_eof425: cs = 425; goto _test_eof;
	_test_eof426: cs = 426; goto _test_eof;
	_test_eof427: cs = 427; goto _test_eof;
	_test_eof428: cs = 428; goto _test_eof;
	_test_eof429: cs = 429; goto _test_eof;
	_test_eof430: cs = 430; goto _test_eof;
	_test_eof431: cs = 431; goto _test_eof;
	_test_eof432: cs = 432; goto _test_eof;
	_test_eof433: cs = 433; goto _test_eof;
	_test_eof434: cs = 434; goto _test_eof;
	_test_eof435: cs = 435; goto _test_eof;
	_test_eof436: cs = 436; goto _test_eof;
	_test_eof437: cs = 437; goto _test_eof;
	_test_eof438: cs = 438; goto _test_eof;
	_test_eof439: cs = 439; goto _test_eof;
	_test_eof440: cs = 440; goto _test_eof;
	_test_eof441: cs = 441; goto _test_eof;
	_test_eof442: cs = 442; goto _test_eof;
	_test_eof443: cs = 443; goto _test_eof;
	_test_eof444: cs = 444; goto _test_eof;
	_test_eof445: cs = 445; goto _test_eof;
	_test_eof446: cs = 446; goto _test_eof;
	_test_eof447: cs = 447; goto _test_eof;
	_test_eof448: cs = 448; goto _test_eof;
	_test_eof449: cs = 449; goto _test_eof;
	_test_eof450: cs = 450; goto _test_eof;
	_test_eof451: cs = 451; goto _test_eof;
	_test_eof452: cs = 452; goto _test_eof;
	_test_eof453: cs = 453; goto _test_eof;
	_test_eof454: cs = 454; goto _test_eof;
	_test_eof455: cs = 455; goto _test_eof;
	_test_eof456: cs = 456; goto _test_eof;
	_test_eof457: cs = 457; goto _test_eof;
	_test_eof458: cs = 458; goto _test_eof;
	_test_eof459: cs = 459; goto _test_eof;
	_test_eof460: cs = 460; goto _test_eof;
	_test_eof461: cs = 461; goto _test_eof;
	_test_eof462: cs = 462; goto _test_eof;
	_test_eof463: cs = 463; goto _test_eof;
	_test_eof464: cs = 464; goto _test_eof;
	_test_eof465: cs = 465; goto _test_eof;
	_test_eof466: cs = 466; goto _test_eof;
	_test_eof467: cs = 467; goto _test_eof;
	_test_eof468: cs = 468; goto _test_eof;
	_test_eof469: cs = 469; goto _test_eof;
	_test_eof470: cs = 470; goto _test_eof;
	_test_eof471: cs = 471; goto _test_eof;
	_test_eof472: cs = 472; goto _test_eof;
	_test_eof473: cs = 473; goto _test_eof;
	_test_eof474: cs = 474; goto _test_eof;
	_test_eof1151: cs = 1151; goto _test_eof;
	_test_eof1152: cs = 1152; goto _test_eof;
	_test_eof1153: cs = 1153; goto _test_eof;
	_test_eof475: cs = 475; goto _test_eof;
	_test_eof476: cs = 476; goto _test_eof;
	_test_eof477: cs = 477; goto _test_eof;
	_test_eof478: cs = 478; goto _test_eof;
	_test_eof479: cs = 479; goto _test_eof;
	_test_eof1154: cs = 1154; goto _test_eof;
	_test_eof480: cs = 480; goto _test_eof;
	_test_eof481: cs = 481; goto _test_eof;
	_test_eof482: cs = 482; goto _test_eof;
	_test_eof483: cs = 483; goto _test_eof;
	_test_eof1155: cs = 1155; goto _test_eof;
	_test_eof1156: cs = 1156; goto _test_eof;
	_test_eof1157: cs = 1157; goto _test_eof;
	_test_eof484: cs = 484; goto _test_eof;
	_test_eof485: cs = 485; goto _test_eof;
	_test_eof1158: cs = 1158; goto _test_eof;
	_test_eof486: cs = 486; goto _test_eof;
	_test_eof487: cs = 487; goto _test_eof;
	_test_eof488: cs = 488; goto _test_eof;
	_test_eof1159: cs = 1159; goto _test_eof;
	_test_eof489: cs = 489; goto _test_eof;
	_test_eof490: cs = 490; goto _test_eof;
	_test_eof491: cs = 491; goto _test_eof;
	_test_eof492: cs = 492; goto _test_eof;
	_test_eof493: cs = 493; goto _test_eof;
	_test_eof494: cs = 494; goto _test_eof;
	_test_eof495: cs = 495; goto _test_eof;
	_test_eof496: cs = 496; goto _test_eof;
	_test_eof497: cs = 497; goto _test_eof;
	_test_eof498: cs = 498; goto _test_eof;
	_test_eof499: cs = 499; goto _test_eof;
	_test_eof500: cs = 500; goto _test_eof;
	_test_eof501: cs = 501; goto _test_eof;
	_test_eof502: cs = 502; goto _test_eof;
	_test_eof503: cs = 503; goto _test_eof;
	_test_eof504: cs = 504; goto _test_eof;
	_test_eof505: cs = 505; goto _test_eof;
	_test_eof506: cs = 506; goto _test_eof;
	_test_eof507: cs = 507; goto _test_eof;
	_test_eof508: cs = 508; goto _test_eof;
	_test_eof509: cs = 509; goto _test_eof;
	_test_eof510: cs = 510; goto _test_eof;
	_test_eof511: cs = 511; goto _test_eof;
	_test_eof512: cs = 512; goto _test_eof;
	_test_eof513: cs = 513; goto _test_eof;
	_test_eof514: cs = 514; goto _test_eof;
	_test_eof515: cs = 515; goto _test_eof;
	_test_eof516: cs = 516; goto _test_eof;
	_test_eof517: cs = 517; goto _test_eof;
	_test_eof518: cs = 518; goto _test_eof;
	_test_eof519: cs = 519; goto _test_eof;
	_test_eof520: cs = 520; goto _test_eof;
	_test_eof521: cs = 521; goto _test_eof;
	_test_eof522: cs = 522; goto _test_eof;
	_test_eof523: cs = 523; goto _test_eof;
	_test_eof524: cs = 524; goto _test_eof;
	_test_eof525: cs = 525; goto _test_eof;
	_test_eof526: cs = 526; goto _test_eof;
	_test_eof527: cs = 527; goto _test_eof;
	_test_eof528: cs = 528; goto _test_eof;
	_test_eof529: cs = 529; goto _test_eof;
	_test_eof530: cs = 530; goto _test_eof;
	_test_eof531: cs = 531; goto _test_eof;
	_test_eof532: cs = 532; goto _test_eof;
	_test_eof533: cs = 533; goto _test_eof;
	_test_eof534: cs = 534; goto _test_eof;
	_test_eof535: cs = 535; goto _test_eof;
	_test_eof536: cs = 536; goto _test_eof;
	_test_eof537: cs = 537; goto _test_eof;
	_test_eof538: cs = 538; goto _test_eof;
	_test_eof539: cs = 539; goto _test_eof;
	_test_eof540: cs = 540; goto _test_eof;
	_test_eof541: cs = 541; goto _test_eof;
	_test_eof542: cs = 542; goto _test_eof;
	_test_eof543: cs = 543; goto _test_eof;
	_test_eof544: cs = 544; goto _test_eof;
	_test_eof545: cs = 545; goto _test_eof;
	_test_eof546: cs = 546; goto _test_eof;
	_test_eof547: cs = 547; goto _test_eof;
	_test_eof548: cs = 548; goto _test_eof;
	_test_eof549: cs = 549; goto _test_eof;
	_test_eof550: cs = 550; goto _test_eof;
	_test_eof551: cs = 551; goto _test_eof;
	_test_eof552: cs = 552; goto _test_eof;
	_test_eof553: cs = 553; goto _test_eof;
	_test_eof554: cs = 554; goto _test_eof;
	_test_eof555: cs = 555; goto _test_eof;
	_test_eof556: cs = 556; goto _test_eof;
	_test_eof557: cs = 557; goto _test_eof;
	_test_eof558: cs = 558; goto _test_eof;
	_test_eof559: cs = 559; goto _test_eof;
	_test_eof560: cs = 560; goto _test_eof;
	_test_eof561: cs = 561; goto _test_eof;
	_test_eof562: cs = 562; goto _test_eof;
	_test_eof563: cs = 563; goto _test_eof;
	_test_eof564: cs = 564; goto _test_eof;
	_test_eof565: cs = 565; goto _test_eof;
	_test_eof566: cs = 566; goto _test_eof;
	_test_eof567: cs = 567; goto _test_eof;
	_test_eof568: cs = 568; goto _test_eof;
	_test_eof569: cs = 569; goto _test_eof;
	_test_eof570: cs = 570; goto _test_eof;
	_test_eof571: cs = 571; goto _test_eof;
	_test_eof572: cs = 572; goto _test_eof;
	_test_eof573: cs = 573; goto _test_eof;
	_test_eof574: cs = 574; goto _test_eof;
	_test_eof575: cs = 575; goto _test_eof;
	_test_eof576: cs = 576; goto _test_eof;
	_test_eof577: cs = 577; goto _test_eof;
	_test_eof578: cs = 578; goto _test_eof;
	_test_eof579: cs = 579; goto _test_eof;
	_test_eof580: cs = 580; goto _test_eof;
	_test_eof581: cs = 581; goto _test_eof;
	_test_eof582: cs = 582; goto _test_eof;
	_test_eof583: cs = 583; goto _test_eof;
	_test_eof584: cs = 584; goto _test_eof;
	_test_eof585: cs = 585; goto _test_eof;
	_test_eof586: cs = 586; goto _test_eof;
	_test_eof587: cs = 587; goto _test_eof;
	_test_eof588: cs = 588; goto _test_eof;
	_test_eof589: cs = 589; goto _test_eof;
	_test_eof590: cs = 590; goto _test_eof;
	_test_eof591: cs = 591; goto _test_eof;
	_test_eof592: cs = 592; goto _test_eof;
	_test_eof1160: cs = 1160; goto _test_eof;
	_test_eof593: cs = 593; goto _test_eof;
	_test_eof594: cs = 594; goto _test_eof;
	_test_eof595: cs = 595; goto _test_eof;
	_test_eof596: cs = 596; goto _test_eof;
	_test_eof597: cs = 597; goto _test_eof;
	_test_eof598: cs = 598; goto _test_eof;
	_test_eof599: cs = 599; goto _test_eof;
	_test_eof600: cs = 600; goto _test_eof;
	_test_eof601: cs = 601; goto _test_eof;
	_test_eof602: cs = 602; goto _test_eof;
	_test_eof603: cs = 603; goto _test_eof;
	_test_eof604: cs = 604; goto _test_eof;
	_test_eof605: cs = 605; goto _test_eof;
	_test_eof606: cs = 606; goto _test_eof;
	_test_eof607: cs = 607; goto _test_eof;
	_test_eof608: cs = 608; goto _test_eof;
	_test_eof609: cs = 609; goto _test_eof;
	_test_eof610: cs = 610; goto _test_eof;
	_test_eof611: cs = 611; goto _test_eof;
	_test_eof612: cs = 612; goto _test_eof;
	_test_eof613: cs = 613; goto _test_eof;
	_test_eof614: cs = 614; goto _test_eof;
	_test_eof615: cs = 615; goto _test_eof;
	_test_eof616: cs = 616; goto _test_eof;
	_test_eof617: cs = 617; goto _test_eof;
	_test_eof618: cs = 618; goto _test_eof;
	_test_eof619: cs = 619; goto _test_eof;
	_test_eof620: cs = 620; goto _test_eof;
	_test_eof621: cs = 621; goto _test_eof;
	_test_eof622: cs = 622; goto _test_eof;
	_test_eof623: cs = 623; goto _test_eof;
	_test_eof624: cs = 624; goto _test_eof;
	_test_eof625: cs = 625; goto _test_eof;
	_test_eof626: cs = 626; goto _test_eof;
	_test_eof627: cs = 627; goto _test_eof;
	_test_eof628: cs = 628; goto _test_eof;
	_test_eof629: cs = 629; goto _test_eof;
	_test_eof630: cs = 630; goto _test_eof;
	_test_eof631: cs = 631; goto _test_eof;
	_test_eof632: cs = 632; goto _test_eof;
	_test_eof633: cs = 633; goto _test_eof;
	_test_eof1161: cs = 1161; goto _test_eof;
	_test_eof634: cs = 634; goto _test_eof;
	_test_eof635: cs = 635; goto _test_eof;
	_test_eof1162: cs = 1162; goto _test_eof;
	_test_eof636: cs = 636; goto _test_eof;
	_test_eof637: cs = 637; goto _test_eof;
	_test_eof638: cs = 638; goto _test_eof;
	_test_eof639: cs = 639; goto _test_eof;
	_test_eof640: cs = 640; goto _test_eof;
	_test_eof641: cs = 641; goto _test_eof;
	_test_eof642: cs = 642; goto _test_eof;
	_test_eof643: cs = 643; goto _test_eof;
	_test_eof644: cs = 644; goto _test_eof;
	_test_eof645: cs = 645; goto _test_eof;
	_test_eof646: cs = 646; goto _test_eof;
	_test_eof647: cs = 647; goto _test_eof;
	_test_eof648: cs = 648; goto _test_eof;
	_test_eof649: cs = 649; goto _test_eof;
	_test_eof1163: cs = 1163; goto _test_eof;
	_test_eof650: cs = 650; goto _test_eof;
	_test_eof651: cs = 651; goto _test_eof;
	_test_eof652: cs = 652; goto _test_eof;
	_test_eof653: cs = 653; goto _test_eof;
	_test_eof654: cs = 654; goto _test_eof;
	_test_eof655: cs = 655; goto _test_eof;
	_test_eof656: cs = 656; goto _test_eof;
	_test_eof657: cs = 657; goto _test_eof;
	_test_eof658: cs = 658; goto _test_eof;
	_test_eof659: cs = 659; goto _test_eof;
	_test_eof660: cs = 660; goto _test_eof;
	_test_eof661: cs = 661; goto _test_eof;
	_test_eof662: cs = 662; goto _test_eof;
	_test_eof663: cs = 663; goto _test_eof;
	_test_eof664: cs = 664; goto _test_eof;
	_test_eof665: cs = 665; goto _test_eof;
	_test_eof666: cs = 666; goto _test_eof;
	_test_eof667: cs = 667; goto _test_eof;
	_test_eof668: cs = 668; goto _test_eof;
	_test_eof669: cs = 669; goto _test_eof;
	_test_eof670: cs = 670; goto _test_eof;
	_test_eof671: cs = 671; goto _test_eof;
	_test_eof1164: cs = 1164; goto _test_eof;
	_test_eof672: cs = 672; goto _test_eof;
	_test_eof673: cs = 673; goto _test_eof;
	_test_eof674: cs = 674; goto _test_eof;
	_test_eof675: cs = 675; goto _test_eof;
	_test_eof676: cs = 676; goto _test_eof;
	_test_eof1165: cs = 1165; goto _test_eof;
	_test_eof677: cs = 677; goto _test_eof;
	_test_eof678: cs = 678; goto _test_eof;
	_test_eof679: cs = 679; goto _test_eof;
	_test_eof680: cs = 680; goto _test_eof;
	_test_eof681: cs = 681; goto _test_eof;
	_test_eof1166: cs = 1166; goto _test_eof;
	_test_eof682: cs = 682; goto _test_eof;
	_test_eof683: cs = 683; goto _test_eof;
	_test_eof684: cs = 684; goto _test_eof;
	_test_eof685: cs = 685; goto _test_eof;
	_test_eof686: cs = 686; goto _test_eof;
	_test_eof1167: cs = 1167; goto _test_eof;
	_test_eof1168: cs = 1168; goto _test_eof;
	_test_eof1169: cs = 1169; goto _test_eof;
	_test_eof687: cs = 687; goto _test_eof;
	_test_eof688: cs = 688; goto _test_eof;
	_test_eof1170: cs = 1170; goto _test_eof;
	_test_eof689: cs = 689; goto _test_eof;
	_test_eof690: cs = 690; goto _test_eof;
	_test_eof691: cs = 691; goto _test_eof;
	_test_eof692: cs = 692; goto _test_eof;
	_test_eof693: cs = 693; goto _test_eof;
	_test_eof694: cs = 694; goto _test_eof;
	_test_eof695: cs = 695; goto _test_eof;
	_test_eof696: cs = 696; goto _test_eof;
	_test_eof697: cs = 697; goto _test_eof;
	_test_eof698: cs = 698; goto _test_eof;
	_test_eof699: cs = 699; goto _test_eof;
	_test_eof700: cs = 700; goto _test_eof;
	_test_eof701: cs = 701; goto _test_eof;
	_test_eof702: cs = 702; goto _test_eof;
	_test_eof703: cs = 703; goto _test_eof;
	_test_eof704: cs = 704; goto _test_eof;
	_test_eof705: cs = 705; goto _test_eof;
	_test_eof706: cs = 706; goto _test_eof;
	_test_eof707: cs = 707; goto _test_eof;
	_test_eof708: cs = 708; goto _test_eof;
	_test_eof709: cs = 709; goto _test_eof;
	_test_eof710: cs = 710; goto _test_eof;
	_test_eof711: cs = 711; goto _test_eof;
	_test_eof712: cs = 712; goto _test_eof;
	_test_eof713: cs = 713; goto _test_eof;
	_test_eof714: cs = 714; goto _test_eof;
	_test_eof715: cs = 715; goto _test_eof;
	_test_eof1171: cs = 1171; goto _test_eof;
	_test_eof1172: cs = 1172; goto _test_eof;
	_test_eof1173: cs = 1173; goto _test_eof;
	_test_eof716: cs = 716; goto _test_eof;
	_test_eof717: cs = 717; goto _test_eof;
	_test_eof718: cs = 718; goto _test_eof;
	_test_eof1174: cs = 1174; goto _test_eof;
	_test_eof1175: cs = 1175; goto _test_eof;
	_test_eof719: cs = 719; goto _test_eof;
	_test_eof720: cs = 720; goto _test_eof;
	_test_eof721: cs = 721; goto _test_eof;
	_test_eof722: cs = 722; goto _test_eof;
	_test_eof1176: cs = 1176; goto _test_eof;
	_test_eof1177: cs = 1177; goto _test_eof;
	_test_eof723: cs = 723; goto _test_eof;
	_test_eof724: cs = 724; goto _test_eof;
	_test_eof725: cs = 725; goto _test_eof;
	_test_eof726: cs = 726; goto _test_eof;
	_test_eof1178: cs = 1178; goto _test_eof;
	_test_eof1179: cs = 1179; goto _test_eof;
	_test_eof727: cs = 727; goto _test_eof;
	_test_eof728: cs = 728; goto _test_eof;
	_test_eof729: cs = 729; goto _test_eof;
	_test_eof730: cs = 730; goto _test_eof;
	_test_eof731: cs = 731; goto _test_eof;
	_test_eof732: cs = 732; goto _test_eof;
	_test_eof733: cs = 733; goto _test_eof;
	_test_eof734: cs = 734; goto _test_eof;
	_test_eof735: cs = 735; goto _test_eof;
	_test_eof736: cs = 736; goto _test_eof;
	_test_eof737: cs = 737; goto _test_eof;
	_test_eof738: cs = 738; goto _test_eof;
	_test_eof739: cs = 739; goto _test_eof;
	_test_eof740: cs = 740; goto _test_eof;
	_test_eof741: cs = 741; goto _test_eof;
	_test_eof742: cs = 742; goto _test_eof;
	_test_eof743: cs = 743; goto _test_eof;
	_test_eof744: cs = 744; goto _test_eof;
	_test_eof745: cs = 745; goto _test_eof;
	_test_eof746: cs = 746; goto _test_eof;
	_test_eof747: cs = 747; goto _test_eof;
	_test_eof748: cs = 748; goto _test_eof;
	_test_eof749: cs = 749; goto _test_eof;
	_test_eof750: cs = 750; goto _test_eof;
	_test_eof751: cs = 751; goto _test_eof;
	_test_eof1180: cs = 1180; goto _test_eof;
	_test_eof752: cs = 752; goto _test_eof;
	_test_eof753: cs = 753; goto _test_eof;
	_test_eof754: cs = 754; goto _test_eof;
	_test_eof755: cs = 755; goto _test_eof;
	_test_eof756: cs = 756; goto _test_eof;
	_test_eof757: cs = 757; goto _test_eof;
	_test_eof758: cs = 758; goto _test_eof;
	_test_eof759: cs = 759; goto _test_eof;
	_test_eof760: cs = 760; goto _test_eof;
	_test_eof761: cs = 761; goto _test_eof;
	_test_eof762: cs = 762; goto _test_eof;
	_test_eof763: cs = 763; goto _test_eof;
	_test_eof764: cs = 764; goto _test_eof;
	_test_eof765: cs = 765; goto _test_eof;
	_test_eof766: cs = 766; goto _test_eof;
	_test_eof1181: cs = 1181; goto _test_eof;
	_test_eof767: cs = 767; goto _test_eof;
	_test_eof768: cs = 768; goto _test_eof;
	_test_eof769: cs = 769; goto _test_eof;
	_test_eof770: cs = 770; goto _test_eof;
	_test_eof771: cs = 771; goto _test_eof;
	_test_eof772: cs = 772; goto _test_eof;
	_test_eof773: cs = 773; goto _test_eof;
	_test_eof774: cs = 774; goto _test_eof;
	_test_eof775: cs = 775; goto _test_eof;
	_test_eof776: cs = 776; goto _test_eof;
	_test_eof777: cs = 777; goto _test_eof;
	_test_eof778: cs = 778; goto _test_eof;
	_test_eof779: cs = 779; goto _test_eof;
	_test_eof1182: cs = 1182; goto _test_eof;
	_test_eof780: cs = 780; goto _test_eof;
	_test_eof781: cs = 781; goto _test_eof;
	_test_eof782: cs = 782; goto _test_eof;
	_test_eof783: cs = 783; goto _test_eof;
	_test_eof784: cs = 784; goto _test_eof;
	_test_eof785: cs = 785; goto _test_eof;
	_test_eof786: cs = 786; goto _test_eof;
	_test_eof787: cs = 787; goto _test_eof;
	_test_eof788: cs = 788; goto _test_eof;
	_test_eof789: cs = 789; goto _test_eof;
	_test_eof790: cs = 790; goto _test_eof;
	_test_eof1183: cs = 1183; goto _test_eof;
	_test_eof1184: cs = 1184; goto _test_eof;
	_test_eof791: cs = 791; goto _test_eof;
	_test_eof792: cs = 792; goto _test_eof;
	_test_eof793: cs = 793; goto _test_eof;
	_test_eof1185: cs = 1185; goto _test_eof;
	_test_eof794: cs = 794; goto _test_eof;
	_test_eof795: cs = 795; goto _test_eof;
	_test_eof796: cs = 796; goto _test_eof;
	_test_eof797: cs = 797; goto _test_eof;
	_test_eof798: cs = 798; goto _test_eof;
	_test_eof799: cs = 799; goto _test_eof;
	_test_eof800: cs = 800; goto _test_eof;
	_test_eof801: cs = 801; goto _test_eof;
	_test_eof802: cs = 802; goto _test_eof;
	_test_eof803: cs = 803; goto _test_eof;
	_test_eof1186: cs = 1186; goto _test_eof;
	_test_eof1187: cs = 1187; goto _test_eof;
	_test_eof1188: cs = 1188; goto _test_eof;
	_test_eof804: cs = 804; goto _test_eof;
	_test_eof805: cs = 805; goto _test_eof;
	_test_eof806: cs = 806; goto _test_eof;
	_test_eof807: cs = 807; goto _test_eof;
	_test_eof808: cs = 808; goto _test_eof;
	_test_eof809: cs = 809; goto _test_eof;
	_test_eof810: cs = 810; goto _test_eof;
	_test_eof811: cs = 811; goto _test_eof;
	_test_eof812: cs = 812; goto _test_eof;
	_test_eof813: cs = 813; goto _test_eof;
	_test_eof814: cs = 814; goto _test_eof;
	_test_eof1189: cs = 1189; goto _test_eof;
	_test_eof1190: cs = 1190; goto _test_eof;
	_test_eof1191: cs = 1191; goto _test_eof;
	_test_eof815: cs = 815; goto _test_eof;
	_test_eof816: cs = 816; goto _test_eof;
	_test_eof817: cs = 817; goto _test_eof;
	_test_eof818: cs = 818; goto _test_eof;
	_test_eof819: cs = 819; goto _test_eof;
	_test_eof820: cs = 820; goto _test_eof;
	_test_eof821: cs = 821; goto _test_eof;
	_test_eof822: cs = 822; goto _test_eof;
	_test_eof823: cs = 823; goto _test_eof;
	_test_eof824: cs = 824; goto _test_eof;
	_test_eof825: cs = 825; goto _test_eof;
	_test_eof826: cs = 826; goto _test_eof;
	_test_eof1192: cs = 1192; goto _test_eof;
	_test_eof827: cs = 827; goto _test_eof;
	_test_eof828: cs = 828; goto _test_eof;
	_test_eof829: cs = 829; goto _test_eof;
	_test_eof1193: cs = 1193; goto _test_eof;
	_test_eof1194: cs = 1194; goto _test_eof;
	_test_eof830: cs = 830; goto _test_eof;
	_test_eof1195: cs = 1195; goto _test_eof;
	_test_eof1196: cs = 1196; goto _test_eof;
	_test_eof831: cs = 831; goto _test_eof;
	_test_eof1197: cs = 1197; goto _test_eof;
	_test_eof1198: cs = 1198; goto _test_eof;
	_test_eof832: cs = 832; goto _test_eof;
	_test_eof833: cs = 833; goto _test_eof;
	_test_eof834: cs = 834; goto _test_eof;
	_test_eof835: cs = 835; goto _test_eof;
	_test_eof836: cs = 836; goto _test_eof;
	_test_eof837: cs = 837; goto _test_eof;
	_test_eof838: cs = 838; goto _test_eof;
	_test_eof839: cs = 839; goto _test_eof;
	_test_eof840: cs = 840; goto _test_eof;
	_test_eof841: cs = 841; goto _test_eof;
	_test_eof842: cs = 842; goto _test_eof;
	_test_eof843: cs = 843; goto _test_eof;
	_test_eof844: cs = 844; goto _test_eof;
	_test_eof845: cs = 845; goto _test_eof;
	_test_eof846: cs = 846; goto _test_eof;
	_test_eof847: cs = 847; goto _test_eof;
	_test_eof848: cs = 848; goto _test_eof;
	_test_eof849: cs = 849; goto _test_eof;
	_test_eof850: cs = 850; goto _test_eof;
	_test_eof851: cs = 851; goto _test_eof;
	_test_eof852: cs = 852; goto _test_eof;
	_test_eof853: cs = 853; goto _test_eof;
	_test_eof854: cs = 854; goto _test_eof;
	_test_eof855: cs = 855; goto _test_eof;
	_test_eof856: cs = 856; goto _test_eof;
	_test_eof857: cs = 857; goto _test_eof;
	_test_eof858: cs = 858; goto _test_eof;
	_test_eof859: cs = 859; goto _test_eof;
	_test_eof860: cs = 860; goto _test_eof;
	_test_eof861: cs = 861; goto _test_eof;
	_test_eof862: cs = 862; goto _test_eof;
	_test_eof863: cs = 863; goto _test_eof;
	_test_eof864: cs = 864; goto _test_eof;
	_test_eof865: cs = 865; goto _test_eof;
	_test_eof866: cs = 866; goto _test_eof;
	_test_eof867: cs = 867; goto _test_eof;
	_test_eof868: cs = 868; goto _test_eof;
	_test_eof869: cs = 869; goto _test_eof;
	_test_eof870: cs = 870; goto _test_eof;
	_test_eof871: cs = 871; goto _test_eof;
	_test_eof872: cs = 872; goto _test_eof;
	_test_eof873: cs = 873; goto _test_eof;
	_test_eof1199: cs = 1199; goto _test_eof;
	_test_eof874: cs = 874; goto _test_eof;
	_test_eof875: cs = 875; goto _test_eof;
	_test_eof876: cs = 876; goto _test_eof;
	_test_eof877: cs = 877; goto _test_eof;
	_test_eof878: cs = 878; goto _test_eof;
	_test_eof879: cs = 879; goto _test_eof;
	_test_eof880: cs = 880; goto _test_eof;
	_test_eof881: cs = 881; goto _test_eof;
	_test_eof882: cs = 882; goto _test_eof;
	_test_eof883: cs = 883; goto _test_eof;
	_test_eof884: cs = 884; goto _test_eof;
	_test_eof885: cs = 885; goto _test_eof;
	_test_eof886: cs = 886; goto _test_eof;
	_test_eof887: cs = 887; goto _test_eof;
	_test_eof888: cs = 888; goto _test_eof;
	_test_eof889: cs = 889; goto _test_eof;
	_test_eof890: cs = 890; goto _test_eof;
	_test_eof891: cs = 891; goto _test_eof;
	_test_eof892: cs = 892; goto _test_eof;
	_test_eof893: cs = 893; goto _test_eof;
	_test_eof894: cs = 894; goto _test_eof;
	_test_eof895: cs = 895; goto _test_eof;
	_test_eof896: cs = 896; goto _test_eof;
	_test_eof897: cs = 897; goto _test_eof;
	_test_eof898: cs = 898; goto _test_eof;
	_test_eof899: cs = 899; goto _test_eof;
	_test_eof900: cs = 900; goto _test_eof;
	_test_eof901: cs = 901; goto _test_eof;
	_test_eof902: cs = 902; goto _test_eof;
	_test_eof903: cs = 903; goto _test_eof;
	_test_eof904: cs = 904; goto _test_eof;
	_test_eof905: cs = 905; goto _test_eof;
	_test_eof906: cs = 906; goto _test_eof;
	_test_eof907: cs = 907; goto _test_eof;
	_test_eof908: cs = 908; goto _test_eof;
	_test_eof909: cs = 909; goto _test_eof;
	_test_eof910: cs = 910; goto _test_eof;
	_test_eof911: cs = 911; goto _test_eof;
	_test_eof912: cs = 912; goto _test_eof;
	_test_eof913: cs = 913; goto _test_eof;
	_test_eof914: cs = 914; goto _test_eof;
	_test_eof915: cs = 915; goto _test_eof;
	_test_eof916: cs = 916; goto _test_eof;
	_test_eof917: cs = 917; goto _test_eof;
	_test_eof918: cs = 918; goto _test_eof;
	_test_eof919: cs = 919; goto _test_eof;
	_test_eof920: cs = 920; goto _test_eof;
	_test_eof921: cs = 921; goto _test_eof;
	_test_eof922: cs = 922; goto _test_eof;
	_test_eof923: cs = 923; goto _test_eof;
	_test_eof924: cs = 924; goto _test_eof;
	_test_eof925: cs = 925; goto _test_eof;
	_test_eof926: cs = 926; goto _test_eof;
	_test_eof927: cs = 927; goto _test_eof;
	_test_eof928: cs = 928; goto _test_eof;
	_test_eof929: cs = 929; goto _test_eof;
	_test_eof930: cs = 930; goto _test_eof;
	_test_eof931: cs = 931; goto _test_eof;
	_test_eof932: cs = 932; goto _test_eof;
	_test_eof933: cs = 933; goto _test_eof;
	_test_eof934: cs = 934; goto _test_eof;
	_test_eof935: cs = 935; goto _test_eof;
	_test_eof936: cs = 936; goto _test_eof;
	_test_eof937: cs = 937; goto _test_eof;
	_test_eof938: cs = 938; goto _test_eof;
	_test_eof939: cs = 939; goto _test_eof;
	_test_eof940: cs = 940; goto _test_eof;
	_test_eof941: cs = 941; goto _test_eof;
	_test_eof942: cs = 942; goto _test_eof;
	_test_eof943: cs = 943; goto _test_eof;
	_test_eof944: cs = 944; goto _test_eof;
	_test_eof945: cs = 945; goto _test_eof;
	_test_eof946: cs = 946; goto _test_eof;
	_test_eof947: cs = 947; goto _test_eof;
	_test_eof948: cs = 948; goto _test_eof;
	_test_eof949: cs = 949; goto _test_eof;
	_test_eof950: cs = 950; goto _test_eof;
	_test_eof951: cs = 951; goto _test_eof;
	_test_eof952: cs = 952; goto _test_eof;
	_test_eof953: cs = 953; goto _test_eof;
	_test_eof954: cs = 954; goto _test_eof;
	_test_eof955: cs = 955; goto _test_eof;
	_test_eof956: cs = 956; goto _test_eof;
	_test_eof957: cs = 957; goto _test_eof;
	_test_eof958: cs = 958; goto _test_eof;
	_test_eof959: cs = 959; goto _test_eof;
	_test_eof960: cs = 960; goto _test_eof;
	_test_eof961: cs = 961; goto _test_eof;
	_test_eof962: cs = 962; goto _test_eof;
	_test_eof963: cs = 963; goto _test_eof;
	_test_eof964: cs = 964; goto _test_eof;
	_test_eof965: cs = 965; goto _test_eof;
	_test_eof966: cs = 966; goto _test_eof;
	_test_eof967: cs = 967; goto _test_eof;
	_test_eof968: cs = 968; goto _test_eof;
	_test_eof969: cs = 969; goto _test_eof;
	_test_eof970: cs = 970; goto _test_eof;
	_test_eof971: cs = 971; goto _test_eof;
	_test_eof972: cs = 972; goto _test_eof;
	_test_eof973: cs = 973; goto _test_eof;
	_test_eof974: cs = 974; goto _test_eof;
	_test_eof975: cs = 975; goto _test_eof;
	_test_eof976: cs = 976; goto _test_eof;
	_test_eof977: cs = 977; goto _test_eof;
	_test_eof978: cs = 978; goto _test_eof;
	_test_eof979: cs = 979; goto _test_eof;
	_test_eof980: cs = 980; goto _test_eof;
	_test_eof981: cs = 981; goto _test_eof;
	_test_eof982: cs = 982; goto _test_eof;
	_test_eof983: cs = 983; goto _test_eof;
	_test_eof984: cs = 984; goto _test_eof;
	_test_eof985: cs = 985; goto _test_eof;
	_test_eof986: cs = 986; goto _test_eof;
	_test_eof987: cs = 987; goto _test_eof;
	_test_eof988: cs = 988; goto _test_eof;
	_test_eof989: cs = 989; goto _test_eof;
	_test_eof990: cs = 990; goto _test_eof;
	_test_eof991: cs = 991; goto _test_eof;
	_test_eof992: cs = 992; goto _test_eof;
	_test_eof993: cs = 993; goto _test_eof;
	_test_eof994: cs = 994; goto _test_eof;
	_test_eof995: cs = 995; goto _test_eof;
	_test_eof996: cs = 996; goto _test_eof;
	_test_eof997: cs = 997; goto _test_eof;
	_test_eof998: cs = 998; goto _test_eof;
	_test_eof999: cs = 999; goto _test_eof;
	_test_eof1000: cs = 1000; goto _test_eof;
	_test_eof1001: cs = 1001; goto _test_eof;
	_test_eof1002: cs = 1002; goto _test_eof;
	_test_eof1003: cs = 1003; goto _test_eof;
	_test_eof1004: cs = 1004; goto _test_eof;
	_test_eof1005: cs = 1005; goto _test_eof;
	_test_eof1006: cs = 1006; goto _test_eof;
	_test_eof1007: cs = 1007; goto _test_eof;
	_test_eof1008: cs = 1008; goto _test_eof;
	_test_eof1009: cs = 1009; goto _test_eof;
	_test_eof1010: cs = 1010; goto _test_eof;
	_test_eof1011: cs = 1011; goto _test_eof;
	_test_eof1012: cs = 1012; goto _test_eof;
	_test_eof1200: cs = 1200; goto _test_eof;
	_test_eof1013: cs = 1013; goto _test_eof;
	_test_eof1014: cs = 1014; goto _test_eof;
	_test_eof1015: cs = 1015; goto _test_eof;
	_test_eof1016: cs = 1016; goto _test_eof;
	_test_eof1017: cs = 1017; goto _test_eof;
	_test_eof1018: cs = 1018; goto _test_eof;
	_test_eof1019: cs = 1019; goto _test_eof;
	_test_eof1020: cs = 1020; goto _test_eof;
	_test_eof1201: cs = 1201; goto _test_eof;
	_test_eof1021: cs = 1021; goto _test_eof;
	_test_eof1022: cs = 1022; goto _test_eof;
	_test_eof1023: cs = 1023; goto _test_eof;
	_test_eof1024: cs = 1024; goto _test_eof;
	_test_eof1025: cs = 1025; goto _test_eof;
	_test_eof1202: cs = 1202; goto _test_eof;
	_test_eof1026: cs = 1026; goto _test_eof;
	_test_eof1027: cs = 1027; goto _test_eof;
	_test_eof1028: cs = 1028; goto _test_eof;
	_test_eof1029: cs = 1029; goto _test_eof;
	_test_eof1030: cs = 1030; goto _test_eof;
	_test_eof1031: cs = 1031; goto _test_eof;
	_test_eof1032: cs = 1032; goto _test_eof;
	_test_eof1033: cs = 1033; goto _test_eof;
	_test_eof1034: cs = 1034; goto _test_eof;
	_test_eof1035: cs = 1035; goto _test_eof;
	_test_eof1036: cs = 1036; goto _test_eof;
	_test_eof1037: cs = 1037; goto _test_eof;
	_test_eof1038: cs = 1038; goto _test_eof;
	_test_eof1039: cs = 1039; goto _test_eof;
	_test_eof1040: cs = 1040; goto _test_eof;
	_test_eof1041: cs = 1041; goto _test_eof;
	_test_eof1042: cs = 1042; goto _test_eof;
	_test_eof1043: cs = 1043; goto _test_eof;
	_test_eof1203: cs = 1203; goto _test_eof;
	_test_eof1044: cs = 1044; goto _test_eof;
	_test_eof1045: cs = 1045; goto _test_eof;
	_test_eof1046: cs = 1046; goto _test_eof;
	_test_eof1047: cs = 1047; goto _test_eof;
	_test_eof1048: cs = 1048; goto _test_eof;
	_test_eof1049: cs = 1049; goto _test_eof;
	_test_eof1050: cs = 1050; goto _test_eof;
	_test_eof1051: cs = 1051; goto _test_eof;
	_test_eof1052: cs = 1052; goto _test_eof;
	_test_eof1053: cs = 1053; goto _test_eof;
	_test_eof1054: cs = 1054; goto _test_eof;
	_test_eof1055: cs = 1055; goto _test_eof;
	_test_eof1056: cs = 1056; goto _test_eof;
	_test_eof1057: cs = 1057; goto _test_eof;
	_test_eof1058: cs = 1058; goto _test_eof;
	_test_eof1059: cs = 1059; goto _test_eof;
	_test_eof1060: cs = 1060; goto _test_eof;
	_test_eof1061: cs = 1061; goto _test_eof;
	_test_eof1062: cs = 1062; goto _test_eof;
	_test_eof1204: cs = 1204; goto _test_eof;
	_test_eof1063: cs = 1063; goto _test_eof;
	_test_eof1064: cs = 1064; goto _test_eof;
	_test_eof1065: cs = 1065; goto _test_eof;
	_test_eof1066: cs = 1066; goto _test_eof;
	_test_eof1067: cs = 1067; goto _test_eof;
	_test_eof1068: cs = 1068; goto _test_eof;
	_test_eof1069: cs = 1069; goto _test_eof;
	_test_eof1070: cs = 1070; goto _test_eof;
	_test_eof1071: cs = 1071; goto _test_eof;
	_test_eof1072: cs = 1072; goto _test_eof;
	_test_eof1073: cs = 1073; goto _test_eof;
	_test_eof1074: cs = 1074; goto _test_eof;
	_test_eof1075: cs = 1075; goto _test_eof;
	_test_eof1076: cs = 1076; goto _test_eof;
	_test_eof1077: cs = 1077; goto _test_eof;
	_test_eof1205: cs = 1205; goto _test_eof;
	_test_eof1206: cs = 1206; goto _test_eof;
	_test_eof1207: cs = 1207; goto _test_eof;
	_test_eof1078: cs = 1078; goto _test_eof;
	_test_eof1079: cs = 1079; goto _test_eof;
	_test_eof1080: cs = 1080; goto _test_eof;
	_test_eof1081: cs = 1081; goto _test_eof;
	_test_eof1082: cs = 1082; goto _test_eof;
	_test_eof1083: cs = 1083; goto _test_eof;
	_test_eof1084: cs = 1084; goto _test_eof;
	_test_eof1208: cs = 1208; goto _test_eof;
	_test_eof1085: cs = 1085; goto _test_eof;
	_test_eof1086: cs = 1086; goto _test_eof;
	_test_eof1087: cs = 1087; goto _test_eof;
	_test_eof1088: cs = 1088; goto _test_eof;
	_test_eof1089: cs = 1089; goto _test_eof;
	_test_eof1090: cs = 1090; goto _test_eof;
	_test_eof1091: cs = 1091; goto _test_eof;
	_test_eof1092: cs = 1092; goto _test_eof;
	_test_eof1093: cs = 1093; goto _test_eof;
	_test_eof1094: cs = 1094; goto _test_eof;
	_test_eof1095: cs = 1095; goto _test_eof;
	_test_eof1096: cs = 1096; goto _test_eof;
	_test_eof1097: cs = 1097; goto _test_eof;
	_test_eof1209: cs = 1209; goto _test_eof;
	_test_eof1098: cs = 1098; goto _test_eof;
	_test_eof1099: cs = 1099; goto _test_eof;
	_test_eof1100: cs = 1100; goto _test_eof;
	_test_eof1101: cs = 1101; goto _test_eof;
	_test_eof1102: cs = 1102; goto _test_eof;
	_test_eof1103: cs = 1103; goto _test_eof;
	_test_eof1104: cs = 1104; goto _test_eof;
	_test_eof1210: cs = 1210; goto _test_eof;
	_test_eof1105: cs = 1105; goto _test_eof;
	_test_eof1106: cs = 1106; goto _test_eof;
	_test_eof1107: cs = 1107; goto _test_eof;
	_test_eof1108: cs = 1108; goto _test_eof;
	_test_eof1109: cs = 1109; goto _test_eof;
	_test_eof1110: cs = 1110; goto _test_eof;
	_test_eof1211: cs = 1211; goto _test_eof;
	_test_eof1111: cs = 1111; goto _test_eof;
	_test_eof1112: cs = 1112; goto _test_eof;
	_test_eof1113: cs = 1113; goto _test_eof;
	_test_eof1114: cs = 1114; goto _test_eof;
	_test_eof1115: cs = 1115; goto _test_eof;
	_test_eof1116: cs = 1116; goto _test_eof;
	_test_eof1212: cs = 1212; goto _test_eof;
	_test_eof1117: cs = 1117; goto _test_eof;
	_test_eof1118: cs = 1118; goto _test_eof;
	_test_eof1119: cs = 1119; goto _test_eof;
	_test_eof1120: cs = 1120; goto _test_eof;
	_test_eof1121: cs = 1121; goto _test_eof;
	_test_eof1122: cs = 1122; goto _test_eof;
	_test_eof1123: cs = 1123; goto _test_eof;
	_test_eof1124: cs = 1124; goto _test_eof;
	_test_eof1213: cs = 1213; goto _test_eof;
	_test_eof1125: cs = 1125; goto _test_eof;
	_test_eof1126: cs = 1126; goto _test_eof;

	_test_eof: {}
	if ( p == eof )
	{
	switch ( cs ) {
	case 7:
	case 8:
	case 151:
	case 175:
	case 177:
	case 227:
	case 241:
	case 264:
	case 318:
	case 319:
	case 325:
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	break;
	case 270:
	case 271:
	case 272:
	case 277:
	case 278:
	{
		WARN(ZS_BAD_DNAME_CHAR);
		p--; {goto st268;}
	}
	break;
	case 160:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	break;
	case 286:
	{
		WARN(ZS_BAD_TEXT);
		p--; {goto st268;}
	}
	break;
	case 291:
	case 299:
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st268;}
	}
	break;
	case 300:
	case 301:
	case 302:
	case 311:
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {goto st268;}
	}
	break;
	case 313:
	{
		ERR(ZS_BAD_INCLUDE_FILENAME);
		p--; {goto st268;}
	}
	break;
	case 316:
	case 326:
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {goto st268;}
	}
	break;
	case 1131:
	{
		NOERR;
	}
	break;
	case 152:
	case 153:
	case 154:
	case 155:
	case 156:
	case 157:
	case 158:
	case 159:
	case 195:
	case 196:
	case 197:
	case 198:
	case 199:
	case 200:
	case 201:
	case 202:
	case 203:
	{
		ERR(ZS_BAD_DIRECTIVE);
		p--; {goto st268;}
	}
	break;
	case 329:
	case 330:
	case 331:
	case 332:
	case 333:
	case 334:
	case 335:
	{
		WARN(ZS_BAD_BASE64_CHAR);
		p--; {goto st268;}
	}
	break;
	case 336:
	case 337:
	case 338:
	case 339:
	case 340:
	case 341:
	case 342:
	case 343:
	case 344:
	case 345:
	case 346:
	case 347:
	case 348:
	case 349:
	case 350:
	case 351:
	case 352:
	case 353:
	case 354:
	case 355:
	case 356:
	case 357:
	case 358:
	case 359:
	case 360:
	case 361:
	case 362:
	case 363:
	case 364:
	case 365:
	case 366:
	case 367:
	case 368:
	case 369:
	case 370:
	case 371:
	case 372:
	case 373:
	case 374:
	case 375:
	case 376:
	case 377:
	case 378:
	case 379:
	case 380:
	case 381:
	case 382:
	case 383:
	case 384:
	case 385:
	case 386:
	case 387:
	case 388:
	case 389:
	case 390:
	case 391:
	case 392:
	case 393:
	case 394:
	case 395:
	case 396:
	case 397:
	case 398:
	case 399:
	case 400:
	case 401:
	case 402:
	case 403:
	case 404:
	case 405:
	case 406:
	case 407:
	case 408:
	case 409:
	case 410:
	case 411:
	case 412:
	case 413:
	case 414:
	case 415:
	case 416:
	case 417:
	case 418:
	case 419:
	case 420:
	case 421:
	case 422:
	case 423:
	case 424:
	case 425:
	case 426:
	case 427:
	case 428:
	case 429:
	case 430:
	case 431:
	case 432:
	case 433:
	case 434:
	case 435:
	case 436:
	case 437:
	case 438:
	case 439:
	case 440:
	case 441:
	case 442:
	case 443:
	case 444:
	case 445:
	case 446:
	case 447:
	case 448:
	case 449:
	case 450:
	case 451:
	case 452:
	case 453:
	case 454:
	case 455:
	case 456:
	case 457:
	case 458:
	case 459:
	case 460:
	case 461:
	case 462:
	case 463:
	case 464:
	case 465:
	case 466:
	case 467:
	{
		WARN(ZS_BAD_BITMAP);
		p--; {goto st268;}
	}
	break;
	case 468:
	case 475:
	case 476:
	case 477:
	case 484:
	case 486:
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st268;}
	}
	break;
	case 487:
	case 488:
	case 489:
	case 490:
	case 491:
	case 492:
	case 493:
	case 494:
	case 495:
	case 496:
	case 497:
	case 498:
	case 499:
	case 500:
	case 501:
	case 502:
	case 503:
	case 504:
	case 505:
	case 506:
	case 507:
	case 508:
	case 509:
	case 510:
	case 511:
	case 512:
	case 513:
	case 514:
	case 515:
	case 516:
	case 517:
	case 518:
	case 519:
	case 520:
	case 521:
	case 522:
	case 523:
	case 524:
	case 525:
	case 526:
	case 527:
	case 528:
	case 529:
	case 530:
	case 531:
	case 532:
	case 533:
	case 534:
	case 535:
	case 536:
	case 537:
	case 538:
	case 539:
	case 540:
	case 541:
	case 542:
	case 543:
	case 544:
	case 545:
	case 546:
	case 547:
	case 548:
	case 549:
	case 550:
	case 551:
	case 552:
	case 553:
	case 554:
	case 555:
	case 556:
	case 557:
	case 558:
	case 559:
	case 560:
	case 561:
	case 562:
	case 563:
	case 564:
	case 565:
	case 566:
	case 567:
	case 568:
	case 569:
	case 570:
	case 571:
	case 572:
	case 573:
	case 574:
	case 575:
	case 576:
	case 577:
	case 578:
	case 579:
	case 580:
	case 581:
	case 582:
	case 583:
	case 584:
	case 585:
	case 586:
	case 587:
	case 588:
	case 589:
	case 590:
	{
		WARN(ZS_BAD_ALGORITHM);
		p--; {goto st268;}
	}
	break;
	case 591:
	case 592:
	case 593:
	case 594:
	case 595:
	case 596:
	case 597:
	case 598:
	case 599:
	case 600:
	case 601:
	case 602:
	case 603:
	case 604:
	case 605:
	case 606:
	case 607:
	case 608:
	case 609:
	case 610:
	case 611:
	case 612:
	case 613:
	case 614:
	case 615:
	case 616:
	case 617:
	case 618:
	case 619:
	case 620:
	case 621:
	case 622:
	case 623:
	case 624:
	case 625:
	case 626:
	case 627:
	case 628:
	case 629:
	case 630:
	case 631:
	{
		WARN(ZS_BAD_CERT_TYPE);
		p--; {goto st268;}
	}
	break;
	case 5:
	case 9:
	case 10:
	case 246:
	case 634:
	case 635:
	case 636:
	case 637:
	case 638:
	case 639:
	case 653:
	case 657:
	case 661:
	case 665:
	case 666:
	case 667:
	case 668:
	case 669:
	case 670:
	case 671:
	case 672:
	case 673:
	case 674:
	case 675:
	case 676:
	case 677:
	case 680:
	case 681:
	case 682:
	case 750:
	case 751:
	case 752:
	case 753:
	case 754:
	case 759:
	case 760:
	case 761:
	case 762:
	case 763:
	case 764:
	case 765:
	case 766:
	case 767:
	case 768:
	case 769:
	case 770:
	case 771:
	case 772:
	case 773:
	case 776:
	case 777:
	case 778:
	case 779:
	case 780:
	case 781:
	case 782:
	case 790:
	case 796:
	case 797:
	case 804:
	case 805:
	case 806:
	case 815:
	case 816:
	case 855:
	case 858:
	case 859:
	case 870:
	case 871:
	case 872:
	case 873:
	case 874:
	case 875:
	case 876:
	case 877:
	case 878:
	case 879:
	case 880:
	case 881:
	case 1010:
	case 1011:
	case 1012:
	case 1017:
	case 1018:
	case 1019:
	case 1020:
	case 1021:
	case 1022:
	case 1023:
	case 1024:
	case 1025:
	case 1043:
	case 1049:
	case 1052:
	case 1053:
	case 1054:
	case 1065:
	case 1066:
	case 1067:
	case 1078:
	case 1079:
	case 1080:
	case 1085:
	case 1098:
	case 1115:
	case 1116:
	case 1117:
	case 1118:
	case 1121:
	case 1122:
	case 1123:
	case 1124:
	case 1125:
	case 1126:
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 3:
	case 17:
	case 18:
	case 22:
	case 23:
	case 24:
	case 25:
	case 26:
	case 29:
	case 30:
	case 32:
	case 33:
	case 34:
	case 36:
	case 37:
	case 38:
	case 39:
	case 41:
	case 42:
	case 43:
	case 45:
	case 46:
	case 47:
	case 50:
	case 51:
	case 52:
	case 53:
	case 55:
	case 57:
	case 58:
	case 59:
	case 60:
	case 62:
	case 63:
	case 64:
	case 65:
	case 66:
	case 67:
	case 68:
	case 69:
	case 70:
	case 71:
	case 73:
	case 74:
	case 77:
	case 78:
	case 80:
	case 82:
	case 85:
	case 86:
	case 87:
	case 88:
	case 91:
	case 92:
	case 93:
	case 94:
	case 96:
	case 99:
	case 102:
	case 103:
	case 104:
	case 105:
	case 107:
	case 108:
	case 110:
	case 112:
	case 113:
	case 114:
	case 117:
	case 118:
	case 120:
	case 122:
	case 124:
	case 125:
	case 126:
	case 128:
	case 129:
	case 130:
	case 132:
	case 134:
	case 135:
	case 138:
	case 139:
	case 141:
	case 143:
	case 144:
	case 150:
	case 165:
	case 169:
	case 230:
	case 231:
	case 233:
	case 234:
	case 235:
	case 237:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	break;
	case 294:
	case 295:
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st268;}
	}
	break;
	case 303:
	case 304:
	case 305:
	case 310:
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {goto st268;}
	}
	break;
	case 273:
	case 274:
	case 275:
	case 276:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_DNAME_CHAR);
		p--; {goto st268;}
	}
	break;
	case 306:
	case 307:
	case 308:
	case 309:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {goto st268;}
	}
	break;
	case 320:
	case 321:
	case 322:
	case 323:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {goto st268;}
	}
	break;
	case 224:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 250:
	case 251:
	case 252:
	case 253:
	case 254:
	case 255:
	case 256:
	case 257:
	case 258:
	case 259:
	case 260:
	case 261:
	case 262:
	case 263:
	case 266:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	break;
	case 292:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st268;}
	}
	break;
	case 469:
	case 470:
	case 478:
	case 479:
	case 485:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st268;}
	}
	break;
	case 640:
	case 641:
	case 642:
	case 644:
	case 646:
	case 648:
	case 678:
	case 679:
	case 744:
	case 745:
	case 746:
	case 747:
	case 748:
	case 749:
	case 755:
	case 756:
	case 757:
	case 758:
	case 774:
	case 775:
	case 794:
	case 795:
	case 798:
	case 799:
	case 807:
	case 808:
	case 809:
	case 810:
	case 817:
	case 818:
	case 860:
	case 861:
	case 862:
	case 863:
	case 868:
	case 869:
	case 1013:
	case 1014:
	case 1015:
	case 1016:
	case 1026:
	case 1027:
	case 1028:
	case 1029:
	case 1030:
	case 1031:
	case 1055:
	case 1056:
	case 1057:
	case 1058:
	case 1059:
	case 1060:
	case 1068:
	case 1069:
	case 1070:
	case 1071:
	case 1072:
	case 1073:
	case 1081:
	case 1082:
	case 1086:
	case 1087:
	case 1111:
	case 1112:
	case 1113:
	case 1114:
	case 1119:
	case 1120:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 136:
	case 145:
	case 161:
	case 164:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	break;
	case 864:
	case 865:
	case 866:
	case 867:
	{
		WARN(ZS_BAD_TIMESTAMP_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 279:
	case 280:
	case 285:
	{
		WARN(ZS_BAD_TEXT_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_TEXT);
		p--; {goto st268;}
	}
	break;
	case 683:
	case 684:
	case 685:
	case 686:
	{
		s->long_string = false;
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 314:
	{
		ERR(ZS_BAD_INCLUDE_FILENAME);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	break;
	case 315:
	case 317:
	case 324:
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	break;
	case 632:
	case 633:
	case 687:
	case 688:
	case 1083:
	case 1084:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 783:
	case 784:
	case 785:
	case 788:
	case 789:
	case 791:
	{
		WARN(ZS_BAD_APL);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 800:
	case 801:
	case 802:
	case 803:
	case 811:
	case 812:
	case 813:
	case 814:
	case 1032:
	case 1033:
	case 1050:
	case 1051:
	case 1061:
	case 1062:
	case 1063:
	case 1064:
	case 1074:
	case 1075:
	case 1076:
	case 1077:
	case 1088:
	case 1089:
	case 1090:
	case 1091:
	case 1093:
	case 1094:
	case 1095:
	case 1096:
	case 1099:
	case 1100:
	case 1102:
	case 1103:
	case 1105:
	case 1106:
	case 1108:
	case 1109:
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 1034:
	case 1035:
	case 1036:
	case 1037:
	case 1038:
	case 1039:
	case 1040:
	case 1041:
	case 1042:
	case 1044:
	case 1045:
	case 1046:
	case 1047:
	case 1048:
	{
		WARN(ZS_BAD_BASE32HEX_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 819:
	case 820:
	case 823:
	case 832:
	case 833:
	case 834:
	case 839:
	case 840:
	case 841:
	case 846:
	case 847:
	case 848:
	case 851:
	case 853:
	case 854:
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 825:
	case 826:
	case 827:
	case 828:
	case 829:
	case 830:
	case 831:
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 856:
	case 857:
	case 882:
	case 883:
	case 884:
	case 885:
	case 886:
	case 887:
	case 888:
	case 889:
	case 890:
	case 891:
	case 892:
	case 893:
	case 894:
	case 895:
	case 896:
	case 897:
	case 898:
	case 899:
	case 900:
	case 901:
	case 902:
	case 903:
	case 904:
	case 905:
	case 906:
	case 907:
	case 908:
	case 909:
	case 910:
	case 911:
	case 912:
	case 913:
	case 914:
	case 915:
	case 916:
	case 917:
	case 918:
	case 919:
	case 920:
	case 921:
	case 922:
	case 923:
	case 924:
	case 925:
	case 926:
	case 927:
	case 928:
	case 929:
	case 930:
	case 931:
	case 932:
	case 933:
	case 934:
	case 935:
	case 936:
	case 937:
	case 938:
	case 939:
	case 940:
	case 941:
	case 942:
	case 943:
	case 944:
	case 945:
	case 946:
	case 947:
	case 948:
	case 949:
	case 950:
	case 951:
	case 952:
	case 953:
	case 954:
	case 955:
	case 956:
	case 957:
	case 958:
	case 959:
	case 960:
	case 961:
	case 962:
	case 963:
	case 964:
	case 965:
	case 966:
	case 967:
	case 968:
	case 969:
	case 970:
	case 971:
	case 972:
	case 973:
	case 974:
	case 975:
	case 976:
	case 977:
	case 978:
	case 979:
	case 980:
	case 981:
	case 982:
	case 983:
	case 984:
	case 985:
	case 986:
	case 987:
	case 988:
	case 989:
	case 990:
	case 991:
	case 992:
	case 993:
	case 994:
	case 995:
	case 996:
	case 997:
	case 998:
	case 999:
	case 1000:
	case 1001:
	case 1002:
	case 1003:
	case 1004:
	case 1007:
	case 1008:
	case 1009:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 689:
	case 690:
	case 691:
	case 692:
	case 693:
	case 694:
	case 695:
	case 696:
	case 697:
	case 698:
	case 699:
	case 700:
	case 701:
	case 702:
	case 703:
	case 704:
	case 705:
	case 706:
	case 707:
	case 708:
	case 709:
	case 710:
	case 711:
	case 712:
	case 713:
	case 714:
	case 715:
	case 716:
	case 717:
	case 718:
	case 719:
	case 720:
	case 721:
	case 722:
	case 723:
	case 724:
	case 725:
	case 726:
	case 727:
	case 728:
	case 729:
	case 730:
	case 731:
	case 732:
	case 733:
	case 734:
	case 735:
	case 736:
	case 737:
	case 738:
	case 739:
	case 740:
	case 741:
	case 742:
	case 743:
	{
		WARN(ZS_BAD_LOC_DATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 6:
	case 11:
	case 12:
	case 13:
	case 20:
	case 21:
	case 173:
	case 176:
	case 193:
	case 226:
	case 228:
	case 229:
	case 239:
	case 240:
	case 242:
	case 243:
	case 244:
	case 245:
	case 248:
	case 249:
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	break;
	case 16:
	case 207:
	case 265:
	case 267:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	break;
	case 4:
	case 19:
	case 27:
	case 28:
	case 31:
	case 35:
	case 40:
	case 44:
	case 48:
	case 49:
	case 54:
	case 56:
	case 61:
	case 72:
	case 75:
	case 76:
	case 79:
	case 81:
	case 83:
	case 84:
	case 89:
	case 90:
	case 95:
	case 97:
	case 98:
	case 100:
	case 101:
	case 106:
	case 109:
	case 111:
	case 115:
	case 116:
	case 119:
	case 121:
	case 123:
	case 127:
	case 131:
	case 133:
	case 140:
	case 232:
	case 236:
	case 238:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 204:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	break;
	case 170:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 821:
	case 822:
	case 835:
	case 836:
	case 842:
	case 843:
	case 849:
	case 850:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 1005:
	case 1006:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 1:
	case 142:
	case 174:
	case 206:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	break;
	case 137:
	case 171:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 297:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st268;}
	}
	break;
	case 643:
	case 645:
	case 647:
	case 649:
	case 650:
	case 651:
	case 652:
	case 654:
	case 655:
	case 656:
	case 658:
	case 659:
	case 660:
	case 662:
	case 663:
	case 664:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 2:
	case 146:
	case 147:
	case 148:
	case 149:
	case 166:
	case 167:
	case 168:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	break;
	case 281:
	case 282:
	case 283:
	case 284:
	case 287:
	case 288:
	case 289:
	case 290:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_TEXT_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_TEXT);
		p--; {goto st268;}
	}
	break;
	case 786:
	case 787:
	case 792:
	case 793:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_APL);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 837:
	case 844:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 471:
	case 472:
	case 473:
	case 474:
	case 480:
	case 481:
	case 482:
	case 483:
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st268;}
	}
	break;
	case 1092:
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_CHAR_COLON);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 824:
	case 852:
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 1101:
	case 1104:
	case 1107:
	case 1110:
	{
		WARN(ZS_BAD_CHAR_DASH);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 1097:
	{
		WARN(ZS_BAD_CHAR_COLON);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 194:
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	{
		ERR(ZS_BAD_DIRECTIVE);
		p--; {goto st268;}
	}
	break;
	case 172:
	case 179:
	case 180:
	case 181:
	case 182:
	case 183:
	case 184:
	case 185:
	case 186:
	case 187:
	case 188:
	case 189:
	case 190:
	case 191:
	case 192:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	break;
	case 163:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	break;
	case 208:
	case 210:
	case 211:
	case 212:
	case 213:
	case 214:
	case 215:
	case 216:
	case 217:
	case 218:
	case 219:
	case 220:
	case 221:
	case 222:
	case 223:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	break;
	case 14:
	case 178:
	case 209:
	case 225:
	case 247:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	break;
	case 293:
	case 296:
	case 298:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st268;}
	}
	break;
	case 838:
	case 845:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	break;
	case 15:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	break;
	case 205:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st268;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st268;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st268;}
	}
	break;
	}
	}

	_out: {}
	}


	// Check if the scanner state machine is in an uncovered state.
	bool extra_error = false;
	if (cs == 0) {
		ERR(ZS_UNCOVERED_STATE);
		extra_error = true;
	// Check for an unclosed multiline record.
	} else if (s->input.eof && s->multiline) {
		ERR(ZS_UNCLOSED_MULTILINE);
		extra_error = true;
	}

	// Treat the extra error.
	if (extra_error) {
		s->error.counter++;
		s->state = ZS_STATE_ERROR;

		// Copy the error context just for the part of the current line.
		s->buffer_length = 0;
		while (p < pe && *p != '\n' && s->buffer_length < 50) {
			s->buffer[s->buffer_length++] = *p++;
		}
		s->buffer[s->buffer_length++] = 0;

		// Execute the error callback.
		if (s->process.automatic && s->process.error != NULL) {
			s->process.error(s);
		}

		return;
	}

	// Storing scanner states.
	s->cs  = cs;
	s->top = top;
	memcpy(s->stack, stack, sizeof(stack));

	// Store the current parser position.
	s->input.current = p;

	// Storing r_data pointer.
	s->r_data_tail = rdata_tail - s->r_data;
}

__attribute__((visibility("default")))
int zs_parse_record(
	zs_scanner_t *s)
{
	if (s == NULL) {
		return -1;
	}

	// Check if parsing is possible.
	switch (s->state) {
	case ZS_STATE_NONE:
	case ZS_STATE_DATA:
	case ZS_STATE_INCLUDE:
		break;
	case ZS_STATE_ERROR:
		if (s->error.fatal) {
			return -1;
		}
		break;
	default:
		// Return if stop or end of file.
		return 0;
	}

	// Check for the end of the input.
	if (s->input.current != s->input.end) {
		// Try to parse another item.
		s->state = ZS_STATE_NONE;
		parse(s);

		// Finish if nothing was parsed.
		if (s->state == ZS_STATE_NONE) {
			// Parse the final block.
			if (zs_set_input_string(s, "\n", 1) != 0) {
				return -1;
			}
			s->input.eof = true;
			parse(s);
			if (s->state == ZS_STATE_NONE) {
				s->state = ZS_STATE_EOF;
			}
		}
	} else {
		s->state = ZS_STATE_EOF;
	}

	return 0;
}

__attribute__((visibility("default")))
int zs_parse_all(
	zs_scanner_t *s)
{
	if (s == NULL) {
		return -1;
	}

	s->process.automatic = true;

	// Parse input block.
	parse(s);

	// Parse trailing newline-char block if it makes sense.
	if (s->state != ZS_STATE_STOP && !s->error.fatal) {
		if (zs_set_input_string(s, "\n", 1) != 0) {
			return -1;
		}
		s->input.eof = true;
		parse(s);
	}

	// Check if any errors have occurred.
	if (s->error.counter > 0) {
		return -1;
	}

	return 0;
}
